"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const typeGuards_1 = require("../utilities/typeGuards");
/**
 * Identify all local identifiers in a provided expression
 */
class LocalResolver {
    /**
     * local resolver constructor
     */
    constructor() {
        this.isTrailer = false;
    }
    /**
     * resolve local identifiers in expression
     * @param expr expression
     */
    resolveExpr(expr) {
        return expr.accept(this);
    }
    /**
     * resolve a suffix term
     * @param suffixTerm suffix term
     */
    resolveSuffixTerm(suffixTerm) {
        return suffixTerm.accept(this);
    }
    /**
     * Visit an invalid expression
     * @param _ invalid expression
     */
    visitExprInvalid(_) {
        return [];
    }
    /**
     * Visit a binary expression
     * @param expr binary expression
     */
    visitBinary(expr) {
        const tokens = this.resolveExpr(expr.left);
        tokens.push(...this.resolveExpr(expr.right));
        return tokens;
    }
    /**
     * Visit a unary expression
     * @param expr unary expression
     */
    visitUnary(expr) {
        return this.resolveExpr(expr.factor);
    }
    /**
     * Vist a factor
     * @param expr factor expression
     */
    visitFactor(expr) {
        const tokens = this.resolveExpr(expr.suffix);
        tokens.push(...this.resolveExpr(expr.exponent));
        return tokens;
    }
    /**
     * Vist a suffix expression
     * @param expr suffix expression
     */
    visitSuffix(expr) {
        const tokens = this.resolveSuffixTerm(expr.suffixTerm);
        if (!typeGuards_1.empty(expr.trailer)) {
            tokens.push(...this.resolveSuffixTerm(expr.trailer));
        }
        return tokens;
    }
    /**
     * Visit a lambda
     * @param _ lambda expression
     */
    visitLambda(_) {
        return [];
    }
    /**
     * Vist an invalid suffix term
     * @param _ invalid suffix term
     */
    visitSuffixTermInvalid(_) {
        return [];
    }
    /**
     * visit a suffix term trailer
     * @param suffixTerm suffix term trailer
     */
    visitSuffixTrailer(suffixTerm) {
        // indicate we're currently in a trailer
        return this.executeAs(true, () => {
            const tokens = this.resolveSuffixTerm(suffixTerm.suffixTerm);
            if (!typeGuards_1.empty(suffixTerm.trailer)) {
                tokens.push(...this.resolveSuffixTerm(suffixTerm.trailer));
            }
            return tokens;
        });
    }
    /**
     * Visit a suffix term
     * @param suffixTerm suffix term
     */
    visitSuffixTerm(suffixTerm) {
        const tokens = this.resolveSuffixTerm(suffixTerm.atom);
        for (const trailer of suffixTerm.trailers) {
            tokens.push(...this.resolveSuffixTerm(trailer));
        }
        return tokens;
    }
    /**
     * Visit a call suffix term trailer
     * @param suffixTerm call trailer
     */
    visitCall(suffixTerm) {
        if (suffixTerm.args.length === 0)
            return [];
        // indicate args are not in a trailer
        return this.executeAs(false, () => {
            const tokens = [];
            for (const arg of suffixTerm.args) {
                tokens.push(...this.resolveExpr(arg));
            }
            return tokens;
        });
    }
    /**
     * Visit an array index
     * @param _ array index trailer
     */
    visitArrayIndex(_) {
        return [];
    }
    /**
     * Visit an array bracket
     * @param suffixTerm array bracket trailer
     */
    visitArrayBracket(suffixTerm) {
        return this.executeAs(false, () => this.resolveExpr(suffixTerm.index));
    }
    /**
     * Visit a delegate
     * @param _ delgate trailer
     */
    visitDelegate(_) {
        return [];
    }
    /**
     * Visit a literal
     * @param _ literal suffix term
     */
    visitLiteral(_) {
        return [];
    }
    /**
     * Visit an identifier
     * @param suffixTerm suffix term identifier
     */
    visitIdentifier(suffixTerm) {
        // if we're a trailer return nothing otherwise identifer
        return this.isTrailer ? [] : [suffixTerm.token];
    }
    /**
     * Visit a suffix term grouping
     * @param suffixTerm suffix term grouping
     */
    visitGrouping(suffixTerm) {
        return this.resolveExpr(suffixTerm.expr);
    }
    /**
     * Execute a node as either a trailer or not. This function
     * returns to it's previous state post execution
     * @param isTrailer is the function a trailer
     * @param func function to execute
     */
    executeAs(isTrailer, func) {
        const old = this.isTrailer;
        this.isTrailer = isTrailer;
        const result = func();
        this.isTrailer = old;
        return result;
    }
}
exports.LocalResolver = LocalResolver;
//# sourceMappingURL=localResolver.js.map