import { IStmtVisitor, IExprVisitor, IStmt, ISuffixTermVisitor, IScript } from '../parser/types';
import * as SuffixTerm from '../parser/suffixTerm';
import * as Expr from '../parser/expr';
import * as Stmt from '../parser/stmt';
import { Var, Lock, Func, Param } from '../parser/declare';
import { SymbolTableBuilder } from './symbolTableBuilder';
import { Diagnostic } from 'vscode-languageserver';
export declare type Diagnostics = Diagnostic[];
/**
 * The pre resolver run prior to the resolver. Its main purpose is to
 * find and store function declaration location of functions is kerboscripts
 * does not matter.
 */
export declare class PreResolver implements IExprVisitor<Diagnostics>, IStmtVisitor<Diagnostics>, ISuffixTermVisitor<Diagnostics> {
    /**
     * current script being processed
     */
    private script;
    /**
     * symbol table builder
     */
    private tableBuilder;
    /**
     * logger
     */
    private readonly logger;
    /**
     * tracer
     */
    private readonly tracer;
    /**
     * function scan to find parameters and return
     * statements
     */
    private readonly functionScan;
    /**
     * Pre resolver constructor
     * @param script pre resolver script
     * @param symbolTableBuilder symbol table builder
     * @param logger logger
     * @param tracer tracer
     */
    constructor(script: IScript, symbolTableBuilder: SymbolTableBuilder, logger?: ILogger, tracer?: ITracer);
    /**
     * Perform an initial resolver pass on the script statements,
     * for function declarations
     */
    resolve(): Diagnostics;
    resolveStmts(stmts: IStmt[]): Diagnostics;
    private resolveStmt;
    private resolveExpr;
    private resolveSuffixTerm;
    visitDeclVariable(decl: Var): Diagnostics;
    visitDeclLock(decl: Lock): Diagnostics;
    /**
     * Because function don't need forward declaration in kerboscript
     * we need to find and add their symbols first
     * @param decl function declaration
     */
    visitDeclFunction(decl: Func): Diagnostics;
    visitDeclParameter(_: Param): Diagnostics;
    visitStmtInvalid(_: Stmt.Invalid): Diagnostics;
    visitBlock(stmt: Stmt.Block): Diagnostics;
    visitExpr(stmt: Stmt.ExprStmt): Diagnostics;
    visitOnOff(stmt: Stmt.OnOff): Diagnostics;
    visitCommand(_: Stmt.Command): Diagnostics;
    visitCommandExpr(stmt: Stmt.CommandExpr): Diagnostics;
    visitUnset(_: Stmt.Unset): Diagnostics;
    visitUnlock(_: Stmt.Unlock): Diagnostics;
    visitSet(stmt: Stmt.Set): Diagnostics;
    visitLazyGlobal(_: Stmt.LazyGlobal): Diagnostics;
    visitIf(stmt: Stmt.If): Diagnostics;
    visitElse(stmt: Stmt.Else): Diagnostics;
    visitUntil(stmt: Stmt.Until): Diagnostics;
    visitFrom(stmt: Stmt.From): Diagnostics;
    visitWhen(stmt: Stmt.When): Diagnostics;
    visitReturn(stmt: Stmt.Return): Diagnostics;
    visitBreak(_: Stmt.Break): Diagnostics;
    visitSwitch(stmt: Stmt.Switch): Diagnostics;
    visitFor(stmt: Stmt.For): Diagnostics;
    visitOn(stmt: Stmt.On): Diagnostics;
    visitToggle(stmt: Stmt.Toggle): Diagnostics;
    visitWait(stmt: Stmt.Wait): Diagnostics;
    visitLog(stmt: Stmt.Log): Diagnostics;
    visitCopy(stmt: Stmt.Copy): Diagnostics;
    visitRename(stmt: Stmt.Rename): Diagnostics;
    visitDelete(stmt: Stmt.Delete): Diagnostics;
    visitRun(stmt: Stmt.Run): Diagnostics;
    visitRunPath(stmt: Stmt.RunPath): Diagnostics;
    visitRunPathOnce(stmt: Stmt.RunOncePath): Diagnostics;
    visitCompile(stmt: Stmt.Compile): Diagnostics;
    visitList(_: Stmt.List): Diagnostics;
    visitEmpty(_: Stmt.Empty): Diagnostics;
    visitPrint(stmt: Stmt.Print): Diagnostics;
    visitExprInvalid(_: Expr.Invalid): Diagnostics;
    visitBinary(expr: Expr.Binary): Diagnostics;
    visitUnary(expr: Expr.Unary): Diagnostics;
    visitFactor(expr: Expr.Factor): Diagnostics;
    visitSuffix(expr: Expr.Suffix): Diagnostics;
    visitLambda(expr: Expr.Lambda): Diagnostics;
    visitSuffixTermInvalid(_: SuffixTerm.Invalid): Diagnostics;
    visitSuffixTrailer(expr: SuffixTerm.SuffixTrailer): Diagnostics;
    visitSuffixTerm(expr: SuffixTerm.SuffixTerm): Diagnostics;
    visitCall(expr: SuffixTerm.Call): Diagnostics;
    visitArrayIndex(_: SuffixTerm.ArrayIndex): Diagnostics;
    visitArrayBracket(expr: SuffixTerm.ArrayBracket): Diagnostics;
    visitDelegate(_: SuffixTerm.Delegate): Diagnostics;
    visitLiteral(_: SuffixTerm.Literal): Diagnostics;
    visitIdentifier(_: SuffixTerm.Identifier): Diagnostics;
    visitGrouping(expr: SuffixTerm.Grouping): Diagnostics;
}
