"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const types_1 = require("../parser/types");
const typeGuards_1 = require("../utilities/typeGuards");
const tokentypes_1 = require("../entities/tokentypes");
const logger_1 = require("../utilities/logger");
const functionScan_1 = require("./functionScan");
/**
 * The pre resolver run prior to the resolver. Its main purpose is to
 * find and store function declaration location of functions is kerboscripts
 * does not matter.
 */
class PreResolver {
    /**
     * Pre resolver constructor
     * @param script pre resolver script
     * @param symbolTableBuilder symbol table builder
     * @param logger logger
     * @param tracer tracer
     */
    constructor(script, symbolTableBuilder, logger = logger_1.mockLogger, tracer = logger_1.mockTracer) {
        this.script = script;
        this.tableBuilder = symbolTableBuilder;
        this.logger = logger;
        this.tracer = tracer;
        this.functionScan = new functionScan_1.FunctionScan();
    }
    /**
     * Perform an initial resolver pass on the script statements,
     * for function declarations
     */
    resolve() {
        try {
            const splits = this.script.uri.split('/');
            const file = splits[splits.length - 1];
            this.logger.info(`Function Resolving started for ${file}.`);
            this.tableBuilder.rewind();
            this.tableBuilder.beginScope(this.script);
            const resolveErrors = this.resolveStmts(this.script.stmts);
            this.tableBuilder.endScope();
            this.logger.info(`Function Resolving finished for ${file}`);
            if (resolveErrors.length) {
                this.logger.warn(`Function Resolver encounted ${resolveErrors.length} errors`);
            }
            return resolveErrors;
        }
        catch (err) {
            this.logger.error(`Error occured in resolver ${err}`);
            this.tracer.log(err);
            return [];
        }
    }
    // resolve the given set of statements
    resolveStmts(stmts) {
        return accumulateErrors(stmts, this.resolveStmt.bind(this));
    }
    // resolve for an statement
    resolveStmt(stmt) {
        return stmt.accept(this);
    }
    // resolve for an expression
    resolveExpr(expr) {
        return expr.accept(this);
    }
    // resolve for an expression
    resolveSuffixTerm(suffixTerm) {
        return suffixTerm.accept(this);
    }
    /* --------------------------------------------
  
    Declarations
  
    ----------------------------------------------*/
    // check variable declaration
    visitDeclVariable(decl) {
        return this.resolveExpr(decl.value);
    }
    // check lock declaration
    visitDeclLock(decl) {
        return this.resolveExpr(decl.value);
    }
    /**
     * Because function don't need forward declaration in kerboscript
     * we need to find and add their symbols first
     * @param decl function declaration
     */
    visitDeclFunction(decl) {
        const scopeToken = decl.scope && decl.scope.scope;
        let scopeType;
        // functions are default global at file scope and local everywhere else
        if (typeGuards_1.empty(scopeToken)) {
            scopeType = this.tableBuilder.isFileScope()
                ? types_1.ScopeKind.global
                : types_1.ScopeKind.local;
        }
        else {
            switch (scopeToken.type) {
                case tokentypes_1.TokenType.local:
                    scopeType = types_1.ScopeKind.local;
                    break;
                case tokentypes_1.TokenType.global:
                    scopeType = types_1.ScopeKind.global;
                    break;
                default:
                    throw new Error('Unexpected scope token encountered. Expected local or global.');
            }
        }
        const result = this.functionScan.scan(decl.block);
        const declareErrors = this.tableBuilder.declareFunction(scopeType, decl.identifier, result.requiredParameters, result.optionalParameters, result.return);
        const stmtErrors = this.resolveStmt(decl.block);
        if (!typeGuards_1.empty(declareErrors)) {
            stmtErrors.push(declareErrors);
        }
        return stmtErrors;
    }
    // check parameter declaration
    visitDeclParameter(_) {
        return [];
    }
    /* --------------------------------------------
  
    Statements
  
    ----------------------------------------------*/
    visitStmtInvalid(_) {
        return [];
    }
    visitBlock(stmt) {
        this.tableBuilder.beginScope(stmt);
        const errors = this.resolveStmts(stmt.stmts);
        this.tableBuilder.endScope();
        return errors;
    }
    visitExpr(stmt) {
        return this.resolveExpr(stmt.suffix);
    }
    visitOnOff(stmt) {
        return this.resolveExpr(stmt.suffix);
    }
    visitCommand(_) {
        return [];
    }
    visitCommandExpr(stmt) {
        return this.resolveExpr(stmt.expr);
    }
    visitUnset(_) {
        return [];
    }
    visitUnlock(_) {
        return [];
    }
    visitSet(stmt) {
        return this.resolveExpr(stmt.value);
    }
    visitLazyGlobal(_) {
        return [];
    }
    visitIf(stmt) {
        const errors = this.resolveExpr(stmt.condition);
        errors.push(...this.resolveStmt(stmt.body));
        if (stmt.elseStmt) {
            errors.push(...this.resolveStmt(stmt.elseStmt));
        }
        return errors;
    }
    visitElse(stmt) {
        return this.resolveStmt(stmt.body);
    }
    visitUntil(stmt) {
        const errors = this.resolveExpr(stmt.condition);
        errors.push(...this.resolveStmt(stmt.body));
        return errors;
    }
    visitFrom(stmt) {
        const errors = this.resolveStmts(stmt.initializer.stmts);
        errors.push(...this.resolveExpr(stmt.condition), ...this.resolveStmts(stmt.increment.stmts), ...this.resolveStmt(stmt.body));
        return errors;
    }
    visitWhen(stmt) {
        const errors = this.resolveExpr(stmt.condition);
        errors.push(...this.resolveStmt(stmt.body));
        return errors;
    }
    visitReturn(stmt) {
        if (stmt.value) {
            return this.resolveExpr(stmt.value);
        }
        return [];
    }
    visitBreak(_) {
        return [];
    }
    visitSwitch(stmt) {
        return this.resolveExpr(stmt.target);
    }
    visitFor(stmt) {
        const errors = this.resolveExpr(stmt.collection);
        errors.push(...this.resolveStmt(stmt.body));
        return errors;
    }
    visitOn(stmt) {
        const errors = this.resolveExpr(stmt.suffix);
        errors.push(...this.resolveStmt(stmt.body));
        return errors;
    }
    visitToggle(stmt) {
        return this.resolveExpr(stmt.suffix);
    }
    visitWait(stmt) {
        return this.resolveExpr(stmt.expr);
    }
    visitLog(stmt) {
        const errors = this.resolveExpr(stmt.expr);
        errors.push(...this.resolveExpr(stmt.target));
        return errors;
    }
    visitCopy(stmt) {
        const errors = this.resolveExpr(stmt.target);
        errors.push(...this.resolveExpr(stmt.destination));
        return errors;
    }
    visitRename(stmt) {
        const errors = this.resolveExpr(stmt.target);
        errors.push(...this.resolveExpr(stmt.alternative));
        return errors;
    }
    visitDelete(stmt) {
        const errors = this.resolveExpr(stmt.target);
        if (typeGuards_1.empty(stmt.volume)) {
            return errors;
        }
        errors.push(...this.resolveExpr(stmt.volume));
        return errors;
    }
    visitRun(stmt) {
        if (typeGuards_1.empty(stmt.args)) {
            return [];
        }
        return accumulateErrors(stmt.args, this.resolveExpr.bind(this));
    }
    visitRunPath(stmt) {
        if (typeGuards_1.empty(stmt.args)) {
            return this.resolveExpr(stmt.expr);
        }
        const errors = this.resolveExpr(stmt.expr);
        errors.push(...accumulateErrors(stmt.args, this.resolveExpr.bind(this)));
        return errors;
    }
    visitRunPathOnce(stmt) {
        if (typeGuards_1.empty(stmt.args)) {
            return this.resolveExpr(stmt.expr);
        }
        const errors = this.resolveExpr(stmt.expr);
        errors.push(...accumulateErrors(stmt.args, this.resolveExpr.bind(this)));
        return errors;
    }
    visitCompile(stmt) {
        if (typeGuards_1.empty(stmt.destination)) {
            return this.resolveExpr(stmt.target);
        }
        const errors = this.resolveExpr(stmt.target);
        errors.push(...this.resolveExpr(stmt.destination));
        return errors;
    }
    visitList(_) {
        return [];
    }
    visitEmpty(_) {
        return [];
    }
    visitPrint(stmt) {
        return this.resolveExpr(stmt.expr);
    }
    /* --------------------------------------------
  
    Expressions
  
    ----------------------------------------------*/
    visitExprInvalid(_) {
        return [];
    }
    visitBinary(expr) {
        const errors = this.resolveExpr(expr.left);
        errors.push(...this.resolveExpr(expr.right));
        return errors;
    }
    visitUnary(expr) {
        return this.resolveExpr(expr.factor);
    }
    visitFactor(expr) {
        const errors = this.resolveExpr(expr.suffix);
        errors.push(...this.resolveExpr(expr.exponent));
        return errors;
    }
    visitSuffix(expr) {
        const errors = this.resolveSuffixTerm(expr.suffixTerm);
        if (!typeGuards_1.empty(expr.trailer)) {
            errors.push(...this.resolveSuffixTerm(expr.trailer));
        }
        return errors;
    }
    visitLambda(expr) {
        return this.resolveStmt(expr.block);
    }
    /* --------------------------------------------
  
    Suffix Terms
  
    ----------------------------------------------*/
    visitSuffixTermInvalid(_) {
        return [];
    }
    visitSuffixTrailer(expr) {
        const errors = this.resolveSuffixTerm(expr.suffixTerm);
        if (!typeGuards_1.empty(expr.trailer)) {
            errors.push(...this.resolveSuffixTerm(expr.trailer));
        }
        return errors;
    }
    visitSuffixTerm(expr) {
        const errors = this.resolveSuffixTerm(expr.atom);
        for (const trailer of expr.trailers) {
            errors.push(...this.resolveSuffixTerm(trailer));
        }
        return errors;
    }
    visitCall(expr) {
        return accumulateErrors(expr.args, this.resolveExpr.bind(this));
    }
    visitArrayIndex(_) {
        return [];
    }
    visitArrayBracket(expr) {
        return this.resolveExpr(expr.index);
    }
    visitDelegate(_) {
        return [];
    }
    visitLiteral(_) {
        return [];
    }
    visitIdentifier(_) {
        return [];
    }
    visitGrouping(expr) {
        return this.resolveExpr(expr.expr);
    }
}
exports.PreResolver = PreResolver;
const accumulateErrors = (items, checker) => {
    const errors = [];
    for (const item of items) {
        errors.push(...checker(item));
    }
    return errors;
};
//# sourceMappingURL=preResolver.js.map