import { IStmtVisitor, IScript, IExprVisitor, ISuffixTermVisitor } from '../parser/types';
import * as SuffixTerm from '../parser/suffixTerm';
import * as Expr from '../parser/expr';
import * as Stmt from '../parser/stmt';
import * as Decl from '../parser/declare';
import { SymbolTableBuilder } from './symbolTableBuilder';
import { Diagnostic } from 'vscode-languageserver';
declare type Diagnostics = Diagnostic[];
export declare class Resolver implements IStmtVisitor<Diagnostics>, IExprVisitor<Diagnostics>, ISuffixTermVisitor<Diagnostics> {
    /**
     * Script the resolver executes on
     */
    private readonly script;
    /**
     * Symbol table builder
     */
    private readonly tableBuilder;
    /**
     * Logger to log messages
     */
    private readonly logger;
    /**
     * Tracer to log traces
     */
    private readonly tracer;
    /**
     * Local resolver to find used local symbols in an expression
     */
    private readonly localResolver;
    /**
     * Set resolver to find set symbols and used symbols
     */
    private readonly setResolver;
    /**
     * Queue of deferred nodes to execute
     */
    private readonly deferred;
    /**
     * Is the script using the lazy global directive
     */
    private lazyGlobal;
    /**
     * Is this the first statement of the script
     */
    private firstStmt;
    /**
     * Should resolution be deferred
     */
    private deferResolve;
    /**
     * How many functions deep is the current location
     */
    private functionDepth;
    /**
     * How many loops deep is the current location
     */
    private loopDepth;
    /**
     * Cached bound method for resolving statement
     */
    private readonly resolveStmtBind;
    /**
     * Cached bound method for using locals in an expreesion
     */
    private readonly useExprLocalsBind;
    /**
     * Cached bound method for resolving an expression
     */
    private readonly resolveExprBind;
    /**
     * Resolver constructor
     * @param script script to resolve
     * @param symbolTableBuilder symbol table builder
     * @param logger logger
     * @param tracer tracer
     */
    constructor(script: IScript, symbolTableBuilder: SymbolTableBuilder, logger?: ILogger, tracer?: ITracer);
    /**
     * Resolve a script
     */
    resolve(): Diagnostics;
    /**
     * Resolve the given set of statements
     * @param stmts statements
     */
    private resolveStmts;
    /**
     * Resolve statement
     * @param stmt resolve statement
     */
    private resolveStmt;
    /**
     * Pass through expression
     * @param expr expression to skip
     */
    private resolveExpr;
    /**
     * Pass through suffix term
     * @param suffixTerm suffix term to skip
     */
    private resolveSuffixTerm;
    /**
     * attempt to use ever variable in the expression
     * @param expr expression to use
     */
    private useExprLocals;
    /**
     * attempt to use ever token in the collection
     * @param tokens local results to use
     */
    private useTokens;
    /**
     * filter to just errors
     * @param maybeError potential error
     */
    private filterError;
    /**
     * Visit the declare variable syntax node
     * @param decl the syntax node
     */
    visitDeclVariable(decl: Decl.Var): Diagnostics;
    /**
     * Visit the declare lock syntax node
     * @param decl the syntax node
     */
    visitDeclLock(decl: Decl.Lock): Diagnostic[];
    /**
     * Visit the declare function syntax node
     * @param decl the syntax node
     */
    visitDeclFunction(decl: Decl.Func): Diagnostic[];
    /**
     * Visit the declare parameter syntax node
     * @param decl the syntax node
     */
    visitDeclParameter(decl: Decl.Param): Diagnostic[];
    /**
     * Visit the Invalid Stmt syntax node
     * @param stmt the syntax node
     */
    visitStmtInvalid(stmt: Stmt.Invalid): Diagnostics;
    /**
     * Visit the Block Stmt syntax node
     * @param stmt the syntax node
     */
    visitBlock(stmt: Stmt.Block): Diagnostics;
    /**
     * Visit the Expr Stmt syntax node
     * @param stmt the syntax node
     */
    visitExpr(stmt: Stmt.ExprStmt): Diagnostics;
    /**
     * Visit the On Off Stmt syntax node
     * @param stmt the syntax node
     */
    visitOnOff(stmt: Stmt.OnOff): Diagnostics;
    /**
     * Visit the Command Stmt syntax node
     * @param _ the syntax node
     */
    visitCommand(_: Stmt.Command): Diagnostics;
    /**
     * Visit the Command Expr Stmt syntax node
     * @param stmt the syntax node
     */
    visitCommandExpr(stmt: Stmt.CommandExpr): Diagnostics;
    /**
     * Visit the Unset Stmt syntax node
     * @param stmt the syntax node
     */
    visitUnset(stmt: Stmt.Unset): Diagnostics;
    /**
     * Visit the Unlock Stmt syntax node
     * @param stmt the syntax node
     */
    visitUnlock(stmt: Stmt.Unlock): Diagnostics;
    /**
     * Visit the Set Stmt syntax node
     * @param stmt the syntax node
     */
    visitSet(stmt: Stmt.Set): Diagnostics;
    /**
     * Visit the Lazy Global Stmt syntax node
     * @param stmt the syntax node
     */
    visitLazyGlobal(stmt: Stmt.LazyGlobal): Diagnostics;
    /**
     * Visit the If Stmt syntax node
     * @param stmt the syntax node
     */
    visitIf(stmt: Stmt.If): Diagnostics;
    /**
     * Visit the Else Stmt syntax node
     * @param stmt the syntax node
     */
    visitElse(stmt: Stmt.Else): Diagnostics;
    /**
     * Visit the When Stmt syntax node
     * @param stmt the syntax node
     */
    visitUntil(stmt: Stmt.Until): Diagnostics;
    /**
     * Visit the From Stmt syntax node
     * @param stmt the syntax node
     */
    visitFrom(stmt: Stmt.From): Diagnostics;
    /**
     * Visit the When Stmt syntax node
     * @param stmt the syntax node
     */
    visitWhen(stmt: Stmt.When): Diagnostics;
    /**
     * Visit the Return Stmt syntax node
     * @param stmt the syntax node
     */
    visitReturn(stmt: Stmt.Return): Diagnostics;
    /**
     * Visit the Break Stmt syntax node
     * @param stmt the syntax node
     */
    visitBreak(stmt: Stmt.Break): Diagnostics;
    /**
     * Visit the Switch Stmt syntax node
     * @param stmt the syntax node
     */
    visitSwitch(stmt: Stmt.Switch): Diagnostics;
    /**
     * Visit the For Stmt syntax node
     * @param stmt the syntax node
     */
    visitFor(stmt: Stmt.For): Diagnostics;
    /**
     * Visit the On Stmt syntax node
     * @param stmt the syntax node
     */
    visitOn(stmt: Stmt.On): Diagnostics;
    /**
     * Visit the Toggle Stmt syntax node
     * @param stmt the syntax node
     */
    visitToggle(stmt: Stmt.Toggle): Diagnostics;
    /**
     * Visit the Wait Stmt syntax node
     * @param stmt the syntax node
     */
    visitWait(stmt: Stmt.Wait): Diagnostics;
    /**
     * Visit the Log Stmt syntax node
     * @param stmt the syntax node
     */
    visitLog(stmt: Stmt.Log): Diagnostics;
    /**
     * Visit the Copy Stmt syntax node
     * @param stmt the syntax node
     */
    visitCopy(stmt: Stmt.Copy): Diagnostics;
    /**
     * Visit the Rename Stmt syntax node
     * @param stmt the syntax node
     */
    visitRename(stmt: Stmt.Rename): Diagnostics;
    /**
     * Visit the Delete Stmt syntax node
     * @param stmt the syntax node
     */
    visitDelete(stmt: Stmt.Delete): Diagnostics;
    /**
     * Visit the Run Stmt syntax node
     * @param stmt the syntax node
     */
    visitRun(stmt: Stmt.Run): Diagnostics;
    /**
     * Visit the RunPath Stmt syntax node
     * @param stmt the syntax node
     */
    visitRunPath(stmt: Stmt.RunPath): Diagnostics;
    /**
     * Visit the RunPathOnce Stmt syntax node
     * @param stmt the syntax node
     */
    visitRunPathOnce(stmt: Stmt.RunOncePath): Diagnostics;
    /**
     * Visit the Compile Stmt syntax node
     * @param stmt the syntax node
     */
    visitCompile(stmt: Stmt.Compile): Diagnostics;
    /**
     * Visit the List Stmt syntax node
     * @param stmt the syntax node
     */
    visitList(stmt: Stmt.List): Diagnostics;
    /**
     * Visit the Empty Stmt syntax node
     * @param stmt the syntax node
     */
    visitEmpty(_: Stmt.Empty): Diagnostics;
    /**
     * Visit the Print Stmt syntax node
     * @param stmt the syntax node
     */
    visitPrint(stmt: Stmt.Print): Diagnostics;
    /**
     * Logic for settings a variable. used by set stmt and list command
     * @param set token to set
     */
    private setBinding;
    visitExprInvalid(_: Expr.Invalid): Diagnostics;
    visitBinary(expr: Expr.Binary): Diagnostics;
    visitUnary(expr: Expr.Unary): Diagnostics;
    visitFactor(expr: Expr.Factor): Diagnostics;
    visitSuffix(expr: Expr.Suffix): Diagnostics;
    visitLambda(expr: Expr.Lambda): Diagnostics;
    visitSuffixTermInvalid(_: SuffixTerm.Invalid): Diagnostics;
    visitSuffixTrailer(suffixTerm: SuffixTerm.SuffixTrailer): Diagnostics;
    visitSuffixTerm(suffixTerm: SuffixTerm.SuffixTerm): Diagnostics;
    visitCall(suffixTerm: SuffixTerm.Call): Diagnostics;
    visitArrayIndex(_: SuffixTerm.ArrayIndex): Diagnostics;
    visitArrayBracket(suffixTerm: SuffixTerm.ArrayBracket): Diagnostics;
    visitDelegate(_: SuffixTerm.Delegate): Diagnostics;
    visitLiteral(_: SuffixTerm.Literal): Diagnostics;
    visitIdentifier(_: SuffixTerm.Identifier): Diagnostics;
    visitGrouping(suffixTerm: SuffixTerm.Grouping): Diagnostics;
    /**
     * Only allow one defer node to be executed at a time
     */
    private executeDeferred;
    /**
     * Track when the function depth has increased or decreased
     * @param functionFunc function body
     */
    private trackFunction;
    /**
     * Tracket when the loop depth has increased or decreased
     * @param loopFunc loop body
     */
    private trackLoop;
    /**
     * Defer a node for later execution
     * @param node node to defer
     */
    private deferNode;
}
export {};
