"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const types_1 = require("../parser/types");
const Expr = __importStar(require("../parser/expr"));
const Stmt = __importStar(require("../parser/stmt"));
const typeGuards_1 = require("../utilities/typeGuards");
const localResolver_1 = require("./localResolver");
const setResolver_1 = require("./setResolver");
const tokentypes_1 = require("../entities/tokentypes");
const logger_1 = require("../utilities/logger");
const vscode_languageserver_1 = require("vscode-languageserver");
const diagnosticsUtils_1 = require("../utilities/diagnosticsUtils");
// tslint:disable-next-line: import-name
const denque_1 = __importDefault(require("denque"));
const arrayUtils_1 = require("../utilities/arrayUtils");
class Resolver {
    /**
     * Resolver constructor
     * @param script script to resolve
     * @param symbolTableBuilder symbol table builder
     * @param logger logger
     * @param tracer tracer
     */
    constructor(script, symbolTableBuilder, logger = logger_1.mockLogger, tracer = logger_1.mockTracer) {
        /**
         * Cached bound method for resolving statement
         */
        this.resolveStmtBind = this.resolveStmt.bind(this);
        /**
         * Cached bound method for using locals in an expreesion
         */
        this.useExprLocalsBind = this.useExprLocals.bind(this);
        /**
         * Cached bound method for resolving an expression
         */
        this.resolveExprBind = this.resolveExpr.bind(this);
        this.script = script;
        this.tableBuilder = symbolTableBuilder;
        this.localResolver = new localResolver_1.LocalResolver();
        this.setResolver = new setResolver_1.SetResolver(this.localResolver);
        this.deferred = new denque_1.default();
        this.lazyGlobal = true;
        this.firstStmt = true;
        this.deferResolve = true;
        this.loopDepth = 0;
        this.functionDepth = 0;
        this.logger = logger;
        this.tracer = tracer;
    }
    /**
     * Resolve a script
     */
    resolve() {
        try {
            const splits = this.script.uri.split('/');
            const file = splits[splits.length - 1];
            this.logger.info(`Resolving started for ${file}.`);
            this.tableBuilder.rewind();
            this.tableBuilder.beginScope(this.script);
            const [firstStmt, ...restStmts] = this.script.stmts;
            // check for lazy global flag
            const errors = this.resolveStmt(firstStmt);
            this.firstStmt = false;
            // resolve reset
            const resolveErrors = this.resolveStmts(restStmts);
            this.tableBuilder.endScope();
            this.script.lazyGlobal = this.lazyGlobal;
            this.deferResolve = false;
            // process all deferred nodes
            let current;
            errors.push(...resolveErrors);
            // process deferred queue
            while ((current = this.deferred.shift())) {
                this.deferResolve = false;
                // set scope path and current depths
                this.tableBuilder.setPath(current.path, current.activeScope);
                this.functionDepth = current.functionDepth;
                this.loopDepth = current.loopDepth;
                // resolve deferred node
                switch (current.node.tag) {
                    case types_1.SyntaxKind.expr:
                        errors.push(...this.resolveExpr(current.node));
                        break;
                    case types_1.SyntaxKind.stmt:
                        errors.push(...this.resolveStmt(current.node));
                        break;
                }
            }
            this.logger.info(`Resolving finished for ${file}`);
            if (errors.length) {
                this.logger.warn(`Resolver encounted ${errors.length} errors`);
            }
            return errors;
        }
        catch (err) {
            this.logger.error(`Error occured in resolver ${err}`);
            this.tracer.log(err);
            return [];
        }
    }
    /**
     * Resolve the given set of statements
     * @param stmts statements
     */
    resolveStmts(stmts) {
        return accumulateErrors(stmts, this.resolveStmtBind);
    }
    /**
     * Resolve statement
     * @param stmt resolve statement
     */
    resolveStmt(stmt) {
        return stmt.accept(this);
    }
    /**
     * Pass through expression
     * @param expr expression to skip
     */
    resolveExpr(expr) {
        return expr.accept(this);
    }
    /**
     * Pass through suffix term
     * @param suffixTerm suffix term to skip
     */
    resolveSuffixTerm(suffixTerm) {
        return suffixTerm.accept(this);
    }
    /**
     * attempt to use ever variable in the expression
     * @param expr expression to use
     */
    useExprLocals(expr) {
        return this.useTokens(this.localResolver.resolveExpr(expr));
    }
    /**
     * attempt to use ever token in the collection
     * @param tokens local results to use
     */
    useTokens(tokens) {
        const errors = [];
        for (const token of tokens) {
            const error = this.tableBuilder.useSymbol(token);
            if (!typeGuards_1.empty(error)) {
                errors.push(error);
            }
        }
        return errors;
    }
    /**
     * filter to just errors
     * @param maybeError potential error
     */
    filterError(maybeError) {
        return !typeGuards_1.empty(maybeError);
    }
    /* --------------------------------------------
  
    Declarations
  
    ----------------------------------------------*/
    /**
     * Visit the declare variable syntax node
     * @param decl the syntax node
     */
    visitDeclVariable(decl) {
        // determine scope type
        const scopeType = !typeGuards_1.empty(decl.scope) ? decl.scope.type : types_1.ScopeKind.global;
        const declareError = this.tableBuilder.declareVariable(scopeType, decl.identifier);
        const errors = this.useExprLocals(decl.value);
        errors.push(...this.resolveExpr(decl.value));
        if (!typeGuards_1.empty(declareError)) {
            errors.push(declareError);
        }
        return errors;
    }
    /**
     * Visit the declare lock syntax node
     * @param decl the syntax node
     */
    visitDeclLock(decl) {
        // determine scope type
        const scopeType = !typeGuards_1.empty(decl.scope) ? decl.scope.type : types_1.ScopeKind.global;
        const lookup = this.tableBuilder.lookupLock(decl.identifier, types_1.ScopeKind.global);
        let declareError = undefined;
        if (typeGuards_1.empty(lookup)) {
            declareError = this.tableBuilder.declareLock(scopeType, decl.identifier);
        }
        const errors = this.useExprLocals(decl.value);
        errors.push(...this.resolveExpr(decl.value));
        if (!typeGuards_1.empty(declareError)) {
            errors.push(declareError);
        }
        return errors;
    }
    /**
     * Visit the declare function syntax node
     * @param decl the syntax node
     */
    visitDeclFunction(decl) {
        if (this.executeDeferred()) {
            return this.deferNode(decl, decl.block);
        }
        return this.trackFunction(() => this.resolveStmt(decl.block));
    }
    /**
     * Visit the declare parameter syntax node
     * @param decl the syntax node
     */
    visitDeclParameter(decl) {
        const errors = [];
        // check that parameter isn't declared global
        if (!typeGuards_1.empty(decl.scope) && !typeGuards_1.empty(decl.scope.scope)) {
            if (decl.scope.scope.type === tokentypes_1.TokenType.global) {
                errors.push(diagnosticsUtils_1.createDiagnostic(decl.scope.scope, 'Parameters cannot be global', vscode_languageserver_1.DiagnosticSeverity.Error));
            }
        }
        // all parameters are local
        const scopeType = types_1.ScopeKind.local;
        // need to check if default paraemter can really be abbitrary expr
        const parameterErrors = decl.requiredParameters
            .map(parameter => this.tableBuilder.declareParameter(scopeType, parameter.identifier, false))
            .filter(this.filterError);
        const defaultParameterErrors = decl.optionalParameters
            .map(parameter => this.tableBuilder.declareParameter(scopeType, parameter.identifier, true))
            .filter(this.filterError);
        const defaultUseErrors = decl.optionalParameters.map(parameter => this.useExprLocals(parameter.value));
        errors.push(...parameterErrors, ...defaultParameterErrors, ...arrayUtils_1.flatten(defaultUseErrors));
        return errors;
    }
    /* --------------------------------------------
  
    Statements
  
    ----------------------------------------------*/
    /**
     * Visit the Invalid Stmt syntax node
     * @param stmt the syntax node
     */
    visitStmtInvalid(stmt) {
        if (typeGuards_1.empty(stmt.partial)) {
            return [];
        }
        const errors = [];
        // check parsed partial nodes
        for (const node of Object.values(stmt.partial)) {
            if (node instanceof Stmt.Stmt) {
                errors.push(...this.resolveStmt(node));
            }
            if (node instanceof Expr.Expr) {
                errors.push(...this.useExprLocalsBind(node));
                errors.push(...this.resolveExpr(node));
            }
        }
        return errors;
    }
    /**
     * Visit the Block Stmt syntax node
     * @param stmt the syntax node
     */
    visitBlock(stmt) {
        this.tableBuilder.beginScope(stmt);
        const errors = this.resolveStmts(stmt.stmts);
        this.tableBuilder.endScope();
        return errors;
    }
    /**
     * Visit the Expr Stmt syntax node
     * @param stmt the syntax node
     */
    visitExpr(stmt) {
        return this.useExprLocals(stmt.suffix).concat(this.resolveExpr(stmt.suffix));
    }
    /**
     * Visit the On Off Stmt syntax node
     * @param stmt the syntax node
     */
    visitOnOff(stmt) {
        return this.useExprLocals(stmt.suffix).concat(this.resolveExpr(stmt.suffix));
    }
    /**
     * Visit the Command Stmt syntax node
     * @param _ the syntax node
     */
    visitCommand(_) {
        return [];
    }
    /**
     * Visit the Command Expr Stmt syntax node
     * @param stmt the syntax node
     */
    visitCommandExpr(stmt) {
        return this.useExprLocals(stmt.expr).concat(this.resolveExpr(stmt.expr));
    }
    /**
     * Visit the Unset Stmt syntax node
     * @param stmt the syntax node
     */
    visitUnset(stmt) {
        const error = this.tableBuilder.useVariable(stmt.identifier);
        return typeGuards_1.empty(error) ? [] : [error];
    }
    /**
     * Visit the Unlock Stmt syntax node
     * @param stmt the syntax node
     */
    visitUnlock(stmt) {
        const error = this.tableBuilder.useLock(stmt.identifier);
        return typeGuards_1.empty(error) ? [] : [error];
    }
    /**
     * Visit the Set Stmt syntax node
     * @param stmt the syntax node
     */
    visitSet(stmt) {
        const { set, used } = this.setResolver.resolveExpr(stmt.suffix);
        // check if a set target exists
        if (typeGuards_1.empty(set)) {
            const [token] = this.localResolver.resolveExpr(stmt.suffix);
            return [
                diagnosticsUtils_1.createDiagnostic(token, `cannot assign to variable ${token.lexeme}`, vscode_languageserver_1.DiagnosticSeverity.Error),
            ];
        }
        const errors = this.setBinding(set);
        errors.push(...this.useExprLocals(stmt.value), ...this.useTokens(used), ...this.resolveExpr(stmt.value));
        return errors;
    }
    /**
     * Visit the Lazy Global Stmt syntax node
     * @param stmt the syntax node
     */
    visitLazyGlobal(stmt) {
        // It is an error if lazy global is not at the start of a file
        if (!this.firstStmt) {
            return [
                diagnosticsUtils_1.createDiagnostic(stmt.lazyGlobal, 'Lazy global was not declared at top of the file', vscode_languageserver_1.DiagnosticSeverity.Error),
            ];
        }
        this.lazyGlobal = stmt.onOff.type === tokentypes_1.TokenType.on;
        return [];
    }
    /**
     * Visit the If Stmt syntax node
     * @param stmt the syntax node
     */
    visitIf(stmt) {
        const errors = this.useExprLocals(stmt.condition);
        errors.push(...this.resolveExpr(stmt.condition), ...this.resolveStmt(stmt.body));
        if (stmt.elseStmt) {
            errors.push(...this.resolveStmt(stmt.elseStmt));
        }
        return errors;
    }
    /**
     * Visit the Else Stmt syntax node
     * @param stmt the syntax node
     */
    visitElse(stmt) {
        return this.resolveStmt(stmt.body);
    }
    /**
     * Visit the When Stmt syntax node
     * @param stmt the syntax node
     */
    visitUntil(stmt) {
        const errors = this.useExprLocals(stmt.condition);
        errors.push(...this.resolveExpr(stmt.condition));
        errors.push(...this.trackLoop(() => this.resolveStmtBind(stmt.body)));
        return errors;
    }
    /**
     * Visit the From Stmt syntax node
     * @param stmt the syntax node
     */
    visitFrom(stmt) {
        // begin hidden loop scope
        this.tableBuilder.beginScope(stmt);
        const errors = this.resolveStmts(stmt.initializer.stmts);
        errors.push(...this.resolveExpr(stmt.condition), ...this.useExprLocalsBind(stmt.condition), ...this.resolveStmts(stmt.increment.stmts));
        errors.push(...this.trackLoop(() => this.resolveStmt(stmt.body)));
        // end hidden loop scope
        this.tableBuilder.endScope();
        return errors;
    }
    /**
     * Visit the When Stmt syntax node
     * @param stmt the syntax node
     */
    visitWhen(stmt) {
        if (this.executeDeferred()) {
            return this.deferNode(stmt, stmt.body);
        }
        const errors = this.useExprLocalsBind(stmt.condition);
        errors.push(...this.resolveExpr(stmt.condition), ...this.resolveStmt(stmt.body));
        return errors;
    }
    /**
     * Visit the Return Stmt syntax node
     * @param stmt the syntax node
     */
    visitReturn(stmt) {
        const errors = [];
        if (stmt.value) {
            errors.push(...this.useExprLocalsBind(stmt.value), ...this.resolveExpr(stmt.value));
        }
        if (this.functionDepth < 1) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.returnToken, 'Return appeared outside of function body', vscode_languageserver_1.DiagnosticSeverity.Error));
        }
        return errors;
    }
    /**
     * Visit the Break Stmt syntax node
     * @param stmt the syntax node
     */
    visitBreak(stmt) {
        return this.loopDepth < 1
            ? [
                diagnosticsUtils_1.createDiagnostic(stmt, 'Break appeared outside of a loop', vscode_languageserver_1.DiagnosticSeverity.Error),
            ]
            : [];
    }
    /**
     * Visit the Switch Stmt syntax node
     * @param stmt the syntax node
     */
    visitSwitch(stmt) {
        return this.useExprLocalsBind(stmt.target).concat(this.resolveExpr(stmt.target));
    }
    /**
     * Visit the For Stmt syntax node
     * @param stmt the syntax node
     */
    visitFor(stmt) {
        return this.trackLoop(() => {
            this.tableBuilder.beginScope(stmt);
            const errors = [];
            const declareError = this.tableBuilder.declareVariable(types_1.ScopeKind.local, stmt.element);
            if (!typeGuards_1.empty(declareError)) {
                errors.push(declareError);
            }
            errors.push(...this.useExprLocalsBind(stmt.collection), ...this.resolveExpr(stmt.collection), ...this.resolveStmt(stmt.body));
            this.tableBuilder.endScope();
            if (!typeGuards_1.empty(declareError)) {
                errors.push(declareError);
            }
            return errors;
        });
    }
    /**
     * Visit the On Stmt syntax node
     * @param stmt the syntax node
     */
    visitOn(stmt) {
        if (this.executeDeferred()) {
            return this.deferNode(stmt, stmt.body);
        }
        return this.useExprLocalsBind(stmt.suffix).concat(this.resolveExpr(stmt.suffix), this.resolveStmt(stmt.body));
    }
    /**
     * Visit the Toggle Stmt syntax node
     * @param stmt the syntax node
     */
    visitToggle(stmt) {
        return this.useExprLocalsBind(stmt.suffix).concat(this.resolveExpr(stmt.suffix));
    }
    /**
     * Visit the Wait Stmt syntax node
     * @param stmt the syntax node
     */
    visitWait(stmt) {
        return this.useExprLocalsBind(stmt.expr).concat(this.resolveExpr(stmt.expr));
    }
    /**
     * Visit the Log Stmt syntax node
     * @param stmt the syntax node
     */
    visitLog(stmt) {
        return this.useExprLocalsBind(stmt.expr).concat(this.resolveExpr(stmt.expr), this.resolveExpr(stmt.target));
    }
    /**
     * Visit the Copy Stmt syntax node
     * @param stmt the syntax node
     */
    visitCopy(stmt) {
        return this.useExprLocalsBind(stmt.target).concat(diagnosticsUtils_1.createDiagnostic(stmt, 'Copy is deprecated as of 1.0.0', vscode_languageserver_1.DiagnosticSeverity.Warning), this.resolveExpr(stmt.target), this.resolveExpr(stmt.destination));
    }
    /**
     * Visit the Rename Stmt syntax node
     * @param stmt the syntax node
     */
    visitRename(stmt) {
        return this.useExprLocalsBind(stmt.target).concat(diagnosticsUtils_1.createDiagnostic(stmt, 'Rename is deprecated as of 1.0.0', vscode_languageserver_1.DiagnosticSeverity.Warning), this.useExprLocalsBind(stmt.alternative), this.resolveExpr(stmt.target), this.resolveExpr(stmt.alternative));
    }
    /**
     * Visit the Delete Stmt syntax node
     * @param stmt the syntax node
     */
    visitDelete(stmt) {
        const deprecated = diagnosticsUtils_1.createDiagnostic(stmt, 'Copy is deprecated as of 1.0.0', vscode_languageserver_1.DiagnosticSeverity.Warning);
        if (typeGuards_1.empty(stmt.volume)) {
            return this.useExprLocalsBind(stmt.target).concat(deprecated, this.resolveExpr(stmt.target));
        }
        return this.useExprLocalsBind(stmt.target).concat(deprecated, this.useExprLocalsBind(stmt.volume), this.resolveExpr(stmt.target), this.resolveExpr(stmt.volume));
    }
    /**
     * Visit the Run Stmt syntax node
     * @param stmt the syntax node
     */
    visitRun(stmt) {
        if (typeGuards_1.empty(stmt.args) && typeGuards_1.empty(stmt.expr)) {
            return [];
        }
        const argError = !typeGuards_1.empty(stmt.args)
            ? accumulateErrors(stmt.args, this.useExprLocalsBind).concat(accumulateErrors(stmt.args, this.resolveExprBind))
            : [];
        if (typeGuards_1.empty(stmt.expr)) {
            return argError;
        }
        return this.useExprLocalsBind(stmt.expr).concat(this.resolveExpr(stmt.expr), argError);
    }
    /**
     * Visit the RunPath Stmt syntax node
     * @param stmt the syntax node
     */
    visitRunPath(stmt) {
        if (typeGuards_1.empty(stmt.args)) {
            return this.useExprLocalsBind(stmt.expr).concat(this.resolveExpr(stmt.expr));
        }
        return this.useExprLocalsBind(stmt.expr).concat(this.resolveExpr(stmt.expr), accumulateErrors(stmt.args, this.useExprLocalsBind), accumulateErrors(stmt.args, this.resolveExprBind));
    }
    /**
     * Visit the RunPathOnce Stmt syntax node
     * @param stmt the syntax node
     */
    visitRunPathOnce(stmt) {
        if (typeGuards_1.empty(stmt.args)) {
            return this.useExprLocalsBind(stmt.expr).concat(this.resolveExpr(stmt.expr));
        }
        return this.useExprLocalsBind(stmt.expr).concat(this.resolveExpr(stmt.expr), accumulateErrors(stmt.args, this.useExprLocalsBind), accumulateErrors(stmt.args, this.resolveExprBind));
    }
    /**
     * Visit the Compile Stmt syntax node
     * @param stmt the syntax node
     */
    visitCompile(stmt) {
        if (typeGuards_1.empty(stmt.destination)) {
            return this.useExprLocalsBind(stmt.target).concat(this.resolveExpr(stmt.target));
        }
        return this.useExprLocalsBind(stmt.target).concat(this.useExprLocalsBind(stmt.destination), this.resolveExpr(stmt.target), this.resolveExpr(stmt.destination));
    }
    /**
     * Visit the List Stmt syntax node
     * @param stmt the syntax node
     */
    visitList(stmt) {
        // list generates new variable when target is used
        if (typeGuards_1.empty(stmt.target)) {
            return [];
        }
        return this.setBinding(stmt.target);
    }
    /**
     * Visit the Empty Stmt syntax node
     * @param stmt the syntax node
     */
    visitEmpty(_) {
        return [];
    }
    /**
     * Visit the Print Stmt syntax node
     * @param stmt the syntax node
     */
    visitPrint(stmt) {
        return this.useExprLocalsBind(stmt.expr).concat(this.resolveExpr(stmt.expr));
    }
    /**
     * Logic for settings a variable. used by set stmt and list command
     * @param set token to set
     */
    setBinding(set) {
        // if variable isn't defined either report error or define
        let defineError = undefined;
        // if we find the symbol just set it
        if (!typeGuards_1.empty(this.tableBuilder.lookupBinding(set, types_1.ScopeKind.global))) {
            defineError = this.tableBuilder.setBinding(set);
            // if we didn't find it and we're not lazy global add error
        }
        else if (!this.lazyGlobal) {
            defineError = diagnosticsUtils_1.createDiagnostic(set, `Attempted to set ${set.lexeme} which has not been declared. ` +
                `Either remove lazy global directive or declare ${set.lexeme}`, vscode_languageserver_1.DiagnosticSeverity.Error);
            // not found and lazy global so declare global
        }
        else {
            defineError = this.tableBuilder.declareVariable(types_1.ScopeKind.global, set);
        }
        if (!typeGuards_1.empty(defineError)) {
            return [defineError];
        }
        return [];
    }
    /* --------------------------------------------
  
    Expressions
  
    ----------------------------------------------*/
    visitExprInvalid(_) {
        return [];
    }
    visitBinary(expr) {
        const errors = this.resolveExpr(expr.left);
        errors.push(...this.resolveExpr(expr.right));
        return errors;
    }
    visitUnary(expr) {
        return this.resolveExpr(expr.factor);
    }
    visitFactor(expr) {
        const errors = this.resolveExpr(expr.suffix);
        errors.push(...this.resolveExpr(expr.exponent));
        return errors;
    }
    visitSuffix(expr) {
        const errors = this.visitSuffixTerm(expr.suffixTerm);
        if (!typeGuards_1.empty(expr.trailer)) {
            errors.push(...this.resolveSuffixTerm(expr.trailer));
        }
        return errors;
    }
    visitLambda(expr) {
        if (this.executeDeferred()) {
            return this.deferNode(expr, expr.block);
        }
        return this.trackFunction(() => this.resolveStmt(expr.block));
    }
    /* --------------------------------------------
  
    Suffix Terms
  
    ----------------------------------------------*/
    visitSuffixTermInvalid(_) {
        return [];
    }
    visitSuffixTrailer(suffixTerm) {
        const errors = this.visitSuffixTerm(suffixTerm.suffixTerm);
        if (!typeGuards_1.empty(suffixTerm.trailer)) {
            errors.push(...(this.resolveSuffixTerm(suffixTerm.trailer)));
        }
        return errors;
    }
    visitSuffixTerm(suffixTerm) {
        const errors = this.resolveSuffixTerm(suffixTerm.atom);
        for (const trailer of suffixTerm.trailers) {
            errors.push(...this.resolveSuffixTerm(trailer));
        }
        return errors;
    }
    visitCall(suffixTerm) {
        return accumulateErrors(suffixTerm.args, this.resolveExprBind);
    }
    visitArrayIndex(_) {
        return [];
    }
    visitArrayBracket(suffixTerm) {
        return this.resolveExpr(suffixTerm.index);
    }
    visitDelegate(_) {
        return [];
    }
    visitLiteral(_) {
        return [];
    }
    visitIdentifier(_) {
        return [];
    }
    visitGrouping(suffixTerm) {
        return this.resolveExpr(suffixTerm.expr);
    }
    /**
     * Only allow one defer node to be executed at a time
     */
    executeDeferred() {
        if (this.deferResolve) {
            return true;
        }
        this.deferResolve = true;
        return false;
    }
    /**
     * Track when the function depth has increased or decreased
     * @param functionFunc function body
     */
    trackFunction(functionFunc) {
        this.functionDepth += 1;
        const result = functionFunc();
        this.functionDepth -= 1;
        return result;
    }
    /**
     * Tracket when the loop depth has increased or decreased
     * @param loopFunc loop body
     */
    trackLoop(loopFunc) {
        this.loopDepth += 1;
        const result = loopFunc();
        this.loopDepth -= 1;
        return result;
    }
    /**
     * Defer a node for later execution
     * @param node node to defer
     */
    deferNode(node, block) {
        this.deferred.push({
            node,
            functionDepth: this.functionDepth,
            loopDepth: this.loopDepth,
            ...this.tableBuilder.getPath(),
        });
        // for now kinda a hack for now may need to look at scope building again
        if (block instanceof Stmt.Block) {
            this.tableBuilder.beginScope(block);
            this.tableBuilder.endScope();
        }
        return [];
    }
}
exports.Resolver = Resolver;
const accumulateErrors = (items, checker) => {
    const errors = [];
    for (const item of items) {
        errors.push(...checker(item));
    }
    return errors;
};
//# sourceMappingURL=resolver.js.map