"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const setResult_1 = require("./setResult");
const typeGuards_1 = require("../utilities/typeGuards");
/**
 * Identify all local sybmbols used and all used symbols
 */
class SetResolver {
    /**
     * Set resolver constructor
     */
    constructor(localResolver) {
        this.localResolver = localResolver;
    }
    /**
     * Resolve expression
     * @param expr expression to resolve
     */
    resolveExpr(expr) {
        return expr.accept(this);
    }
    /**
     * Resolve suffix term
     * @param suffixTerm suffix term to resolve
     */
    resolveSuffixTerm(suffixTerm) {
        return suffixTerm.accept(this);
    }
    /**
     * Resolve invalid expression, i.e. do nothing
     * @param _ invalid expression
     */
    visitExprInvalid(_) {
        return setResult_1.setResult();
    }
    /**
     * Resolve binary expresiosn. Binary expression cannot be set
     * @param _ binary expression
     */
    visitBinary(_) {
        return setResult_1.setResult();
    }
    /**
     * Resolve unary expression. Unary expression cannot be set
     * @param _ unary expression
     */
    visitUnary(_) {
        return setResult_1.setResult();
    }
    /**
     * Resolve factor expression. Factor cannot be set.
     * @param _ factor expression
     */
    visitFactor(_) {
        return setResult_1.setResult();
    }
    /**
     * Resolve suffix expression.
     * @param expr suffix expression
     */
    visitSuffix(expr) {
        const result = this.resolveSuffixTerm(expr.suffixTerm);
        if (typeGuards_1.empty(expr.trailer)) {
            return setResult_1.setResult(result.set, result.used);
        }
        return setResult_1.setResult(result.set, result.used, this.localResolver.resolveSuffixTerm(expr.trailer));
    }
    /**
     * Resolve lambda expression. Lambda cannot be set
     * @param _ lambda expression
     */
    visitLambda(_) {
        return setResult_1.setResult();
    }
    /**
     * Resolve invalid suffix term. i.e. do nothing
     * @param _ invalid suffix term
     */
    visitSuffixTermInvalid(_) {
        throw setResult_1.setResult();
    }
    /**
     * Resolve suffix term trailer
     * @param expr suffix term trailer
     */
    visitSuffixTrailer(expr) {
        const result = this.resolveSuffixTerm(expr.suffixTerm);
        if (typeGuards_1.empty(expr.trailer)) {
            return setResult_1.setResult(result.set, result.used);
        }
        return setResult_1.setResult(result.set, result.used, this.localResolver.resolveSuffixTerm(expr.trailer));
    }
    /**
     * Resolve suffix term
     * @param expr suffix term
     */
    visitSuffixTerm(expr) {
        const result = this.resolveSuffixTerm(expr.atom);
        if (expr.trailers.length === 0) {
            return setResult_1.setResult(result.set, result.used);
        }
        return setResult_1.setResult(result.set, result.used, expr.trailers.reduce((acc, curr) => acc.concat(this.localResolver.resolveSuffixTerm(curr)), []));
    }
    /**
     * Resolve suffix term call trailer
     * @param _ suffix term call trailer
     */
    visitCall(_) {
        return setResult_1.setResult();
    }
    /**
     * Resolve array index trailer
     * @param _ suffix term array index trailer
     */
    visitArrayIndex(_) {
        return setResult_1.setResult();
    }
    /**
     * Resolve array bracket trailer
     * @param _ suffix term array bracket trailer
     */
    visitArrayBracket(_) {
        return setResult_1.setResult();
    }
    /**
     * Resolve delegate trailer
     * @param _ suffix term delegate
     */
    visitDelegate(_) {
        return setResult_1.setResult();
    }
    /**
     * Resolve literal
     * @param _ literal
     */
    visitLiteral(_) {
        return setResult_1.setResult();
    }
    /**
     * Resolve identifier
     * @param expr identifier suffix term
     */
    visitIdentifier(expr) {
        return setResult_1.setResult(expr.token);
    }
    /**
     * Resolve grouping suffix term. Grouping cannot be set
     * @param _ grouping suffix term
     */
    visitGrouping(_) {
        return setResult_1.setResult();
    }
}
exports.SetResolver = SetResolver;
//# sourceMappingURL=setResolver.js.map