import { EnvironmentNode, KsSymbol, GraphNode, SymbolTrackerBase, KsBaseSymbol } from './types';
import { Position } from 'vscode-languageserver';
import { KsFunction } from '../entities/function';
import { KsLock } from '../entities/lock';
import { KsVariable } from '../entities/variable';
import { KsParameter } from '../entities/parameter';
import { BasicTracker } from './tracker';
/**
 * The Symbol table is used to update and retreive symbol type infromation
 */
export declare class SymbolTable implements GraphNode<SymbolTable> {
    /**
     * The root scope of this files symbol table
     */
    readonly rootScope: EnvironmentNode;
    /**
     * The parent symbol tables to this symbol table
     */
    readonly parentSymbolTables: Set<SymbolTable>;
    /**
     * the child symbol tables to this symbol table
     */
    readonly childSymbolTables: Set<SymbolTable>;
    /**
     * file uri for this symbol table
     */
    readonly uri: string;
    /**
     * the symbol table logger
     */
    readonly logger: ILogger;
    /**
     * construct a symbol table. typically this is done by the symbol table builder
     * @param rootScope the root scope of the symbol table
     * @param childSymbolTables child scopes for this symbol table
     * @param uri the file uri for this symbol table
     * @param logger a logger for the symbol table
     */
    constructor(rootScope: EnvironmentNode, childSymbolTables: Set<SymbolTable>, uri: string, logger?: ILogger);
    /**
     * return itself for the graph interface
     */
    readonly value: SymbolTable;
    /**
     * get all adjacent nodes to this symbol table
     */
    readonly adjacentNodes: GraphNode<SymbolTable>[];
    /**
     * add a new child symbol table to this symbol table
     * @param symbolTable the new child
     */
    addScope(symbolTable: SymbolTable): void;
    /**
     * should be called when the associated file is closed. It will remove it's reference
     * if another symbol table doesn't reference it
     */
    closeSelf(): void;
    /**
     * This should be called when the associated symbol table is to be deleted
     */
    removeSelf(): void;
    /**
     * get every symbol in the file
     */
    fileSymbols(): KsBaseSymbol[];
    /**
     * Get global symbols that match the requested symbol name
     * @param name symbol name
     */
    globalTrackers(name: string): SymbolTrackerBase[];
    /**
     * Get the tracker for a function symbol
     * @param pos position of symbol
     * @param name name of the symbol
     */
    scopedFunctionTracker(pos: Position, name: string): Maybe<SymbolTrackerBase<KsFunction>>;
    /**
     * Get the tracker for a lock symbol
     * @param pos position of symbol
     * @param name name of the symbol
     */
    scopedLockTracker(pos: Position, name: string): Maybe<SymbolTrackerBase<KsLock>>;
    /**
     * Get the tracker for a variable symbol
     * @param pos position of the symbol
     * @param name name of the symbol
     */
    scopedVariableTracker(pos: Position, name: string): Maybe<SymbolTrackerBase<KsVariable>>;
    /**
     * Get the tracker for a parameter symbol
     * @param pos position of the symbol
     * @param name name of the symbol
     */
    scopedParameterTracker(pos: Position, name: string): Maybe<SymbolTrackerBase<KsParameter>>;
    /**
     * Get all symbols available at the provide position
     * @param pos position to check
     */
    scopedSymbols(pos: Position): KsSymbol[];
    /**
     * Get a tracker for a given symbol under a filter condition and position
     * @param pos position to check
     * @param name name of the symbol
     * @param symbolFilter filter function for the symbol
     */
    scopedNamedTracker(pos: Position, name: string, symbolFilter?: (x: BasicTracker) => boolean): Maybe<BasicTracker>;
    /**
     * recursively move up the scope to get every file symbol
     * @param nodes nodes to retrive symbols from
     */
    private fileSymbolsDepth;
    /**
     * get all symbol trackers in scope at a positition
     * @param pos the position to retrive symbols from
     */
    private scopedTrackers;
    /**
     * get a symbol tracker under the conditions of a provided filter
     * @param pos the posiition to check for symbols
     * @param trackerFilter the tracker filter
     */
    private scopedTracker;
    /**
     * recursively move up scopes to find most relevant symbols
     * @param pos position to search
     * @param nodes scope nodes to check
     */
    private scopedTrackersDepth;
    /**
     * recursively move up scope for more relevant symbols
     * @param pos position to search
     * @param nodes scope nodes to check
     * @param trackerFilter filter for acceptable symbols
     */
    private scopedTrackerDepth;
}
