"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const positionUtils_1 = require("../utilities/positionUtils");
const logger_1 = require("../utilities/logger");
const typeGuards_1 = require("../utilities/typeGuards");
const entityHelpers_1 = require("../entities/entityHelpers");
const types_1 = require("../parser/types");
/**
 * The Symbol table is used to update and retreive symbol type infromation
 */
class SymbolTable {
    /**
     * construct a symbol table. typically this is done by the symbol table builder
     * @param rootScope the root scope of the symbol table
     * @param childSymbolTables child scopes for this symbol table
     * @param uri the file uri for this symbol table
     * @param logger a logger for the symbol table
     */
    constructor(rootScope, childSymbolTables, uri, logger = logger_1.mockLogger) {
        this.rootScope = rootScope;
        this.childSymbolTables = childSymbolTables;
        this.uri = uri;
        this.logger = logger;
        this.parentSymbolTables = new Set();
        for (const symbolTable of childSymbolTables) {
            symbolTable.parentSymbolTables.add(this);
        }
    }
    /**
     * return itself for the graph interface
     */
    get value() {
        return this;
    }
    /**
     * get all adjacent nodes to this symbol table
     */
    get adjacentNodes() {
        return Array.from(this.childSymbolTables);
    }
    /**
     * add a new child symbol table to this symbol table
     * @param symbolTable the new child
     */
    addScope(symbolTable) {
        this.childSymbolTables.add(symbolTable);
        symbolTable.parentSymbolTables.add(this);
    }
    /**
     * should be called when the associated file is closed. It will remove it's reference
     * if another symbol table doesn't reference it
     */
    closeSelf() {
        // remove childern if no parents
        if (this.parentSymbolTables.size === 0) {
            // remove references from child scopes
            for (const child of this.childSymbolTables) {
                child.parentSymbolTables.delete(this);
            }
            // clear own references
            this.childSymbolTables.clear();
        }
    }
    /**
     * This should be called when the associated symbol table is to be deleted
     */
    removeSelf() {
        // remove refernces from parent scopes
        for (const parent of this.parentSymbolTables) {
            parent.childSymbolTables.delete(this);
        }
        // remove references from child scopes
        for (const child of this.childSymbolTables) {
            child.parentSymbolTables.delete(this);
        }
        // clear own references
        this.childSymbolTables.clear();
        this.parentSymbolTables.clear();
    }
    /**
     * get every symbol in the file
     */
    fileSymbols() {
        return Array.from(this.rootScope.environment.symbols()).concat(this.fileSymbolsDepth(this.rootScope.children));
    }
    /**
     * Get global symbols that match the requested symbol name
     * @param name symbol name
     */
    globalTrackers(name) {
        return Array.from(this.rootScope.environment.values()).filter(tracker => tracker.declared.symbol.name.lookup === name);
    }
    /**
     * Get the tracker for a function symbol
     * @param pos position of symbol
     * @param name name of the symbol
     */
    scopedFunctionTracker(pos, name) {
        const tracker = this.scopedNamedTracker(pos, name, tracker => entityHelpers_1.isKsFunction(tracker.declared.symbol));
        if (!typeGuards_1.empty(tracker)) {
            return tracker;
        }
        return undefined;
    }
    /**
     * Get the tracker for a lock symbol
     * @param pos position of symbol
     * @param name name of the symbol
     */
    scopedLockTracker(pos, name) {
        const tracker = this.scopedNamedTracker(pos, name, trackers => entityHelpers_1.isKsLock(trackers.declared.symbol));
        if (!typeGuards_1.empty(tracker)) {
            return tracker;
        }
        return undefined;
    }
    /**
     * Get the tracker for a variable symbol
     * @param pos position of the symbol
     * @param name name of the symbol
     */
    scopedVariableTracker(pos, name) {
        const tracker = this.scopedNamedTracker(pos, name, trackers => entityHelpers_1.isVariable(trackers.declared.symbol));
        if (!typeGuards_1.empty(tracker)) {
            return tracker;
        }
        return undefined;
    }
    /**
     * Get the tracker for a parameter symbol
     * @param pos position of the symbol
     * @param name name of the symbol
     */
    scopedParameterTracker(pos, name) {
        const tracker = this.scopedNamedTracker(pos, name, tracker => entityHelpers_1.isParameter(tracker.declared.symbol));
        if (!typeGuards_1.empty(tracker)) {
            return tracker;
        }
        return undefined;
    }
    /**
     * Get all symbols available at the provide position
     * @param pos position to check
     */
    scopedSymbols(pos) {
        return this.scopedTrackers(pos).map(tracker => tracker.declared.symbol);
    }
    /**
     * Get a tracker for a given symbol under a filter condition and position
     * @param pos position to check
     * @param name name of the symbol
     * @param symbolFilter filter function for the symbol
     */
    scopedNamedTracker(pos, name, symbolFilter) {
        // our base filter is to match the symbol lexeme
        const baseFilter = (trackers) => trackers.declared.symbol.name.lookup === name;
        // our compound filter checkes for any and the other requested filtering operations
        const compoundFilter = typeGuards_1.empty(symbolFilter)
            ? baseFilter
            : (trackers) => baseFilter(trackers) && symbolFilter(trackers);
        return this.scopedTracker(pos, compoundFilter);
    }
    /**
     * recursively move up the scope to get every file symbol
     * @param nodes nodes to retrive symbols from
     */
    fileSymbolsDepth(nodes) {
        const symbols = [];
        for (const node of nodes) {
            symbols.push(...node.environment.symbols(), ...this.fileSymbolsDepth(node.children));
        }
        return symbols;
    }
    /**
     * get all symbol trackers in scope at a positition
     * @param pos the position to retrive symbols from
     */
    scopedTrackers(pos) {
        const scoped = this.scopedTrackersDepth(pos, this.rootScope.children);
        const fileGlobal = Array.from(this.rootScope.environment.values());
        const importedGlobals = Array.from(this.childSymbolTables.values()).map(scope => Array.from(scope.rootScope.environment.values()));
        return scoped.concat(fileGlobal, ...importedGlobals);
    }
    /**
     * get a symbol tracker under the conditions of a provided filter
     * @param pos the posiition to check for symbols
     * @param trackerFilter the tracker filter
     */
    scopedTracker(pos, trackerFilter = _ => true) {
        const scoped = this.scopedTrackerDepth(pos, this.rootScope.children, trackerFilter);
        if (!typeGuards_1.empty(scoped)) {
            return scoped;
        }
        const fileGlobal = Array.from(this.rootScope.environment.values());
        const importedGlobals = Array.from(this.childSymbolTables.values()).map(scope => Array.from(scope.rootScope.environment.values()));
        const [match] = fileGlobal.concat(...importedGlobals).filter(trackerFilter);
        return match;
    }
    /**
     * recursively move up scopes to find most relevant symbols
     * @param pos position to search
     * @param nodes scope nodes to check
     */
    scopedTrackersDepth(pos, nodes) {
        for (const node of nodes) {
            const { position } = node;
            switch (position.kind) {
                // if global it is available
                case types_1.ScopeKind.global:
                    return this.scopedTrackersDepth(pos, node.children).concat(Array.from(node.environment.values()));
                // if the scope has a real position check if we're in the bounds
                case types_1.ScopeKind.local:
                    if (positionUtils_1.rangeContainsPos(position, pos)) {
                        return this.scopedTrackersDepth(pos, node.children).concat(Array.from(node.environment.values()));
                    }
                    break;
            }
        }
        return [];
    }
    /**
     * recursively move up scope for more relevant symbols
     * @param pos position to search
     * @param nodes scope nodes to check
     * @param trackerFilter filter for acceptable symbols
     */
    scopedTrackerDepth(pos, nodes, trackerFilter) {
        let childSymbol = undefined;
        for (const node of nodes) {
            const { position } = node;
            switch (position.kind) {
                // if global it is available
                case types_1.ScopeKind.global:
                    childSymbol = this.scopedTrackerDepth(pos, node.children, trackerFilter);
                    if (!typeGuards_1.empty(childSymbol)) {
                        return childSymbol;
                    }
                    const currentSymbols = Array.from(node.environment.values()).filter(trackerFilter);
                    if (currentSymbols.length === 1) {
                        return currentSymbols[0];
                    }
                    break;
                // if the scope has a real position check if we're in the bounds
                case types_1.ScopeKind.local:
                    if (positionUtils_1.rangeContainsPos(position, pos)) {
                        childSymbol = this.scopedTrackerDepth(pos, node.children, trackerFilter);
                        if (!typeGuards_1.empty(childSymbol)) {
                            return childSymbol;
                        }
                        const currentSymbols = Array.from(node.environment.values()).filter(trackerFilter);
                        if (currentSymbols.length === 1) {
                            return currentSymbols[0];
                        }
                    }
                    break;
            }
        }
        return undefined;
    }
}
exports.SymbolTable = SymbolTable;
//# sourceMappingURL=symbolTable.js.map