import { ScopeKind } from '../parser/types';
import { KsVariable } from '../entities/variable';
import { EnvironmentNode, SymbolTrackerBase, KsSymbolKind, EnvironmentPath } from './types';
import { KsFunction } from '../entities/function';
import { KsLock } from '../entities/lock';
import { Range, Diagnostic } from 'vscode-languageserver';
import { BasicTracker } from './tracker';
import { ArgumentType, IFunctionType } from '../typeChecker/types/types';
import { SymbolTable } from './symbolTable';
import { Token } from '../entities/token';
import { KsParameter } from '../entities/parameter';
/**
 * The Symbol table builder is used to declare new symbols and track new symbols
 * in the kerboscript program
 */
export declare class SymbolTableBuilder {
    readonly uri: string;
    /**
     * The global environment for this file
     */
    private readonly global;
    /**
     * The root environment include the global environment and it's children
     */
    private readonly rootNode;
    /**
     * The active environment node
     */
    private activeNode;
    /**
     * The path in the environment
     */
    private path;
    /**
     * A set of child symbols tables relative to this symbols table builder
     */
    childSymbolTables: Set<SymbolTable>;
    /**
     * logger class to reporting errors and infomation
     */
    logger: ILogger;
    /**
     * construct a symbol table builder
     * @param uri the file uri
     * @param logger a logger object
     */
    constructor(uri: string, logger?: ILogger);
    /**
     * Generate a symbol table
     */
    build(): SymbolTable;
    /**
     * Rewind the current scope path for another pass
     */
    rewind(): void;
    /**
     * Get the current scope path TODO consider recomputing active scope
     * when set
     */
    getPath(): {
        path: EnvironmentPath;
        activeScope: EnvironmentNode;
    };
    /**
     * Set the current scope path
     * @param path current path
     * @param activeScope current scope
     */
    setPath(path: EnvironmentPath, activeScope: EnvironmentNode): void;
    /**
     * Add a child symbol table to this symbol table builder
     * @param symbolTable the child symbol table
     */
    linkTable(symbolTable: SymbolTable): void;
    /**
     * Push a new scope onto the scope stack
     * @param range the range of the new scope
     */
    beginScope(range: Range): void;
    /**
     * Pop the current scope off the stack
     */
    endScope(): void;
    /**
     * Indicate if the current scope if file scope
     */
    isFileScope(): boolean;
    /**
     * Indicate if the current scope if global scope
     */
    isGlobalScope(): boolean;
    /**
     * Returns the current scope depth
     */
    scopeDepth(): number;
    /**
     * Find all unused symbols in this symbol table builder
     */
    findUnused(): Diagnostic[];
    /**
     * Use a symbol in the symbol table
     * @param name token for the current symbol
     */
    useSymbol(name: Token): Maybe<Diagnostic>;
    /**
     * Use a variable symbol
     * @param name token for the current variable
     */
    useVariable(name: Token): Maybe<Diagnostic>;
    /**
     * Use a function symbol
     * @param name token for the current function
     */
    useFunction(name: Token): Maybe<Diagnostic>;
    /**
     * Use a lock symbol
     * @param name token for the current lock
     */
    useLock(name: Token): Maybe<Diagnostic>;
    /**
     * Use a parameter symbol
     * @param name token for the current parameter
     */
    useParameter(name: Token): Maybe<Diagnostic>;
    /**
     * Set a variable symbol
     * @param token token for the variable to set
     */
    setBinding(token: Token): Maybe<Diagnostic>;
    /**
     * Add a new variable symbol to the table
     * @param scopeType the requested scope type
     * @param token token for the requested variable
     * @param type type to declare variable as
     */
    declareVariable(scopeType: ScopeKind, token: Token, type?: ArgumentType): Maybe<Diagnostic>;
    /**
     * Add a new function symbol to the table
     * @param scopeType the requested scope type
     * @param token token for the requested function
     * @param requiredParameters required parameters for the function
     * @param optionalParameters optional parameters for the function
     * @param returnValue does the function have a return type
     * @param type type to declare function as
     */
    declareFunction(scopeType: ScopeKind, token: Token, requiredParameters: number, optionalParameters: number, returnValue: boolean, type?: IFunctionType): Maybe<Diagnostic>;
    /**
     * Add a new lock symbol to the table
     * @param scopeType the requested scope type
     * @param token token for the requested lock
     * @param type type to declare lock as
     */
    declareLock(scopeType: ScopeKind, token: Token, type?: ArgumentType): Maybe<Diagnostic>;
    /**
     * Add a new parameter symbol to the table
     * @param scopeType the requested scope type
     * @param token token for the requested parameter
     * @param defaulted is the parameter defaulted
     */
    declareParameter(scopeType: ScopeKind, token: Token, defaulted: boolean): Maybe<Diagnostic>;
    /**
     * check if the symbol exist and add a usage if it does
     * @param token token for the requested symbol
     * @param tracker symbol tracker
     * @param symbolType symbol type
     * @param expr expression context
     */
    checkUseSymbol(token: Token, tracker: Maybe<BasicTracker>, symbolType: KsSymbolKind): Maybe<Diagnostic>;
    /**
     * Lookup a variable
     * @param token token for the requested variable
     * @param scope requested scope lookup
     */
    lookupVariable(token: Token, scope: ScopeKind): Maybe<KsVariable>;
    /**
     * lookup a function
     * @param token token for the requested function
     * @param scope requested scope lookup
     */
    lookupFunction(token: Token, scope: ScopeKind): Maybe<KsFunction>;
    /**
     * lookup a lock
     * @param token token for the requested lock
     * @param scope requested scope lookup
     */
    lookupLock(token: Token, scope: ScopeKind): Maybe<KsLock>;
    /**
     * lookup a parameter
     * @param token token for the requested parameter
     * @param scope requested scope lookup
     */
    lookupParameter(token: Token, scope: ScopeKind): Maybe<KsParameter>;
    /**
     * lookup a parameter
     * @param token token for the requested parameter
     * @param scope requested scope lookup
     */
    lookupBinding(token: Token, scope: ScopeKind): Maybe<KsParameter | KsVariable>;
    /**
     * lookup a variable tracker
     * @param token token for the requested variable
     * @param scopeKind requested scope lookup
     */
    lookupVariableTracker(token: Token, scopeKind: ScopeKind): Maybe<BasicTracker<KsVariable>>;
    /**
     * lookup a function tracker
     * @param token token for the requested function
     * @param scopeKind requested scope lookup
     */
    lookupFunctionTracker(token: Token, scopeKind: ScopeKind): Maybe<BasicTracker<KsFunction>>;
    /**
     * lookup a lock tracker
     * @param token token for the requested lock
     * @param scopeKind requested scope lookup
     */
    lookupLockTracker(token: Token, scopeKind: ScopeKind): Maybe<BasicTracker<KsLock>>;
    /**
     * lookup a parameter tracker
     * @param token token for the requested parameter
     * @param scopeKind requested scope lookup
     */
    lookupParameterTracker(token: Token, scopeKind: ScopeKind): Maybe<BasicTracker<KsParameter>>;
    /**
     * lookup a parameter tracker
     * @param token token for the requested parameter
     * @param scopeKind requested scope lookup
     */
    lookupBindingTracker(token: Token, scopeKind: ScopeKind): Maybe<SymbolTrackerBase<KsParameter | KsVariable>>;
    /**
     * Find all unused symbol in a scope node that aren't used and
     * it's children
     * @param node scope node
     * @param errors cumulative errors
     */
    private findScopeNodeUnused;
    /**
     * Find all unused symbols in an environment that aren't used
     * @param environment environment to check
     * @param errors cumulative errors
     */
    private findScopeUnused;
    /**
     * Determine if a symbol tracker is in the current path
     * @param lookup token lookup string
     * @param scopeKind scope kind for the lookup local or global
     */
    private lookup;
    /**
     * Determine if a scope in the current path contains a symbol of interest
     * @param lookup token lookup string
     * @param scopeKind scope kind for the lookup local or global
     */
    private lookupScopeNode;
    /**
     * Retrieve the appropriate scope
     * @param type scope type global or local
     */
    private selectScopeNode;
    /**
     * Retrieve the active environment
     */
    private peekScope;
    /**
     * generate a resolver error when a declare symbol shadows with an existing on
     * @param name token for the requested symbol
     * @param symbol collided symbol
     */
    private shadowSymbolHint;
    /**
     * generate a resolver error when a declare symbol collides with an existing on
     * @param name token for the requested symbol
     * @param symbol collided symbol
     */
    private localConflictError;
}
