"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const typeGuards_1 = require("../utilities/typeGuards");
const types_1 = require("../parser/types");
const variable_1 = require("../entities/variable");
const types_2 = require("./types");
const function_1 = require("../entities/function");
const lock_1 = require("../entities/lock");
const vscode_languageserver_1 = require("vscode-languageserver");
const scopePosition_1 = require("./scopePosition");
const logger_1 = require("../utilities/logger");
const scope_1 = require("./scope");
const tracker_1 = require("./tracker");
const symbolTable_1 = require("./symbolTable");
const entityHelpers_1 = require("../entities/entityHelpers");
const positionUtils_1 = require("../utilities/positionUtils");
const diagnosticsUtils_1 = require("../utilities/diagnosticsUtils");
const constants_1 = require("../utilities/constants");
const stringUtils_1 = require("../utilities/stringUtils");
const clean_1 = require("../utilities/clean");
const parameter_1 = require("../entities/parameter");
/**
 * The Symbol table builder is used to declare new symbols and track new symbols
 * in the kerboscript program
 */
class SymbolTableBuilder {
    /**
     * construct a symbol table builder
     * @param uri the file uri
     * @param logger a logger object
     */
    constructor(uri, logger = logger_1.mockLogger) {
        this.uri = uri;
        this.logger = logger;
        this.global = new scope_1.Environment();
        this.rootNode = {
            parent: undefined,
            environment: this.global,
            children: [],
            position: { kind: types_1.ScopeKind.global },
        };
        this.activeNode = this.rootNode;
        this.path = {
            active: [],
            backTrack: [],
        };
        this.childSymbolTables = new Set();
    }
    /**
     * Generate a symbol table
     */
    build() {
        return new symbolTable_1.SymbolTable(this.rootNode, this.childSymbolTables, this.uri, this.logger);
    }
    /**
     * Rewind the current scope path for another pass
     */
    rewind() {
        this.path = {
            active: [],
            backTrack: [],
        };
        this.activeNode = this.rootNode;
    }
    /**
     * Get the current scope path TODO consider recomputing active scope
     * when set
     */
    getPath() {
        return {
            activeScope: this.activeNode,
            path: {
                active: [...this.path.active],
                backTrack: [...this.path.backTrack],
            },
        };
    }
    /**
     * Set the current scope path
     * @param path current path
     * @param activeScope current scope
     */
    setPath(path, activeScope) {
        this.path = path;
        this.activeNode = activeScope;
    }
    /**
     * Add a child symbol table to this symbol table builder
     * @param symbolTable the child symbol table
     */
    linkTable(symbolTable) {
        this.childSymbolTables.add(symbolTable);
    }
    /**
     * Push a new scope onto the scope stack
     * @param range the range of the new scope
     */
    beginScope(range) {
        // determine next scope node
        const depth = this.path.active.length - 1;
        const next = !typeGuards_1.empty(this.path.backTrack[depth + 1])
            ? this.path.backTrack[depth + 1] + 1
            : 0;
        const currentScope = this.activeNode;
        let activeScope;
        // generate a new scope node if it doesn't exist
        if (typeGuards_1.empty(currentScope.children[next])) {
            activeScope = {
                parent: currentScope,
                environment: new scope_1.Environment(),
                position: new scopePosition_1.ScopePosition(range.start, range.end),
                children: [],
            };
            currentScope.children.push(activeScope);
            // or get the existing one
        }
        else {
            activeScope = currentScope.children[next];
        }
        this.activeNode = activeScope;
        this.logger.verbose(`begin scope at ${positionUtils_1.positionToString(range.start)}`);
        this.path.active.push(next);
        this.path.backTrack = [...this.path.active];
    }
    /**
     * Pop the current scope off the stack
     */
    endScope() {
        const currentScope = this.activeNode;
        if (typeGuards_1.empty(currentScope.parent)) {
            throw new Error('Attempted to pop the global scope');
        }
        this.activeNode = currentScope.parent;
        this.path.active.pop();
        if (currentScope.position.kind === types_1.ScopeKind.local) {
            this.logger.verbose(`end scope at ${positionUtils_1.positionToString(currentScope.position.end)}`);
        }
    }
    /**
     * Indicate if the current scope if file scope
     */
    isFileScope() {
        return this.scopeDepth() === 1;
    }
    /**
     * Indicate if the current scope if global scope
     */
    isGlobalScope() {
        return this.activeNode === this.rootNode;
    }
    /**
     * Returns the current scope depth
     */
    scopeDepth() {
        return this.path.active.length;
    }
    /**
     * Find all unused symbols in this symbol table builder
     */
    findUnused() {
        const errors = [];
        this.findScopeNodeUnused(this.rootNode, errors);
        return errors;
    }
    /**
     * Use a symbol in the symbol table
     * @param name token for the current symbol
     */
    useSymbol(name) {
        const tracker = this.lookup(name.lookup, types_1.ScopeKind.global);
        // check if symbols exists
        if (typeGuards_1.empty(tracker)) {
            return diagnosticsUtils_1.createDiagnostic(name, `Symbol ${name.lexeme} may not exist`, vscode_languageserver_1.DiagnosticSeverity.Warning);
        }
        return this.checkUseSymbol(name, tracker, tracker.declared.symbol.tag);
    }
    /**
     * Use a variable symbol
     * @param name token for the current variable
     */
    useVariable(name) {
        const variable = this.lookupVariableTracker(name, types_1.ScopeKind.global);
        return this.checkUseSymbol(name, variable, types_2.KsSymbolKind.variable);
    }
    /**
     * Use a function symbol
     * @param name token for the current function
     */
    useFunction(name) {
        const func = this.lookupFunctionTracker(name, types_1.ScopeKind.global);
        return this.checkUseSymbol(name, func, types_2.KsSymbolKind.function);
    }
    /**
     * Use a lock symbol
     * @param name token for the current lock
     */
    useLock(name) {
        const lock = this.lookupLockTracker(name, types_1.ScopeKind.global);
        return this.checkUseSymbol(name, lock, types_2.KsSymbolKind.lock);
    }
    /**
     * Use a parameter symbol
     * @param name token for the current parameter
     */
    useParameter(name) {
        const parameter = this.lookupParameterTracker(name, types_1.ScopeKind.global);
        return this.checkUseSymbol(name, parameter, types_2.KsSymbolKind.parameter);
    }
    /**
     * Set a variable symbol
     * @param token token for the variable to set
     */
    setBinding(token) {
        const tracker = this.lookup(token.lookup, types_1.ScopeKind.global);
        // check if variable has already been defined
        if (typeGuards_1.empty(tracker)) {
            return diagnosticsUtils_1.createDiagnostic(token, `${token.lexeme} may not exist`, vscode_languageserver_1.DiagnosticSeverity.Warning);
        }
        token.tracker = tracker;
        tracker.sets.push(tracker_1.createSymbolSet(token));
        this.logger.verbose(`set variable ${token.lexeme} at ${positionUtils_1.rangeToString(token)}`);
        return undefined;
    }
    /**
     * Add a new variable symbol to the table
     * @param scopeType the requested scope type
     * @param token token for the requested variable
     * @param type type to declare variable as
     */
    declareVariable(scopeType, token, type) {
        const conflictTracker = this.lookup(token.lookup, scopeType);
        // check if variable has already been defined
        if (!typeGuards_1.empty(conflictTracker)) {
            return this.localConflictError(token, conflictTracker.declared.symbol);
        }
        let diagnostic;
        // if global we can't shadow
        if (scopeType === types_1.ScopeKind.global) {
            diagnostic = undefined;
            // if local check for shadowing hints
        }
        else {
            const shadowTracker = this.lookup(token.lookup, types_1.ScopeKind.global);
            diagnostic = typeGuards_1.empty(shadowTracker)
                ? undefined
                : this.shadowSymbolHint(token, types_2.KsSymbolKind.variable, shadowTracker.declared.symbol);
        }
        const scopeNode = this.selectScopeNode(scopeType);
        const tracker = new tracker_1.BasicTracker(new variable_1.KsVariable(scopeType, token), this.uri, type);
        token.tracker = tracker;
        scopeNode.environment.set(token.lookup, tracker);
        this.logger.verbose(`declare variable ${token.lexeme} at ${positionUtils_1.rangeToString(token)}`);
        return diagnostic;
    }
    /**
     * Add a new function symbol to the table
     * @param scopeType the requested scope type
     * @param token token for the requested function
     * @param requiredParameters required parameters for the function
     * @param optionalParameters optional parameters for the function
     * @param returnValue does the function have a return type
     * @param type type to declare function as
     */
    declareFunction(scopeType, token, requiredParameters, optionalParameters, returnValue, type) {
        const conflictTracker = this.lookup(token.lookup, scopeType);
        // check if variable has already been defined
        if (!typeGuards_1.empty(conflictTracker)) {
            return this.localConflictError(token, conflictTracker.declared.symbol);
        }
        let diagnostic;
        // if global we can't shadow
        if (scopeType === types_1.ScopeKind.global) {
            diagnostic = undefined;
            // if local check for shadowing hints
        }
        else {
            const shadowTracker = this.lookup(token.lookup, types_1.ScopeKind.global);
            diagnostic = typeGuards_1.empty(shadowTracker)
                ? undefined
                : this.shadowSymbolHint(token, types_2.KsSymbolKind.function, shadowTracker.declared.symbol);
        }
        const scopeNode = this.selectScopeNode(scopeType);
        const tracker = new tracker_1.BasicTracker(new function_1.KsFunction(scopeType, token, requiredParameters, optionalParameters, returnValue), this.uri, type);
        token.tracker = tracker;
        scopeNode.environment.set(token.lookup, tracker);
        this.logger.verbose(`declare function ${token.lexeme} at ${positionUtils_1.rangeToString(token)}`);
        return diagnostic;
    }
    /**
     * Add a new lock symbol to the table
     * @param scopeType the requested scope type
     * @param token token for the requested lock
     * @param type type to declare lock as
     */
    declareLock(scopeType, token, type) {
        const conflictTracker = this.lookup(token.lookup, scopeType);
        // check if variable has already been defined
        if (!typeGuards_1.empty(conflictTracker)) {
            return this.localConflictError(token, conflictTracker.declared.symbol);
        }
        let diagnostic;
        // if global we can't shadow
        if (scopeType === types_1.ScopeKind.global) {
            diagnostic = undefined;
            // if local check for shadowing hints
        }
        else {
            const shadowTracker = this.lookup(token.lookup, types_1.ScopeKind.global);
            diagnostic = typeGuards_1.empty(shadowTracker)
                ? undefined
                : this.shadowSymbolHint(token, types_2.KsSymbolKind.lock, shadowTracker.declared.symbol);
        }
        const scopeNode = this.selectScopeNode(scopeType);
        const tracker = new tracker_1.BasicTracker(new lock_1.KsLock(scopeType, token), this.uri, type);
        token.tracker = tracker;
        scopeNode.environment.set(token.lookup, tracker);
        this.logger.verbose(`declare lock ${token.lexeme} at ${positionUtils_1.rangeToString(token)}`);
        return diagnostic;
    }
    /**
     * Add a new parameter symbol to the table
     * @param scopeType the requested scope type
     * @param token token for the requested parameter
     * @param defaulted is the parameter defaulted
     */
    declareParameter(scopeType, token, defaulted) {
        const conflictTracker = this.lookup(token.lookup, scopeType);
        // check if variable has already been defined
        if (!typeGuards_1.empty(conflictTracker)) {
            return this.localConflictError(token, conflictTracker.declared.symbol);
        }
        let diagnostic;
        // if global we can't shadow
        if (scopeType === types_1.ScopeKind.global) {
            diagnostic = undefined;
            // if local check for shadowing hints
        }
        else {
            const shadowTracker = this.lookup(token.lookup, types_1.ScopeKind.global);
            diagnostic = typeGuards_1.empty(shadowTracker)
                ? undefined
                : this.shadowSymbolHint(token, types_2.KsSymbolKind.parameter, shadowTracker.declared.symbol);
        }
        const scopeNode = this.selectScopeNode(scopeType);
        const tracker = new tracker_1.BasicTracker(new parameter_1.KsParameter(token, defaulted), this.uri);
        token.tracker = tracker;
        scopeNode.environment.set(token.lookup, tracker);
        this.logger.verbose(`declare parameter ${token.lexeme} at ${positionUtils_1.rangeToString(token)}`);
        return diagnostic;
    }
    /**
     * check if the symbol exist and add a usage if it does
     * @param token token for the requested symbol
     * @param tracker symbol tracker
     * @param symbolType symbol type
     * @param expr expression context
     */
    checkUseSymbol(token, tracker, symbolType) {
        // check that variable has already been defined
        if (typeGuards_1.empty(tracker)) {
            return diagnosticsUtils_1.createDiagnostic(token, `${symbolType} ${token.lexeme} may not exist.`, vscode_languageserver_1.DiagnosticSeverity.Warning);
        }
        // indicate usage
        token.tracker = tracker;
        tracker.usages.push(clean_1.cleanLocation(token));
        this.logger.verbose(`Use ${symbolType} ${token.lexeme} at ${positionUtils_1.rangeToString(token)}`);
        // check if a variable may not be defined in a runtime situtation
        // functions will always be availbe because they are executed
        // during a pre pass
        if (positionUtils_1.rangeBefore(token, tracker.declared.range.start) &&
            tracker.declared.symbol.tag !== types_2.KsSymbolKind.function) {
            return diagnosticsUtils_1.createDiagnostic(token, `${symbolType} ${token.lexeme} may not exist at script runtime.`, vscode_languageserver_1.DiagnosticSeverity.Hint, undefined, [
                vscode_languageserver_1.DiagnosticRelatedInformation.create({ uri: tracker.declared.uri, range: tracker.declared.range }, `${token.lexeme} is declared after this use`),
            ]);
        }
        return undefined;
    }
    /**
     * Lookup a variable
     * @param token token for the requested variable
     * @param scope requested scope lookup
     */
    lookupVariable(token, scope) {
        const tracker = this.lookupVariableTracker(token, scope);
        return tracker && tracker.declared.symbol;
    }
    /**
     * lookup a function
     * @param token token for the requested function
     * @param scope requested scope lookup
     */
    lookupFunction(token, scope) {
        const tracker = this.lookupFunctionTracker(token, scope);
        return tracker && tracker.declared.symbol;
    }
    /**
     * lookup a lock
     * @param token token for the requested lock
     * @param scope requested scope lookup
     */
    lookupLock(token, scope) {
        const tracker = this.lookupLockTracker(token, scope);
        return tracker && tracker.declared.symbol;
    }
    /**
     * lookup a parameter
     * @param token token for the requested parameter
     * @param scope requested scope lookup
     */
    lookupParameter(token, scope) {
        const tracker = this.lookupParameterTracker(token, scope);
        return tracker && tracker.declared.symbol;
    }
    /**
     * lookup a parameter
     * @param token token for the requested parameter
     * @param scope requested scope lookup
     */
    lookupBinding(token, scope) {
        const tracker = this.lookupBindingTracker(token, scope);
        return tracker && tracker.declared.symbol;
    }
    /**
     * lookup a variable tracker
     * @param token token for the requested variable
     * @param scopeKind requested scope lookup
     */
    lookupVariableTracker(token, scopeKind) {
        const tracker = this.lookup(token.lookup, scopeKind);
        return !typeGuards_1.empty(tracker) && entityHelpers_1.isVariable(tracker.declared.symbol)
            ? tracker
            : undefined;
    }
    /**
     * lookup a function tracker
     * @param token token for the requested function
     * @param scopeKind requested scope lookup
     */
    lookupFunctionTracker(token, scopeKind) {
        const tracker = this.lookup(token.lookup, scopeKind);
        return !typeGuards_1.empty(tracker) && entityHelpers_1.isVariable(tracker.declared.symbol)
            ? tracker
            : undefined;
    }
    /**
     * lookup a lock tracker
     * @param token token for the requested lock
     * @param scopeKind requested scope lookup
     */
    lookupLockTracker(token, scopeKind) {
        const tracker = this.lookup(token.lookup, scopeKind);
        return !typeGuards_1.empty(tracker) && entityHelpers_1.isKsLock(tracker.declared.symbol)
            ? tracker
            : undefined;
    }
    /**
     * lookup a parameter tracker
     * @param token token for the requested parameter
     * @param scopeKind requested scope lookup
     */
    lookupParameterTracker(token, scopeKind) {
        const tracker = this.lookup(token.lookup, scopeKind);
        return !typeGuards_1.empty(tracker) && entityHelpers_1.isParameter(tracker.declared.symbol)
            ? tracker
            : undefined;
    }
    /**
     * lookup a parameter tracker
     * @param token token for the requested parameter
     * @param scopeKind requested scope lookup
     */
    lookupBindingTracker(token, scopeKind) {
        const tracker = this.lookup(token.lookup, scopeKind);
        return !typeGuards_1.empty(tracker) &&
            (entityHelpers_1.isParameter(tracker.declared.symbol) ||
                entityHelpers_1.isVariable(tracker.declared.symbol))
            ? tracker
            : undefined;
    }
    /**
     * Find all unused symbol in a scope node that aren't used and
     * it's children
     * @param node scope node
     * @param errors cumulative errors
     */
    findScopeNodeUnused(node, errors) {
        this.findScopeUnused(node.environment, errors);
        for (const childNode of node.children) {
            this.findScopeNodeUnused(childNode, errors);
        }
    }
    /**
     * Find all unused symbols in an environment that aren't used
     * @param environment environment to check
     * @param errors cumulative errors
     */
    findScopeUnused(environment, errors) {
        for (const tracker of environment.values()) {
            switch (tracker.declared.symbol.tag) {
                case types_2.KsSymbolKind.function:
                    break;
                case types_2.KsSymbolKind.parameter:
                    if (tracker.usages.length === 0) {
                        errors.push(diagnosticsUtils_1.createDiagnostic(tracker.declared.symbol.name, `Parameter ${tracker.declared.symbol.name.lexeme} was not used.`, vscode_languageserver_1.DiagnosticSeverity.Warning));
                    }
                    break;
                case types_2.KsSymbolKind.lock:
                    if (!tracker.declared.symbol.cooked && tracker.usages.length === 0) {
                        errors.push(diagnosticsUtils_1.createDiagnostic(tracker.declared.symbol.name, `Lock ${tracker.declared.symbol.name.lexeme} was not used.`, vscode_languageserver_1.DiagnosticSeverity.Warning));
                    }
                    break;
                case types_2.KsSymbolKind.variable:
                    if (tracker.usages.length === 0) {
                        errors.push(diagnosticsUtils_1.createDiagnostic(tracker.declared.symbol.name, `Variable ${tracker.declared.symbol.name.lexeme} was not used.`, vscode_languageserver_1.DiagnosticSeverity.Warning));
                    }
                    break;
                default:
                    throw new Error('Unknown symbol found');
            }
        }
    }
    /**
     * Determine if a symbol tracker is in the current path
     * @param lookup token lookup string
     * @param scopeKind scope kind for the lookup local or global
     */
    lookup(lookup, scopeKind) {
        const scopeNode = this.lookupScopeNode(lookup, scopeKind);
        if (typeGuards_1.empty(scopeNode)) {
            return undefined;
        }
        return scopeNode.environment.get(lookup);
    }
    /**
     * Determine if a scope in the current path contains a symbol of interest
     * @param lookup token lookup string
     * @param scopeKind scope kind for the lookup local or global
     */
    lookupScopeNode(lookup, scopeKind) {
        if (scopeKind === types_1.ScopeKind.local) {
            return this.peekScope().has(lookup) ? this.activeNode : undefined;
        }
        let currentScope = this.activeNode;
        while (!typeGuards_1.empty(currentScope)) {
            if (currentScope.environment.has(lookup)) {
                return currentScope;
            }
            currentScope = currentScope.parent;
        }
        // check child scopes symbol is in another file
        for (const child of this.childSymbolTables) {
            if (child.rootScope.environment.has(lookup)) {
                return child.rootScope;
            }
        }
        return undefined;
    }
    /**
     * Retrieve the appropriate scope
     * @param type scope type global or local
     */
    selectScopeNode(type) {
        return type === types_1.ScopeKind.global ? this.rootNode : this.activeNode;
    }
    /**
     * Retrieve the active environment
     */
    peekScope() {
        return this.activeNode.environment;
    }
    /**
     * generate a resolver error when a declare symbol shadows with an existing on
     * @param name token for the requested symbol
     * @param symbol collided symbol
     */
    shadowSymbolHint(name, kind, symbol) {
        return diagnosticsUtils_1.createDiagnostic(name, `${stringUtils_1.toCase(3 /* pascalcase */, types_2.KsSymbolKind[symbol.tag])} ${symbol.name.lexeme} ` + `already exists here. This ${types_2.KsSymbolKind[kind]} shadows it.`, vscode_languageserver_1.DiagnosticSeverity.Warning, undefined, [
            vscode_languageserver_1.DiagnosticRelatedInformation.create({ uri: this.uri, range: symbol.name }, symbol.name.uri === constants_1.builtIn
                ? `${symbol.name.lexeme} is a built in ${types_2.KsSymbolKind[symbol.tag]}`
                : 'Orignally declared here'),
        ]);
    }
    /**
     * generate a resolver error when a declare symbol collides with an existing on
     * @param name token for the requested symbol
     * @param symbol collided symbol
     */
    localConflictError(name, symbol) {
        return diagnosticsUtils_1.createDiagnostic(name, `${stringUtils_1.toCase(3 /* pascalcase */, types_2.KsSymbolKind[symbol.tag])} ${symbol.name.lexeme} already exists.`, vscode_languageserver_1.DiagnosticSeverity.Warning, undefined, [
            vscode_languageserver_1.DiagnosticRelatedInformation.create({ uri: this.uri, range: symbol.name }, symbol.name.uri === constants_1.builtIn
                ? `${symbol.name.lexeme} is a built in ${types_2.KsSymbolKind[symbol.tag]}`
                : 'Orignally declared here'),
        ]);
    }
}
exports.SymbolTableBuilder = SymbolTableBuilder;
//# sourceMappingURL=symbolTableBuilder.js.map