/// <reference types="node" />
import { Position, Location, Range } from 'vscode-languageserver';
import { IDocumentInfo, IDiagnosticUri } from './types';
import { PerformanceObserver } from 'perf_hooks';
import { KsSymbol, SymbolTracker } from './analysis/types';
import { PathResolver } from './utilities/pathResolver';
import { Token } from './entities/token';
export declare class Analyzer {
    workspaceFolder?: string;
    private standardLibrary;
    private bodyLibrary;
    readonly pathResolver: PathResolver;
    readonly documentInfos: Map<string, IDocumentInfo>;
    readonly logger: ILogger;
    readonly tracer: ITracer;
    readonly observer: PerformanceObserver;
    constructor(caseKind?: CaseKind, logger?: ILogger, tracer?: ITracer);
    /**
     * Set the volume 0 path for the analyzer
     * @param path path of volume 0
     */
    setPath(path: string): void;
    /**
     * Set the volume 0 uri for the analyzer
     * @param uri uri to volume 0
     */
    setUri(uri: string): void;
    /**
     * Set the case of the body library and standard library
     * @param caseKind case to set
     */
    setCase(caseKind: CaseKind): void;
    /**
     * Validate a document in asynchronous stages. This produces diagnostics about known errors or
     * potential problems in the provided script
     * @param uri uri of the document
     * @param text source text of the document
     * @param depth TODO remove: current depth of the document
     */
    validateDocument(uri: string, text: string, depth?: number): AsyncIterableIterator<IDiagnosticUri[]>;
    /**
     * Main validation function for a document. Lexically and semantically understands a document.
     * Will additionally perform the same analysis on other run scripts found in this script
     * @param uri uri of the document
     * @param text source text of the document
     * @param depth TODO remove: current depth of the document
     */
    private validateDocument_;
    /**
     * Get the symbol table corresponding active set of celetrial bodies for the user.
     * This allows for bodies other than that in stock ksp to be incorporated
     */
    private activeBodyLibrary;
    /**
     * Get the token at the provided position in the text document
     * @param pos position in the text document
     * @param uri uri of the text document
     */
    getToken(pos: Position, uri: string): Maybe<Token>;
    /**
     * Get the declaration location for the token at the provided position
     * @param pos position in the document
     * @param uri uri of the document
     */
    getDeclarationLocation(pos: Position, uri: string): Maybe<Location>;
    /**
     * Get all usage locations in all files
     * @param pos position in document
     * @param uri uri of document
     */
    getUsageLocations(pos: Position, uri: string): Maybe<Location[]>;
    /**
     * Get all usage ranges in a provide file
     * @param pos position in document
     * @param uri uri of document
     */
    getFileUsageRanges(pos: Position, uri: string): Maybe<Range[]>;
    /**
     * Get all symbols in scope at a particulare location in the file
     * @param pos position in document
     * @param uri document uri
     */
    getScopedSymbols(pos: Position, uri: string): KsSymbol[];
    /**
     * Get all symbols in a provided file
     * @param uri document uri
     */
    getAllFileSymbols(uri: string): KsSymbol[];
    getFunctionAtPosition(pos: Position, uri: string): Maybe<{
        tracker: SymbolTracker;
        index: number;
    }>;
    /**
     * Generate a ast from the provided source text
     * @param uri uri to document
     * @param text source text of document
     */
    private parseDocument;
    private getValidUri;
    private loadAndValidateDocument;
    /**
     * Try to find a document in the workspace
     * @param path path to file
     * @param uri uri to file
     */
    private tryFindDocument;
}
