"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const vscode_languageserver_1 = require("vscode-languageserver");
const perf_hooks_1 = require("perf_hooks");
const parser_1 = require("./parser/parser");
const preResolver_1 = require("./analysis/preResolver");
const scanner_1 = require("./scanner/scanner");
const resolver_1 = require("./analysis/resolver");
const types_1 = require("./analysis/types");
const logger_1 = require("./utilities/logger");
const typeGuards_1 = require("./utilities/typeGuards");
const scriptFind_1 = require("./parser/scriptFind");
const Expr = __importStar(require("./parser/expr"));
const Stmt = __importStar(require("./parser/stmt"));
const SuffixTerm = __importStar(require("./parser/suffixTerm"));
const pathResolver_1 = require("./utilities/pathResolver");
const fs_1 = require("fs");
const path_1 = require("path");
const fsUtils_1 = require("./utilities/fsUtils");
const standardLibrary_1 = require("./analysis/standardLibrary");
const constants_1 = require("./utilities/constants");
const symbolTableBuilder_1 = require("./analysis/symbolTableBuilder");
const typeChecker_1 = require("./typeChecker/typeChecker");
const positionUtils_1 = require("./utilities/positionUtils");
class Analyzer {
    constructor(caseKind = 2 /* camelcase */, logger = logger_1.mockLogger, tracer = logger_1.mockTracer) {
        this.pathResolver = new pathResolver_1.PathResolver();
        this.logger = logger;
        this.tracer = tracer;
        this.documentInfos = new Map();
        this.workspaceFolder = undefined;
        this.standardLibrary = standardLibrary_1.standardLibraryBuilder(caseKind);
        this.bodyLibrary = standardLibrary_1.bodyLibraryBuilder(caseKind);
        this.observer = new perf_hooks_1.PerformanceObserver(list => {
            this.logger.info('');
            this.logger.info('-------- Performance ---------');
            for (const entry of list.getEntries()) {
                this.logger.info(`${entry.name} took ${entry.duration} ms`);
            }
            this.logger.info('------------------------------');
        });
        this.observer.observe({ entryTypes: ['measure'], buffered: true });
    }
    /**
     * Set the volume 0 path for the analyzer
     * @param path path of volume 0
     */
    setPath(path) {
        this.pathResolver.volume0Path = path;
        this.workspaceFolder = path;
    }
    /**
     * Set the volume 0 uri for the analyzer
     * @param uri uri to volume 0
     */
    setUri(uri) {
        this.pathResolver.volume0Uri = uri;
    }
    /**
     * Set the case of the body library and standard library
     * @param caseKind case to set
     */
    setCase(caseKind) {
        this.standardLibrary = standardLibrary_1.standardLibraryBuilder(caseKind);
        this.bodyLibrary = standardLibrary_1.bodyLibraryBuilder(caseKind);
    }
    /**
     * Validate a document in asynchronous stages. This produces diagnostics about known errors or
     * potential problems in the provided script
     * @param uri uri of the document
     * @param text source text of the document
     * @param depth TODO remove: current depth of the document
     */
    async *validateDocument(uri, text, depth = 0) {
        for await (const result of this.validateDocument_(uri, text, depth)) {
            if (Array.isArray(result)) {
                yield result;
            }
        }
    }
    /**
     * Main validation function for a document. Lexically and semantically understands a document.
     * Will additionally perform the same analysis on other run scripts found in this script
     * @param uri uri of the document
     * @param text source text of the document
     * @param depth TODO remove: current depth of the document
     */
    async *validateDocument_(uri, text, depth) {
        const { script, parseErrors, scanErrors } = await this.parseDocument(uri, text);
        const symbolTables = [];
        const scanDiagnostics = scanErrors.map(scanError => addDiagnosticsUri(scanError, uri));
        const parserDiagnostics = parseErrors.length === 0
            ? []
            : parseErrors
                .map(error => error.inner.concat(error))
                .reduce((acc, current) => acc.concat(current))
                .map(error => parseToDiagnostics(error, uri));
        yield scanDiagnostics.concat(parserDiagnostics);
        // if any run statement exist get uri then load
        if (script.runStmts.length > 0 && this.pathResolver.ready) {
            const loadDatas = this.getValidUri(uri, script.runStmts);
            // for each document run validate and yield any results
            for (const loadData of loadDatas) {
                for await (const result of this.loadAndValidateDocument(uri, loadData, depth + 1)) {
                    if (Array.isArray(result)) {
                        yield result;
                    }
                    else {
                        symbolTables.push(result);
                    }
                }
            }
        }
        this.logger.info('');
        this.logger.info('-------------Semantic Analysis------------');
        // generate a scope manager for resolving
        const symbolTableBuilder = new symbolTableBuilder_1.SymbolTableBuilder(uri, this.logger);
        // add child scopes
        for (const symbolTable of symbolTables) {
            symbolTableBuilder.linkTable(symbolTable);
        }
        // add standard library
        symbolTableBuilder.linkTable(this.standardLibrary);
        symbolTableBuilder.linkTable(this.activeBodyLibrary());
        // generate resolvers
        const preResolver = new preResolver_1.PreResolver(script, symbolTableBuilder, this.logger, this.tracer);
        const resolver = new resolver_1.Resolver(script, symbolTableBuilder, this.logger, this.tracer);
        // traverse the ast to find functions to pre populate symbol table
        perf_hooks_1.performance.mark('pre-resolver-start');
        const preDiagnostics = preResolver
            .resolve()
            .map(error => addDiagnosticsUri(error, uri));
        yield preDiagnostics;
        perf_hooks_1.performance.mark('pre-resolver-end');
        // traverse the ast again to resolve the remaning symbols
        perf_hooks_1.performance.mark('resolver-start');
        const resolverDiagnostics = resolver
            .resolve()
            .map(error => addDiagnosticsUri(error, uri));
        yield resolverDiagnostics;
        // find scopes were symbols were never used
        const unusedDiagnostics = symbolTableBuilder
            .findUnused()
            .map(error => addDiagnosticsUri(error, uri));
        yield unusedDiagnostics;
        perf_hooks_1.performance.mark('resolver-end');
        // build the final symbol table
        const symbolTable = symbolTableBuilder.build();
        // perform type checking
        const typeChecker = new typeChecker_1.TypeChecker(script, this.logger, this.tracer);
        perf_hooks_1.performance.mark('type-checking-start');
        typeChecker.check().map(error => addDiagnosticsUri(error, uri));
        // yield typeDiagnostics;
        perf_hooks_1.performance.mark('type-checking-end');
        // measure performance
        perf_hooks_1.performance.measure('Pre Resolver', 'pre-resolver-start', 'pre-resolver-end');
        perf_hooks_1.performance.measure('Resolver', 'resolver-start', 'resolver-end');
        perf_hooks_1.performance.measure('Type Checking', 'type-checking-start', 'type-checking-end');
        // make sure to delete references so scope manager can be gc'ed
        let documentInfo = this.documentInfos.get(uri);
        if (!typeGuards_1.empty(documentInfo)) {
            documentInfo.symbolsTable.removeSelf();
            documentInfo = undefined;
        }
        this.documentInfos.set(uri, {
            script,
            symbolsTable: symbolTable,
            diagnostics: scanDiagnostics.concat(parserDiagnostics, preDiagnostics, resolverDiagnostics),
        });
        this.logger.info('--------------------------------------');
        perf_hooks_1.performance.clearMarks();
        yield symbolTable;
    }
    /**
     * Get the symbol table corresponding active set of celetrial bodies for the user.
     * This allows for bodies other than that in stock ksp to be incorporated
     */
    activeBodyLibrary() {
        /** TODO actually load other bodies */
        return this.bodyLibrary;
    }
    /**
     * Get the token at the provided position in the text document
     * @param pos position in the text document
     * @param uri uri of the text document
     */
    getToken(pos, uri) {
        const documentInfo = this.documentInfos.get(uri);
        if (typeGuards_1.empty(documentInfo)) {
            return undefined;
        }
        // try to find an symbol at the position
        const { script } = documentInfo;
        const finder = new scriptFind_1.ScriptFind();
        const result = finder.find(script, pos);
        return result && result.token;
    }
    /**
     * Get the declaration location for the token at the provided position
     * @param pos position in the document
     * @param uri uri of the document
     */
    getDeclarationLocation(pos, uri) {
        const documentInfo = this.documentInfos.get(uri);
        if (typeGuards_1.empty(documentInfo)) {
            return undefined;
        }
        // try to find an symbol at the position
        const { symbolsTable, script } = documentInfo;
        const finder = new scriptFind_1.ScriptFind();
        const result = finder.find(script, pos);
        if (typeGuards_1.empty(result)) {
            return undefined;
        }
        // check if symbols exists
        const { token } = result;
        const symbol = symbolsTable.scopedNamedTracker(pos, token.lookup);
        if (typeGuards_1.empty(symbol)) {
            return undefined;
        }
        // exit if undefiend
        if (symbol.declared.uri === constants_1.builtIn) {
            return undefined;
        }
        return symbol.declared.symbol.name;
    }
    /**
     * Get all usage locations in all files
     * @param pos position in document
     * @param uri uri of document
     */
    getUsageLocations(pos, uri) {
        const documentInfo = this.documentInfos.get(uri);
        if (typeGuards_1.empty(documentInfo) ||
            typeGuards_1.empty(documentInfo.symbolsTable) ||
            typeGuards_1.empty(documentInfo.script)) {
            return undefined;
        }
        // try to find the symbol at the position
        const { symbolsTable, script } = documentInfo;
        const finder = new scriptFind_1.ScriptFind();
        const result = finder.find(script, pos);
        if (typeGuards_1.empty(result)) {
            return undefined;
        }
        // try to find the tracker at a given position
        const { token } = result;
        const tracker = symbolsTable.scopedNamedTracker(pos, token.lookup);
        if (typeGuards_1.empty(tracker)) {
            return undefined;
        }
        return tracker.usages
            .map(usage => usage)
            .concat(tracker.declared.symbol.name)
            .filter(location => location.uri !== constants_1.builtIn);
    }
    /**
     * Get all usage ranges in a provide file
     * @param pos position in document
     * @param uri uri of document
     */
    getFileUsageRanges(pos, uri) {
        const locations = this.getUsageLocations(pos, uri);
        if (typeGuards_1.empty(locations)) {
            return locations;
        }
        return locations.filter(loc => loc.uri === uri).map(loc => loc.range);
    }
    /**
     * Get all symbols in scope at a particulare location in the file
     * @param pos position in document
     * @param uri document uri
     */
    getScopedSymbols(pos, uri) {
        const documentInfo = this.documentInfos.get(uri);
        if (!typeGuards_1.empty(documentInfo) && !typeGuards_1.empty(documentInfo.symbolsTable)) {
            return documentInfo.symbolsTable.scopedSymbols(pos);
        }
        return [];
    }
    /**
     * Get all symbols in a provided file
     * @param uri document uri
     */
    getAllFileSymbols(uri) {
        const documentInfo = this.documentInfos.get(uri);
        if (!typeGuards_1.empty(documentInfo) && !typeGuards_1.empty(documentInfo.symbolsTable)) {
            return documentInfo.symbolsTable.fileSymbols();
        }
        return [];
    }
    // get function at position
    getFunctionAtPosition(pos, uri) {
        // we need the document info to lookup a signature
        const documentInfo = this.documentInfos.get(uri);
        if (typeGuards_1.empty(documentInfo))
            return undefined;
        const { script } = documentInfo;
        const finder = new scriptFind_1.ScriptFind();
        // attempt to find a token here get surround invalid inst context
        const outerResult = finder.find(script, pos, Expr.Suffix, Stmt.Invalid);
        const innerResult = finder.find(script, pos, SuffixTerm.Call);
        let index = 0;
        if (!typeGuards_1.empty(innerResult) &&
            !typeGuards_1.empty(innerResult.node) &&
            innerResult.node instanceof SuffixTerm.Call) {
            index = innerResult.node.args.length > 0
                ? innerResult.node.args.length - 1
                : 0;
        }
        // currently we only support invalid statements for signature completion
        // we could possible support call expressions as well
        if (typeGuards_1.empty(outerResult) || typeGuards_1.empty(outerResult.node)) {
            return undefined;
        }
        // determine the identifier of the invalid statement and parameter index
        const { node } = outerResult;
        // check if suffix
        if (node instanceof Expr.Suffix) {
            const tracker = node.mostResolveTracker();
            if (typeGuards_1.empty(tracker)) {
                return undefined;
            }
            switch (tracker.declared.symbol.tag) {
                case types_1.KsSymbolKind.function:
                case types_1.KsSymbolKind.suffix:
                    return {
                        index,
                        tracker,
                    };
                default:
                    return undefined;
            }
        }
        // check if invalid statment
        if (node instanceof Stmt.Invalid) {
            const { ranges } = node;
            const indices = positionUtils_1.binarySearchIndex(ranges, pos);
            const start = Array.isArray(indices) ? indices[0] : indices;
            for (let i = start; i >= 0; i -= 1) {
                const element = ranges[i];
                if (element instanceof Expr.Suffix) {
                    const tracker = element.mostResolveTracker();
                    if (!typeGuards_1.empty(tracker)) {
                        switch (tracker.declared.symbol.tag) {
                            case types_1.KsSymbolKind.function:
                            case types_1.KsSymbolKind.suffix:
                                return {
                                    index,
                                    tracker,
                                };
                            default:
                                return undefined;
                        }
                    }
                }
            }
        }
        return undefined;
    }
    /**
     * Generate a ast from the provided source text
     * @param uri uri to document
     * @param text source text of document
     */
    async parseDocument(uri, text) {
        this.logger.info('');
        this.logger.info('-------------Lexical Analysis------------');
        perf_hooks_1.performance.mark('scanner-start');
        const scanner = new scanner_1.Scanner(text, uri, this.logger, this.tracer);
        const { tokens, scanErrors } = scanner.scanTokens();
        perf_hooks_1.performance.mark('scanner-end');
        // if scanner found errors report those immediately
        if (scanErrors.length > 0) {
            this.logger.warn(`Scanning encountered ${scanErrors.length} Errors.`);
        }
        perf_hooks_1.performance.mark('parser-start');
        const parser = new parser_1.Parser(uri, tokens, this.logger, this.tracer);
        const result = parser.parse();
        perf_hooks_1.performance.mark('parser-end');
        // measure performance
        perf_hooks_1.performance.measure('Scanner', 'scanner-start', 'scanner-end');
        perf_hooks_1.performance.measure('Parser', 'parser-start', 'parser-end');
        perf_hooks_1.performance.clearMarks();
        this.logger.info('--------------------------------------');
        return {
            scanErrors,
            ...result,
        };
    }
    // get usable file uri from run statements
    getValidUri(uri, runStmts) {
        // generate uris then remove empty or preloaded documents
        return runStmts
            .map(inst => this.pathResolver.resolveUri({
            uri,
            range: { start: inst.start, end: inst.end },
        }, pathResolver_1.runPath(inst)))
            .filter(typeGuards_1.notEmpty);
        // .filter(uriStmts => !this.documentInfos.has(uriStmts.uri));
    }
    // load an validate a file from disk
    async *loadAndValidateDocument(parentUri, { uri, caller, path }, depth) {
        try {
            // non ideal fix for depedency cycle
            // TODO we need to actually check for cycles and do something else
            if (depth > 10) {
                return { diagnostics: [] };
            }
            // if cache not found attempt to find file from disk
            const validated = this.tryFindDocument(path, uri);
            if (typeGuards_1.empty(validated)) {
                return {
                    diagnostics: [
                        {
                            uri: parentUri,
                            range: caller,
                            severity: vscode_languageserver_1.DiagnosticSeverity.Error,
                            message: `Unable to find ${path}`,
                        },
                    ],
                };
            }
            // attempt to read file from disk
            const fileResult = await fsUtils_1.readFileAsync(validated.path, 'utf-8');
            yield* this.validateDocument_(validated.uri, fileResult, depth + 1);
        }
        catch (err) {
            // if we already checked for the file exists but failed anyways??
            return {
                diagnostics: [
                    {
                        uri: parentUri,
                        range: caller,
                        severity: vscode_languageserver_1.DiagnosticSeverity.Error,
                        message: `Unable to read ${path}`,
                    },
                ],
            };
        }
    }
    /**
     * Try to find a document in the workspace
     * @param path path to file
     * @param uri uri to file
     */
    tryFindDocument(path, uri) {
        const ext = path_1.extname(path);
        switch (ext) {
            case '.ks':
                // case '.ksm': probably need to report we can't read ksm files
                if (fs_1.existsSync(path)) {
                    return { path, uri };
                }
                return undefined;
            case '':
                if (fs_1.existsSync(`${path}.ks`)) {
                    return { path: `${path}.ks`, uri: `${uri}.ks` };
                }
                return undefined;
            default:
                return undefined;
        }
    }
}
exports.Analyzer = Analyzer;
// convert parse error to diagnostic
const parseToDiagnostics = (error, uri) => {
    return {
        uri,
        severity: vscode_languageserver_1.DiagnosticSeverity.Error,
        range: { start: error.start, end: error.end },
        message: error.message,
        source: 'kos-language-server',
    };
};
// convert resolver error to diagnostic
const addDiagnosticsUri = (error, uri) => {
    return { uri, ...error };
};
//# sourceMappingURL=analyzer.js.map