"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const scanner_1 = require("../scanner/scanner");
const fsUtils_1 = require("../utilities/fsUtils");
const fs_1 = require("fs");
const path_1 = require("path");
const perf_hooks_1 = require("perf_hooks");
const parser_1 = require("../parser/parser");
const symbolTableBuilder_1 = require("../analysis/symbolTableBuilder");
const preResolver_1 = require("../analysis/preResolver");
const resolver_1 = require("../analysis/resolver");
const standardLibrary_1 = require("../analysis/standardLibrary");
const typeChecker_1 = require("../typeChecker/typeChecker");
const testDir = path_1.join(__dirname, '../../../kerboscripts/parser_valid/');
const logger = fs_1.createWriteStream('bench.txt', { flags: 'a' });
const log = (line) => {
    console.log(line);
    // tslint:disable-next-line: prefer-template
    logger.write(line + '\n');
};
const standardLibrary = standardLibrary_1.standardLibraryBuilder(0 /* lowercase */);
const scriptSources = [];
// jit warm up
fsUtils_1.walkDir(testDir, (filePath) => {
    const size = fs_1.statSync(filePath).size;
    const kosFile = fs_1.readFileSync(filePath, 'utf8');
    scriptSources.push({ size, kosFile, filePath });
    const scanner = new scanner_1.Scanner(kosFile);
    const { tokens } = scanner.scanTokens();
    const parser = new parser_1.Parser('', tokens);
    const { script } = parser.parse();
    const symbolTableBuilder = new symbolTableBuilder_1.SymbolTableBuilder('');
    symbolTableBuilder.linkTable(standardLibrary);
    const preResolver = new preResolver_1.PreResolver(script, symbolTableBuilder);
    const resolver = new resolver_1.Resolver(script, symbolTableBuilder);
    preResolver.resolve();
    resolver.resolve();
    const typeChecker = new typeChecker_1.TypeChecker(script);
    typeChecker.check();
});
const overallResults = [];
let scanResults = [];
let parseResults = [];
let resolverResults = [];
let typeCheckerResults = [];
for (let i = 0; i < 10; i += 1) {
    for (const { kosFile, size, filePath } of scriptSources) {
        const scannerStart = perf_hooks_1.performance.now();
        const scanner = new scanner_1.Scanner(kosFile);
        const { tokens } = scanner.scanTokens();
        const scannerEnd = perf_hooks_1.performance.now();
        const parserStart = perf_hooks_1.performance.now();
        const parser = new parser_1.Parser('', tokens);
        const { script } = parser.parse();
        const parserEnd = perf_hooks_1.performance.now();
        const resolverStart = perf_hooks_1.performance.now();
        const symbolTableBuilder = new symbolTableBuilder_1.SymbolTableBuilder('');
        symbolTableBuilder.linkTable(standardLibrary);
        const preResolver = new preResolver_1.PreResolver(script, symbolTableBuilder);
        const resolver = new resolver_1.Resolver(script, symbolTableBuilder);
        preResolver.resolve();
        resolver.resolve();
        symbolTableBuilder.findUnused();
        const resolveEnd = perf_hooks_1.performance.now();
        const typeCheckerStart = perf_hooks_1.performance.now();
        const typeChecker = new typeChecker_1.TypeChecker(script);
        typeChecker.check();
        const typeCheckerEnd = perf_hooks_1.performance.now();
        scanResults.push({
            filePath,
            size: size / 1024,
            time: scannerEnd - scannerStart,
            rate: size / ((scannerEnd - scannerStart) / 1000 * 1024),
        });
        parseResults.push({
            filePath,
            size: size / 1024,
            time: parserEnd - parserStart,
            rate: size / ((parserEnd - parserStart) / 1000 * 1024),
        });
        resolverResults.push({
            filePath,
            size: size / 1024,
            time: resolveEnd - resolverStart,
            rate: size / ((resolveEnd - resolverStart) / 1000 * 1024),
        });
        typeCheckerResults.push({
            filePath,
            size: size / 1024,
            time: typeCheckerEnd - typeCheckerStart,
            rate: size / ((typeCheckerEnd - typeCheckerStart) / 1000 * 1024),
        });
        overallResults.push({
            filePath,
            size: size / 1024,
            time: typeCheckerEnd - scannerStart,
            rate: size / ((typeCheckerEnd - scannerStart) / 1000 * 1024),
        });
    }
}
scanResults = scanResults.sort((x, y) => (y.rate - x.rate));
parseResults = parseResults.sort((x, y) => (y.rate - x.rate));
resolverResults = resolverResults.sort((x, y) => (y.rate - x.rate));
typeCheckerResults = typeCheckerResults.sort((x, y) => (y.rate - x.rate));
for (const result of scanResults) {
    // tslint:disable-next-line:max-line-length
    log(`file: ${path_1.relative(testDir, result.filePath)} size: ${result.size.toFixed(1)} KB, time ${result.time.toFixed(2)} ms ${result.rate.toFixed(2)} KB/s`);
}
log('');
log('--------------------------Parser------------------------------');
log('');
for (const result of parseResults) {
    // tslint:disable-next-line:max-line-length
    log(`file: ${path_1.relative(testDir, result.filePath)} size: ${result.size.toFixed(1)} KB, time ${result.time.toFixed(2)} ms ${result.rate.toFixed(2)} KB/s`);
}
log('');
log('--------------------------Resolver------------------------------');
log('');
for (const result of resolverResults) {
    // tslint:disable-next-line:max-line-length
    console.log(`file: ${path_1.relative(testDir, result.filePath)} size: ${result.size.toFixed(1)} KB, time ${result.time.toFixed(2)} ms ${result.rate.toFixed(2)} KB/s`);
}
log('');
log('--------------------------Type Checker------------------------------');
log('');
for (const result of typeCheckerResults) {
    // tslint:disable-next-line:max-line-length
    log(`file: ${path_1.relative(testDir, result.filePath)} size: ${result.size.toFixed(1)} KB, time ${result.time.toFixed(2)} ms ${result.rate.toFixed(2)} KB/s`);
}
let scanTime = 0;
let parseTime = 0;
let resolveTime = 0;
let typeCheckerTime = 0;
let overallTime = 0;
let size = 0;
for (const result of scanResults) {
    scanTime += result.time;
    size += result.size;
}
for (const result of parseResults) {
    parseTime += result.time;
}
for (const result of resolverResults) {
    resolveTime += result.time;
}
for (const result of typeCheckerResults) {
    typeCheckerTime += result.time;
}
for (const result of overallResults) {
    overallTime += result.time;
}
log('');
log('--------------------------Overall------------------------------');
log('');
log(`Overall scan rate ${size / (scanTime / 1000)} KB/s`);
log(`Overall parse rate ${size / (parseTime / 1000)} KB/s`);
log(`Overall resolver rate ${size / (resolveTime / 1000)} KB/s`);
log(`Overall type check rate ${size / (typeCheckerTime / 1000)} KB/s`);
log(`Overall rate ${size / (overallTime / 1000)} KB/s`);
//# sourceMappingURL=fullBench.js.map