"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
// import * as Decl from '../parser/declare';
const Expr = __importStar(require("../parser/expr"));
const Stmt = __importStar(require("../parser/stmt"));
const randomUtils_1 = require("../utilities/randomUtils");
const logger_1 = require("../utilities/logger");
const tokentypes_1 = require("../entities/tokentypes");
const constants_1 = require("../utilities/constants");
const jStat = require('jstat').jStat;
const ALPHA_CHARSET = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
const ALPHA_NUMERIC_CHARSET = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
const ASCII_LOW = 35;
const ASCII_HIGH = 126;
const ID_GAMMA_SHAPE = 12;
const ID_GAMMA_SCALE = 0.75;
const STRING_GAMMA_SHAPE = 18;
const STRING_GAMMA_SCALE = 0.75;
const DOUBLE_LOGNORM_MEAN = 0;
const DOUBLE_LOGNORM_SIGMA = 5;
class Generator {
    constructor(mutation, logger = logger_1.mockLogger, tracer = logger_1.mockTracer) {
        this.mutation = mutation;
        this.logger = logger;
        this.tracer = tracer;
    }
    generateExpr(exprType) {
        try {
            const exprClass = exprType || this.randomEntity(Expr.validExprTypes);
            return this.genExpr(exprClass);
        }
        catch (err) {
            this.logger.error(`Error occured in resolver ${err}`);
            this.tracer.log(err);
            return '';
        }
    }
    // resolve for an expression
    genExpr(exprClass) {
        return exprClass.classAccept(this);
    }
    draw(dist) {
        switch (dist.tag) {
            case 'constant':
                return dist.value;
            case 'exp':
                const expDraw = jStat.exponential.sample(dist.rate);
                return expDraw > 0 ? expDraw : 0;
            case 'gamma':
                const grammDraw = jStat.gamma.sample(dist.shape, dist.scale);
                return grammDraw > 0 ? grammDraw : 0;
            case 'normal':
                const normalDraw = jStat.normal.sample(dist.mean, dist.std);
                return normalDraw > 0 ? normalDraw : 0;
            default:
                throw new Error('Unsupported distribution');
        }
    }
    // given a list of expression constructors pick one
    randomEntity(entities) {
        const draws = entities.map(([_, dist]) => this.draw(dist));
        const cummulative = [draws[0]];
        for (const draw of draws.slice(1)) {
            cummulative.push(cummulative[cummulative.length - 1] + draw);
        }
        const indexDraw = Math.random() * cummulative[cummulative.length - 1];
        for (let i = 0; i < cummulative.length; i += 1) {
            if (indexDraw <= cummulative[i]) {
                return entities[i][0];
            }
        }
        throw new Error('');
    }
    visitExprInvalid(exprClass) {
        return this.generateGrammarNodes(exprClass.grammar);
    }
    visitBinary(exprClass) {
        return this.generateGrammarNodes(exprClass.grammar);
    }
    visitUnary(exprClass) {
        return this.generateGrammarNodes(exprClass.grammar);
    }
    visitFactor(exprClass) {
        return this.generateGrammarNodes(exprClass.grammar);
    }
    visitSuffix(exprClass) {
        return this.generateGrammarNodes(exprClass.grammar);
    }
    visitAnonymousFunction(exprClass) {
        return this.generateGrammarNodes(exprClass.grammar);
    }
    visitSuffixTerm(exprClass) {
        throw this.generateGrammarNodes(exprClass.grammar);
    }
    visitCall(exprClass) {
        return this.generateGrammarNodes(exprClass.grammar);
    }
    visitArrayIndex(exprClass) {
        return this.generateGrammarNodes(exprClass.grammar);
    }
    visitArrayBracket(exprClass) {
        return this.generateGrammarNodes(exprClass.grammar);
    }
    visitDelegate(exprClass) {
        return this.generateGrammarNodes(exprClass.grammar);
    }
    visitLiteral(exprClass) {
        return this.generateGrammarNodes(exprClass.grammar);
    }
    visitVariable(exprClass) {
        return this.generateGrammarNodes(exprClass.grammar);
    }
    visitGrouping(exprClass) {
        return this.generateGrammarNodes(exprClass.grammar);
    }
    generateGrammarNode(node) {
        if (isNode(node)) {
            if (node instanceof Object) {
                if (node.prototype instanceof Expr.Expr) {
                    return this.genExpr(node);
                }
                if (node.prototype instanceof Stmt.Stmt) {
                    throw new Error();
                }
                throw new Error();
            }
            return this.generateToken(node);
        }
        switch (node.tag) {
            case 'optional':
                return this.generateOptional(node);
            case 'union':
                return this.generateUnion(node);
            case 'repeat':
                return this.generateRepeat(node);
        }
    }
    generateUnion(union) {
        const node = this.randomEntity(union.node);
        return this.generateGrammarNode(node);
    }
    generateOptional(optional) {
        const threshold = this.draw(optional.dist);
        const value = Math.random();
        if (value > threshold) {
            return this.generateGrammarNodes(optional.nodes);
        }
        return '';
    }
    generateRepeat(repeat) {
        const draw = Math.floor(this.draw(repeat.dist));
        const results = [];
        for (let i = 0; i < draw; i += 1) {
            results.push(this.generateGrammarNodes(repeat.nodes));
        }
        return results.join('');
    }
    generateGrammarNodes(nodes) {
        return nodes.map(node => this.generateGrammarNode(node)).join('');
    }
    generateInteger() {
        return randomUtils_1.getRandomInt(1e6).toFixed(0);
    }
    generateDouble() {
        const number = jStat.lognormal.sample(DOUBLE_LOGNORM_MEAN, DOUBLE_LOGNORM_SIGMA);
        let numberStr = number.toExponential(randomUtils_1.getRandomInt(20));
        const removeSign = Math.random();
        if (removeSign < 0.1) {
            numberStr = numberStr.replace(/(\+|\-)/gi, '');
        }
        return numberStr;
    }
    generateRandomString(length) {
        let rngString = '';
        for (let i = 0; i < length; i += 1) {
            const rngChar = Math.floor((Math.random() * (ASCII_HIGH - ASCII_LOW)) + ASCII_LOW);
            rngString += String.fromCharCode(rngChar);
        }
        return rngString;
    }
    generateIdentifier() {
        const length = Math.ceil(jStat.gamma.sample(ID_GAMMA_SHAPE, ID_GAMMA_SCALE));
        let identifier = ALPHA_CHARSET[Math.floor(Math.random() * ALPHA_CHARSET.length)];
        for (let i = 1; i < length; i += 1) {
            identifier += ALPHA_NUMERIC_CHARSET[Math.floor(Math.random() * ALPHA_NUMERIC_CHARSET.length)];
        }
        if (constants_1.keywords.has(identifier.toLowerCase())) {
            return this.generateIdentifier();
        }
        return identifier;
    }
    generateString() {
        const length = jStat.gamma.sample(STRING_GAMMA_SHAPE, STRING_GAMMA_SCALE);
        return `"${this.generateRandomString(length)}"`;
    }
    generateToken(tokenType) {
        switch (tokenType) {
            case tokentypes_1.TokenType.plus:
                return ' + ';
            case tokentypes_1.TokenType.minus:
                return ' - ';
            case tokentypes_1.TokenType.multi:
                return ' * ';
            case tokentypes_1.TokenType.div:
                return ' / ';
            case tokentypes_1.TokenType.power:
                return '^';
            case tokentypes_1.TokenType.not:
                return 'not ';
            case tokentypes_1.TokenType.and:
                return ' and ';
            case tokentypes_1.TokenType.or:
                return ' or ';
            case tokentypes_1.TokenType.true:
                return 'true';
            case tokentypes_1.TokenType.false:
                return 'false';
            case tokentypes_1.TokenType.equal:
                return ' = ';
            case tokentypes_1.TokenType.notEqual:
                return ' <> ';
            case tokentypes_1.TokenType.greaterEqual:
                return ' >= ';
            case tokentypes_1.TokenType.greater:
                return ' > ';
            case tokentypes_1.TokenType.lessEqual:
                return ' <= ';
            case tokentypes_1.TokenType.less:
                return ' < ';
            case tokentypes_1.TokenType.set:
                return 'set';
            case tokentypes_1.TokenType.unset:
                return 'unset';
            case tokentypes_1.TokenType.to:
                return 'to';
            case tokentypes_1.TokenType.is:
                return 'is';
            case tokentypes_1.TokenType.until:
                return 'until';
            case tokentypes_1.TokenType.if:
                return 'if';
            case tokentypes_1.TokenType.else:
                return 'else';
            case tokentypes_1.TokenType.for:
                return 'for';
            case tokentypes_1.TokenType.when:
                return 'when';
            case tokentypes_1.TokenType.then:
                return 'then';
            case tokentypes_1.TokenType.from:
                return 'from';
            case tokentypes_1.TokenType.do:
                return 'do';
            case tokentypes_1.TokenType.at:
                return 'at';
            case tokentypes_1.TokenType.on:
                return 'on';
            case tokentypes_1.TokenType.in:
                return 'in';
            case tokentypes_1.TokenType.lock:
                return 'lock';
            case tokentypes_1.TokenType.unlock:
                return 'unlock';
            case tokentypes_1.TokenType.print:
                return 'print';
            case tokentypes_1.TokenType.arrayIndex:
                return '#';
            case tokentypes_1.TokenType.integer:
                return this.generateInteger();
            case tokentypes_1.TokenType.double:
                return this.generateDouble();
            case tokentypes_1.TokenType.string:
                return this.generateString();
            case tokentypes_1.TokenType.local:
                return 'local';
            case tokentypes_1.TokenType.global:
                return 'global';
            case tokentypes_1.TokenType.parameter:
                return 'parameter';
            case tokentypes_1.TokenType.function:
                return 'function';
            case tokentypes_1.TokenType.preserve:
                return 'preserve';
            case tokentypes_1.TokenType.break:
                return 'break';
            case tokentypes_1.TokenType.return:
                return 'return';
            case tokentypes_1.TokenType.declare:
                return 'declare';
            case tokentypes_1.TokenType.defined:
                return 'defined ';
            case tokentypes_1.TokenType.bracketOpen:
                return '(';
            case tokentypes_1.TokenType.bracketClose:
                return ')';
            case tokentypes_1.TokenType.curlyOpen:
                return '{';
            case tokentypes_1.TokenType.curlyClose:
                return '}';
            case tokentypes_1.TokenType.squareOpen:
                return '[';
            case tokentypes_1.TokenType.squareClose:
                return ']';
            case tokentypes_1.TokenType.comma:
                return ',';
            case tokentypes_1.TokenType.colon:
                return ':';
            case tokentypes_1.TokenType.period:
                return '.';
            case tokentypes_1.TokenType.atSign:
                return '@';
            case tokentypes_1.TokenType.eof:
                return '';
            case tokentypes_1.TokenType.toggle:
                return 'toggle';
            case tokentypes_1.TokenType.wait:
                return 'wait';
            case tokentypes_1.TokenType.off:
                return 'off';
            case tokentypes_1.TokenType.list:
                return 'list';
            case tokentypes_1.TokenType.clearscreen:
                return 'clearscreen';
            case tokentypes_1.TokenType.stage:
                return 'stage';
            case tokentypes_1.TokenType.add:
                return 'add';
            case tokentypes_1.TokenType.remove:
                return 'remove';
            case tokentypes_1.TokenType.log:
                return 'log';
            case tokentypes_1.TokenType.step:
                return 'step';
            case tokentypes_1.TokenType.switch:
                return 'switch';
            case tokentypes_1.TokenType.copy:
                return 'copy';
            case tokentypes_1.TokenType.rename:
                return 'rename';
            case tokentypes_1.TokenType.volume:
                return 'volume';
            case tokentypes_1.TokenType.file:
                return 'file';
            case tokentypes_1.TokenType.delete:
                return 'delete';
            case tokentypes_1.TokenType.edit:
                return 'edit';
            case tokentypes_1.TokenType.all:
                return 'all';
            case tokentypes_1.TokenType.run:
                return 'run';
            case tokentypes_1.TokenType.runPath:
                return 'runpath';
            case tokentypes_1.TokenType.runOncePath:
                return 'runoncepath';
            case tokentypes_1.TokenType.once:
                return 'once';
            case tokentypes_1.TokenType.compile:
                return 'compile';
            case tokentypes_1.TokenType.reboot:
                return 'reboot';
            case tokentypes_1.TokenType.shutdown:
                return 'shutdown';
            case tokentypes_1.TokenType.identifier:
                return this.generateIdentifier();
            case tokentypes_1.TokenType.fileIdentifier:
                return `${this.generateIdentifier()}.${this.generateIdentifier()}`;
            case tokentypes_1.TokenType.lazyGlobal:
                return 'lazyglobal';
            default:
                throw new Error(`Unknown token ${tokenType} encounted`);
        }
    }
}
exports.Generator = Generator;
const isNode = (node) => {
    if (node.hasOwnProperty('tag')) {
        return false;
    }
    return true;
};
//# sourceMappingURL=generator.js.map