import { IExprClass, IExprVisitor, IExpr, IExprClassVisitor, GrammarNode, Distribution, IExprPasser, SyntaxKind } from './types';
import * as SuffixTerm from './suffixTerm';
import * as Stmt from './stmt';
import { Range, Position } from 'vscode-languageserver';
import { NodeBase } from './base';
import { Token } from '../entities/token';
import { SymbolTracker } from '../analysis/types';
/**
 * Expression base class
 */
export declare abstract class Expr extends NodeBase implements IExpr {
    /**
     * Return the tree node type of expression
     */
    readonly tag: SyntaxKind.expr;
    /**
     * All expressions implement the pass method
     * Called when the node should be passed through
     * @param visitor visitor object
     */
    abstract pass<T>(visitor: IExprPasser<T>): T;
    /**
     * All expressions implement the accept method
     * Called when the node should execute the visitors methods
     * @param visitor visitor object
     */
    abstract accept<T>(visitor: IExprVisitor<T>): T;
}
/**
 * Container for tokens constituting an invalid expression
 */
export declare class Invalid extends Expr {
    readonly tokens: Token[];
    /**
     * Invalid expression constructor
     * @param tokens all tokens in the invalid range
     */
    constructor(tokens: Token[]);
    readonly start: Position;
    readonly end: Position;
    readonly ranges: Range[];
    toLines(): string[];
    accept<T>(visitor: IExprVisitor<T>): T;
    pass<T>(visitor: IExprPasser<T>): T;
    static classAccept<T>(visitor: IExprClassVisitor<T>): T;
}
/**
 * Class holding all valid binary expressions in KOS
 */
export declare class Binary extends Expr {
    readonly left: IExpr;
    readonly operator: Token;
    readonly right: IExpr;
    /**
     * Grammar for the binary expression
     */
    static grammar: GrammarNode[];
    /**
     * Constructor for all binary expressions
     * @param left left expression of the operation
     * @param operator the operator
     * @param right right expression of the operation
     */
    constructor(left: IExpr, operator: Token, right: IExpr);
    readonly start: Position;
    readonly end: Position;
    readonly ranges: Range[];
    toString(): string;
    toLines(): string[];
    accept<T>(visitor: IExprVisitor<T>): T;
    pass<T>(visitor: IExprPasser<T>): T;
    static classAccept<T>(visitor: IExprClassVisitor<T>): T;
}
/**
 * Class holding all valid unary expressions in KOS
 */
export declare class Unary extends Expr {
    readonly operator: Token;
    readonly factor: IExpr;
    /**
     * Grammar for the unary expressions
     */
    static grammar: GrammarNode[];
    /**
     * Unary expression constructor
     * @param operator unary operator
     * @param factor factor
     */
    constructor(operator: Token, factor: IExpr);
    readonly start: Position;
    readonly end: Position;
    readonly ranges: Range[];
    toString(): string;
    toLines(): string[];
    accept<T>(visitor: IExprVisitor<T>): T;
    pass<T>(visitor: IExprPasser<T>): T;
    static classAccept<T>(visitor: IExprClassVisitor<T>): T;
}
/**
 * Class holding expression with exponents
 */
export declare class Factor extends Expr {
    readonly suffix: IExpr;
    readonly power: Token;
    readonly exponent: IExpr;
    /**
     * Grammer for factor expressions
     */
    static grammar: GrammarNode[];
    /**
     * Factor constructor
     * @param suffix base expression
     * @param power exponent token
     * @param exponent exponent expression
     */
    constructor(suffix: IExpr, power: Token, exponent: IExpr);
    readonly start: Position;
    readonly end: Position;
    readonly ranges: Range[];
    toLines(): string[];
    accept<T>(visitor: IExprVisitor<T>): T;
    pass<T>(visitor: IExprPasser<T>): T;
    static classAccept<T>(visitor: IExprClassVisitor<T>): T;
}
/**
 * Class holding all anonymous functions
 */
export declare class Lambda extends Expr {
    readonly block: Stmt.Block;
    /**
     * Grammar for anonymous functions
     */
    static grammar: GrammarNode[];
    /**
     * Anonymous Function constructor
     * @param block the scope for the lambda
     */
    constructor(block: Stmt.Block);
    readonly start: Position;
    readonly end: Position;
    readonly ranges: Range[];
    toLines(): string[];
    accept<T>(visitor: IExprVisitor<T>): T;
    pass<T>(visitor: IExprPasser<T>): T;
    static classAccept<T>(visitor: IExprClassVisitor<T>): T;
}
/**
 * Class holding all kos suffixes
 */
export declare class Suffix extends Expr {
    readonly suffixTerm: SuffixTerm.SuffixTerm;
    colon?: Token | undefined;
    trailer?: SuffixTerm.SuffixTrailer | undefined;
    /**
     * Grammar for suffixes
     */
    static grammar: GrammarNode[];
    /**
     * Suffix constructor
     * @param suffixTerm base suffix term
     * @param colon optional suffix color
     * @param trailer optional suffix trailer
     */
    constructor(suffixTerm: SuffixTerm.SuffixTerm, colon?: Token | undefined, trailer?: SuffixTerm.SuffixTrailer | undefined);
    readonly start: Position;
    readonly end: Position;
    readonly ranges: Range[];
    /**
     * Method indicating if the suffix ends with a settable trailer
     */
    isSettable(): boolean;
    /**
     * Get the most resolved type on this suffix
     */
    mostResolveTracker(): Maybe<SymbolTracker>;
    toLines(): string[];
    accept<T>(visitor: IExprVisitor<T>): T;
    pass<T>(visitor: IExprPasser<T>): T;
    static classAccept<T>(visitor: IExprClassVisitor<T>): T;
}
/**
 * All valid expressions
 */
export declare const validExpressions: Constructor<Expr>[];
export declare const validExprTypes: [IExprClass, Distribution][];
export declare const expr: import("./types").IGrammarUnion;
