"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const types_1 = require("./types");
const SuffixTerm = __importStar(require("./suffixTerm"));
const tokentypes_1 = require("../entities/tokentypes");
const grammarNodes_1 = require("./grammarNodes");
const typeGuards_1 = require("../utilities/typeGuards");
const base_1 = require("./base");
const toStringUtils_1 = require("./toStringUtils");
/**
 * Expression base class
 */
class Expr extends base_1.NodeBase {
    /**
     * Return the tree node type of expression
     */
    get tag() {
        return types_1.SyntaxKind.expr;
    }
}
exports.Expr = Expr;
/**
 * Container for tokens constituting an invalid expression
 */
class Invalid extends Expr {
    /**
     * Invalid expression constructor
     * @param tokens all tokens in the invalid range
     */
    constructor(tokens) {
        super();
        this.tokens = tokens;
    }
    get start() {
        return this.tokens[0].start;
    }
    get end() {
        return this.tokens[this.tokens.length - 1].end;
    }
    get ranges() {
        return [...this.tokens];
    }
    toLines() {
        return [this.tokens.map(t => t.lexeme).join('')];
    }
    accept(visitor) {
        return visitor.visitExprInvalid(this);
    }
    pass(visitor) {
        return visitor.passExprInvalid(this);
    }
    static classAccept(visitor) {
        return visitor.visitExprInvalid(this);
    }
}
exports.Invalid = Invalid;
/**
 * Class holding all valid binary expressions in KOS
 */
class Binary extends Expr {
    /**
     * Constructor for all binary expressions
     * @param left left expression of the operation
     * @param operator the operator
     * @param right right expression of the operation
     */
    constructor(left, operator, right) {
        super();
        this.left = left;
        this.operator = operator;
        this.right = right;
    }
    get start() {
        return this.left.start;
    }
    get end() {
        return this.right.end;
    }
    get ranges() {
        return [this.left, this.operator, this.right];
    }
    toString() {
        return `${this.left.toString()} ${this.operator.lexeme} ${this.right.toString()}`;
    }
    toLines() {
        const leftLines = this.left.toLines();
        const rightLines = this.right.toLines();
        return toStringUtils_1.joinLines(` ${this.operator.lexeme} `, leftLines, rightLines);
    }
    accept(visitor) {
        return visitor.visitBinary(this);
    }
    pass(visitor) {
        return visitor.passBinary(this);
    }
    static classAccept(visitor) {
        return visitor.visitBinary(this);
    }
}
exports.Binary = Binary;
/**
 * Class holding all valid unary expressions in KOS
 */
class Unary extends Expr {
    /**
     * Unary expression constructor
     * @param operator unary operator
     * @param factor factor
     */
    constructor(operator, factor) {
        super();
        this.operator = operator;
        this.factor = factor;
    }
    get start() {
        return this.operator.start;
    }
    get end() {
        return this.factor.end;
    }
    get ranges() {
        return [this.operator, this.factor];
    }
    toString() {
        return `${this.operator.lexeme} ${this.factor.toString()}`;
    }
    toLines() {
        const lines = this.factor.toLines();
        switch (this.operator.type) {
            case tokentypes_1.TokenType.plus:
            case tokentypes_1.TokenType.minus:
                lines[0] = `${this.operator.lexeme}${lines[0]}`;
                return lines;
            default:
                lines[0] = `${this.operator.lexeme} ${lines[0]}`;
                return lines;
        }
    }
    accept(visitor) {
        return visitor.visitUnary(this);
    }
    pass(visitor) {
        return visitor.passUnary(this);
    }
    static classAccept(visitor) {
        return visitor.visitUnary(this);
    }
}
exports.Unary = Unary;
/**
 * Class holding expression with exponents
 */
class Factor extends Expr {
    /**
     * Factor constructor
     * @param suffix base expression
     * @param power exponent token
     * @param exponent exponent expression
     */
    constructor(suffix, power, exponent) {
        super();
        this.suffix = suffix;
        this.power = power;
        this.exponent = exponent;
    }
    get start() {
        return this.suffix.start;
    }
    get end() {
        return this.exponent.end;
    }
    get ranges() {
        return [this.suffix, this.power, this.exponent];
    }
    toLines() {
        return toStringUtils_1.joinLines(this.power.lexeme, this.suffix.toLines(), this.exponent.toLines());
    }
    accept(visitor) {
        return visitor.visitFactor(this);
    }
    pass(visitor) {
        return visitor.passFactor(this);
    }
    static classAccept(visitor) {
        return visitor.visitFactor(this);
    }
}
exports.Factor = Factor;
/**
 * Class holding all anonymous functions
 */
class Lambda extends Expr {
    /**
     * Anonymous Function constructor
     * @param block the scope for the lambda
     */
    constructor(block) {
        super();
        this.block = block;
    }
    get start() {
        return this.block.start;
    }
    get end() {
        return this.block.end;
    }
    get ranges() {
        return this.block.ranges;
    }
    toLines() {
        return this.block.toLines();
    }
    accept(visitor) {
        return visitor.visitLambda(this);
    }
    pass(visitor) {
        return visitor.passAnonymousFunction(this);
    }
    static classAccept(visitor) {
        return visitor.visitAnonymousFunction(this);
    }
}
exports.Lambda = Lambda;
/**
 * Class holding all kos suffixes
 */
class Suffix extends Expr {
    /**
     * Suffix constructor
     * @param suffixTerm base suffix term
     * @param colon optional suffix color
     * @param trailer optional suffix trailer
     */
    constructor(suffixTerm, colon, trailer) {
        super();
        this.suffixTerm = suffixTerm;
        this.colon = colon;
        this.trailer = trailer;
    }
    get start() {
        return this.suffixTerm.start;
    }
    get end() {
        return typeGuards_1.empty(this.trailer) ? this.suffixTerm.end : this.trailer.end;
    }
    get ranges() {
        if (!typeGuards_1.empty(this.colon) && !typeGuards_1.empty(this.trailer)) {
            return [this.suffixTerm, this.colon, this.trailer];
        }
        return [this.suffixTerm];
    }
    /**
     * Method indicating if the suffix ends with a settable trailer
     */
    isSettable() {
        // if no trailer check suffix term
        if (typeGuards_1.empty(this.trailer)) {
            const { atom, trailers } = this.suffixTerm;
            // check for suffix term trailers
            if (trailers.length > 0) {
                const lastTrailer = trailers[trailers.length - 1];
                if (lastTrailer instanceof SuffixTerm.Identifier) {
                    return true;
                }
                if (lastTrailer instanceof SuffixTerm.ArrayBracket) {
                    return true;
                }
                if (lastTrailer instanceof SuffixTerm.ArrayIndex) {
                    return true;
                }
            }
            if (atom instanceof SuffixTerm.Identifier) {
                return true;
            }
            return false;
        }
        // check nested trailers
        if (this.trailer instanceof SuffixTerm.SuffixTrailer) {
            return this.trailer.isSettable();
        }
        return false;
    }
    /**
     * Get the most resolved type on this suffix
     */
    mostResolveTracker() {
        // if no trailer check suffix term
        if (typeGuards_1.empty(this.trailer)) {
            const { atom, trailers } = this.suffixTerm;
            // check for suffix term trailers
            if (trailers.length > 0) {
                const lastTrailer = trailers[trailers.length - 1];
                if (lastTrailer instanceof SuffixTerm.ArrayBracket) {
                    return lastTrailer.open.tracker;
                }
                if (lastTrailer instanceof SuffixTerm.ArrayIndex) {
                    return undefined;
                }
                if (lastTrailer instanceof SuffixTerm.Call) {
                    return lastTrailer.open.tracker;
                }
                return undefined;
            }
            // check nested trailers
            if (atom instanceof SuffixTerm.Identifier) {
                return atom.token.tracker;
            }
            return undefined;
        }
        // check nested trailers
        if (!typeGuards_1.empty(this.trailer)) {
            return this.trailer.mostResolveTracker();
        }
        return undefined;
    }
    toLines() {
        const suffixTermLines = this.suffixTerm.toLines();
        if (!typeGuards_1.empty(this.colon) && !typeGuards_1.empty(this.trailer)) {
            const trailerLines = this.trailer.toLines();
            return toStringUtils_1.joinLines(this.colon.lexeme, suffixTermLines, trailerLines);
        }
        return suffixTermLines;
    }
    accept(visitor) {
        return visitor.visitSuffix(this);
    }
    pass(visitor) {
        return visitor.passSuffix(this);
    }
    static classAccept(visitor) {
        return visitor.visitSuffix(this);
    }
}
exports.Suffix = Suffix;
/**
 * All valid expressions
 */
exports.validExpressions = [
    Binary,
    Unary,
    Factor,
    Suffix,
    Lambda,
];
exports.validExprTypes = [
    [Binary, grammarNodes_1.createConstant(1.0)],
    [Unary, grammarNodes_1.createConstant(0.5)],
    [Factor, grammarNodes_1.createConstant(0.5)],
    [Suffix, grammarNodes_1.createConstant(3)],
    // TODO update when insts included
    [Lambda, grammarNodes_1.createConstant(0)],
];
exports.expr = grammarNodes_1.createGrammarUnion(...exports.validExprTypes);
Binary.grammar = [
    exports.expr,
    grammarNodes_1.createGrammarUnion([tokentypes_1.TokenType.plus, grammarNodes_1.createConstant(1)], [tokentypes_1.TokenType.minus, grammarNodes_1.createConstant(1)], [tokentypes_1.TokenType.multi, grammarNodes_1.createConstant(1)], [tokentypes_1.TokenType.div, grammarNodes_1.createConstant(1)], [tokentypes_1.TokenType.equal, grammarNodes_1.createConstant(1)], [tokentypes_1.TokenType.notEqual, grammarNodes_1.createConstant(1)], [tokentypes_1.TokenType.less, grammarNodes_1.createConstant(1)], [tokentypes_1.TokenType.lessEqual, grammarNodes_1.createConstant(1)], [tokentypes_1.TokenType.greater, grammarNodes_1.createConstant(1)], [tokentypes_1.TokenType.greaterEqual, grammarNodes_1.createConstant(1)], [tokentypes_1.TokenType.or, grammarNodes_1.createConstant(1)], [tokentypes_1.TokenType.and, grammarNodes_1.createConstant(1)]),
    exports.expr,
];
Unary.grammar = [
    grammarNodes_1.createGrammarOptional(grammarNodes_1.createConstant(0.2), grammarNodes_1.createGrammarUnion([tokentypes_1.TokenType.plus, grammarNodes_1.createConstant(1)], [tokentypes_1.TokenType.minus, grammarNodes_1.createConstant(1)], [tokentypes_1.TokenType.not, grammarNodes_1.createConstant(1)], [tokentypes_1.TokenType.defined, grammarNodes_1.createConstant(1)])),
    grammarNodes_1.createGrammarUnion([Factor, grammarNodes_1.createConstant(1)], [Suffix, grammarNodes_1.createConstant(3)]),
];
Factor.grammar = [
    Suffix,
    grammarNodes_1.createGrammarOptional(grammarNodes_1.createExponential(2), tokentypes_1.TokenType.power, Suffix),
];
Suffix.grammar = [
    SuffixTerm.SuffixTerm,
    grammarNodes_1.createGrammarRepeat(grammarNodes_1.createExponential(2), tokentypes_1.TokenType.colon, SuffixTerm.SuffixTerm),
];
Lambda.grammar = [
    tokentypes_1.TokenType.curlyOpen,
    //
    tokentypes_1.TokenType.curlyClose,
];
//# sourceMappingURL=expr.js.map