"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const tokentypes_1 = require("../entities/tokentypes");
const parserError_1 = require("./parserError");
const Expr = __importStar(require("./expr"));
const SuffixTerm = __importStar(require("./suffixTerm"));
const Stmt = __importStar(require("./stmt"));
const Decl = __importStar(require("./declare"));
const typeGuards_1 = require("../utilities/typeGuards");
const script_1 = require("../entities/script");
const parseResult_1 = require("./parseResult");
const token_1 = require("../entities/token");
const logger_1 = require("../utilities/logger");
const arrayUtils_1 = require("../utilities/arrayUtils");
const marker_1 = require("../entities/marker");
class Parser {
    constructor(uri, tokens, logger = logger_1.mockLogger, tracer = logger_1.mockTracer) {
        this.andBind = this.and.bind(this);
        this.eqaulityBind = this.equality.bind(this);
        this.comparisonBind = this.comparison.bind(this);
        this.additionBind = this.addition.bind(this);
        this.multiplicationBind = this.multiplication.bind(this);
        this.unaryBind = this.unary.bind(this);
        // binary expression parser
        this.binaryExpression = (recurse, types) => {
            let expr = recurse();
            while (this.matchTokens(types)) {
                const operator = this.previous();
                const right = recurse();
                expr = parseResult_1.nodeResult(new Expr.Binary(expr.value, operator, right.value), arrayUtils_1.flatten([expr.errors, right.errors]));
            }
            return expr;
        };
        this.uri = uri;
        this.tokens = tokens.concat(this.eof(tokens));
        this.current = 0;
        this.runStmts = [];
        this.logger = logger;
        this.tracer = tracer;
    }
    // parse tokens
    parse() {
        try {
            const splits = this.uri.split('/');
            const file = splits[splits.length - 1];
            this.logger.info(`Parsing started for ${file} with ${this.tokens.length} tokens.`);
            const statements = [];
            const parseErrors = [];
            while (!this.isAtEnd()) {
                const { value, errors } = this.declaration();
                statements.push(value);
                parseErrors.push(...errors);
            }
            this.logger.info(`Parsing finished for ${file} with ` +
                `${this.runStmts.length} run statements`);
            if (parseErrors.length > 0) {
                this.logger.warn(`Parser encounted ${parseErrors.length} errors`);
            }
            return {
                parseErrors,
                script: new script_1.Script(this.uri, statements, this.runStmts),
            };
        }
        catch (err) {
            this.logger.error(`Error occured in parser ${err}`);
            this.tracer.log(err);
            return {
                parseErrors: [],
                script: new script_1.Script(this.uri, [], []),
            };
        }
    }
    // testing function / utility
    parseStatement() {
        return this.declaration();
    }
    // testing function / utility
    parseExpression() {
        try {
            return this.expression();
        }
        catch (error) {
            if (error instanceof parserError_1.ParseError) {
                this.synchronize(error.failed);
                return {
                    errors: [error],
                    value: new Expr.Invalid(this.tokens.slice(0, this.current)),
                };
            }
            throw error;
        }
    }
    parseArgCount() {
        return this.partialArgumentsCount();
    }
    // generate a placholder token as a fake end of file
    eof(tokens) {
        if (tokens.length === 0) {
            return new token_1.Token(tokentypes_1.TokenType.eof, '', undefined, new marker_1.Marker(0, 0), new marker_1.Marker(0, 1), '');
        }
        const last = tokens[tokens.length - 1];
        return new token_1.Token(tokentypes_1.TokenType.eof, '', undefined, new marker_1.Marker(last.end.line + 1, 0), new marker_1.Marker(last.end.line + 1, 1), last.uri);
    }
    // parse declaration attempt to synchronize
    declaration() {
        const start = this.current;
        try {
            if ([
                tokentypes_1.TokenType.declare,
                tokentypes_1.TokenType.local,
                tokentypes_1.TokenType.global,
                tokentypes_1.TokenType.parameter,
                tokentypes_1.TokenType.function,
                tokentypes_1.TokenType.lock,
            ].some(t => this.check(t))) {
                return this.define();
            }
            return this.statement();
        }
        catch (error) {
            if (error instanceof parserError_1.ParseError) {
                this.logger.verbose(JSON.stringify(error.partial));
                this.synchronize(error.failed);
                const tokens = this.tokens.slice(start, this.current);
                return {
                    errors: [error],
                    value: new Stmt.Invalid(start === this.current
                        ? this.peek().end
                        : tokens[0].start, tokens, error.partial),
                };
            }
            throw error;
        }
    }
    // parse declaration statements
    define() {
        // attempt to find scoping
        const declare = this.matchToken(tokentypes_1.TokenType.declare)
            ? this.previous()
            : undefined;
        const scope = this.matchTokens([tokentypes_1.TokenType.local, tokentypes_1.TokenType.global])
            ? this.previous()
            : undefined;
        const scopeDeclare = declare || scope ? new Decl.Scope(scope, declare) : undefined;
        // match declaration
        if (this.matchToken(tokentypes_1.TokenType.function)) {
            return this.declareFunction(scopeDeclare);
        }
        if (this.matchToken(tokentypes_1.TokenType.parameter)) {
            return this.declareParameter(scopeDeclare);
        }
        if (this.matchToken(tokentypes_1.TokenType.lock)) {
            return this.declareLock(scopeDeclare);
        }
        if (!typeGuards_1.empty(scopeDeclare)) {
            return this.declareVariable(scopeDeclare);
        }
        throw this.error(this.peek(), Decl.Scope, 'Expected function parameter or variable declaration.', 'Example: "local function exampleFunc { ... }", "global x is 0"');
    }
    // parse function declaration
    declareFunction(scope) {
        const builder = {
            scope,
            functionToken: undefined,
            identifier: undefined,
            block: undefined,
        };
        builder.functionToken = this.previous();
        builder.identifier = this.consumeIdentifierThrow('Expected identifier', Decl.Func, builder);
        // match function body
        if (this.matchToken(tokentypes_1.TokenType.curlyOpen)) {
            const block = this.block();
            builder.block = block.value;
            this.matchToken(tokentypes_1.TokenType.period);
            return parseResult_1.nodeResult(new Decl.Func(builder), block.errors);
        }
        throw this.error(this.peek(), Decl.Func, 'Expected function statement block starting with "{"', 'Example: local function { print "hi". }', builder);
    }
    // parse parameter declaration
    declareParameter(scope) {
        const builder = {
            scope,
            parameterToken: undefined,
            requiredParameters: undefined,
            optionalParameters: undefined,
        };
        builder.parameterToken = this.previous();
        builder.requiredParameters = this.declareNormalParameters();
        const defaultParameters = this.declaredDefaultedParameters();
        builder.optionalParameters = defaultParameters.value;
        this.terminal(Decl.Param);
        return parseResult_1.nodeResult(new Decl.Param(builder), defaultParameters.errors);
    }
    // parse regular parameters
    declareNormalParameters() {
        const parameters = [];
        // parse paremter until defaulted
        do {
            // break if this parameter is defaulted
            if (this.checkNext(tokentypes_1.TokenType.is) || this.checkNext(tokentypes_1.TokenType.to))
                break;
            const identifer = this.consumeIdentifierThrow('Expected additional identiifer following comma.', Decl.Parameter);
            parameters.push(new Decl.Parameter(identifer));
        } while (this.matchToken(tokentypes_1.TokenType.comma));
        return parameters;
    }
    // parse defaulted parameters
    declaredDefaultedParameters() {
        const defaultParameters = [];
        const errors = [];
        // parse until no additional parameters exist
        do {
            if (!this.checkNext(tokentypes_1.TokenType.is) && !this.checkNext(tokentypes_1.TokenType.to))
                break;
            const identifer = this.consumeIdentifierThrow('Expected identifier following comma.', Decl.DefaultParam);
            const toIs = this.consumeTokensThrow('Expected default parameter using keyword "to" or "is".', Decl.DefaultParam, [tokentypes_1.TokenType.to, tokentypes_1.TokenType.is]);
            const valueResult = this.expression();
            defaultParameters.push(new Decl.DefaultParam(identifer, toIs, valueResult.value));
            errors.push(valueResult.errors);
        } while (this.matchToken(tokentypes_1.TokenType.comma));
        return parseResult_1.nodeResult(defaultParameters, arrayUtils_1.flatten(errors));
    }
    // parse lock statement
    declareLock(scope) {
        const builder = {
            scope,
            lock: undefined,
            identifier: undefined,
            value: undefined,
            to: undefined,
        };
        builder.lock = this.previous();
        builder.identifier = this.consumeIdentifierThrow('Expected identifier following lock keyword.', Decl.Lock, builder);
        builder.to = this.consumeTokenThrow('Expected keyword "to" following lock.', Decl.Lock, tokentypes_1.TokenType.to, builder);
        const valueResult = this.expression(Decl.Lock);
        builder.value = valueResult.value;
        this.terminal(Decl.Lock, builder);
        return parseResult_1.nodeResult(new Decl.Lock(builder), valueResult.errors);
    }
    // parse a variable declaration, scoping occurs elseware
    declareVariable(scope) {
        const builder = {
            scope,
            identifier: undefined,
            value: undefined,
            toIs: undefined,
        };
        builder.identifier = this.consumeIdentifierThrow('Expected identifier.', Decl.Var, builder);
        builder.toIs = this.consumeTokensThrow('Expected keyword "to" or "is" following declare.', Decl.Var, [tokentypes_1.TokenType.to, tokentypes_1.TokenType.is], builder);
        const valueResult = this.expression(Decl.Var);
        builder.value = valueResult.value;
        this.terminal(Decl.Var, builder);
        return parseResult_1.nodeResult(new Decl.Var(builder), valueResult.errors);
    }
    // parse statement
    statement() {
        switch (this.peek().type) {
            case tokentypes_1.TokenType.curlyOpen:
                this.advance();
                return this.block();
            case tokentypes_1.TokenType.integer:
            case tokentypes_1.TokenType.double:
            case tokentypes_1.TokenType.true:
            case tokentypes_1.TokenType.false:
            case tokentypes_1.TokenType.identifier:
            case tokentypes_1.TokenType.fileIdentifier:
            case tokentypes_1.TokenType.bracketOpen:
            case tokentypes_1.TokenType.string:
                // note we don't advance the token index here
                // TODO see if there exists a more general solution
                return this.identifierLedStatement();
            case tokentypes_1.TokenType.stage:
            case tokentypes_1.TokenType.clearscreen:
            case tokentypes_1.TokenType.preserve:
            case tokentypes_1.TokenType.reboot:
            case tokentypes_1.TokenType.shutdown:
                this.advance();
                return this.command();
            case tokentypes_1.TokenType.edit:
            case tokentypes_1.TokenType.add:
            case tokentypes_1.TokenType.remove:
                this.advance();
                return this.commandExpression();
            case tokentypes_1.TokenType.unset:
                this.advance();
                return this.unset();
            case tokentypes_1.TokenType.unlock:
                this.advance();
                return this.unlock();
            case tokentypes_1.TokenType.set:
                this.advance();
                return this.set();
            case tokentypes_1.TokenType.atSign:
                this.advance();
                return this.lazyGlobal();
            case tokentypes_1.TokenType.if:
                this.advance();
                return this.ifStmt();
            case tokentypes_1.TokenType.until:
                this.advance();
                return this.until();
            case tokentypes_1.TokenType.from:
                this.advance();
                return this.from();
            case tokentypes_1.TokenType.when:
                this.advance();
                return this.when();
            case tokentypes_1.TokenType.return:
                this.advance();
                return this.returnStmt();
            case tokentypes_1.TokenType.break:
                this.advance();
                return this.breakStmt();
            case tokentypes_1.TokenType.switch:
                this.advance();
                return this.switchStmt();
            case tokentypes_1.TokenType.for:
                this.advance();
                return this.forStmt();
            case tokentypes_1.TokenType.on:
                this.advance();
                return this.on();
            case tokentypes_1.TokenType.toggle:
                this.advance();
                return this.toggle();
            case tokentypes_1.TokenType.wait:
                this.advance();
                return this.wait();
            case tokentypes_1.TokenType.log:
                this.advance();
                return this.log();
            case tokentypes_1.TokenType.copy:
                this.advance();
                return this.copy();
            case tokentypes_1.TokenType.rename:
                this.advance();
                return this.rename();
            case tokentypes_1.TokenType.delete:
                this.advance();
                return this.delete();
            case tokentypes_1.TokenType.run:
                this.advance();
                return this.run();
            case tokentypes_1.TokenType.runPath:
                this.advance();
                return this.runPath();
            case tokentypes_1.TokenType.runOncePath:
                this.advance();
                return this.runOncePath();
            case tokentypes_1.TokenType.compile:
                this.advance();
                return this.compile();
            case tokentypes_1.TokenType.list:
                this.advance();
                return this.list();
            case tokentypes_1.TokenType.print:
                this.advance();
                return this.print();
            case tokentypes_1.TokenType.period:
                return parseResult_1.nodeResult(new Stmt.Empty(this.advance()), []);
            default:
                throw this.error(this.peek(), undefined, 'Unknown statement found', 'Examples: "print "hi"", "LIST.", "RUN "example.ks""');
        }
    }
    // parse a block of statements
    block() {
        const builder = {
            open: undefined,
            stmts: undefined,
            close: undefined,
        };
        builder.open = this.previous();
        const declarations = [];
        const parseErrors = [];
        // while not at end and until closing curly keep parsing statements
        while (!this.check(tokentypes_1.TokenType.curlyClose) && !this.isAtEnd()) {
            const { value, errors } = this.declaration();
            declarations.push(value);
            parseErrors.push(...errors);
        }
        builder.stmts = declarations;
        // check closing curly is found
        const close = this.consumeTokenReturn('Expected "}" to finish statement block', Stmt.Block, [tokentypes_1.TokenType.curlyClose]);
        // throw and bundle inner error if close not found
        if (close.tag === 'parseError') {
            close.inner = parseErrors;
            throw close;
        }
        builder.close = close;
        return parseResult_1.nodeResult(new Stmt.Block(builder), parseErrors);
    }
    // parse an statement lead with a identifier
    identifierLedStatement() {
        const suffix = this.suffixCatch(Stmt.ExprStmt);
        if (this.matchTokens([tokentypes_1.TokenType.on, tokentypes_1.TokenType.off])) {
            const onOff = this.onOff(suffix.value);
            return parseResult_1.nodeResult(onOff.value, arrayUtils_1.flatten([suffix.errors, onOff.errors]));
        }
        this.terminal(Stmt.ExprStmt);
        return parseResult_1.nodeResult(new Stmt.ExprStmt(suffix.value), suffix.errors);
    }
    // parse on off statement
    onOff(suffix) {
        const builder = {
            suffix,
            onOff: this.previous(),
        };
        this.terminal(Stmt.OnOff, builder);
        return parseResult_1.nodeResult(new Stmt.OnOff(builder), []);
    }
    // parse command statement
    command() {
        const command = this.previous();
        // All commands besides preserve can be called as a function
        if (command.type !== tokentypes_1.TokenType.preserve &&
            this.check(tokentypes_1.TokenType.bracketOpen)) {
            // Note this back only exists because of identifier led statement quarks
            this.backup();
            return this.identifierLedStatement();
        }
        this.terminal(Stmt.Command);
        return parseResult_1.nodeResult(new Stmt.Command(command), []);
    }
    // parse command statement
    commandExpression() {
        const builder = {
            command: this.previous(),
            expr: undefined,
        };
        const expr = this.expression(Stmt.CommandExpr);
        builder.expr = expr.value;
        this.terminal(Stmt.CommandExpr, builder);
        return parseResult_1.nodeResult(new Stmt.CommandExpr(builder), expr.errors);
    }
    /**
     * Parse unset statement
     */
    unset() {
        const builder = {
            unset: this.previous(),
            identifier: undefined,
        };
        builder.identifier = this.consumeTokensThrow('Excpeted identifier or "all" following keyword "unset".', Stmt.Unset, [tokentypes_1.TokenType.identifier, tokentypes_1.TokenType.all], builder);
        this.terminal(Stmt.Unset);
        return parseResult_1.nodeResult(new Stmt.Unset(builder), []);
    }
    /**
     * Parse unlock statement
     */
    unlock() {
        const builder = {
            unlock: this.previous(),
            identifier: undefined,
        };
        builder.identifier = this.consumeTokensThrow('Excpeted identifier or "all" following keyword "unlock".', Stmt.Unlock, [tokentypes_1.TokenType.identifier, tokentypes_1.TokenType.all], builder);
        this.terminal(Stmt.Unlock);
        return parseResult_1.nodeResult(new Stmt.Unlock(builder), []);
    }
    /**
     * Parse the set statement
     */
    set() {
        const builder = {
            set: undefined,
            suffix: undefined,
            to: undefined,
            value: undefined,
        };
        builder.set = this.previous();
        const suffix = this.suffixCatch(Stmt.Set);
        builder.suffix = suffix.value;
        builder.to = this.consumeTokenThrow('Expected "to" following keyword "set".', Stmt.Set, tokentypes_1.TokenType.to, builder);
        const valueResult = this.expression(Stmt.Set);
        builder.value = valueResult.value;
        this.terminal(Stmt.Set, builder);
        return parseResult_1.nodeResult(new Stmt.Set(builder), arrayUtils_1.flatten([suffix.errors, valueResult.errors]));
    }
    /**
     * Parse lazy global statement
     */
    lazyGlobal() {
        const builder = {
            atSign: this.previous(),
            lazyGlobal: undefined,
            onOff: undefined,
        };
        builder.lazyGlobal = this.consumeTokenThrow('Expected keyword "lazyGlobal" following @.', Stmt.LazyGlobal, tokentypes_1.TokenType.lazyGlobal, builder);
        builder.onOff = this.consumeTokensThrow('Expected "on" or "off" following lazy global directive.', Stmt.LazyGlobal, [tokentypes_1.TokenType.on, tokentypes_1.TokenType.off], builder);
        this.terminal(Stmt.LazyGlobal, builder);
        return parseResult_1.nodeResult(new Stmt.LazyGlobal(builder), []);
    }
    /**
     * Parse if statement
     */
    ifStmt() {
        const builder = {
            ifToken: undefined,
            condition: undefined,
            body: undefined,
            elseStmt: undefined,
        };
        builder.ifToken = this.previous();
        const conditionResult = this.expression(Stmt.If);
        builder.condition = conditionResult.value;
        const declare = this.declaration();
        builder.body = declare.value;
        this.matchToken(tokentypes_1.TokenType.period);
        // if else if found parse that branch
        if (this.matchToken(tokentypes_1.TokenType.else)) {
            const elseBuilder = {
                elseToken: undefined,
                body: undefined,
            };
            elseBuilder.elseToken = this.previous();
            const elseResult = this.declaration();
            elseBuilder.body = elseResult.value;
            const elseStmt = new Stmt.Else(elseBuilder);
            builder.elseStmt = elseStmt;
            this.matchToken(tokentypes_1.TokenType.period);
            return parseResult_1.nodeResult(new Stmt.If(builder), arrayUtils_1.flatten([conditionResult.errors, declare.errors, elseResult.errors]));
        }
        return parseResult_1.nodeResult(new Stmt.If(builder), arrayUtils_1.flatten([conditionResult.errors, declare.errors]));
    }
    // parse until statement
    until() {
        const builder = {
            until: this.previous(),
            condition: undefined,
            body: undefined,
        };
        const conditionResult = this.expression(Stmt.Until);
        builder.condition = conditionResult.value;
        const declare = this.declaration();
        builder.body = declare.value;
        this.matchToken(tokentypes_1.TokenType.period);
        return parseResult_1.nodeResult(new Stmt.Until(builder), arrayUtils_1.flatten([conditionResult.errors, declare.errors]));
    }
    // parse from statement
    from() {
        const builder = {
            from: this.previous(),
            initializer: undefined,
            until: undefined,
            condition: undefined,
            step: undefined,
            increment: undefined,
            doToken: undefined,
            body: undefined,
        };
        if (this.matchToken(tokentypes_1.TokenType.curlyOpen)) {
            const initResult = this.block();
            builder.initializer = initResult.value;
            builder.until = this.consumeTokenThrow('Expected "until" expression following from.', Stmt.From, tokentypes_1.TokenType.until, builder);
            const conditionResult = this.expression(Stmt.From);
            builder.condition = conditionResult.value;
            builder.step = this.consumeTokenThrow('Expected "step" statement following until.', Stmt.From, tokentypes_1.TokenType.step, builder);
            if (this.matchToken(tokentypes_1.TokenType.curlyOpen)) {
                const incrementResult = this.block();
                builder.increment = incrementResult.value;
                builder.doToken = this.consumeTokenThrow('Expected "do" block following step.', Stmt.From, tokentypes_1.TokenType.do, builder);
                const declare = this.declaration();
                builder.body = declare.value;
                return parseResult_1.nodeResult(new Stmt.From(builder), arrayUtils_1.flatten([
                    initResult.errors,
                    conditionResult.errors,
                    incrementResult.errors,
                    declare.errors,
                ]));
            }
            throw this.error(this.peek(), Stmt.From, 'Expected "{" followed by step block logic.', 'Example: FROM {LOCAL x is 0.} UNTIL x >= 10 STEP { set x to x + 1. } { print x. }');
        }
        throw this.error(this.peek(), Stmt.From, 'Expected "{" followed by initializer logic', 'Example: FROM {LOCAL x is 0.} UNTIL x >= 10 STEP { set x to x + 1. } { print x. }');
    }
    // parse when statement
    when() {
        const builder = {
            when: this.previous(),
            condition: undefined,
            then: undefined,
            body: undefined,
        };
        const conditionResult = this.expression(Stmt.When);
        builder.condition = conditionResult.value;
        builder.then = this.consumeTokenThrow('Expected "then" following "when" condition.', Stmt.When, tokentypes_1.TokenType.then, builder);
        const declare = this.declaration();
        builder.body = declare.value;
        this.matchToken(tokentypes_1.TokenType.period);
        return parseResult_1.nodeResult(new Stmt.When(builder), arrayUtils_1.flatten([conditionResult.errors, declare.errors]));
    }
    // parse return statement
    returnStmt() {
        const builder = {
            returnToken: this.previous(),
            value: undefined,
        };
        const valueResult = !this.check(tokentypes_1.TokenType.period)
            ? this.expression(Stmt.Return)
            : undefined;
        let errors = [];
        this.terminal(Stmt.Return, builder);
        if (!typeGuards_1.empty(valueResult)) {
            builder.value = valueResult.value;
            errors = valueResult.errors;
        }
        return parseResult_1.nodeResult(new Stmt.Return(builder), errors);
    }
    // parse return statement
    breakStmt() {
        const breakToken = this.previous();
        this.terminal(Stmt.Break);
        return parseResult_1.nodeResult(new Stmt.Break(breakToken), []);
    }
    // parse switch statement
    switchStmt() {
        const builder = {
            switchToken: this.previous(),
            to: undefined,
            target: undefined,
        };
        builder.to = this.consumeTokenThrow('Expected "to" following keyword "switch".', Stmt.Switch, tokentypes_1.TokenType.to, builder);
        const targetResult = this.expression(Stmt.Switch);
        builder.target = targetResult.value;
        this.terminal(Stmt.Switch, builder);
        return parseResult_1.nodeResult(new Stmt.Switch(builder), targetResult.errors);
    }
    // parse for statement
    forStmt() {
        const builder = {
            forToken: undefined,
            element: undefined,
            inToken: undefined,
            collection: undefined,
            body: undefined,
        };
        builder.forToken = this.previous();
        builder.element = this.consumeIdentifierThrow('Expected identifier. following keyword "for"', Stmt.For, builder);
        builder.inToken = this.consumeTokenThrow('Expected "in" after "for" loop variable.', Stmt.For, tokentypes_1.TokenType.in, builder);
        const collection = this.suffixCatch(Stmt.For);
        builder.collection = collection.value;
        const declare = this.declaration();
        builder.body = declare.value;
        this.matchToken(tokentypes_1.TokenType.period);
        return parseResult_1.nodeResult(new Stmt.For(builder), arrayUtils_1.flatten([collection.errors, declare.errors]));
    }
    // parse on statement
    on() {
        const builder = {
            on: this.previous(),
            suffix: undefined,
            body: undefined,
        };
        const suffix = this.suffixCatch(Stmt.On);
        builder.suffix = suffix.value;
        const declare = this.declaration();
        builder.body = declare.value;
        return parseResult_1.nodeResult(new Stmt.On(builder), arrayUtils_1.flatten([suffix.errors, declare.errors]));
    }
    // parse toggle statement
    toggle() {
        const builder = {
            toggle: this.previous(),
            suffix: undefined,
        };
        const suffix = this.suffixCatch(Stmt.Toggle);
        builder.suffix = suffix.value;
        this.terminal(Stmt.Toggle, builder);
        return parseResult_1.nodeResult(new Stmt.Toggle(builder), suffix.errors);
    }
    // parse wait statement
    wait() {
        const builder = {
            wait: this.previous(),
            until: undefined,
            expr: undefined,
        };
        builder.until = this.matchToken(tokentypes_1.TokenType.until)
            ? this.previous()
            : undefined;
        const expr = this.expression(Stmt.Wait);
        builder.expr = expr.value;
        this.terminal(Stmt.Wait, builder);
        return parseResult_1.nodeResult(new Stmt.Wait(builder), expr.errors);
    }
    // parse log statement
    log() {
        const builder = {
            log: this.previous(),
            expr: undefined,
            to: undefined,
            target: undefined,
        };
        const expr = this.expression(Stmt.Log);
        builder.expr = expr.value;
        builder.to = this.consumeTokenThrow('Expected "to" following "log" expression.', Stmt.Log, tokentypes_1.TokenType.to, builder);
        const targetResult = this.expression(Stmt.Log);
        builder.target = targetResult.value;
        this.terminal(Stmt.Log, builder);
        return parseResult_1.nodeResult(new Stmt.Log(builder), arrayUtils_1.flatten([expr.errors, targetResult.errors]));
    }
    // parse copy statement
    copy() {
        const builder = {
            copy: this.previous(),
            target: undefined,
            toFrom: undefined,
            destination: undefined,
        };
        builder.copy = this.previous();
        const targetResult = this.expression(Stmt.Copy);
        builder.target = targetResult.value;
        builder.toFrom = this.consumeTokensThrow('Expected "to" or "from" following "copy" expression.', Stmt.Copy, [tokentypes_1.TokenType.from, tokentypes_1.TokenType.to], builder);
        const destinationResult = this.expression(Stmt.Copy);
        builder.destination = destinationResult.value;
        this.terminal(Stmt.Copy, builder);
        return parseResult_1.nodeResult(new Stmt.Copy(builder), arrayUtils_1.flatten([targetResult.errors, destinationResult.errors]));
    }
    // parse rename statement
    rename() {
        const builder = {
            rename: this.previous(),
            fileVolume: undefined,
            identifier: undefined,
            target: undefined,
            to: undefined,
            alternative: undefined,
        };
        builder.fileVolume = this.consumeTokensThrow('Expected file or volume following keyword "rename"', Stmt.Rename, [tokentypes_1.TokenType.volume, tokentypes_1.TokenType.file], builder);
        builder.identifier = this.consumeTokensThrow('Expected identifier or file identifier following keyword "rename"', Stmt.Rename, [tokentypes_1.TokenType.identifier, tokentypes_1.TokenType.fileIdentifier], builder);
        const targetResult = this.expression(Stmt.Rename);
        builder.target = targetResult.value;
        builder.to = this.consumeTokenThrow('Expected "to" following keyword "rename".', Stmt.Rename, tokentypes_1.TokenType.to, builder);
        const alternativeResult = this.expression(Stmt.Rename);
        builder.alternative = alternativeResult.value;
        this.terminal(Stmt.Rename, builder);
        return parseResult_1.nodeResult(new Stmt.Rename(builder), arrayUtils_1.flatten([targetResult.errors, alternativeResult.errors]));
    }
    // parse delete statement
    delete() {
        const builder = {
            deleteToken: this.previous(),
            target: undefined,
            from: undefined,
            volume: undefined,
        };
        const targetResult = this.expression(Stmt.Delete);
        builder.target = targetResult.value;
        if (this.matchToken(tokentypes_1.TokenType.from)) {
            builder.from = this.previous();
            const volumeResult = this.expression(Stmt.Delete);
            builder.volume = volumeResult.value;
            this.terminal(Stmt.Delete, builder);
            return parseResult_1.nodeResult(new Stmt.Delete(builder), arrayUtils_1.flatten([targetResult.errors, volumeResult.errors]));
        }
        this.terminal(Stmt.Delete);
        return parseResult_1.nodeResult(new Stmt.Delete(builder), targetResult.errors);
    }
    /**
     * Parse run statement
     */
    run() {
        const builder = {
            run: this.previous(),
            once: undefined,
            identifier: undefined,
            open: undefined,
            args: undefined,
            close: undefined,
            on: undefined,
            expr: undefined,
        };
        const errors = [];
        builder.once = this.matchToken(tokentypes_1.TokenType.once)
            ? this.previous()
            : undefined;
        builder.identifier = this.consumeTokensThrow('Expected string or fileidentifier following keyword "run".', Stmt.Run, [tokentypes_1.TokenType.string, tokentypes_1.TokenType.identifier, tokentypes_1.TokenType.fileIdentifier]);
        // parse arguments if found
        if (this.matchToken(tokentypes_1.TokenType.bracketOpen)) {
            builder.open = this.previous();
            const argsResult = this.arguments(Stmt.Run);
            builder.args = argsResult.value;
            errors.push(...argsResult.errors);
            builder.close = this.consumeTokenThrow('Expected ")" after "run" arguments.', Stmt.Run, tokentypes_1.TokenType.bracketClose);
        }
        // parse arguments if found
        if (this.matchToken(tokentypes_1.TokenType.on)) {
            builder.on = this.previous();
            const expr = this.expression(Stmt.Run);
            errors.push(...expr.errors);
            builder.expr = expr.value;
        }
        this.terminal(Stmt.Run);
        // handle all the cases
        return this.addRunStmts(new Stmt.Run(builder), []);
    }
    /**
     * Parse run path statement
     */
    runPath() {
        const builder = {
            runPath: this.previous(),
            open: undefined,
            expr: undefined,
            args: undefined,
            close: undefined,
        };
        builder.open = this.consumeTokenThrow('Expected "(" after keyword "runPath".', Stmt.RunPath, tokentypes_1.TokenType.bracketOpen);
        const exprResult = this.expression(Stmt.RunPath);
        builder.expr = exprResult.value;
        const args = this.matchToken(tokentypes_1.TokenType.comma)
            ? this.arguments(Stmt.RunPath)
            : undefined;
        builder.close = this.consumeTokenThrow('Expected ")" after runPath arguments.', Stmt.RunPath, tokentypes_1.TokenType.bracketClose);
        this.terminal(Stmt.RunPath);
        if (typeGuards_1.empty(args)) {
            return this.addRunStmts(new Stmt.RunPath(builder), exprResult.errors);
        }
        builder.args = args.value;
        return this.addRunStmts(new Stmt.RunPath(builder), arrayUtils_1.flatten([exprResult.errors, args.errors]));
    }
    /**
     * Parse run path once statement
     */
    runOncePath() {
        const builder = {
            runPath: this.previous(),
            open: undefined,
            expr: undefined,
            args: undefined,
            close: undefined,
        };
        const errors = [];
        builder.open = this.consumeTokenThrow('Expected "(" after keyword "runPathOnce".', Stmt.RunOncePath, tokentypes_1.TokenType.bracketOpen);
        const exprResult = this.expression(Stmt.RunOncePath);
        builder.expr = exprResult.value;
        const args = this.matchToken(tokentypes_1.TokenType.comma)
            ? this.arguments(Stmt.RunOncePath)
            : undefined;
        builder.close = this.consumeTokenThrow('Expected ")" after runPathOnce arugments.', Stmt.RunOncePath, tokentypes_1.TokenType.bracketClose);
        this.terminal(Stmt.RunOncePath);
        if (!typeGuards_1.empty(args)) {
            builder.args = args.value;
            errors.push(...args.errors);
        }
        return this.addRunStmts(new Stmt.RunOncePath(builder), arrayUtils_1.flatten([exprResult.errors, errors]));
    }
    /**
     * parse a compile statement
     */
    compile() {
        const builder = {
            compile: this.previous(),
            target: undefined,
            to: undefined,
            destination: undefined,
        };
        const targetResult = this.expression(Stmt.Compile);
        builder.target = targetResult.value;
        const errors = targetResult.errors;
        if (this.matchToken(tokentypes_1.TokenType.to)) {
            builder.to = this.previous();
            const destinationResult = this.expression(Stmt.Compile);
            builder.destination = destinationResult.value;
            errors.push(...destinationResult.errors);
        }
        this.terminal(Stmt.Compile);
        return parseResult_1.nodeResult(new Stmt.Compile(builder), errors);
    }
    /**
     * Parse a list statement
     */
    list() {
        const builder = {
            list: this.previous(),
            collection: undefined,
            inToken: undefined,
            target: undefined,
        };
        if (this.matchIdentifier()) {
            builder.collection = this.previous();
            if (this.matchToken(tokentypes_1.TokenType.in)) {
                builder.inToken = this.previous();
                builder.target = this.consumeIdentifierThrow('Expected identifier after "in" keyword in "list" command', Stmt.List, builder);
            }
        }
        this.terminal(Stmt.List, builder);
        return parseResult_1.nodeResult(new Stmt.List(builder), []);
    }
    /**
     * Parse a print statement
     */
    print() {
        const builder = {
            print: undefined,
            expr: undefined,
            at: undefined,
            open: undefined,
            x: undefined,
            y: undefined,
            close: undefined,
        };
        builder.print = this.previous();
        // if we find function variant of print use that instead
        if (this.check(tokentypes_1.TokenType.bracketOpen)) {
            // Note this back only exists because of identifier led statement quarks
            this.backup();
            return this.identifierLedStatement();
        }
        const expr = this.expression(Stmt.Print);
        builder.expr = expr.value;
        if (this.matchToken(tokentypes_1.TokenType.at)) {
            builder.at = this.previous();
            builder.open = this.consumeTokenThrow('Expected "(".', Stmt.Print, tokentypes_1.TokenType.bracketOpen, builder);
            const xResult = this.expression(Stmt.Print);
            builder.x = xResult.value;
            this.consumeTokenThrow('Expected ",".', Stmt.Print, tokentypes_1.TokenType.comma, builder);
            const yResult = this.expression(Stmt.Print);
            builder.y = yResult.value;
            builder.close = this.consumeTokenThrow('Expected ")".', Stmt.Print, tokentypes_1.TokenType.bracketClose, builder);
            this.terminal(Stmt.Print);
            return parseResult_1.nodeResult(new Stmt.Print(builder), arrayUtils_1.flatten([expr.errors, xResult.errors, yResult.errors]));
        }
        this.terminal(Stmt.Print, builder);
        return parseResult_1.nodeResult(new Stmt.Print(builder), expr.errors);
    }
    /**
     * Parse an expression
     * @param stmt the statment context
     */
    expression(stmt) {
        try {
            // match anonymous function
            if (this.matchToken(tokentypes_1.TokenType.curlyOpen)) {
                return this.lambda();
            }
            // other match conditional
            return this.or();
        }
        catch (error) {
            if (error instanceof parserError_1.ParseError) {
                error.failed.stmt = stmt;
                throw error;
            }
            throw error;
        }
    }
    // parse or expression
    or() {
        return this.binaryExpression(this.andBind, [tokentypes_1.TokenType.or]);
    }
    // parse and expression
    and() {
        return this.binaryExpression(this.eqaulityBind, [tokentypes_1.TokenType.and]);
    }
    // parse equality expression
    equality() {
        return this.binaryExpression(this.comparisonBind, [
            tokentypes_1.TokenType.equal,
            tokentypes_1.TokenType.notEqual,
        ]);
    }
    // parse comparison expression
    comparison() {
        return this.binaryExpression(this.additionBind, [
            tokentypes_1.TokenType.less,
            tokentypes_1.TokenType.greater,
            tokentypes_1.TokenType.lessEqual,
            tokentypes_1.TokenType.greaterEqual,
        ]);
    }
    // parse addition expression
    addition() {
        return this.binaryExpression(this.multiplicationBind, [
            tokentypes_1.TokenType.plus,
            tokentypes_1.TokenType.minus,
        ]);
    }
    // parse multiplication expression
    multiplication() {
        return this.binaryExpression(this.unaryBind, [
            tokentypes_1.TokenType.multi,
            tokentypes_1.TokenType.div,
        ]);
    }
    // parse unary expression
    unary() {
        // if unary token found parse as unary
        if (this.matchTokens([
            tokentypes_1.TokenType.plus,
            tokentypes_1.TokenType.minus,
            tokentypes_1.TokenType.not,
            tokentypes_1.TokenType.defined,
        ])) {
            const operator = this.previous();
            const unary = this.unary();
            return parseResult_1.nodeResult(new Expr.Unary(operator, unary.value), unary.errors);
        }
        // else parse plain factor
        return this.factor();
    }
    // parse factor expression
    factor() {
        // parse suffix
        let expr = this.suffix();
        // parse seqeunce of factors if they exist
        while (this.matchToken(tokentypes_1.TokenType.power)) {
            const power = this.previous();
            const exponenent = this.suffix();
            expr = parseResult_1.nodeResult(new Expr.Factor(expr.value, power, exponenent.value), exponenent.errors);
        }
        return expr;
    }
    // parse suffix for use in stmt directly, will catch
    suffixCatch(stmt) {
        try {
            return this.suffix();
        }
        catch (error) {
            if (error instanceof parserError_1.ParseError) {
                error.failed.stmt = stmt;
                throw error;
            }
            throw error;
        }
    }
    // parse suffix
    suffix() {
        const suffixTerm = this.suffixTerm(false);
        const suffix = new Expr.Suffix(suffixTerm.value);
        const errors = suffixTerm.errors;
        // check to see if expr is really a suffix
        if (this.matchToken(tokentypes_1.TokenType.colon)) {
            // parse first suffix term
            let colon = this.previous();
            let suffixTerm = this.suffixTerm(true);
            errors.push(...suffixTerm.errors);
            // patch suffix with new trailer
            const suffixTrailer = new SuffixTerm.SuffixTrailer(suffixTerm.value);
            suffix.colon = colon;
            suffix.trailer = suffixTrailer;
            let current = suffixTrailer;
            // while there are more trailer parse down
            while (this.matchToken(tokentypes_1.TokenType.colon)) {
                colon = this.previous();
                suffixTerm = this.suffixTerm(true);
                const suffixTrailer = new SuffixTerm.SuffixTrailer(suffixTerm.value);
                // patch curren trailer with trailer update current
                current.colon = colon;
                current.trailer = suffixTrailer;
                current = suffixTrailer;
                errors.push(...suffixTerm.errors);
            }
        }
        return parseResult_1.nodeResult(suffix, errors);
    }
    // parse suffix term expression
    suffixTerm(isTrailer) {
        // parse atom
        const atom = this.atom(isTrailer);
        const trailers = [];
        const parseErrors = atom.errors;
        const isValid = !(atom.value instanceof SuffixTerm.Invalid);
        // parse any trailers that exist
        while (isValid) {
            if (this.matchToken(tokentypes_1.TokenType.arrayIndex)) {
                const index = this.arrayIndex();
                trailers.push(index.value);
                parseErrors.push(...index.errors);
            }
            else if (this.matchToken(tokentypes_1.TokenType.squareOpen)) {
                const bracket = this.arrayBracket();
                trailers.push(bracket.value);
                parseErrors.push(...bracket.errors);
            }
            else if (this.matchToken(tokentypes_1.TokenType.bracketOpen)) {
                const trailer = this.functionTrailer();
                trailers.push(trailer.value);
                parseErrors.push(...trailer.errors);
            }
            else if (this.matchToken(tokentypes_1.TokenType.atSign)) {
                trailers.push(new SuffixTerm.Delegate(this.previous()));
                break;
            }
            else {
                break;
            }
        }
        return parseResult_1.nodeResult(new SuffixTerm.SuffixTerm(atom.value, trailers), parseErrors);
    }
    // function call
    functionTrailer() {
        const open = this.previous();
        const args = this.arguments(SuffixTerm.Call);
        const close = this.consumeTokenThrow('Expect ")" after arguments.', SuffixTerm.Call, tokentypes_1.TokenType.bracketClose);
        return parseResult_1.nodeResult(new SuffixTerm.Call(open, args.value, close), args.errors);
    }
    // get an argument list
    partialArgumentsCount() {
        let count = -1;
        if (!this.check(tokentypes_1.TokenType.bracketClose)) {
            do {
                count += 1;
                if (this.isAtEnd())
                    break;
                this.expression();
            } while (this.matchToken(tokentypes_1.TokenType.comma));
        }
        return parseResult_1.nodeResult(count < 0 ? 0 : count, []);
    }
    // get an argument list
    arguments(context) {
        const args = [];
        const errors = [];
        if (!this.isAtEnd() && !this.check(tokentypes_1.TokenType.bracketClose)) {
            do {
                // if we are expecting an argument but find a closing
                // braket we report an error and stop arguments
                if (this.check(tokentypes_1.TokenType.bracketClose)) {
                    args.push(new Expr.Invalid([this.previous()]));
                    errors.push([this.error(this.previous(), context, 'Expected another argument.')]);
                    break;
                }
                const arg = this.expression();
                args.push(arg.value);
                errors.push(arg.errors);
            } while (this.matchToken(tokentypes_1.TokenType.comma));
        }
        return parseResult_1.nodeResult(args, arrayUtils_1.flatten(errors));
    }
    // generate array bracket expression
    arrayBracket() {
        const open = this.previous();
        const index = this.expression();
        const close = this.consumeTokenThrow('Expected "]" at end of array index.', SuffixTerm.ArrayBracket, tokentypes_1.TokenType.squareClose);
        return parseResult_1.nodeResult(new SuffixTerm.ArrayBracket(open, index.value, close), index.errors);
    }
    // generate array index expression
    arrayIndex() {
        const indexer = this.previous();
        // check for integer or identifier
        const index = this.consumeTokensThrow('Expected integer or identifer.', SuffixTerm.ArrayIndex, [tokentypes_1.TokenType.integer, tokentypes_1.TokenType.identifier]);
        return parseResult_1.nodeResult(new SuffixTerm.ArrayIndex(indexer, index), []);
    }
    // parse lambda expression
    lambda() {
        const { value, errors } = this.block();
        return parseResult_1.nodeResult(new Expr.Lambda(value), errors);
    }
    // match atom expressions literals, identifers, list, and parenthesis
    atom(isTrailer) {
        // match all literals
        if (this.matchTokens([
            tokentypes_1.TokenType.false,
            tokentypes_1.TokenType.true,
            tokentypes_1.TokenType.fileIdentifier,
            tokentypes_1.TokenType.string,
            tokentypes_1.TokenType.integer,
            tokentypes_1.TokenType.double,
        ])) {
            return parseResult_1.nodeResult(new SuffixTerm.Literal(this.previous()), []);
        }
        // match identifiers TODO identifier all keywords that can be used here
        if (tokentypes_1.isValidIdentifier(this.peek().type)) {
            return parseResult_1.nodeResult(new SuffixTerm.Identifier(this.advance()), []);
        }
        // match grouping expression
        if (this.matchToken(tokentypes_1.TokenType.bracketOpen)) {
            const open = this.previous();
            const expr = this.expression();
            const close = this.consumeTokenThrow('Expect ")" after expression', SuffixTerm.Grouping, tokentypes_1.TokenType.bracketClose);
            return parseResult_1.nodeResult(new SuffixTerm.Grouping(open, expr.value, close), expr.errors);
        }
        if (isTrailer) {
            const previous = this.previous();
            return parseResult_1.nodeResult(new SuffixTerm.Invalid(previous.end), [
                this.error(previous, undefined, 'Expected suffix', undefined),
            ]);
        }
        // valid expression not found
        throw this.error(this.peek(), undefined, 'Expected expression.');
    }
    addRunStmts(stmt, errors) {
        this.runStmts.push(stmt);
        return parseResult_1.nodeResult(stmt, errors);
    }
    /**
     * Check the the statement is terminated
     * @param failed failed constructor context
     * @param partialNode partially constructed node
     */
    terminal(failed, partialNode) {
        return this.consumeTokenThrow('Expected ".".', failed, tokentypes_1.TokenType.period, partialNode);
    }
    /**
     * Check for any valid identifier throws
     * error if incorrect token is found
     * @param message message if token is not found
     * @param failed failed construcotor
     * @param partialNode partially constructed node
     */
    consumeIdentifierThrow(message, failed, partialNode) {
        if (this.matchIdentifier())
            return this.previous();
        throw this.error(this.previous(), failed, message, undefined, partialNode);
    }
    /**
     * Consume current token if it matches the supplied type
     * throw error if incorrect token is found
     * @param message error message
     * @param failed fail context
     * @param tokenType token type
     * @param partialNode partially constructed node
     */
    consumeTokenThrow(message, failed, tokenType, partialNode) {
        if (this.matchToken(tokenType))
            return this.previous();
        throw this.error(this.previous(), failed, message, undefined, partialNode);
    }
    /**
     * Consume current token if it matches one of the supplied types
     * throw error if incorrect token is found
     * @param message error message
     * @param failed fail context
     * @param tokenTypes token types
     * @param partialNode partially constructed node
     */
    consumeTokensThrow(message, failed, tokenTypes, partialNode) {
        if (this.matchTokens(tokenTypes))
            return this.previous();
        throw this.error(this.previous(), failed, message, undefined, partialNode);
    }
    /**
     * Consume current token if it matches one of the supplied types
     * returns error if incorrect token is found
     * @param message error message
     * @param failed fail context
     * @param tokenTypes token types
     * @param partialNode partially constructed node
     */
    consumeTokenReturn(message, failed, tokenType, partialNode) {
        if (this.matchTokens(tokenType))
            return this.previous();
        return this.error(this.previous(), failed, message, undefined, partialNode);
    }
    // was identifier matched
    matchIdentifier() {
        const found = this.identifierCheck();
        if (found)
            this.advance();
        return found;
    }
    matchToken(type) {
        const found = this.check(type);
        if (found)
            this.advance();
        return found;
    }
    // determine if current token matches a set of tokens
    matchTokens(types) {
        if (this.isAtEnd())
            return false;
        const current = this.peek().type;
        for (let i = 0; i < types.length; i += 1) {
            if (current === types[i]) {
                this.advance();
                return true;
            }
        }
        return false;
    }
    // check if current token can be an identifier
    identifierCheck() {
        if (this.isAtEnd())
            return false;
        return tokentypes_1.isValidIdentifier(this.peek().type);
    }
    // check if current token matches expected type
    check(tokenType) {
        if (this.isAtEnd())
            return false;
        return this.peek().type === tokenType;
    }
    // check if the next token matches expected type
    checkNext(tokenType) {
        const nextToken = this.peekNext();
        if (typeGuards_1.empty(nextToken))
            return false;
        return nextToken.type === tokenType;
    }
    // return current token and advance
    advance() {
        if (!this.isAtEnd()) {
            this.current += 1;
        }
        return this.previous();
    }
    // TODO REMOVE ME
    // return current token and backup
    backup() {
        const current = this.peek();
        if (this.current !== 0) {
            this.current -= 1;
        }
        return current;
    }
    // is parse at the end of file
    isAtEnd() {
        return this.peek().type === tokentypes_1.TokenType.eof;
    }
    // peek current token
    peek() {
        return this.tokens[this.current];
    }
    // peek next token
    peekNext() {
        const nextToken = this.tokens[this.current + 1];
        if (typeGuards_1.empty(nextToken) || nextToken.type === tokentypes_1.TokenType.eof)
            return undefined;
        return nextToken;
    }
    // retrieve previous token
    previous() {
        return this.tokens[this.current - 1];
    }
    // report parse error
    error(token, failed, message, moreInfo, partialNode) {
        if (typeGuards_1.empty(failed)) {
            return new parserError_1.ParseError(token, parserError_1.failedUnknown(), message, moreInfo, partialNode);
        }
        if (failed.prototype instanceof Expr.Expr) {
            return new parserError_1.ParseError(token, parserError_1.failedExpr(failed), message, moreInfo, partialNode);
        }
        if (failed.prototype instanceof Stmt.Stmt) {
            return new parserError_1.ParseError(token, parserError_1.failedStmt(failed), message, moreInfo, partialNode);
        }
        return new parserError_1.ParseError(token, parserError_1.failedUnknown(), message, moreInfo, partialNode);
    }
    // attempt to synchronize parser
    synchronize(failed) {
        // need to confirm this is the only case
        if (typeGuards_1.empty(failed.stmt)) {
            this.advance();
        }
        while (!this.isAtEnd()) {
            if (this.previous().type === tokentypes_1.TokenType.period)
                return;
            switch (this.peek().type) {
                // declarations
                case tokentypes_1.TokenType.declare:
                case tokentypes_1.TokenType.function:
                case tokentypes_1.TokenType.parameter:
                case tokentypes_1.TokenType.lock:
                case tokentypes_1.TokenType.local:
                case tokentypes_1.TokenType.global:
                // commands
                case tokentypes_1.TokenType.stage:
                case tokentypes_1.TokenType.clearscreen:
                case tokentypes_1.TokenType.preserve:
                case tokentypes_1.TokenType.reboot:
                case tokentypes_1.TokenType.shutdown:
                // command expressions
                case tokentypes_1.TokenType.edit:
                case tokentypes_1.TokenType.add:
                case tokentypes_1.TokenType.remove:
                // variable statements
                case tokentypes_1.TokenType.unset:
                case tokentypes_1.TokenType.unlock:
                case tokentypes_1.TokenType.set:
                // control flow
                case tokentypes_1.TokenType.if:
                case tokentypes_1.TokenType.until:
                case tokentypes_1.TokenType.from:
                case tokentypes_1.TokenType.when:
                case tokentypes_1.TokenType.return:
                case tokentypes_1.TokenType.break:
                case tokentypes_1.TokenType.switch:
                case tokentypes_1.TokenType.for:
                case tokentypes_1.TokenType.on:
                case tokentypes_1.TokenType.toggle:
                case tokentypes_1.TokenType.wait:
                // io statements
                case tokentypes_1.TokenType.log:
                case tokentypes_1.TokenType.copy:
                case tokentypes_1.TokenType.rename:
                case tokentypes_1.TokenType.delete:
                case tokentypes_1.TokenType.run:
                case tokentypes_1.TokenType.runPath:
                case tokentypes_1.TokenType.runOncePath:
                case tokentypes_1.TokenType.compile:
                case tokentypes_1.TokenType.list:
                case tokentypes_1.TokenType.print:
                // close scope
                case tokentypes_1.TokenType.curlyClose:
                    return;
                default:
                    break;
            }
            this.advance();
        }
    }
}
exports.Parser = Parser;
//# sourceMappingURL=parser.js.map