"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const types_1 = require("./types");
const typeGuards_1 = require("../utilities/typeGuards");
const base_1 = require("./base");
const toStringUtils_1 = require("./toStringUtils");
const arrayUtils_1 = require("../utilities/arrayUtils");
const positionUtils_1 = require("../utilities/positionUtils");
/**
 * Statement base class
 */
class Stmt extends base_1.NodeBase {
    /**
     * Return the tree node type of statement
     */
    get tag() {
        return types_1.SyntaxKind.stmt;
    }
}
exports.Stmt = Stmt;
class Invalid extends Stmt {
    constructor(pos, tokens, partial) {
        super();
        this.pos = pos;
        this.tokens = tokens;
        this.partial = partial;
    }
    toLines() {
        return [this.tokens.map(t => t.lexeme).join(' ')];
    }
    get start() {
        return this.tokens.length > 0
            ? this.tokens[0].start
            : this.pos;
    }
    get end() {
        return this.tokens.length > 0
            ? this.tokens[this.tokens.length - 1].end
            : this.pos;
    }
    get ranges() {
        if (typeGuards_1.empty(this.partial)) {
            return this.tokens;
        }
        // order each piece of the partial by it's range
        let segments = Object.values(this.partial)
            .filter(typeGuards_1.notEmpty)
            .sort(positionUtils_1.rangeOrder);
        // attempt to add tokens to segments
        for (const token of this.tokens) {
            let overlap = false;
            // check for overlap
            for (const segment of segments) {
                if (positionUtils_1.rangeContains(segment, token)) {
                    overlap = true;
                    break;
                }
            }
            // if no overall add to segments
            if (!overlap) {
                segments.push(token);
            }
        }
        // reorder segments
        segments = segments.sort(positionUtils_1.rangeOrder);
        return segments;
    }
    pass(visitor) {
        return visitor.passStmtInvalid(this);
    }
    accept(visitor) {
        return visitor.visitStmtInvalid(this);
    }
}
exports.Invalid = Invalid;
class Block extends Stmt {
    constructor(builder) {
        super();
        this.open = typeGuards_1.unWrap(builder.open);
        this.stmts = typeGuards_1.unWrap(builder.stmts);
        this.close = typeGuards_1.unWrap(builder.close);
    }
    toLines() {
        const lines = arrayUtils_1.flatten(this.stmts.map(stmt => stmt.toLines()));
        if (lines.length === 0) {
            return [`${this.open.lexeme} ${this.close.lexeme}`];
        }
        if (lines.length === 1) {
            return [`${this.open.lexeme} ${lines[0]} ${this.close.lexeme}`];
        }
        return [`${this.open.lexeme}`].concat(...lines.map(line => `    ${line}`), `${this.close.lexeme}`);
    }
    get start() {
        return this.open.start;
    }
    get end() {
        return this.close.end;
    }
    get ranges() {
        return [this.open, ...this.stmts, this.close];
    }
    pass(visitor) {
        return visitor.passBlock(this);
    }
    accept(visitor) {
        return visitor.visitBlock(this);
    }
}
exports.Block = Block;
class ExprStmt extends Stmt {
    constructor(suffix) {
        super();
        this.suffix = suffix;
    }
    toLines() {
        const suffixLines = this.suffix.toLines();
        suffixLines[suffixLines.length - 1] = `${suffixLines[suffixLines.length - 1]}.`;
        return suffixLines;
    }
    get start() {
        return this.suffix.start;
    }
    get end() {
        return this.suffix.end;
    }
    get ranges() {
        return [this.suffix];
    }
    pass(visitor) {
        return visitor.passExpr(this);
    }
    accept(visitor) {
        return visitor.visitExpr(this);
    }
}
exports.ExprStmt = ExprStmt;
class OnOff extends Stmt {
    constructor(builder) {
        super();
        this.suffix = typeGuards_1.unWrap(builder.suffix);
        this.onOff = typeGuards_1.unWrap(builder.onOff);
    }
    toLines() {
        const lines = this.suffix.toLines();
        lines[lines.length - 1] = `${lines[lines.length - 1]} ${this.onOff.lexeme}.`;
        return lines;
    }
    get start() {
        return this.suffix.start;
    }
    get end() {
        return this.onOff.end;
    }
    get ranges() {
        return [this.suffix, this.onOff];
    }
    pass(visitor) {
        return visitor.passOnOff(this);
    }
    accept(visitor) {
        return visitor.visitOnOff(this);
    }
}
exports.OnOff = OnOff;
class Command extends Stmt {
    constructor(command) {
        super();
        this.command = command;
    }
    toLines() {
        return [`${this.command.lexeme}.`];
    }
    get start() {
        return this.command.start;
    }
    get end() {
        return this.command.end;
    }
    get ranges() {
        return [this.command];
    }
    pass(visitor) {
        return visitor.passCommand(this);
    }
    accept(visitor) {
        return visitor.visitCommand(this);
    }
}
exports.Command = Command;
class CommandExpr extends Stmt {
    constructor(builder) {
        super();
        this.command = typeGuards_1.unWrap(builder.command);
        this.expr = typeGuards_1.unWrap(builder.expr);
    }
    toLines() {
        const lines = this.expr.toLines();
        lines[0] = `${this.command.lexeme} ${lines[0]}`;
        lines[lines.length - 1] = `${lines[lines.length - 1]}.`;
        return lines;
    }
    get start() {
        return this.command.start;
    }
    get end() {
        return this.expr.end;
    }
    get ranges() {
        return [this.command, this.expr];
    }
    pass(visitor) {
        return visitor.passCommandExpr(this);
    }
    accept(visitor) {
        return visitor.visitCommandExpr(this);
    }
}
exports.CommandExpr = CommandExpr;
class Unset extends Stmt {
    constructor(builder) {
        super();
        this.unset = typeGuards_1.unWrap(builder.unset);
        this.identifier = typeGuards_1.unWrap(builder.identifier);
    }
    toLines() {
        return [`${this.unset.lexeme} ${this.identifier.lexeme}.`];
    }
    get start() {
        return this.unset.start;
    }
    get end() {
        return this.identifier.end;
    }
    get ranges() {
        return [this.unset, this.identifier];
    }
    pass(visitor) {
        return visitor.passUnset(this);
    }
    accept(visitor) {
        return visitor.visitUnset(this);
    }
}
exports.Unset = Unset;
class Unlock extends Stmt {
    constructor(builder) {
        super();
        this.unlock = typeGuards_1.unWrap(builder.unlock);
        this.identifier = typeGuards_1.unWrap(builder.identifier);
    }
    toLines() {
        return [`${this.unlock.lexeme} ${this.identifier.lexeme}.`];
    }
    get start() {
        return this.unlock.start;
    }
    get end() {
        return this.identifier.end;
    }
    get ranges() {
        return [this.unlock, this.identifier];
    }
    pass(visitor) {
        return visitor.passUnlock(this);
    }
    accept(visitor) {
        return visitor.visitUnlock(this);
    }
}
exports.Unlock = Unlock;
class Set extends Stmt {
    constructor(builder) {
        super();
        this.set = typeGuards_1.unWrap(builder.set);
        this.suffix = typeGuards_1.unWrap(builder.suffix);
        this.to = typeGuards_1.unWrap(builder.to);
        this.value = typeGuards_1.unWrap(builder.value);
    }
    toLines() {
        const suffixLines = this.suffix.toLines();
        const valueLines = this.value.toLines();
        suffixLines[0] = `${this.set.lexeme} ${suffixLines[0]}`;
        const lines = toStringUtils_1.joinLines(` ${this.to.lexeme} `, suffixLines, valueLines);
        lines[lines.length - 1] = `${lines[lines.length - 1]}.`;
        return lines;
    }
    get start() {
        return this.set.start;
    }
    get end() {
        return this.value.end;
    }
    get ranges() {
        return [this.set, this.suffix, this.to, this.value];
    }
    pass(visitor) {
        return visitor.passSet(this);
    }
    accept(visitor) {
        return visitor.visitSet(this);
    }
}
exports.Set = Set;
class LazyGlobal extends Stmt {
    constructor(builder) {
        super();
        this.atSign = typeGuards_1.unWrap(builder.atSign);
        this.lazyGlobal = typeGuards_1.unWrap(builder.lazyGlobal);
        this.onOff = typeGuards_1.unWrap(builder.onOff);
    }
    toLines() {
        return [
            `${this.atSign.lexeme}${this.lazyGlobal.lexeme} ${this.onOff.lexeme}.`,
        ];
    }
    get start() {
        return this.atSign.start;
    }
    get end() {
        return this.onOff.end;
    }
    get ranges() {
        return [this.atSign, this.lazyGlobal, this.onOff];
    }
    pass(visitor) {
        return visitor.passLazyGlobal(this);
    }
    accept(visitor) {
        return visitor.visitLazyGlobal(this);
    }
}
exports.LazyGlobal = LazyGlobal;
class If extends Stmt {
    constructor(builder) {
        super();
        this.ifToken = typeGuards_1.unWrap(builder.ifToken);
        this.condition = typeGuards_1.unWrap(builder.condition);
        this.body = typeGuards_1.unWrap(builder.body);
        this.elseStmt = builder.elseStmt;
    }
    toLines() {
        const conditionLines = this.condition.toLines();
        const stmtLines = this.body.toLines();
        conditionLines[0] = `${this.ifToken.lexeme} ${conditionLines[0]}`;
        const lines = toStringUtils_1.joinLines(' ', conditionLines, stmtLines);
        if (!typeGuards_1.empty(this.elseStmt)) {
            const elseLines = this.elseStmt.toLines();
            return toStringUtils_1.joinLines(' ', lines, elseLines);
        }
        return lines;
    }
    get start() {
        return this.ifToken.start;
    }
    get end() {
        return typeGuards_1.empty(this.elseStmt) ? this.body.end : this.elseStmt.end;
    }
    get ranges() {
        const ranges = [this.ifToken, this.condition, this.body];
        if (!typeGuards_1.empty(this.elseStmt)) {
            ranges.push(this.elseStmt);
        }
        return ranges;
    }
    pass(visitor) {
        return visitor.passIf(this);
    }
    accept(visitor) {
        return visitor.visitIf(this);
    }
}
exports.If = If;
class Else extends Stmt {
    constructor(builder) {
        super();
        this.elseToken = typeGuards_1.unWrap(builder.elseToken);
        this.body = typeGuards_1.unWrap(builder.body);
    }
    toLines() {
        const lines = this.body.toLines();
        lines[0] = `${this.elseToken.lexeme} ${lines[0]}`;
        return lines;
    }
    get start() {
        return this.elseToken.start;
    }
    get end() {
        return this.body.end;
    }
    get ranges() {
        return [this.elseToken, this.body];
    }
    pass(visitor) {
        return visitor.passElse(this);
    }
    accept(visitor) {
        return visitor.visitElse(this);
    }
}
exports.Else = Else;
class Until extends Stmt {
    constructor(builder) {
        super();
        this.until = typeGuards_1.unWrap(builder.until);
        this.condition = typeGuards_1.unWrap(builder.condition);
        this.body = typeGuards_1.unWrap(builder.body);
    }
    toLines() {
        const conditionLines = this.condition.toLines();
        const bodyLines = this.body.toLines();
        conditionLines[0] = `${this.until.lexeme} ${conditionLines[0]}`;
        return toStringUtils_1.joinLines(' ', conditionLines, bodyLines);
    }
    get start() {
        return this.until.start;
    }
    get end() {
        return this.body.end;
    }
    get ranges() {
        return [this.until, this.condition, this.body];
    }
    pass(visitor) {
        return visitor.passUntil(this);
    }
    accept(visitor) {
        return visitor.visitUntil(this);
    }
}
exports.Until = Until;
class From extends Stmt {
    constructor(builder) {
        super();
        this.from = typeGuards_1.unWrap(builder.from);
        this.initializer = typeGuards_1.unWrap(builder.initializer);
        this.until = typeGuards_1.unWrap(builder.until);
        this.condition = typeGuards_1.unWrap(builder.condition);
        this.step = typeGuards_1.unWrap(builder.step);
        this.increment = typeGuards_1.unWrap(builder.increment);
        this.doToken = typeGuards_1.unWrap(builder.doToken);
        this.body = typeGuards_1.unWrap(builder.body);
    }
    toLines() {
        const initializerLines = this.initializer.toLines();
        const conditionLines = this.condition.toLines();
        const incrementLines = this.increment.toLines();
        const bodyLines = this.body.toLines();
        initializerLines[0] = `${this.from.lexeme} ${initializerLines[0]}`;
        conditionLines[0] = `${this.until.lexeme} ${conditionLines[0]}`;
        incrementLines[0] = `${this.step.lexeme} ${incrementLines[0]}`;
        bodyLines[0] = `${this.doToken.lexeme} ${bodyLines[0]}`;
        return toStringUtils_1.joinLines(' ', initializerLines, conditionLines, incrementLines, bodyLines);
    }
    get start() {
        return this.from.start;
    }
    get end() {
        return this.body.end;
    }
    get ranges() {
        return [
            this.from,
            this.initializer,
            this.until,
            this.condition,
            this.step,
            this.increment,
            this.doToken,
            this.body,
        ];
    }
    pass(visitor) {
        return visitor.passFrom(this);
    }
    accept(visitor) {
        return visitor.visitFrom(this);
    }
}
exports.From = From;
class When extends Stmt {
    constructor(builder) {
        super();
        this.when = typeGuards_1.unWrap(builder.when);
        this.condition = typeGuards_1.unWrap(builder.condition);
        this.then = typeGuards_1.unWrap(builder.then);
        this.body = typeGuards_1.unWrap(builder.body);
    }
    toLines() {
        const conditionLines = this.condition.toLines();
        const bodyLines = this.body.toLines();
        conditionLines[0] = `${this.when.lexeme} ${conditionLines[0]}`;
        bodyLines[0] = `${this.then.lexeme} ${bodyLines[0]}`;
        return toStringUtils_1.joinLines(' ', conditionLines, bodyLines);
    }
    get start() {
        return this.when.start;
    }
    get end() {
        return this.body.end;
    }
    get ranges() {
        return [this.when, this.condition, this.then, this.body];
    }
    pass(visitor) {
        return visitor.passWhen(this);
    }
    accept(visitor) {
        return visitor.visitWhen(this);
    }
}
exports.When = When;
class Return extends Stmt {
    constructor(builder) {
        super();
        this.returnToken = typeGuards_1.unWrap(builder.returnToken);
        this.value = builder.value;
    }
    toLines() {
        if (!typeGuards_1.empty(this.value)) {
            const exprLines = this.value.toLines();
            exprLines[0] = `${this.returnToken.lexeme} ${exprLines[0]}`;
            exprLines[exprLines.length - 1] = `${exprLines[exprLines.length - 1]}.`;
            return exprLines;
        }
        return [`${this.returnToken.lexeme}.`];
    }
    get start() {
        return this.returnToken.start;
    }
    get end() {
        return typeGuards_1.empty(this.value) ? this.returnToken.end : this.value.end;
    }
    get ranges() {
        let ranges = [this.returnToken];
        if (!typeGuards_1.empty(this.value)) {
            ranges = ranges.concat(this.value.ranges);
        }
        return ranges;
    }
    pass(visitor) {
        return visitor.passReturn(this);
    }
    accept(visitor) {
        return visitor.visitReturn(this);
    }
}
exports.Return = Return;
class Break extends Stmt {
    constructor(breakToken) {
        super();
        this.breakToken = breakToken;
    }
    toLines() {
        return [`${this.breakToken.lexeme}.`];
    }
    get start() {
        return this.breakToken.start;
    }
    get end() {
        return this.breakToken.end;
    }
    get ranges() {
        return [this.breakToken];
    }
    pass(visitor) {
        return visitor.passBreak(this);
    }
    accept(visitor) {
        return visitor.visitBreak(this);
    }
}
exports.Break = Break;
class Switch extends Stmt {
    constructor(builder) {
        super();
        this.switchToken = typeGuards_1.unWrap(builder.switchToken);
        this.to = typeGuards_1.unWrap(builder.to);
        this.target = typeGuards_1.unWrap(builder.target);
    }
    toLines() {
        const targetLines = this.target.toLines();
        targetLines[0] = `${this.switchToken.lexeme} ${this.to.lexeme} ${targetLines[0]}`;
        targetLines[targetLines.length - 1] = `${targetLines[targetLines.length - 1]}.`;
        return targetLines;
    }
    get start() {
        return this.switchToken.start;
    }
    get end() {
        return this.target.end;
    }
    get ranges() {
        return [this.switchToken, this.to, this.target];
    }
    pass(visitor) {
        return visitor.passSwitch(this);
    }
    accept(visitor) {
        return visitor.visitSwitch(this);
    }
}
exports.Switch = Switch;
class For extends Stmt {
    constructor(builder) {
        super();
        this.forToken = typeGuards_1.unWrap(builder.forToken);
        this.element = typeGuards_1.unWrap(builder.element);
        this.inToken = typeGuards_1.unWrap(builder.inToken);
        this.collection = typeGuards_1.unWrap(builder.collection);
        this.body = typeGuards_1.unWrap(builder.body);
    }
    toLines() {
        const suffixLines = this.collection.toLines();
        const bodyLines = this.body.toLines();
        suffixLines[0] =
            `${this.forToken.lexeme} ${this.element.lexeme} ` +
                `${this.inToken.lexeme} ${suffixLines[0]}`;
        return toStringUtils_1.joinLines(' ', suffixLines, bodyLines);
    }
    get start() {
        return this.forToken.start;
    }
    get end() {
        return this.body.end;
    }
    get ranges() {
        return [
            this.forToken,
            this.element,
            this.inToken,
            this.collection,
            this.body,
        ];
    }
    pass(visitor) {
        return visitor.passFor(this);
    }
    accept(visitor) {
        return visitor.visitFor(this);
    }
}
exports.For = For;
class On extends Stmt {
    constructor(builder) {
        super();
        this.on = typeGuards_1.unWrap(builder.on);
        this.suffix = typeGuards_1.unWrap(builder.suffix);
        this.body = typeGuards_1.unWrap(builder.body);
    }
    toLines() {
        const suffixLines = this.suffix.toLines();
        const bodyLines = this.body.toLines();
        suffixLines[0] = `${this.on.lexeme} ${suffixLines[0]}`;
        return toStringUtils_1.joinLines(' ', suffixLines, bodyLines);
    }
    get start() {
        return this.on.start;
    }
    get end() {
        return this.body.end;
    }
    get ranges() {
        return [this.on, this.suffix, this.body];
    }
    pass(visitor) {
        return visitor.passOn(this);
    }
    accept(visitor) {
        return visitor.visitOn(this);
    }
}
exports.On = On;
class Toggle extends Stmt {
    constructor(builder) {
        super();
        this.toggle = typeGuards_1.unWrap(builder.toggle);
        this.suffix = typeGuards_1.unWrap(builder.suffix);
    }
    toLines() {
        const suffixLines = this.suffix.toLines();
        suffixLines[0] = `${this.toggle.lexeme} ${suffixLines[0]}`;
        suffixLines[suffixLines.length - 1] = `${suffixLines[suffixLines.length - 1]}.`;
        return suffixLines;
    }
    get start() {
        return this.toggle.start;
    }
    get end() {
        return this.suffix.end;
    }
    get ranges() {
        return [this.toggle, this.suffix];
    }
    pass(visitor) {
        return visitor.passToggle(this);
    }
    accept(visitor) {
        return visitor.visitToggle(this);
    }
}
exports.Toggle = Toggle;
class Wait extends Stmt {
    constructor(builder) {
        super();
        this.wait = typeGuards_1.unWrap(builder.wait);
        this.until = builder.until;
        this.expr = typeGuards_1.unWrap(builder.expr);
    }
    toLines() {
        const exprLines = this.expr.toLines();
        exprLines[0] = typeGuards_1.empty(this.until)
            ? `${this.wait.lexeme} ${exprLines[0]}`
            : `${this.wait.lexeme} ${this.until.lexeme} ${exprLines[0]}`;
        exprLines[exprLines.length - 1] = `${exprLines[exprLines.length - 1]}.`;
        return exprLines;
    }
    get start() {
        return this.wait.start;
    }
    get end() {
        return this.expr.end;
    }
    get ranges() {
        if (typeGuards_1.empty(this.until)) {
            return [this.wait, this.expr];
        }
        return [this.wait, this.until, this.expr];
    }
    pass(visitor) {
        return visitor.passWait(this);
    }
    accept(visitor) {
        return visitor.visitWait(this);
    }
}
exports.Wait = Wait;
class Log extends Stmt {
    constructor(builder) {
        super();
        this.log = typeGuards_1.unWrap(builder.log);
        this.expr = typeGuards_1.unWrap(builder.expr);
        this.to = typeGuards_1.unWrap(builder.to);
        this.target = typeGuards_1.unWrap(builder.target);
    }
    toLines() {
        const exprLines = this.expr.toLines();
        const targetLines = this.target.toLines();
        exprLines[0] = `${this.log.lexeme} ${exprLines[0]}`;
        targetLines[0] = `${this.to.lexeme} ${targetLines[0]}`;
        targetLines[targetLines.length - 1] = `${targetLines[targetLines.length - 1]}.`;
        return toStringUtils_1.joinLines(' ', exprLines, targetLines);
    }
    get start() {
        return this.log.start;
    }
    get end() {
        return this.target.end;
    }
    get ranges() {
        return [this.log, this.expr, this.to, this.target];
    }
    pass(visitor) {
        return visitor.passLog(this);
    }
    accept(visitor) {
        return visitor.visitLog(this);
    }
}
exports.Log = Log;
class Copy extends Stmt {
    constructor(builder) {
        super();
        this.copy = typeGuards_1.unWrap(builder.copy);
        this.target = typeGuards_1.unWrap(builder.target);
        this.toFrom = typeGuards_1.unWrap(builder.toFrom);
        this.destination = typeGuards_1.unWrap(builder.destination);
    }
    toLines() {
        const targetLines = this.target.toLines();
        const destinationLines = this.target.toLines();
        targetLines[0] = `${this.copy.lexeme} ${targetLines[0]}`;
        destinationLines[0] = `${this.toFrom.lexeme} ${destinationLines[0]}.`;
        return toStringUtils_1.joinLines(' ', targetLines, destinationLines);
    }
    get start() {
        return this.copy.start;
    }
    get end() {
        return this.destination.end;
    }
    get ranges() {
        return [this.copy, this.target, this.toFrom, this.destination];
    }
    pass(visitor) {
        return visitor.passCopy(this);
    }
    accept(visitor) {
        return visitor.visitCopy(this);
    }
}
exports.Copy = Copy;
class Rename extends Stmt {
    constructor(builder) {
        super();
        this.rename = typeGuards_1.unWrap(builder.rename);
        this.fileVolume = typeGuards_1.unWrap(builder.fileVolume);
        this.identifier = typeGuards_1.unWrap(builder.identifier);
        this.target = typeGuards_1.unWrap(builder.target);
        this.to = typeGuards_1.unWrap(builder.to);
        this.alternative = typeGuards_1.unWrap(builder.alternative);
    }
    toLines() {
        const targetLines = this.target.toLines();
        const alternativeLines = this.alternative.toLines();
        targetLines[0] =
            `${this.rename.lexeme} ${this.fileVolume.lexeme} ` +
                `${this.identifier.lexeme} ${targetLines[0]}`;
        alternativeLines[0] = `${this.to.lexeme} ${alternativeLines[0]}.`;
        return toStringUtils_1.joinLines(' ', targetLines, alternativeLines);
    }
    get start() {
        return this.rename.start;
    }
    get end() {
        return this.alternative.end;
    }
    get ranges() {
        return [
            this.rename,
            this.identifier,
            this.target,
            this.to,
            this.alternative,
        ];
    }
    pass(visitor) {
        return visitor.passRename(this);
    }
    accept(visitor) {
        return visitor.visitRename(this);
    }
}
exports.Rename = Rename;
class Delete extends Stmt {
    constructor(builder) {
        super();
        this.deleteToken = typeGuards_1.unWrap(builder.deleteToken);
        this.target = typeGuards_1.unWrap(builder.target);
        this.from = builder.from;
        this.volume = builder.volume;
    }
    toLines() {
        const targetLines = this.target.toLines();
        targetLines[0] = `${this.deleteToken.lexeme} ${targetLines[0]}`;
        if (!typeGuards_1.empty(this.from) && !typeGuards_1.empty(this.volume)) {
            const volumeLines = this.volume.toLines();
            volumeLines[0] = `${this.from.lexeme} ${volumeLines[0]}`;
            volumeLines[volumeLines.length - 1] = `${volumeLines[volumeLines.length - 1]}.`;
            return toStringUtils_1.joinLines(' ', targetLines, volumeLines);
        }
        targetLines[targetLines.length - 1] = `${targetLines[targetLines.length - 1]}.`;
        return targetLines;
    }
    get start() {
        return this.deleteToken.start;
    }
    get end() {
        return typeGuards_1.empty(this.volume) ? this.target.end : this.volume.end;
    }
    get ranges() {
        const ranges = [this.deleteToken, this.target];
        if (!typeGuards_1.empty(this.from) && !typeGuards_1.empty(this.volume)) {
            ranges.push(this.from);
            ranges.push(this.volume);
        }
        return ranges;
    }
    pass(visitor) {
        return visitor.passDelete(this);
    }
    accept(visitor) {
        return visitor.visitDelete(this);
    }
}
exports.Delete = Delete;
class Run extends Stmt {
    constructor(builder) {
        super();
        this.run = typeGuards_1.unWrap(builder.run);
        this.identifier = typeGuards_1.unWrap(builder.identifier);
        this.once = builder.once;
        this.open = builder.open;
        this.args = builder.args;
        this.close = builder.close;
        this.on = builder.on;
        this.expr = builder.expr;
    }
    toLines() {
        let lines = typeGuards_1.empty(this.once)
            ? [`${this.run.lexeme} ${this.identifier.lexeme}`]
            : [`${this.run.lexeme} ${this.once.lexeme} ${this.identifier.lexeme}`];
        if (!typeGuards_1.empty(this.open) && !typeGuards_1.empty(this.args) && !typeGuards_1.empty(this.close)) {
            const argsLines = toStringUtils_1.joinLines(', ', ...this.args.map(arg => arg.toLines()));
            argsLines[0] = `${this.open.lexeme}${argsLines[0]}`;
            argsLines[argsLines.length - 1] = `${argsLines[argsLines.length - 1]}${this.close.lexeme}`;
            lines = toStringUtils_1.joinLines(' ', lines, argsLines);
        }
        if (!typeGuards_1.empty(this.on) && !typeGuards_1.empty(this.expr)) {
            const exprLines = this.expr.toLines();
            exprLines[0] = `${this.on.lexeme} ${exprLines[0]}`;
            lines = toStringUtils_1.joinLines(' ', lines, exprLines);
        }
        lines[lines.length - 1] = `${lines[lines.length - 1]}.`;
        return lines;
    }
    get start() {
        return this.run.start;
    }
    get end() {
        return !typeGuards_1.empty(this.expr)
            ? this.expr.end
            : !typeGuards_1.empty(this.args)
                ? this.args[this.args.length - 1].end
                : this.identifier.end;
    }
    get ranges() {
        const ranges = [this.run];
        if (!typeGuards_1.empty(this.once)) {
            ranges.push(this.once);
        }
        ranges.push(this.identifier);
        if (!typeGuards_1.empty(this.open) && !typeGuards_1.empty(this.args) && !typeGuards_1.empty(this.close)) {
            ranges.push(this.open);
            for (const arg of this.args) {
                ranges.push(arg);
            }
            ranges.push(this.close);
        }
        if (!typeGuards_1.empty(this.on) && !typeGuards_1.empty(this.expr)) {
            ranges.push(this.on);
            ranges.push(this.expr);
        }
        return ranges;
    }
    pass(visitor) {
        return visitor.passRun(this);
    }
    accept(visitor) {
        return visitor.visitRun(this);
    }
}
exports.Run = Run;
class RunPath extends Stmt {
    constructor(builder) {
        super();
        this.runPath = typeGuards_1.unWrap(builder.runPath);
        this.open = typeGuards_1.unWrap(builder.open);
        this.expr = typeGuards_1.unWrap(builder.expr);
        this.close = typeGuards_1.unWrap(builder.close);
        this.args = builder.args;
    }
    toLines() {
        let lines = this.expr.toLines();
        if (!typeGuards_1.empty(this.args)) {
            lines = toStringUtils_1.joinLines(', ', lines, ...this.args.map(arg => arg.toLines()));
        }
        lines[0] = `${this.runPath.lexeme}${this.open.lexeme}${lines[0]}`;
        lines[lines.length - 1] = `${lines[lines.length - 1]}${this.close.lexeme}.`;
        return lines;
    }
    get start() {
        return this.runPath.start;
    }
    get end() {
        return this.close.end;
    }
    get ranges() {
        const ranges = [this.runPath, this.open, this.expr];
        if (!typeGuards_1.empty(this.args)) {
            for (const arg of this.args) {
                ranges.push(arg);
            }
        }
        ranges.push(this.close);
        return ranges;
    }
    pass(visitor) {
        return visitor.passRunPath(this);
    }
    accept(visitor) {
        return visitor.visitRunPath(this);
    }
}
exports.RunPath = RunPath;
class RunOncePath extends Stmt {
    constructor(builder) {
        super();
        this.runPath = typeGuards_1.unWrap(builder.runPath);
        this.open = typeGuards_1.unWrap(builder.open);
        this.expr = typeGuards_1.unWrap(builder.expr);
        this.close = typeGuards_1.unWrap(builder.close);
        this.args = builder.args;
    }
    toLines() {
        let lines = this.expr.toLines();
        if (!typeGuards_1.empty(this.args)) {
            lines = toStringUtils_1.joinLines(', ', lines, ...this.args.map(arg => arg.toLines()));
        }
        lines[0] = `${this.runPath.lexeme}${this.open.lexeme}${lines[0]}`;
        lines[lines.length - 1] = `${lines[lines.length - 1]}${this.close.lexeme}.`;
        return lines;
    }
    get start() {
        return this.runPath.start;
    }
    get end() {
        return this.close.end;
    }
    get ranges() {
        const ranges = [this.runPath, this.open, this.expr];
        if (!typeGuards_1.empty(this.args)) {
            for (const arg of this.args) {
                ranges.push(arg);
            }
        }
        ranges.push(this.close);
        return ranges;
    }
    pass(visitor) {
        return visitor.passRunPathOnce(this);
    }
    accept(visitor) {
        return visitor.visitRunPathOnce(this);
    }
}
exports.RunOncePath = RunOncePath;
class Compile extends Stmt {
    constructor(builder) {
        super();
        this.compile = typeGuards_1.unWrap(builder.compile);
        this.target = typeGuards_1.unWrap(builder.target);
        this.to = builder.to;
        this.destination = builder.destination;
    }
    toLines() {
        let lines = this.target.toLines();
        if (!typeGuards_1.empty(this.destination) && !typeGuards_1.empty(this.to)) {
            lines = toStringUtils_1.joinLines(` ${this.to.lexeme} `, this.destination.toLines());
        }
        lines[0] = `${this.compile.lexeme} ${lines[0]}`;
        lines[lines.length - 1] = `${lines[lines.length - 1]}.`;
        return lines;
    }
    get start() {
        return this.compile.start;
    }
    get end() {
        return typeGuards_1.empty(this.destination) ? this.target.end : this.destination.end;
    }
    get ranges() {
        const ranges = [this.compile, this.target];
        if (!typeGuards_1.empty(this.to) && !typeGuards_1.empty(this.destination)) {
            ranges.push(this.to);
            ranges.push(this.destination);
        }
        return ranges;
    }
    pass(visitor) {
        return visitor.passCompile(this);
    }
    accept(visitor) {
        return visitor.visitCompile(this);
    }
}
exports.Compile = Compile;
class List extends Stmt {
    constructor(builder) {
        super();
        this.list = typeGuards_1.unWrap(builder.list);
        this.collection = builder.collection;
        this.inToken = builder.inToken;
        this.target = builder.target;
    }
    toLines() {
        if (!typeGuards_1.empty(this.collection) &&
            !typeGuards_1.empty(this.inToken) &&
            !typeGuards_1.empty(this.target)) {
            return [
                `${this.list.lexeme} ${this.collection.lexeme} ` +
                    `${this.inToken.lexeme} ${this.target.lexeme}.`,
            ];
        }
        if (!typeGuards_1.empty(this.collection)) {
            return [`${this.list.lexeme} ${this.collection.lexeme}.`];
        }
        return [`${this.list.lexeme}.`];
    }
    get start() {
        return this.list.start;
    }
    get end() {
        return !typeGuards_1.empty(this.target)
            ? this.target.end
            : !typeGuards_1.empty(this.collection)
                ? this.collection.end
                : this.list.end;
    }
    get ranges() {
        const ranges = [this.list];
        if (!typeGuards_1.empty(this.collection)) {
            ranges.push(this.collection);
            if (!typeGuards_1.empty(this.inToken) && !typeGuards_1.empty(this.target)) {
                ranges.push(this.inToken);
                ranges.push(this.target);
            }
        }
        return ranges;
    }
    pass(visitor) {
        return visitor.passList(this);
    }
    accept(visitor) {
        return visitor.visitList(this);
    }
}
exports.List = List;
class Empty extends Stmt {
    constructor(empty) {
        super();
        this.empty = empty;
    }
    toLines() {
        return ['.'];
    }
    get start() {
        return this.empty.start;
    }
    get end() {
        return this.empty.end;
    }
    get ranges() {
        return [this.empty];
    }
    pass(visitor) {
        return visitor.passEmpty(this);
    }
    accept(visitor) {
        return visitor.visitEmpty(this);
    }
}
exports.Empty = Empty;
class Print extends Stmt {
    constructor(builder) {
        super();
        this.print = typeGuards_1.unWrap(builder.print);
        this.expr = typeGuards_1.unWrap(builder.expr);
        this.at = builder.at;
        this.open = builder.open;
        this.x = builder.x;
        this.y = builder.y;
        this.close = builder.close;
    }
    toLines() {
        let lines = this.expr.toLines();
        lines[0] = `${this.print.lexeme} ${lines[0]}`;
        if (!typeGuards_1.empty(this.at) &&
            !typeGuards_1.empty(this.open) &&
            !typeGuards_1.empty(this.x) &&
            !typeGuards_1.empty(this.y) &&
            !typeGuards_1.empty(this.close)) {
            const xLines = this.x.toLines();
            const yLines = this.y.toLines();
            xLines[0] = `${this.at.lexeme} ${this.open.lexeme}${xLines[0]}`;
            yLines[yLines.length - 1] = `${yLines[yLines.length - 1]}${this.close.lexeme}`;
            const argLines = toStringUtils_1.joinLines(', ', xLines, yLines);
            lines = toStringUtils_1.joinLines(' ', lines, argLines);
        }
        lines[lines.length - 1] = `${lines[lines.length - 1]}.`;
        return lines;
    }
    get start() {
        return this.print.start;
    }
    get end() {
        return typeGuards_1.empty(this.close) ? this.expr.end : this.close.end;
    }
    get ranges() {
        const ranges = [this.print, this.expr];
        if (!typeGuards_1.empty(this.at) &&
            !typeGuards_1.empty(this.open) &&
            !typeGuards_1.empty(this.x) &&
            !typeGuards_1.empty(this.y) &&
            !typeGuards_1.empty(this.close)) {
            ranges.push(this.at);
            ranges.push(this.open);
            ranges.push(this.x);
            ranges.push(this.y);
            ranges.push(this.close);
        }
        return ranges;
    }
    pass(visitor) {
        return visitor.passPrint(this);
    }
    accept(visitor) {
        return visitor.visitPrint(this);
    }
}
exports.Print = Print;
exports.validStatements = [
    Block,
    ExprStmt,
    OnOff,
    Command,
    CommandExpr,
    Unset,
    Unlock,
    Set,
    LazyGlobal,
    If,
    Else,
    Until,
    From,
    When,
    Return,
    Break,
    Switch,
    For,
    On,
    Toggle,
    Wait,
    Log,
    Copy,
    Rename,
    Delete,
    Run,
    RunPath,
    RunOncePath,
    Compile,
    List,
    Empty,
    Print,
];
//# sourceMappingURL=stmt.js.map