import { ISuffixTerm, ISuffixTermVisitor, GrammarNode, IExpr, Atom, SuffixTermTrailer, ISuffixTermClassVisitor, ISuffixTermParamVisitor, ISuffixTermPasser, SyntaxKind } from './types';
import { Range, Position } from 'vscode-languageserver';
import { NodeBase } from './base';
import { Token } from '../entities/token';
import { SymbolTracker } from '../analysis/types';
/**
 * Base class for all suffix terms
 */
export declare abstract class SuffixTermBase extends NodeBase implements ISuffixTerm {
    /**
     * Tag used to denote syntax node of the instance
     */
    readonly tag: SyntaxKind.suffixTerm;
    /**
     * Require all subclasses to implement the accept method
     * Called when the node should execute the visitors methods
     * @param visitor visitor object
     */
    abstract accept<T>(visitor: ISuffixTermVisitor<T>): T;
    /**
     * Require all subclass to implement the pass method
     * Call when the node should be passed through
     * @param visitor visitor object
     */
    abstract pass<T>(visitor: ISuffixTermPasser<T>): T;
    abstract acceptParam<TP, TR>(visitor: ISuffixTermParamVisitor<TP, TR>, param: TP): TR;
}
/**
 * Container for tokens constituting an invalid suffix term
 */
export declare class Invalid extends SuffixTermBase {
    readonly position: Position;
    /**
     * Invalid suffix term constructor
     * @param tokens tokens in the invalid range
     */
    constructor(position: Position);
    readonly start: Position;
    readonly end: Position;
    readonly ranges: Range[];
    toLines(): string[];
    acceptParam<TP, TR>(visitor: ISuffixTermParamVisitor<TP, TR>, param: TP): TR;
    pass<T>(visitor: ISuffixTermPasser<T>): T;
    accept<T>(visitor: ISuffixTermVisitor<T>): T;
    static classAccept<T>(visitor: ISuffixTermClassVisitor<T>): T;
}
/**
 * Class holding all suffix trailers
 */
export declare class SuffixTrailer extends SuffixTermBase {
    readonly suffixTerm: SuffixTerm;
    colon?: Token | undefined;
    trailer?: SuffixTrailer | undefined;
    /**
     * Grammar for the suffix trailers
     */
    static grammar: GrammarNode[];
    /**
     * Constructor for the suffix trailer
     * @param suffixTerm base suffix term
     * @param colon colon separating the base from the trailer
     * @param trailer the suffix trailer
     */
    constructor(suffixTerm: SuffixTerm, colon?: Token | undefined, trailer?: SuffixTrailer | undefined);
    readonly start: Position;
    readonly end: Position;
    readonly ranges: Range[];
    /**
     * Method indicating if the suffix ends with a settable trailer
     */
    isSettable(): boolean;
    /**
     * Get the most resolved type on this suffix
     */
    mostResolveTracker(): Maybe<SymbolTracker>;
    toLines(): string[];
    acceptParam<TP, TR>(visitor: ISuffixTermParamVisitor<TP, TR>, param: TP): TR;
    accept<T>(visitor: ISuffixTermVisitor<T>): T;
    pass<T>(visitor: ISuffixTermPasser<T>): T;
    static classAccept<T>(visitor: ISuffixTermClassVisitor<T>): T;
}
/**
 * Class holding all valid suffix terms
 */
export declare class SuffixTerm extends SuffixTermBase {
    readonly atom: Atom;
    readonly trailers: SuffixTermTrailer[];
    /**
     * Grammer for the suffix terms
     */
    static grammar: GrammarNode[];
    /**
     * Constructor for suffix terms
     * @param atom base item of the suffix term
     * @param trailers trailers present in the suffixterm
     */
    constructor(atom: Atom, trailers: SuffixTermTrailer[]);
    readonly ranges: Range[];
    readonly start: Position;
    readonly end: Position;
    toLines(): string[];
    acceptParam<TP, TR>(visitor: ISuffixTermParamVisitor<TP, TR>, param: TP): TR;
    accept<T>(visitor: ISuffixTermVisitor<T>): T;
    pass<T>(visitor: ISuffixTermPasser<T>): T;
    static classAccept<T>(visitor: ISuffixTermClassVisitor<T>): T;
}
/**
 * Class containing all valid call suffixterm trailers
 */
export declare class Call extends SuffixTermBase {
    readonly open: Token;
    readonly args: IExpr[];
    readonly close: Token;
    /**
     * Grammer for the call trailers
     */
    static grammar: GrammarNode[];
    /**
     * Constructor for the suffix term trailers
     * @param open open paren of the call
     * @param args arguments for the call
     * @param close close paren of the call
     */
    constructor(open: Token, args: IExpr[], close: Token);
    readonly start: Position;
    readonly end: Position;
    readonly ranges: Range[];
    toLines(): string[];
    acceptParam<TP, TR>(visitor: ISuffixTermParamVisitor<TP, TR>, param: TP): TR;
    accept<T>(visitor: ISuffixTermVisitor<T>): T;
    pass<T>(visitor: ISuffixTermPasser<T>): T;
    static classAccept<T>(visitor: ISuffixTermClassVisitor<T>): T;
}
/**
 * Class containing all array index suffix term trailers
 */
export declare class ArrayIndex extends SuffixTermBase {
    readonly indexer: Token;
    readonly index: Token;
    /**
     * Grammar for the array index suffix term trailers
     */
    static grammar: GrammarNode[];
    /**
     * Constructor for the suffix term trailer
     * @param indexer "#" token indicating a index
     * @param index index to be used
     */
    constructor(indexer: Token, index: Token);
    readonly start: Position;
    readonly end: Position;
    readonly ranges: Range[];
    toLines(): string[];
    acceptParam<TP, TR>(visitor: ISuffixTermParamVisitor<TP, TR>, param: TP): TR;
    accept<T>(visitor: ISuffixTermVisitor<T>): T;
    pass<T>(visitor: ISuffixTermPasser<T>): T;
    static classAccept<T>(visitor: ISuffixTermClassVisitor<T>): T;
}
/**
 * Class containing all valid array bracket suffix term trailers
 */
export declare class ArrayBracket extends SuffixTermBase {
    readonly open: Token;
    readonly index: IExpr;
    readonly close: Token;
    /**
     * Grammar for the array bracket suffix term
     */
    static grammar: GrammarNode[];
    /**
     * Constructor for the array bracket suffix term trailer
     * @param open open bracket
     * @param index index into the collection
     * @param close close bracket
     */
    constructor(open: Token, index: IExpr, close: Token);
    readonly start: Position;
    readonly end: Position;
    readonly ranges: Range[];
    toLines(): string[];
    acceptParam<TP, TR>(visitor: ISuffixTermParamVisitor<TP, TR>, param: TP): TR;
    accept<T>(visitor: ISuffixTermVisitor<T>): T;
    pass<T>(visitor: ISuffixTermPasser<T>): T;
    static classAccept<T>(visitor: ISuffixTermClassVisitor<T>): T;
}
/**
 * Class containing function delgate creation suffix terms
 */
export declare class Delegate extends SuffixTermBase {
    readonly atSign: Token;
    /**
     * Grammar for the delgate
     */
    static grammar: GrammarNode[];
    /**
     * Constructor for the function delegate
     * @param atSign at sign indicating that function should create a delgate
     */
    constructor(atSign: Token);
    readonly start: Position;
    readonly end: Position;
    readonly ranges: Range[];
    toLines(): string[];
    acceptParam<TP, TR>(visitor: ISuffixTermParamVisitor<TP, TR>, param: TP): TR;
    accept<T>(visitor: ISuffixTermVisitor<T>): T;
    pass<T>(visitor: ISuffixTermPasser<T>): T;
    static classAccept<T>(visitor: ISuffixTermClassVisitor<T>): T;
}
/**
 * Class containing literal suffix terms
 */
export declare class Literal extends SuffixTermBase {
    readonly token: Token;
    /**
     * Grammar for literal suffix terms
     */
    static grammar: GrammarNode[];
    /**
     * Constructor for literal suffix term
     * @param token token for the literal
     */
    constructor(token: Token);
    readonly start: Position;
    readonly end: Position;
    readonly ranges: Range[];
    toLines(): string[];
    acceptParam<TP, TR>(visitor: ISuffixTermParamVisitor<TP, TR>, param: TP): TR;
    accept<T>(visitor: ISuffixTermVisitor<T>): T;
    pass<T>(visitor: ISuffixTermPasser<T>): T;
    static classAccept<T>(visitor: ISuffixTermClassVisitor<T>): T;
}
/**
 * Class containing all valid identifiers
 */
export declare class Identifier extends SuffixTermBase {
    readonly token: Token;
    /**
     * Grammar for valid identifiers
     */
    static grammar: GrammarNode[];
    /**
     * Constructor for suffix term identifiers
     * @param token identifier token
     */
    constructor(token: Token);
    readonly start: Position;
    readonly end: Position;
    readonly ranges: Range[];
    readonly isKeyword: boolean;
    toLines(): string[];
    acceptParam<TP, TR>(visitor: ISuffixTermParamVisitor<TP, TR>, param: TP): TR;
    accept<T>(visitor: ISuffixTermVisitor<T>): T;
    pass<T>(visitor: ISuffixTermPasser<T>): T;
    static classAccept<T>(visitor: ISuffixTermClassVisitor<T>): T;
}
/**
 * Class containing all valid groupings
 */
export declare class Grouping extends SuffixTermBase {
    readonly open: Token;
    readonly expr: IExpr;
    readonly close: Token;
    /**
     * Grammar for all valid groupings
     */
    static grammar: GrammarNode[];
    /**
     * Grouping constructor
     * @param open open paren token
     * @param expr expression within the grouping
     * @param close close paren token
     */
    constructor(open: Token, expr: IExpr, close: Token);
    readonly start: Position;
    readonly end: Position;
    readonly ranges: Range[];
    toString(): string;
    toLines(): string[];
    acceptParam<TP, TR>(visitor: ISuffixTermParamVisitor<TP, TR>, param: TP): TR;
    accept<T>(visitor: ISuffixTermVisitor<T>): T;
    pass<T>(visitor: ISuffixTermPasser<T>): T;
    static classAccept<T>(visitor: ISuffixTermClassVisitor<T>): T;
}
/**
 * All valid suffix terms
 */
export declare const validSuffixTerms: Constructor<SuffixTermBase>[];
