"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const types_1 = require("./types");
const tokentypes_1 = require("../entities/tokentypes");
const grammarNodes_1 = require("./grammarNodes");
const expr_1 = require("./expr");
const base_1 = require("./base");
const typeGuards_1 = require("../utilities/typeGuards");
const toStringUtils_1 = require("./toStringUtils");
/**
 * Base class for all suffix terms
 */
class SuffixTermBase extends base_1.NodeBase {
    /**
     * Tag used to denote syntax node of the instance
     */
    get tag() {
        return types_1.SyntaxKind.suffixTerm;
    }
}
exports.SuffixTermBase = SuffixTermBase;
/**
 * Container for tokens constituting an invalid suffix term
 */
class Invalid extends SuffixTermBase {
    /**
     * Invalid suffix term constructor
     * @param tokens tokens in the invalid range
     */
    constructor(position) {
        super();
        this.position = position;
    }
    get start() {
        return this.position;
    }
    get end() {
        return this.position;
    }
    get ranges() {
        return [];
    }
    toLines() {
        return [''];
    }
    acceptParam(visitor, param) {
        return visitor.visitSuffixTermInvalid(this, param);
    }
    pass(visitor) {
        return visitor.passSuffixTermInvalid(this);
    }
    accept(visitor) {
        return visitor.visitSuffixTermInvalid(this);
    }
    static classAccept(visitor) {
        return visitor.visitSuffixTermInvalid(this);
    }
}
exports.Invalid = Invalid;
/**
 * Class holding all suffix trailers
 */
class SuffixTrailer extends SuffixTermBase {
    /**
     * Constructor for the suffix trailer
     * @param suffixTerm base suffix term
     * @param colon colon separating the base from the trailer
     * @param trailer the suffix trailer
     */
    constructor(suffixTerm, colon, trailer) {
        super();
        this.suffixTerm = suffixTerm;
        this.colon = colon;
        this.trailer = trailer;
    }
    get start() {
        return this.suffixTerm.start;
    }
    get end() {
        return typeGuards_1.empty(this.trailer) ? this.suffixTerm.end : this.trailer.end;
    }
    get ranges() {
        if (!typeGuards_1.empty(this.colon) && !typeGuards_1.empty(this.trailer)) {
            return [this.suffixTerm, this.colon, this.trailer];
        }
        return [this.suffixTerm];
    }
    /**
     * Method indicating if the suffix ends with a settable trailer
     */
    isSettable() {
        // if no trailer check suffix term
        if (typeGuards_1.empty(this.trailer)) {
            const { atom, trailers } = this.suffixTerm;
            // check for suffix term trailers
            if (trailers.length > 0) {
                const lastTrailer = trailers[trailers.length - 1];
                if (lastTrailer instanceof Identifier) {
                    return true;
                }
                if (lastTrailer instanceof ArrayBracket) {
                    return true;
                }
                if (lastTrailer instanceof ArrayIndex) {
                    return true;
                }
            }
            if (atom instanceof Identifier) {
                return true;
            }
            return false;
        }
        // check nested trailers
        if (this.trailer instanceof SuffixTrailer) {
            return this.trailer.isSettable();
        }
        return false;
    }
    /**
     * Get the most resolved type on this suffix
     */
    mostResolveTracker() {
        // if no trailer check suffix term
        if (typeGuards_1.empty(this.trailer)) {
            const { atom, trailers } = this.suffixTerm;
            // check for suffix term trailers
            if (trailers.length > 0) {
                const lastTrailer = trailers[trailers.length - 1];
                if (lastTrailer instanceof ArrayBracket) {
                    return lastTrailer.open.tracker;
                }
                if (lastTrailer instanceof ArrayIndex) {
                    return undefined;
                }
                if (lastTrailer instanceof Call) {
                    return lastTrailer.open.tracker;
                }
                return undefined;
            }
            // check nested trailers
            if (atom instanceof Identifier) {
                return atom.token.tracker;
            }
            return undefined;
        }
        // check nested trailers
        if (!typeGuards_1.empty(this.trailer)) {
            return this.trailer.mostResolveTracker();
        }
        return undefined;
    }
    toLines() {
        const suffixTermLines = this.suffixTerm.toLines();
        if (!typeGuards_1.empty(this.colon) && !typeGuards_1.empty(this.trailer)) {
            const [joinLine, ...restLines] = this.trailer.toLines();
            if (suffixTermLines.length === 1) {
                return [`${suffixTermLines[0]}${this.colon.lexeme}${joinLine}`].concat(restLines);
            }
            return suffixTermLines
                .slice(0, suffixTermLines.length - 2)
                .concat(`${suffixTermLines[0]}${this.colon.lexeme}${joinLine}`, restLines);
        }
        return suffixTermLines;
    }
    acceptParam(visitor, param) {
        return visitor.visitSuffixTrailer(this, param);
    }
    accept(visitor) {
        return visitor.visitSuffixTrailer(this);
    }
    pass(visitor) {
        return visitor.passSuffixTrailer(this);
    }
    static classAccept(visitor) {
        return visitor.visitSuffixTrailer(this);
    }
}
exports.SuffixTrailer = SuffixTrailer;
/**
 * Class holding all valid suffix terms
 */
class SuffixTerm extends SuffixTermBase {
    /**
     * Constructor for suffix terms
     * @param atom base item of the suffix term
     * @param trailers trailers present in the suffixterm
     */
    constructor(atom, trailers) {
        super();
        this.atom = atom;
        this.trailers = trailers;
    }
    get ranges() {
        return [this.atom, ...this.trailers];
    }
    get start() {
        return this.atom.start;
    }
    get end() {
        if (this.trailers.length > 0) {
            return this.trailers[this.trailers.length - 1].end;
        }
        return this.atom.end;
    }
    toLines() {
        const atomLines = this.atom.toLines();
        const trailersLines = this.trailers.map(t => t.toLines());
        return toStringUtils_1.joinLines('', atomLines, ...trailersLines);
    }
    acceptParam(visitor, param) {
        return visitor.visitSuffixTerm(this, param);
    }
    accept(visitor) {
        return visitor.visitSuffixTerm(this);
    }
    pass(visitor) {
        return visitor.passSuffixTerm(this);
    }
    static classAccept(visitor) {
        return visitor.visitSuffixTerm(this);
    }
}
exports.SuffixTerm = SuffixTerm;
/**
 * Class containing all valid call suffixterm trailers
 */
class Call extends SuffixTermBase {
    /**
     * Constructor for the suffix term trailers
     * @param open open paren of the call
     * @param args arguments for the call
     * @param close close paren of the call
     */
    constructor(open, args, close) {
        super();
        this.open = open;
        this.args = args;
        this.close = close;
    }
    get start() {
        return this.open.start;
    }
    get end() {
        return this.close.end;
    }
    get ranges() {
        return [this.open, ...this.args, this.close];
    }
    toLines() {
        if (this.args.length === 0) {
            return [`${this.open.lexeme}${this.close.lexeme}`];
        }
        const argsLines = this.args.map(a => a.toLines());
        const argsResult = toStringUtils_1.joinLines(',', ...argsLines);
        argsResult[0] = `${this.open.lexeme}${argsResult[0]}`;
        argsResult[argsResult.length - 1] = `${argsResult[argsResult.length - 1]}${this.close.lexeme}`;
        return argsResult;
    }
    acceptParam(visitor, param) {
        return visitor.visitCall(this, param);
    }
    accept(visitor) {
        return visitor.visitCall(this);
    }
    pass(visitor) {
        return visitor.passCall(this);
    }
    static classAccept(visitor) {
        return visitor.visitCall(this);
    }
}
exports.Call = Call;
/**
 * Class containing all array index suffix term trailers
 */
class ArrayIndex extends SuffixTermBase {
    /**
     * Constructor for the suffix term trailer
     * @param indexer "#" token indicating a index
     * @param index index to be used
     */
    constructor(indexer, index) {
        super();
        this.indexer = indexer;
        this.index = index;
    }
    get start() {
        return this.indexer.start;
    }
    get end() {
        return this.index.end;
    }
    get ranges() {
        return [this.indexer, this.index];
    }
    toLines() {
        return [`${this.indexer.lexeme}${this.index.lexeme}`];
    }
    acceptParam(visitor, param) {
        return visitor.visitArrayIndex(this, param);
    }
    accept(visitor) {
        return visitor.visitArrayIndex(this);
    }
    pass(visitor) {
        return visitor.passArrayIndex(this);
    }
    static classAccept(visitor) {
        return visitor.visitArrayIndex(this);
    }
}
exports.ArrayIndex = ArrayIndex;
/**
 * Class containing all valid array bracket suffix term trailers
 */
class ArrayBracket extends SuffixTermBase {
    /**
     * Constructor for the array bracket suffix term trailer
     * @param open open bracket
     * @param index index into the collection
     * @param close close bracket
     */
    constructor(open, index, close) {
        super();
        this.open = open;
        this.index = index;
        this.close = close;
    }
    get start() {
        return this.open.start;
    }
    get end() {
        return this.close.end;
    }
    get ranges() {
        return [this.open, this.index, this.close];
    }
    toLines() {
        const lines = this.index.toLines();
        lines[0] = `${this.open.lexeme}${lines[0]}`;
        lines[lines.length - 1] = `${lines[lines.length - 1]}${this.close.lexeme}`;
        return lines;
    }
    acceptParam(visitor, param) {
        return visitor.visitArrayBracket(this, param);
    }
    accept(visitor) {
        return visitor.visitArrayBracket(this);
    }
    pass(visitor) {
        return visitor.passArrayBracket(this);
    }
    static classAccept(visitor) {
        return visitor.visitArrayBracket(this);
    }
}
exports.ArrayBracket = ArrayBracket;
/**
 * Class containing function delgate creation suffix terms
 */
class Delegate extends SuffixTermBase {
    /**
     * Constructor for the function delegate
     * @param atSign at sign indicating that function should create a delgate
     */
    constructor(atSign) {
        super();
        this.atSign = atSign;
    }
    get start() {
        return this.atSign.start;
    }
    get end() {
        return this.atSign.end;
    }
    get ranges() {
        return [this.atSign];
    }
    toLines() {
        return [this.atSign.lexeme];
    }
    acceptParam(visitor, param) {
        return visitor.visitDelegate(this, param);
    }
    accept(visitor) {
        return visitor.visitDelegate(this);
    }
    pass(visitor) {
        return visitor.passDelegate(this);
    }
    static classAccept(visitor) {
        return visitor.visitDelegate(this);
    }
}
exports.Delegate = Delegate;
/**
 * Class containing literal suffix terms
 */
class Literal extends SuffixTermBase {
    /**
     * Constructor for literal suffix term
     * @param token token for the literal
     */
    constructor(token) {
        super();
        this.token = token;
    }
    get start() {
        return this.token.start;
    }
    get end() {
        return this.token.end;
    }
    get ranges() {
        return [this.token];
    }
    toLines() {
        return [`${this.token.lexeme}`];
    }
    acceptParam(visitor, param) {
        return visitor.visitLiteral(this, param);
    }
    accept(visitor) {
        return visitor.visitLiteral(this);
    }
    pass(visitor) {
        return visitor.passLiteral(this);
    }
    static classAccept(visitor) {
        return visitor.visitLiteral(this);
    }
}
exports.Literal = Literal;
/**
 * Class containing all valid identifiers
 */
class Identifier extends SuffixTermBase {
    /**
     * Constructor for suffix term identifiers
     * @param token identifier token
     */
    constructor(token) {
        super();
        this.token = token;
    }
    get start() {
        return this.token.start;
    }
    get end() {
        return this.token.end;
    }
    get ranges() {
        return [this.token];
    }
    get isKeyword() {
        return !(this.token.type === tokentypes_1.TokenType.identifier ||
            this.token.type === tokentypes_1.TokenType.fileIdentifier);
    }
    toLines() {
        return [`${this.token.lexeme}`];
    }
    acceptParam(visitor, param) {
        return visitor.visitIdentifier(this, param);
    }
    accept(visitor) {
        return visitor.visitIdentifier(this);
    }
    pass(visitor) {
        return visitor.passIdentifier(this);
    }
    static classAccept(visitor) {
        return visitor.visitVariable(this);
    }
}
exports.Identifier = Identifier;
/**
 * Class containing all valid groupings
 */
class Grouping extends SuffixTermBase {
    /**
     * Grouping constructor
     * @param open open paren token
     * @param expr expression within the grouping
     * @param close close paren token
     */
    constructor(open, expr, close) {
        super();
        this.open = open;
        this.expr = expr;
        this.close = close;
    }
    get start() {
        return this.open.start;
    }
    get end() {
        return this.close.end;
    }
    get ranges() {
        return [this.open, this.expr, this.close];
    }
    toString() {
        return `${this.open.lexeme}${this.expr.toString()}${this.close.lexeme}`;
    }
    toLines() {
        const lines = this.expr.toLines();
        lines[0] = `${this.open.lexeme}${lines[0]}`;
        lines[lines.length - 1] = `${lines[lines.length - 1]}${this.close.lexeme}`;
        return lines;
    }
    acceptParam(visitor, param) {
        return visitor.visitGrouping(this, param);
    }
    accept(visitor) {
        return visitor.visitGrouping(this);
    }
    pass(visitor) {
        return visitor.passGrouping(this);
    }
    static classAccept(visitor) {
        return visitor.visitGrouping(this);
    }
}
exports.Grouping = Grouping;
/**
 * All valid suffix terms
 */
exports.validSuffixTerms = [
    SuffixTrailer,
    SuffixTerm,
    Call,
    ArrayIndex,
    ArrayBracket,
    Delegate,
    Literal,
    Identifier,
    Grouping,
];
const atomTypes = [
    [Literal, grammarNodes_1.createConstant(0.8)],
    [Identifier, grammarNodes_1.createConstant(1)],
    [Grouping, grammarNodes_1.createConstant(0.3)],
];
const suffixTermTrailers = [
    [Call, grammarNodes_1.createConstant(0.8)],
    [ArrayBracket, grammarNodes_1.createConstant(1)],
    [ArrayIndex, grammarNodes_1.createConstant(0.3)],
];
const atom = grammarNodes_1.createGrammarUnion(...atomTypes);
const suffixTermTrailer = grammarNodes_1.createGrammarUnion(...suffixTermTrailers);
SuffixTerm.grammar = [
    atom,
    grammarNodes_1.createGrammarRepeat(grammarNodes_1.createExponential(1.5), suffixTermTrailer),
];
Call.grammar = [
    tokentypes_1.TokenType.bracketOpen,
    grammarNodes_1.createGrammarOptional(grammarNodes_1.createExponential(3), expr_1.expr, grammarNodes_1.createGrammarRepeat(grammarNodes_1.createGamma(1.5, 0.4), tokentypes_1.TokenType.comma, expr_1.expr)),
    tokentypes_1.TokenType.bracketClose,
];
ArrayIndex.grammar = [
    tokentypes_1.TokenType.arrayIndex,
    grammarNodes_1.createGrammarUnion([tokentypes_1.TokenType.integer, grammarNodes_1.createNormal(3, 1)], [tokentypes_1.TokenType.identifier, grammarNodes_1.createNormal(1, 1)]),
];
ArrayBracket.grammar = [tokentypes_1.TokenType.bracketOpen, expr_1.expr, tokentypes_1.TokenType.bracketClose];
Delegate.grammar = [tokentypes_1.TokenType.atSign];
Literal.grammar = [
    grammarNodes_1.createGrammarUnion([tokentypes_1.TokenType.integer, grammarNodes_1.createConstant(1)], [tokentypes_1.TokenType.double, grammarNodes_1.createConstant(1.5)], [tokentypes_1.TokenType.true, grammarNodes_1.createConstant(0.5)], [tokentypes_1.TokenType.false, grammarNodes_1.createConstant(0.5)], [tokentypes_1.TokenType.fileIdentifier, grammarNodes_1.createConstant(0.1)], [tokentypes_1.TokenType.string, grammarNodes_1.createConstant(2)]),
];
Identifier.grammar = [tokentypes_1.TokenType.identifier];
Grouping.grammar = [tokentypes_1.TokenType.bracketOpen, expr_1.expr, tokentypes_1.TokenType.bracketClose];
//# sourceMappingURL=suffixTerm.js.map