"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const typeGuards_1 = require("../utilities/typeGuards");
class TreeTraverse {
    constructor() { }
    // find an statement
    stmtAction(stmt) {
        return stmt.accept(this);
    }
    // find an expression
    exprAction(expr) {
        return expr.accept(this);
    }
    // find an expression
    suffixTermAction(suffixTerm) {
        return suffixTerm.accept(this);
    }
    visitDeclVariable(decl) {
        if (this.nodeAction(decl)) {
            this.exprAction(decl.value);
        }
    }
    visitDeclLock(decl) {
        if (this.nodeAction(decl)) {
            this.exprAction(decl.value);
        }
    }
    visitDeclFunction(decl) {
        if (this.nodeAction(decl)) {
            this.stmtAction(decl.block);
        }
    }
    visitDeclParameter(decl) {
        this.nodeAction(decl);
    }
    visitStmtInvalid(stmt) {
        this.nodeAction(stmt);
    }
    visitBlock(stmt) {
        if (this.nodeAction(stmt)) {
            for (const childStmt of stmt.stmts) {
                this.stmtAction(childStmt);
            }
        }
    }
    visitExpr(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.suffix);
        }
    }
    visitOnOff(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.suffix);
        }
    }
    visitCommand(stmt) {
        this.nodeAction(stmt);
    }
    visitCommandExpr(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.expr);
        }
    }
    visitUnset(stmt) {
        this.nodeAction(stmt);
    }
    visitUnlock(stmt) {
        this.nodeAction(stmt);
    }
    visitSet(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.suffix);
            this.exprAction(stmt.value);
        }
    }
    visitLazyGlobal(stmt) {
        this.nodeAction(stmt);
    }
    visitIf(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.condition);
            this.stmtAction(stmt.body);
            if (!typeGuards_1.empty(stmt.elseStmt)) {
                this.stmtAction(stmt.elseStmt);
            }
        }
    }
    visitElse(stmt) {
        if (this.nodeAction(stmt)) {
            this.stmtAction(stmt.body);
        }
    }
    visitUntil(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.condition);
            this.stmtAction(stmt.body);
        }
    }
    visitFrom(stmt) {
        if (this.nodeAction(stmt)) {
            this.stmtAction(stmt.initializer);
            this.exprAction(stmt.condition);
            this.stmtAction(stmt.increment);
            this.stmtAction(stmt.body);
        }
    }
    visitWhen(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.condition);
            this.stmtAction(stmt.body);
        }
    }
    visitReturn(stmt) {
        if (this.nodeAction(stmt)) {
            if (!typeGuards_1.empty(stmt.value)) {
                this.exprAction(stmt.value);
            }
        }
    }
    visitBreak(stmt) {
        this.nodeAction(stmt);
    }
    visitSwitch(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.target);
        }
    }
    visitFor(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.collection);
            this.stmtAction(stmt.body);
        }
    }
    visitOn(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.suffix);
            this.stmtAction(stmt.body);
        }
    }
    visitToggle(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.suffix);
        }
    }
    visitWait(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.expr);
        }
    }
    visitLog(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.expr);
            this.exprAction(stmt.target);
        }
    }
    visitCopy(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.target);
            this.exprAction(stmt.destination);
        }
    }
    visitRename(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.target);
            this.exprAction(stmt.alternative);
        }
    }
    visitDelete(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.target);
            if (!typeGuards_1.empty(stmt.volume)) {
                this.exprAction(stmt.volume);
            }
        }
    }
    visitRun(stmt) {
        if (this.nodeAction(stmt)) {
            if (!typeGuards_1.empty(stmt.args)) {
                for (const arg of stmt.args) {
                    this.exprAction(arg);
                }
            }
            if (!typeGuards_1.empty(stmt.expr)) {
                this.exprAction(stmt.expr);
            }
        }
    }
    visitRunPath(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.expr);
            if (!typeGuards_1.empty(stmt.args)) {
                for (const arg of stmt.args) {
                    this.exprAction(arg);
                }
            }
        }
    }
    visitRunPathOnce(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.expr);
            if (!typeGuards_1.empty(stmt.args)) {
                for (const arg of stmt.args) {
                    this.exprAction(arg);
                }
            }
        }
    }
    visitCompile(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.target);
            if (!typeGuards_1.empty(stmt.destination)) {
                this.exprAction(stmt.destination);
            }
        }
    }
    visitList(stmt) {
        this.nodeAction(stmt);
    }
    visitEmpty(stmt) {
        this.nodeAction(stmt);
    }
    visitPrint(stmt) {
        if (this.nodeAction(stmt)) {
            this.exprAction(stmt.expr);
            if (!typeGuards_1.empty(stmt.x)) {
                this.exprAction(stmt.x);
            }
            if (!typeGuards_1.empty(stmt.y)) {
                this.exprAction(stmt.y);
            }
        }
    }
    visitExprInvalid(expr) {
        this.nodeAction(expr);
    }
    visitBinary(expr) {
        if (this.nodeAction(expr)) {
            this.exprAction(expr.left);
            this.exprAction(expr.right);
        }
    }
    visitUnary(expr) {
        if (this.nodeAction(expr)) {
            this.exprAction(expr.factor);
        }
    }
    visitFactor(expr) {
        if (this.nodeAction(expr)) {
            this.exprAction(expr.suffix);
            this.exprAction(expr.exponent);
        }
    }
    visitSuffix(expr) {
        if (this.nodeAction(expr)) {
            this.suffixTermAction(expr.suffixTerm);
            if (!typeGuards_1.empty(expr.trailer)) {
                this.suffixTermAction(expr.trailer);
            }
        }
    }
    visitLambda(expr) {
        if (this.nodeAction(expr)) {
            this.stmtAction(expr.block);
        }
    }
    visitSuffixTermInvalid(suffixTerm) {
        this.nodeAction(suffixTerm);
    }
    visitSuffixTrailer(suffixTerm) {
        if (this.nodeAction(suffixTerm)) {
            this.suffixTermAction(suffixTerm.suffixTerm);
            if (!typeGuards_1.empty(suffixTerm.trailer)) {
                this.suffixTermAction(suffixTerm.trailer);
            }
        }
    }
    visitSuffixTerm(suffixTerm) {
        if (this.nodeAction(suffixTerm)) {
            this.suffixTermAction(suffixTerm.atom);
            for (const trailer of suffixTerm.trailers) {
                this.suffixTermAction(trailer);
            }
        }
    }
    visitCall(suffixTerm) {
        if (this.nodeAction(suffixTerm)) {
            for (const arg of suffixTerm.args) {
                this.exprAction(arg);
            }
        }
    }
    visitArrayIndex(suffixTerm) {
        this.nodeAction(suffixTerm);
    }
    visitArrayBracket(suffixTerm) {
        if (this.nodeAction(suffixTerm)) {
            this.exprAction(suffixTerm.index);
        }
    }
    visitDelegate(suffixTerm) {
        this.nodeAction(suffixTerm);
    }
    visitLiteral(suffixTerm) {
        this.nodeAction(suffixTerm);
    }
    visitIdentifier(suffixTerm) {
        this.nodeAction(suffixTerm);
    }
    visitGrouping(suffixTerm) {
        if (this.nodeAction(suffixTerm)) {
            this.exprAction(suffixTerm.expr);
        }
    }
}
exports.TreeTraverse = TreeTraverse;
//# sourceMappingURL=treeTraverse.js.map