'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
const vscode_languageserver_1 = require("vscode-languageserver");
const typeGuards_1 = require("./utilities/typeGuards");
const analyzer_1 = require("./analyzer");
const types_1 = require("./analysis/types");
const serverUtils_1 = require("./utilities/serverUtils");
const logger_1 = require("./utilities/logger");
const initialize_1 = require("./typeChecker/types/primitives/initialize");
const initialize_2 = require("./typeChecker/types/orbital/initialize");
const clean_1 = require("./utilities/clean");
const constants_1 = require("./utilities/constants");
const typeUitlities_1 = require("./typeChecker/typeUitlities");
// get connection primitives based on command argument
const { reader, writer } = serverUtils_1.getConnectionPrimitives(process.argv[2]);
// Create a connection for the server. The connection uses Node's IPC as a transport.
// Also include all preview / proposed LSP features.
exports.connection = vscode_languageserver_1.createConnection(vscode_languageserver_1.ProposedFeatures.all, reader, writer);
// REMOVE ME TODO probably need to refactor the type modules as
// structure and the primitives have a dependnecy loop
initialize_1.primitiveInitializer();
initialize_2.oribitalInitializer();
// default client configuration
const defaultClientConfiguration = {
    kerbalSpaceProgramPath: undefined,
    telnetHost: '127.0.0.1',
    telnetPort: 5410,
    lspPort: 7000,
    completionCase: 'camelcase',
    trace: {
        server: {
            verbosity: 'off',
            format: 'text',
            level: 'error',
        },
    },
};
// create server options object
const server = {
    reader,
    writer,
    workspaceFolder: '',
    workspaceUri: '',
    clientCapability: {
        hasConfiguration: false,
        hasWorkspaceFolder: false,
    },
    keywords: constants_1.keywordCompletions(2 /* camelcase */),
    clientConfig: defaultClientConfiguration,
    analyzer: new analyzer_1.Analyzer(2 /* camelcase */, new logger_1.Logger(exports.connection.console, 1 /* info */), exports.connection.tracer),
};
// Create a simple text document manager. The text document manager
// supports full document sync only
const documents = new vscode_languageserver_1.TextDocuments();
/**
 * Initialize the server from the client
 */
exports.connection.onInitialize((params) => {
    const { capabilities, rootPath, rootUri } = params;
    exports.connection.console.log(`[Server(${process.pid}) ${JSON.stringify(capabilities, undefined, 2)}] Started and initialize received`);
    // does the client support configurations
    server.clientCapability.hasConfiguration = !!(capabilities.workspace && !!capabilities.workspace.configuration);
    // does the client support workspace folders
    server.clientCapability.hasWorkspaceFolder = !!(capabilities.workspace && !!capabilities.workspace.workspaceFolders);
    // get root path if it exists
    if (rootPath) {
        server.analyzer.setPath(rootPath);
        server.workspaceFolder = rootPath;
    }
    // get root uri if it exists
    if (rootUri) {
        server.analyzer.setUri(rootUri);
        server.workspaceUri = rootUri;
    }
    exports.connection.console.log(`[Server(${process.pid}) ${server.workspaceFolder}] Started and initialize received`);
    return {
        capabilities: {
            textDocumentSync: documents.syncKind,
            // Tell the client that the server supports code completion
            completionProvider: {
                resolveProvider: true,
                triggerCharacters: [':', '(', ', '],
            },
            signatureHelpProvider: {
                triggerCharacters: ['(', ',', ', '],
            },
            documentHighlightProvider: true,
            hoverProvider: true,
            referencesProvider: true,
            documentSymbolProvider: true,
            definitionProvider: true,
        },
    };
});
/**
 * Once connection is initialized make additional registrations based
 * on client capability
 */
exports.connection.onInitialized(async () => {
    const { clientCapability } = server;
    // register for all configuration changes.
    if (clientCapability.hasConfiguration) {
        exports.connection.client.register(vscode_languageserver_1.DidChangeConfigurationNotification.type, {
            section: constants_1.serverName,
        });
    }
    // register workspace changes
    if (clientCapability.hasWorkspaceFolder) {
        exports.connection.workspace.onDidChangeWorkspaceFolders(_ => {
            exports.connection.console.log('Workspace folder change event received.');
        });
    }
    server.clientConfig = await getDocumentSettings();
    serverUtils_1.updateServer(server);
});
/**
 * When the client changes a configuration update our server settings
 */
exports.connection.onDidChangeConfiguration(change => {
    const { clientCapability } = server;
    if (clientCapability.hasConfiguration) {
        if (change.settings && constants_1.serverName in change.settings) {
            Object.assign(server.clientConfig, defaultClientConfiguration, change.settings[constants_1.serverName]);
        }
        // update server on client config
        serverUtils_1.updateServer(server);
    }
});
// monitor when the editor opens a document
exports.connection.onDidOpenTextDocument((param) => {
    exports.connection.console.info(`We received ${param.textDocument.uri} was opened`);
});
// monitor when a text document is changed
exports.connection.onDidChangeTextDocument((param) => {
    exports.connection.console.info(`We received ${param.contentChanges.length} file changes`);
});
// monitor file change events
exports.connection.onDidChangeWatchedFiles((change) => {
    exports.connection.console.info(`We received ${change.changes.length} file change events`);
});
// monitor when a text document is closed
exports.connection.onDidCloseTextDocument((param) => {
    exports.connection.console.info(`We received ${param.textDocument.uri} was closed`);
});
// The content of a text document has changed. This event is emitted
// when the text document first opened or when its content has changed.
documents.onDidChangeContent(async (change) => {
    try {
        const diagnosticResults = server.analyzer.validateDocument(change.document.uri, change.document.getText());
        let total = 0;
        const diagnosticMap = new Map();
        // retrieve diagnostics from analyzer
        for await (const diagnostics of diagnosticResults) {
            total += diagnostics.length;
            for (const diagnostic of diagnostics) {
                const uriDiagnostics = diagnosticMap.get(diagnostic.uri);
                if (typeGuards_1.empty(uriDiagnostics)) {
                    diagnosticMap.set(diagnostic.uri, [clean_1.cleanDiagnostic(diagnostic)]);
                }
                else {
                    uriDiagnostics.push(clean_1.cleanDiagnostic(diagnostic));
                }
            }
        }
        // send diagnostics to each document reported
        for (const [uri, diagnostics] of diagnosticMap.entries()) {
            exports.connection.sendDiagnostics({
                uri,
                diagnostics,
            });
        }
        // if not problems found clear out diagnostics
        if (total === 0) {
            exports.connection.sendDiagnostics({
                uri: change.document.uri,
                diagnostics: [],
            });
        }
    }
    catch (e) {
        exports.connection.console.error('kos-language-server Error occured:');
        if (e instanceof Error) {
            exports.connection.console.error(e.message);
            if (!typeGuards_1.empty(e.stack)) {
                exports.connection.console.error(e.stack);
            }
        }
        else {
            exports.connection.console.error(JSON.stringify(e));
        }
    }
});
/**
 * This handler provide completition items capability
 */
exports.connection.onCompletion((completionParams) => {
    const { context } = completionParams;
    try {
        // check if suffix completion
        if (!typeGuards_1.empty(context) && !typeGuards_1.empty(context.triggerCharacter)) {
            const { triggerCharacter } = context;
            if (triggerCharacter === ':') {
                return serverUtils_1.suffixCompletionItems(server.analyzer, completionParams);
            }
        }
        // complete base symbols
        return serverUtils_1.symbolCompletionItems(server.analyzer, completionParams, server.keywords);
        // catch any errors
    }
    catch (err) {
        if (err instanceof Error) {
            exports.connection.console.warn(`${err.message} ${err.stack}`);
        }
        return [];
    }
});
/**
 * This handler provides document highlighting capability
 */
exports.connection.onDocumentHighlight((positionParams) => {
    const { position } = positionParams;
    const { uri } = positionParams.textDocument;
    const locations = server.analyzer.getFileUsageRanges(position, uri);
    return typeGuards_1.empty(locations)
        ? []
        : locations.map(range => ({ range: clean_1.cleanRange(range) }));
});
/**
 * This handlers provides on hover capability
 */
exports.connection.onHover((positionParams) => {
    const { position } = positionParams;
    const { uri } = positionParams.textDocument;
    const token = server.analyzer.getToken(position, uri);
    if (typeGuards_1.empty(token)) {
        return undefined;
    }
    const type = typeUitlities_1.tokenTrackedType(token);
    const { tracker } = token;
    let label;
    let symbolKind;
    if (!typeGuards_1.empty(tracker)) {
        symbolKind = types_1.KsSymbolKind[tracker.declared.symbol.tag];
        label =
            tracker.kind === 0 /* basic */
                ? tracker.declared.symbol.name.lexeme
                : tracker.declared.symbol.name;
    }
    else {
        symbolKind = 'literal';
        label = token.lexeme;
    }
    if (typeGuards_1.empty(type)) {
        return undefined;
    }
    return {
        contents: {
            // Note doesn't does do much other than format it as code
            // may look into adding type def syntax highlighting
            language: 'kos',
            value: `(${symbolKind}) ${label}: ${type.toTypeString()} `,
        },
        range: {
            start: clean_1.cleanPosition(token.start),
            end: clean_1.cleanPosition(token.end),
        },
    };
});
/**
 * This handler provides the find all reference capability
 */
exports.connection.onReferences((referenceParams) => {
    const { position } = referenceParams;
    const { uri } = referenceParams.textDocument;
    const locations = server.analyzer.getUsageLocations(position, uri);
    return locations && locations.map(loc => clean_1.cleanLocation(loc));
});
// This handler provides signature help
exports.connection.onSignatureHelp((documentPosition) => {
    const { position } = documentPosition;
    const { uri } = documentPosition.textDocument;
    const result = server.analyzer.getFunctionAtPosition(position, uri);
    if (typeGuards_1.empty(result))
        return defaultSignature();
    const { tracker, index } = result;
    let label = typeof tracker.declared.symbol.name === 'string'
        ? tracker.declared.symbol.name
        : tracker.declared.symbol.name.lexeme;
    const type = tracker.getType({
        uri,
        range: { start: position, end: position },
    });
    if (typeGuards_1.empty(type)) {
        return defaultSignature();
    }
    switch (type.kind) {
        case 3 /* function */:
        case 2 /* suffix */:
            let start = label.length + 1;
            const { params } = type;
            const paramInfos = [];
            // check if normal or variadic type
            if (Array.isArray(params)) {
                // generate normal labels
                if (params.length > 0) {
                    const labels = [];
                    for (let i = 0; i < params.length - 1; i += 1) {
                        const paramLabel = `${params[i].toTypeString()}, `;
                        paramInfos.push(vscode_languageserver_1.ParameterInformation.create([
                            start,
                            start + paramLabel.length - 2,
                        ]));
                        labels.push(paramLabel);
                        start = start + paramLabel.length;
                    }
                    const paramLabel = `${params[params.length - 1].toTypeString()}`;
                    paramInfos.push(vscode_languageserver_1.ParameterInformation.create([
                        start,
                        start + paramLabel.length,
                    ]));
                    labels.push(paramLabel);
                    label = `${label}(${labels.join('')})`;
                }
            }
            else {
                // generate variadic labels
                const variadicLabel = params.toTypeString();
                paramInfos.push(vscode_languageserver_1.ParameterInformation.create([start, start + variadicLabel.length]));
                label = `${label}(${variadicLabel})`;
            }
            return {
                signatures: [
                    vscode_languageserver_1.SignatureInformation.create(label, undefined, ...paramInfos),
                ],
                activeParameter: index,
                activeSignature: null,
            };
        default:
            return defaultSignature();
    }
});
/**
 * This handler provides document symbols capability
 */
exports.connection.onDocumentSymbol((documentSymbol) => {
    return serverUtils_1.documentSymbols(server.analyzer, documentSymbol);
});
/**
 * This handler provides defintition capability
 */
exports.connection.onDefinition((documentPosition) => {
    const { position } = documentPosition;
    const { uri } = documentPosition.textDocument;
    const location = server.analyzer.getDeclarationLocation(position, uri);
    return location && clean_1.cleanLocation(location);
});
/**
 * This handler provider compleition item resolution capability. This provides
 * additional information for the currently compeltion item selection
 */
exports.connection.onCompletionResolve((item) => {
    try {
        const token = item.data;
        if (!typeGuards_1.empty(token)) {
        }
        return item;
    }
    catch (err) {
        if (err instanceof Error) {
            exports.connection.console.error(`${err.message} ${err.stack}`);
        }
        return item;
    }
});
const getDocumentSettings = () => {
    if (!server.clientCapability.hasConfiguration) {
        return Promise.resolve(defaultClientConfiguration);
    }
    return exports.connection.workspace.getConfiguration({
        scopeUri: server.workspaceUri,
        section: constants_1.serverName,
    });
};
const defaultSignature = () => ({
    signatures: [],
    activeParameter: null,
    activeSignature: null,
});
// Make the text document manager listen on the connection
// for open, change and close text document events
documents.listen(exports.connection);
// Listen on the connection
exports.connection.listen();
//# sourceMappingURL=server.js.map