"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = require("fs");
const path_1 = require("path");
const scanner_1 = require("../scanner/scanner");
const types_1 = require("../parser/types");
const parser_1 = require("../parser/parser");
const tokenCheck_1 = require("../parser/tokenCheck");
const arrayUtils_1 = require("../utilities/arrayUtils");
const tokentypes_1 = require("../entities/tokentypes");
const Expr = __importStar(require("../parser/expr"));
const Decl = __importStar(require("../parser/declare"));
const typeGuards_1 = require("../utilities/typeGuards");
const SuffixTerm = __importStar(require("../parser/suffixTerm"));
// scan source file
const scan = (source) => {
    const scanner = new scanner_1.Scanner(source);
    return scanner.scanTokens();
};
// parse source expression
const parseExpression = (source) => {
    const { tokens, scanErrors } = scan(source);
    const parser = new parser_1.Parser('', tokens);
    return [parser.parseExpression(), scanErrors];
};
// parse source expression
const parse = (source) => {
    const { tokens, scanErrors } = scan(source);
    expect(scanErrors.length).toBe(0);
    const parser = new parser_1.Parser('', tokens);
    return parser.parse();
};
const testDir = path_1.join(__dirname, '../../../kerboscripts/parser_valid/');
const tokenCheck = new tokenCheck_1.TokenCheck();
const walkDir = (dir, callback) => {
    fs_1.readdirSync(dir).forEach(f => {
        const dirPath = path_1.join(dir, f);
        const isDirectory = fs_1.statSync(dirPath).isDirectory();
        isDirectory ? walkDir(dirPath, callback) : callback(path_1.join(dir, f));
    });
};
describe('Parse all test files', () => {
    test('parse all', () => {
        walkDir(testDir, filePath => {
            const kosFile = fs_1.readFileSync(filePath, 'utf8');
            const scanner = new scanner_1.Scanner(kosFile, filePath);
            const { tokens, scanErrors } = scanner.scanTokens();
            expect(scanErrors.length === 0).toBe(true);
            const parser = new parser_1.Parser('', tokens);
            const { parseErrors } = parser.parse();
            expect(parseErrors.length === 0).toBe(true);
        });
    });
    test('parse all validate', () => {
        walkDir(testDir, filePath => {
            const kosFile = fs_1.readFileSync(filePath, 'utf8');
            const scanner1 = new scanner_1.Scanner(kosFile, filePath);
            const scanResults1 = scanner1.scanTokens();
            expect(scanResults1.scanErrors.length).toBe(0);
            const parser1 = new parser_1.Parser('', scanResults1.tokens);
            const parseResults1 = parser1.parse();
            expect(parseResults1.parseErrors.length).toBe(0);
            const prettyKosFile = parseResults1.script.toString();
            const scanner2 = new scanner_1.Scanner(prettyKosFile, filePath);
            const scanResults2 = scanner2.scanTokens();
            expect(scanResults1.scanErrors.length).toBe(0);
            const parser2 = new parser_1.Parser('', scanResults2.tokens);
            const parseResults2 = parser2.parse();
            expect(parseResults2.parseErrors.length).toBe(0);
            const zipped = arrayUtils_1.zip(tokenCheck.orderedTokens(parseResults1.script), tokenCheck.orderedTokens(parseResults2.script));
            for (const [token1, token2] of zipped) {
                expect(token1.lexeme).toBe(token2.lexeme);
                expect(token1.type).toBe(token2.type);
            }
        });
    });
});
const atomTest = (source, type, literal) => {
    return {
        source,
        type,
        literal,
    };
};
const testAtom = (value, testFunct) => {
    expect(value instanceof Expr.Suffix).toBe(true);
    if (value instanceof Expr.Suffix) {
        expect(typeGuards_1.empty(value.trailer)).toBe(true);
        expect(value.suffixTerm instanceof SuffixTerm.SuffixTerm).toBe(true);
        if (value.suffixTerm instanceof SuffixTerm.SuffixTerm) {
            expect(0).toBe(value.suffixTerm.trailers.length);
            testFunct(value.suffixTerm.atom);
        }
    }
};
const testSuffixTerm = (value, atomTest, ...trailerTests) => {
    expect(value instanceof Expr.Suffix).toBe(true);
    if (value instanceof Expr.Suffix) {
        expect(typeGuards_1.empty(value.trailer)).toBe(true);
        expect(value.suffixTerm instanceof SuffixTerm.SuffixTerm).toBe(true);
        if (value.suffixTerm instanceof SuffixTerm.SuffixTerm) {
            expect(trailerTests.length).toBe(value.suffixTerm.trailers.length);
            atomTest(value.suffixTerm.atom);
            for (const [trailerTest, trailer] of arrayUtils_1.zip(trailerTests, value.suffixTerm.trailers)) {
                trailerTest(trailer);
            }
        }
    }
};
const callTest = (source, callee, args) => ({ source, callee, args });
const unaryTest = (source, operator, base, baseLiteral) => {
    return {
        source,
        operator,
        base: {
            expr: base,
            literal: baseLiteral,
        },
    };
};
const binaryTest = (source, operator, left, leftLiteral, right, rightLiteral) => {
    return {
        source,
        operator,
        leftArm: {
            expr: left,
            literal: leftLiteral,
        },
        rightArm: {
            expr: right,
            literal: rightLiteral,
        },
    };
};
describe('Parse expressions', () => {
    // test basic literal
    test('basic valid literal', () => {
        const validExpressions = [
            atomTest('5', tokentypes_1.TokenType.integer, 5),
            atomTest('10e6', tokentypes_1.TokenType.double, 10e6),
            atomTest('"Test string"', tokentypes_1.TokenType.string, 'Test string'),
            atomTest('"true if until"', tokentypes_1.TokenType.string, 'true if until'),
            atomTest('true', tokentypes_1.TokenType.true, true),
            atomTest('false', tokentypes_1.TokenType.false, false),
            atomTest('fileVariable.thing', tokentypes_1.TokenType.fileIdentifier, 'filevariable.thing'),
        ];
        for (const expression of validExpressions) {
            const [{ value, errors }, scanErrors] = parseExpression(expression.source);
            expect(errors.length).toBe(0);
            expect(scanErrors.length).toBe(0);
            testAtom(value, atom => {
                expect(atom instanceof SuffixTerm.Literal).toBe(true);
                if (atom instanceof SuffixTerm.Literal) {
                    expect(atom.token.type).toBe(expression.type);
                    expect(atom.token.literal).toBe(expression.literal);
                }
            });
        }
    });
    // test basic literal
    test('basic invalid literal', () => {
        const validExpressions = [
            atomTest('-', tokentypes_1.TokenType.integer, 5),
            atomTest('"Test string', tokentypes_1.TokenType.string, 'test string'),
        ];
        for (const expression of validExpressions) {
            const [{ errors }, scanErrors] = parseExpression(expression.source);
            expect(errors.length > 0 || scanErrors.length > 0).toBe(true);
        }
    });
    // test basic identifier
    test('basic valid identifier', () => {
        const validExpressions = [
            atomTest('α', tokentypes_1.TokenType.identifier, undefined),
            atomTest('until123OtherStuff', tokentypes_1.TokenType.identifier, undefined),
            atomTest('_variableName', tokentypes_1.TokenType.identifier, undefined),
        ];
        for (const expression of validExpressions) {
            const [{ value, errors }, scanErrors] = parseExpression(expression.source);
            expect(errors.length).toBe(0);
            expect(scanErrors.length).toBe(0);
            testAtom(value, atom => {
                expect(atom instanceof SuffixTerm.Identifier).toBe(true);
                if (atom instanceof SuffixTerm.Identifier) {
                    expect(expression.type).toBe(atom.token.type);
                    expect(expression.literal).toBe(atom.token.literal);
                }
            });
        }
    });
    // test basic identifier
    test('basic invalid identifier', () => {
        const validExpressions = [
            atomTest('11α', tokentypes_1.TokenType.identifier, undefined),
            atomTest('+until123OtherStuff', tokentypes_1.TokenType.identifier, undefined),
            atomTest(',БНЯД', tokentypes_1.TokenType.fileIdentifier, undefined),
            atomTest('БНЯД.БНЯД', tokentypes_1.TokenType.fileIdentifier, undefined),
        ];
        for (const expression of validExpressions) {
            const [{ value, errors }] = parseExpression(expression.source);
            expect(value instanceof SuffixTerm.Identifier).toBe(false);
            expect(errors.length >= 0).toBe(true);
        }
    });
    // test basic identifier
    test('valid call', () => {
        const validExpressions = [
            callTest('test(4, "car")', 'test', [
                SuffixTerm.Literal,
                SuffixTerm.Literal,
            ]),
            callTest('БНЯД(varName, 14.3)', 'бняд', [
                SuffixTerm.Identifier,
                SuffixTerm.Literal,
            ]),
            callTest('_variableName()', '_variablename', []),
        ];
        for (const expression of validExpressions) {
            const [{ value, errors }, scannerErrors] = parseExpression(expression.source);
            expect(errors.length).toBe(0);
            expect(scannerErrors.length).toBe(0);
            testSuffixTerm(value, atom => {
                expect(atom instanceof SuffixTerm.Identifier).toBe(true);
                if (atom instanceof SuffixTerm.Identifier) {
                    expect(expression.callee).toBe(atom.token.lookup);
                }
            }, trailer => {
                expect(trailer instanceof SuffixTerm.Call).toBe(true);
                if (trailer instanceof SuffixTerm.Call) {
                    expect(expression.args.length).toBe(trailer.args.length);
                    for (let i = 0; i < expression.args.length; i += 1) {
                        testAtom(trailer.args[i], atom => {
                            expect(atom instanceof expression.args[i]).toBe(true);
                        });
                    }
                }
            });
        }
    });
    // test basic identifier
    test('invalid call', () => {
        const validExpressions = [
            atomTest('11α', tokentypes_1.TokenType.identifier, undefined),
            atomTest('+until123OtherStuff', tokentypes_1.TokenType.identifier, undefined),
            atomTest(',БНЯД', tokentypes_1.TokenType.fileIdentifier, undefined),
        ];
        for (const expression of validExpressions) {
            const [{ value }, scannerErrors] = parseExpression(expression.source);
            expect(value instanceof SuffixTerm.Identifier).toBe(false);
            expect(scannerErrors.length === 0).toBe(true);
        }
    });
    test('valid unary', () => {
        const validExpressions = [
            unaryTest('+ 5', tokentypes_1.TokenType.plus, SuffixTerm.Literal, 5),
            unaryTest('defined other', tokentypes_1.TokenType.defined, SuffixTerm.Identifier, undefined),
            unaryTest('not false', tokentypes_1.TokenType.not, SuffixTerm.Literal, false),
            unaryTest('-suffix:call(10, 5)', tokentypes_1.TokenType.minus, Expr.Suffix, undefined),
        ];
        for (const expression of validExpressions) {
            const [{ value, errors }, scannerErrors] = parseExpression(expression.source);
            expect(value instanceof Expr.Unary).toBe(true);
            expect(errors.length === 0).toBe(true);
            expect(scannerErrors.length === 0).toBe(true);
            if (value instanceof Expr.Unary) {
                expect(value.operator.type).toBe(expression.operator);
                if (expression.base.expr.prototype instanceof SuffixTerm.SuffixTermBase) {
                    testSuffixTerm(value.factor, atom => {
                        expect(atom instanceof expression.base.expr).toBe(true);
                        if (atom instanceof SuffixTerm.Literal) {
                            expect(expression.base.literal).toBe(atom.token.literal);
                        }
                    });
                }
                else {
                    expect(value.factor instanceof expression.base.expr).toBe(true);
                }
            }
        }
    });
    // test basic binary
    test('valid binary', () => {
        const validExpressions = [
            binaryTest('10 + 5', tokentypes_1.TokenType.plus, SuffixTerm.Literal, 10, SuffixTerm.Literal, 5),
            binaryTest('"example" + "other"', tokentypes_1.TokenType.plus, SuffixTerm.Literal, 'example', SuffixTerm.Literal, 'other'),
            binaryTest('variable <> false', tokentypes_1.TokenType.notEqual, SuffixTerm.Identifier, undefined, SuffixTerm.Literal, false),
            binaryTest('suffix:call(10, 5) <= 10 ^ 3', tokentypes_1.TokenType.lessEqual, Expr.Suffix, undefined, Expr.Factor, undefined),
        ];
        for (const expression of validExpressions) {
            const [{ value, errors }, scannerErrors] = parseExpression(expression.source);
            expect(value instanceof Expr.Binary).toBe(true);
            expect(errors.length === 0).toBe(true);
            expect(scannerErrors.length === 0).toBe(true);
            if (value instanceof Expr.Binary) {
                expect(value.operator.type).toBe(expression.operator);
                if (expression.leftArm.expr.prototype instanceof SuffixTerm.SuffixTermBase) {
                    testSuffixTerm(value.left, atom => {
                        expect(atom instanceof expression.leftArm.expr).toBe(true);
                        if (atom instanceof SuffixTerm.Literal) {
                            expect(expression.leftArm.literal).toBe(atom.token.literal);
                        }
                    });
                }
                else {
                    expect(value.left instanceof expression.leftArm.expr).toBe(true);
                }
                if (expression.rightArm.expr.prototype instanceof
                    SuffixTerm.SuffixTermBase) {
                    testSuffixTerm(value.right, atom => {
                        expect(atom instanceof expression.rightArm.expr).toBe(true);
                        if (atom instanceof SuffixTerm.Literal) {
                            expect(expression.rightArm.literal).toBe(atom.token.literal);
                        }
                    });
                }
                else {
                    expect(value.right instanceof expression.rightArm.expr).toBe(true);
                }
            }
        }
    });
});
const varDeclareTest = (source, identifier, scope, value) => {
    return {
        source,
        identifier,
        scope,
        value,
    };
};
const lockDeclareTest = (source, identifier, value, scope) => {
    return {
        source,
        identifier,
        scope,
        value,
    };
};
describe('Parse statement', () => {
    test('valid variable declarations', () => {
        const validDeclarations = [
            varDeclareTest('local a is { return 10. }.', 'a', types_1.ScopeKind.local, Expr.Lambda),
            varDeclareTest('declare other to "example" + "another".', 'other', types_1.ScopeKind.local, Expr.Binary),
            varDeclareTest('declare global another is thing:withSuffix[10].', 'another', types_1.ScopeKind.global, Expr.Suffix),
        ];
        for (const declaration of validDeclarations) {
            const { script, parseErrors } = parse(declaration.source);
            expect(parseErrors.length).toBe(0);
            expect(script.stmts.length).toBe(1);
            expect(script.runStmts.length).toBe(0);
            const [stmt] = script.stmts;
            expect(stmt instanceof Decl.Var).toBe(true);
            if (stmt instanceof Decl.Var) {
                expect(stmt.identifier.lexeme).toBe(declaration.identifier);
                expect(stmt.scope.type).toBe(declaration.scope);
                expect(stmt.value instanceof declaration.value).toBe(true);
            }
        }
    });
    test('valid lock declarations', () => {
        const validDeclarations = [
            lockDeclareTest('lock a to { return 10. }.', 'a', Expr.Lambda, undefined),
            lockDeclareTest('local lock other to "example" + "another".', 'other', Expr.Binary, types_1.ScopeKind.local),
            lockDeclareTest('declare global lock another to thing:withSuffix[10].', 'another', Expr.Suffix, types_1.ScopeKind.global),
        ];
        for (const declaration of validDeclarations) {
            const { script, parseErrors } = parse(declaration.source);
            expect(parseErrors.length).toBe(0);
            expect(script.stmts.length).toBe(1);
            expect(script.runStmts.length).toBe(0);
            const [stmt] = script.stmts;
            expect(stmt instanceof Decl.Lock).toBe(true);
            if (stmt instanceof Decl.Lock) {
                expect(stmt.identifier.lexeme).toBe(declaration.identifier);
                if (typeGuards_1.empty(declaration.scope)) {
                    expect(stmt.scope).toBeUndefined();
                }
                else {
                    if (typeGuards_1.empty(stmt.scope)) {
                        fail();
                    }
                    else {
                        expect(stmt.scope.type).toBe(declaration.scope);
                    }
                }
                expect(stmt.value instanceof declaration.value).toBe(true);
            }
        }
    });
});
//# sourceMappingURL=parser.spec.js.map