"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const parser_1 = require("../parser/parser");
const scanner_1 = require("../scanner/scanner");
const typeGuards_1 = require("../utilities/typeGuards");
const positionUtils_1 = require("../utilities/positionUtils");
const vscode_languageserver_1 = require("vscode-languageserver");
const structure_1 = require("../typeChecker/types/primitives/structure");
const fs_1 = require("fs");
const path_1 = require("path");
const arrayUtils_1 = require("../utilities/arrayUtils");
const symbolTableBuilder_1 = require("../analysis/symbolTableBuilder");
const preResolver_1 = require("../analysis/preResolver");
const types_1 = require("../analysis/types");
const resolver_1 = require("../analysis/resolver");
const functionScan_1 = require("../analysis/functionScan");
const Decl = __importStar(require("../parser/declare"));
const standardLibrary_1 = require("../analysis/standardLibrary");
const localResolver_1 = require("../analysis/localResolver");
const Stmt = __importStar(require("../parser/stmt"));
const setResolver_1 = require("../analysis/setResolver");
const marker_1 = require("../entities/marker");
const fakeUri = 'C:\\fake.ks';
// parse source
const parseSource = (source) => {
    const scanner = new scanner_1.Scanner(source, fakeUri);
    const scan = scanner.scanTokens();
    const parser = new parser_1.Parser(fakeUri, scan.tokens);
    const parse = parser.parse();
    return { scan, parse };
};
const resolveSource = (source, standardLib = false) => {
    const result = parseSource(source);
    const symbolTableBuilder = new symbolTableBuilder_1.SymbolTableBuilder(fakeUri);
    if (standardLib) {
        symbolTableBuilder.linkTable(standardLibrary_1.standardLibraryBuilder(0 /* lowercase */));
    }
    const functionResolver = new preResolver_1.PreResolver(result.parse.script, symbolTableBuilder);
    const resolver = new resolver_1.Resolver(result.parse.script, symbolTableBuilder);
    const preResolverError = functionResolver.resolve();
    const resolverErrors = resolver.resolve();
    const ususedErrors = symbolTableBuilder.findUnused();
    return {
        ...result,
        resolveDiagnostics: preResolverError.concat(resolverErrors, ususedErrors),
        table: symbolTableBuilder.build(),
    };
};
const makeRange = (sLine, schar, eLine, echar) => {
    return {
        start: {
            line: sLine,
            character: schar,
        },
        end: {
            line: eLine,
            character: echar,
        },
    };
};
const noParseErrors = (result) => {
    expect(result.scan.scanErrors.length).toBe(0);
    expect(result.parse.parseErrors.length).toBe(0);
};
const noErrors = (result) => {
    expect(result.scan.scanErrors.length).toBe(0);
    expect(result.parse.parseErrors.length).toBe(0);
    expect(result.resolveDiagnostics.length).toBe(0);
};
const setSource = `
set x to 10.
set x to "cat".

set y to false.
set x to y.
print x.
`;
const lazyListSource = `
function example {
  parameter y.
  list files in y.
  print y.
}

list files in x.
print x.
`;
describe('Resolver tracking', () => {
    // test basic identifier
    test('basic set test', () => {
        const results = resolveSource(setSource);
        noErrors(results);
        const { table } = results;
        const symbols = table.fileSymbols();
        const names = new Map(symbols.map((s) => [s.name.lexeme, s]));
        expect(names.has('x')).toBe(true);
        expect(names.has('y')).toBe(true);
        const xWrap = names.get('x');
        const yWrap = names.get('y');
        expect(typeGuards_1.empty(xWrap)).toBe(false);
        expect(typeGuards_1.empty(yWrap)).toBe(false);
        const x = typeGuards_1.unWrap(xWrap);
        const y = typeGuards_1.unWrap(yWrap);
        expect(x.name.lexeme).toBe('x');
        expect(y.name.lexeme).toBe('y');
        expect(x.tag).toBe(types_1.KsSymbolKind.variable);
        expect(y.tag).toBe(types_1.KsSymbolKind.variable);
    });
    // test basic identifier
    test('basic tracker set test', () => {
        const results = resolveSource(setSource);
        noErrors(results);
        const { table } = results;
        const xTrack = table.scopedNamedTracker({ line: 0, character: 0 }, 'x');
        const yTrack = table.scopedNamedTracker({ line: 0, character: 0 }, 'y');
        expect(typeGuards_1.empty(xTrack)).toBe(false);
        expect(typeGuards_1.empty(yTrack)).toBe(false);
        const [x, y] = typeGuards_1.unWrapMany(xTrack, yTrack);
        const xRange = {
            start: { line: 1, character: 4 },
            end: { line: 1, character: 5 },
        };
        const yRange = {
            start: { line: 4, character: 4 },
            end: { line: 4, character: 5 },
        };
        expect(positionUtils_1.rangeEqual(x.declared.range, xRange)).toBe(true);
        expect(positionUtils_1.rangeEqual(y.declared.range, yRange)).toBe(true);
        // check the tracked declared properties
        expect(x.declared.uri).toBe(fakeUri);
        expect(y.declared.uri).toBe(fakeUri);
        // should only be structure as it is the default placeholder
        expect(x.declared.type).toBe(structure_1.structureType);
        expect(y.declared.type).toBe(structure_1.structureType);
        expect(x.declared.symbol.name.lexeme).toBe('x');
        expect(y.declared.symbol.name.lexeme).toBe('y');
        expect(x.declared.symbol.tag).toBe(types_1.KsSymbolKind.variable);
        expect(y.declared.symbol.tag).toBe(types_1.KsSymbolKind.variable);
        // check the tracked set properties
        expect(x.sets.length).toBe(2);
        expect(y.sets.length).toBe(0);
        // check the tracked set properties
        expect(x.usages.length).toBe(1);
        expect(y.usages.length).toBe(1);
        const [yUsage] = y.usages;
        const [xUsage] = x.usages;
        const xUsageRange = {
            start: { line: 6, character: 6 },
            end: { line: 6, character: 7 },
        };
        const yUsageRange = {
            start: { line: 5, character: 9 },
            end: { line: 5, character: 10 },
        };
        expect(positionUtils_1.rangeEqual(xUsage.range, xUsageRange)).toBe(true);
        expect(positionUtils_1.rangeEqual(yUsage.range, yUsageRange)).toBe(true);
        expect(xUsage.uri).toBe(fakeUri);
        expect(yUsage.uri).toBe(fakeUri);
    });
    test('lazy list test', () => {
        const results = resolveSource(lazyListSource);
        noErrors(results);
        const { table } = results;
        const symbols = table.fileSymbols();
        const names = new Map(symbols.map((s) => [s.name.lexeme, s]));
        expect(names.has('x')).toBe(true);
        expect(names.has('y')).toBe(true);
        const xWrap = names.get('x');
        const yWrap = names.get('y');
        expect(typeGuards_1.empty(xWrap)).toBe(false);
        expect(typeGuards_1.empty(yWrap)).toBe(false);
        const x = typeGuards_1.unWrap(xWrap);
        const y = typeGuards_1.unWrap(yWrap);
        expect(x.name.lexeme).toBe('x');
        expect(y.name.lexeme).toBe('y');
        expect(x.tag).toBe(types_1.KsSymbolKind.variable);
        expect(y.tag).toBe(types_1.KsSymbolKind.parameter);
    });
    const symbolKindPath = path_1.join(__dirname, '../../../kerboscripts/parser_valid/unitTests/scannertest.ks');
    // test basic identifier
    test('symbol kind', () => {
        const symbolKindSource = fs_1.readFileSync(symbolKindPath, 'utf8');
        const results = resolveSource(symbolKindSource, true);
        noErrors(results);
        let outOfScope;
        let range;
        const { table } = results;
        const testTracker = table.scopedNamedTracker({ line: 0, character: 0 }, 'test');
        // check test function
        expect(testTracker).not.toBeUndefined();
        if (testTracker !== undefined) {
            const { name, tag } = testTracker.declared.symbol;
            expect(name.lexeme).toBe('test');
            range = makeRange(3, 16, 3, 20);
            expect(positionUtils_1.rangeEqual(name, range)).toBe(true);
            expect(tag).toBe(types_1.KsSymbolKind.function);
            expect(testTracker.usages.length).toBe(1);
            expect(testTracker.sets.length).toBe(0);
        }
        // check param a
        const aParamTracker = table.scopedNamedTracker({ line: 4, character: 0 }, 'a');
        outOfScope = table.scopedNamedTracker(vscode_languageserver_1.Position.create(0, 0), 'a');
        expect(outOfScope).toBeUndefined();
        expect(aParamTracker).not.toBeUndefined();
        if (aParamTracker !== undefined) {
            const { name, tag } = aParamTracker.declared.symbol;
            expect(name.lexeme).toBe('a');
            range = makeRange(4, 14, 4, 15);
            expect(positionUtils_1.rangeEqual(name, range)).toBeTruthy();
            expect(tag).toBe(types_1.KsSymbolKind.parameter);
            expect(aParamTracker.usages.length).toBe(1);
            expect(aParamTracker.sets.length).toBe(0);
        }
        // check param b
        const bParamTracker = table.scopedNamedTracker({ line: 4, character: 0 }, 'b');
        outOfScope = table.scopedNamedTracker(vscode_languageserver_1.Position.create(0, 0), 'b');
        expect(outOfScope).toBeUndefined();
        expect(bParamTracker).not.toBeUndefined();
        if (bParamTracker !== undefined) {
            const { name, tag } = bParamTracker.declared.symbol;
            expect(name.lexeme).toBe('b');
            range = makeRange(4, 17, 4, 18);
            expect(positionUtils_1.rangeEqual(name, range)).toBeTruthy();
            expect(tag).toBe(types_1.KsSymbolKind.parameter);
            expect(bParamTracker.usages.length).toBe(1);
            expect(bParamTracker.sets.length).toBe(0);
        }
        // check loop i
        const iTracker = table.scopedNamedTracker({ line: 8, character: 0 }, 'i');
        outOfScope = table.scopedNamedTracker(vscode_languageserver_1.Position.create(6, 0), 'i');
        expect(outOfScope).toBeUndefined();
        expect(iTracker).not.toBeUndefined();
        if (iTracker !== undefined) {
            const { name, tag } = iTracker.declared.symbol;
            expect(name.lexeme).toBe('i');
            range = makeRange(7, 8, 7, 9);
            expect(positionUtils_1.rangeEqual(name, range)).toBeTruthy();
            expect(tag).toBe(types_1.KsSymbolKind.variable);
            expect(iTracker.usages.length).toBe(1);
            expect(iTracker.sets.length).toBe(0);
        }
        // check loop x
        const xTracker = table.scopedNamedTracker({ line: 13, character: 0 }, 'x');
        outOfScope = table.scopedNamedTracker(vscode_languageserver_1.Position.create(15, 0), 'x');
        expect(outOfScope).toBeUndefined();
        expect(xTracker).not.toBeUndefined();
        if (xTracker !== undefined) {
            const { name, tag } = xTracker.declared.symbol;
            expect(name.lexeme).toBe('x');
            range = makeRange(11, 17, 11, 18);
            expect(positionUtils_1.rangeEqual(name, range)).toBeTruthy();
            expect(tag).toBe(types_1.KsSymbolKind.variable);
            expect(xTracker.usages.length).toBe(3);
            expect(xTracker.sets.length).toBe(1);
        }
        // check t lock
        const tTracker = table.scopedNamedTracker({ line: 16, character: 0 }, 't');
        expect(tTracker).not.toBeUndefined();
        if (tTracker !== undefined) {
            const { name, tag } = tTracker.declared.symbol;
            expect(name.lexeme).toBe('t');
            range = makeRange(17, 5, 17, 6);
            expect(positionUtils_1.rangeEqual(name, range)).toBeTruthy();
            expect(tag).toBe(types_1.KsSymbolKind.lock);
            expect(tTracker.usages.length).toBe(2);
            expect(tTracker.sets.length).toBe(0);
        }
    });
    const allNodePath = path_1.join(__dirname, '../../../kerboscripts/parser_valid/unitTests/allLanguage.ks');
    const allNodeLocations = [
        { start: new marker_1.Marker(8, 4), end: new marker_1.Marker(8, 35) },
        { start: new marker_1.Marker(9, 4), end: new marker_1.Marker(9, 48) },
        { start: new marker_1.Marker(10, 4), end: new marker_1.Marker(10, 30) },
    ];
    test('all tree nodes', () => {
        const symbolKindSource = fs_1.readFileSync(allNodePath, 'utf8');
        const results = resolveSource(symbolKindSource, true);
        expect(results.scan.scanErrors.length).toBe(0);
        expect(results.parse.parseErrors.length).toBe(0);
        expect(results.resolveDiagnostics.length).toBe(allNodeLocations.length);
        // spot each deprecated error
        for (const [error, location] of arrayUtils_1.zip(results.resolveDiagnostics, allNodeLocations)) {
            expect(vscode_languageserver_1.DiagnosticSeverity.Warning).toBe(error.severity);
            expect(location.start).toEqual(error.range.start);
            expect(location.end).toEqual(error.range.end);
        }
        const { table } = results;
        const shipTracker = table.scopedNamedTracker({ line: 0, character: 0 }, 'ship');
        // check test function
        expect(shipTracker).not.toBeUndefined();
        if (shipTracker !== undefined) {
            const { name, tag } = shipTracker.declared.symbol;
            expect(name.lexeme).toBe('ship');
            expect(tag).toBe(types_1.KsSymbolKind.variable);
            expect(shipTracker.usages.length).toBe(5);
            expect(shipTracker.sets.length).toBe(0);
        }
        const bodyTracker = table.scopedNamedTracker({ line: 0, character: 0 }, 'body');
        // check test function
        expect(bodyTracker).not.toBeUndefined();
        if (bodyTracker !== undefined) {
            const { name, tag } = bodyTracker.declared.symbol;
            expect(name.lexeme).toBe('body');
            expect(tag).toBe(types_1.KsSymbolKind.variable);
            expect(bodyTracker.usages.length).toBe(2);
            expect(bodyTracker.sets.length).toBe(0);
        }
        const funcTracker = table.scopedNamedTracker({ line: 0, character: 0 }, 'func');
        // check test function
        expect(funcTracker).not.toBeUndefined();
        if (funcTracker !== undefined) {
            const { name, tag } = funcTracker.declared.symbol;
            expect(name.lexeme).toBe('func');
            expect(tag).toBe(types_1.KsSymbolKind.function);
            expect(funcTracker.usages.length).toBe(1);
            expect(funcTracker.sets.length).toBe(0);
        }
    });
});
const listSource = `
@lazyglobal off.
list files in x.
`;
describe('Resolver errors', () => {
    const listLocations = [
        { start: new marker_1.Marker(2, 14), end: new marker_1.Marker(2, 15) },
    ];
    // test basic identifier
    test('list command', () => {
        const results = resolveSource(listSource);
        expect(results.scan.scanErrors.length).toBe(0);
        expect(results.parse.parseErrors.length).toBe(0);
        expect(results.resolveDiagnostics.length).toBe(listLocations.length);
        for (const [error, location] of arrayUtils_1.zip(results.resolveDiagnostics, listLocations)) {
            expect(vscode_languageserver_1.DiagnosticSeverity.Error).toBe(error.severity);
            expect(location.start).toEqual(error.range.start);
            expect(location.end).toEqual(error.range.end);
        }
    });
    const definedPath = path_1.join(__dirname, '../../../kerboscripts/parser_valid/unitTests/definedtest.ks');
    const definedLocations = [
        { start: new marker_1.Marker(7, 42), end: new marker_1.Marker(7, 46) },
        { start: new marker_1.Marker(10, 42), end: new marker_1.Marker(10, 46) },
        { start: new marker_1.Marker(26, 43), end: new marker_1.Marker(26, 59) },
        { start: new marker_1.Marker(32, 41), end: new marker_1.Marker(32, 51) },
        { start: new marker_1.Marker(33, 41), end: new marker_1.Marker(33, 57) },
        { start: new marker_1.Marker(51, 43), end: new marker_1.Marker(51, 54) },
        { start: new marker_1.Marker(55, 41), end: new marker_1.Marker(55, 52) },
        { start: new marker_1.Marker(56, 41), end: new marker_1.Marker(56, 52) },
    ];
    // test basic identifier
    test('defined command', () => {
        const defineSource = fs_1.readFileSync(definedPath, 'utf8');
        const results = resolveSource(defineSource);
        expect(results.scan.scanErrors.length).toBe(0);
        expect(results.parse.parseErrors.length).toBe(0);
        expect(results.resolveDiagnostics.length).toBe(definedLocations.length);
        const sortedErrors = results.resolveDiagnostics
            .sort((a, b) => a.range.start.line - b.range.start.line);
        for (const [error, location] of arrayUtils_1.zip(sortedErrors, definedLocations)) {
            expect(error.severity).toBe(vscode_languageserver_1.DiagnosticSeverity.Warning);
            expect(location.start).toEqual(error.range.start);
            expect(location.end).toEqual(error.range.end);
        }
    });
    const usedPath = path_1.join(__dirname, '../../../kerboscripts/parser_valid/unitTests/usedtest.ks');
    const usedLocations = [
        { start: new marker_1.Marker(1, 6), end: new marker_1.Marker(1, 7) },
        { start: new marker_1.Marker(2, 6), end: new marker_1.Marker(2, 7) },
    ];
    // test basic identifier
    test('used symbols', () => {
        const usedSource = fs_1.readFileSync(usedPath, 'utf8');
        const results = resolveSource(usedSource);
        expect(results.scan.scanErrors.length).toBe(0);
        expect(results.parse.parseErrors.length).toBe(0);
        expect(results.resolveDiagnostics.length > 0).toBe(true);
        for (const [error, location] of arrayUtils_1.zip(results.resolveDiagnostics, usedLocations)) {
            expect(vscode_languageserver_1.DiagnosticSeverity.Warning).toBe(error.severity);
            expect(location.start).toEqual(error.range.start);
            expect(location.end).toEqual(error.range.end);
        }
    });
    const shadowPath = path_1.join(__dirname, '../../../kerboscripts/parser_valid/unitTests/shadowtest.ks');
    const shadowedLocations = [
        { start: new marker_1.Marker(4, 10), end: new marker_1.Marker(4, 11) },
    ];
    // test shadowing
    test('shadowed symbols', () => {
        const shadowedSource = fs_1.readFileSync(shadowPath, 'utf8');
        const results = resolveSource(shadowedSource);
        expect(results.scan.scanErrors.length).toBe(0);
        expect(results.parse.parseErrors.length).toBe(0);
        expect(results.resolveDiagnostics.length > 0).toBe(true);
        for (const [error, location] of arrayUtils_1.zip(results.resolveDiagnostics, shadowedLocations)) {
            expect(vscode_languageserver_1.DiagnosticSeverity.Warning).toBe(error.severity);
            expect(location.start).toEqual(error.range.start);
            expect(location.end).toEqual(error.range.end);
        }
    });
    const deferredPath = path_1.join(__dirname, '../../../kerboscripts/parser_valid/unitTests/deferredtest.ks');
    const deferredLocations = [
        { start: new marker_1.Marker(3, 10), end: new marker_1.Marker(3, 11) },
        { start: new marker_1.Marker(4, 10), end: new marker_1.Marker(4, 11) },
        { start: new marker_1.Marker(7, 3), end: new marker_1.Marker(7, 4) },
        { start: new marker_1.Marker(8, 10), end: new marker_1.Marker(8, 11) },
        { start: new marker_1.Marker(16, 14), end: new marker_1.Marker(16, 15) },
    ];
    // test deferred resolving
    test('deferred nodes', () => {
        const deferredSource = fs_1.readFileSync(deferredPath, 'utf8');
        const results = resolveSource(deferredSource);
        expect(results.scan.scanErrors.length).toBe(0);
        expect(results.parse.parseErrors.length).toBe(0);
        expect(results.resolveDiagnostics.length > 0).toBe(true);
        for (const [error, location] of arrayUtils_1.zip(results.resolveDiagnostics, deferredLocations)) {
            expect(error.severity).toBe(vscode_languageserver_1.DiagnosticSeverity.Hint);
            expect(location.start).toEqual(error.range.start);
            expect(location.end).toEqual(error.range.end);
        }
    });
    const breakPath = path_1.join(__dirname, '../../../kerboscripts/parser_valid/unitTests/breaktest.ks');
    const breakLocations = [
        { start: new marker_1.Marker(9, 0), end: new marker_1.Marker(9, 5) },
        { start: new marker_1.Marker(21, 0), end: new marker_1.Marker(21, 5) },
    ];
    // invalid break locations
    test('break locations', () => {
        const deferredSource = fs_1.readFileSync(breakPath, 'utf8');
        const results = resolveSource(deferredSource);
        expect(results.scan.scanErrors.length).toBe(0);
        expect(results.parse.parseErrors.length).toBe(0);
        expect(results.resolveDiagnostics.length > 0).toBe(true);
        for (const [error, location] of arrayUtils_1.zip(results.resolveDiagnostics, breakLocations)) {
            expect(error.severity).toBe(vscode_languageserver_1.DiagnosticSeverity.Error);
            expect(location.start).toEqual(error.range.start);
            expect(location.end).toEqual(error.range.end);
        }
    });
    const returnPath = path_1.join(__dirname, '../../../kerboscripts/parser_valid/unitTests/returntest.ks');
    const returnLocations = [
        { start: new marker_1.Marker(14, 4), end: new marker_1.Marker(14, 10) },
    ];
    // invalid return locations
    test('return locations', () => {
        const deferredSource = fs_1.readFileSync(returnPath, 'utf8');
        const results = resolveSource(deferredSource);
        expect(results.scan.scanErrors.length).toBe(0);
        expect(results.parse.parseErrors.length).toBe(0);
        expect(results.resolveDiagnostics.length > 0).toBe(true);
        for (const [error, location] of arrayUtils_1.zip(results.resolveDiagnostics, returnLocations)) {
            expect(error.severity).toBe(vscode_languageserver_1.DiagnosticSeverity.Error);
            expect(location.start).toEqual(error.range.start);
            expect(location.end).toEqual(error.range.end);
        }
    });
});
const emptyFunction = `
function empty { }
`;
const returnFunction = `
function returning {
  return 10.
}
`;
const complicatedFunction = `
function complicated {
  parameter first.
  parameter x, y, other is 10.

  if x < 10 {
    return other.
  }

  local total is 0.
  for i in y {
    for j in first {
      set total to total + other + j.

      if total > 1e3 {
        return total.
      }
    }
  }

  local function inner {
    parameter a.
    print(a).

    return a.
  }

  return 10.
}
`;
describe('Function Scan', () => {
    test('empty function', () => {
        const parseResult = parseSource(emptyFunction);
        noParseErrors(parseResult);
        const { script } = parseResult.parse;
        expect(script.stmts.length).toBe(1);
        const [funcStmt] = script.stmts;
        expect(funcStmt).toBeInstanceOf(Decl.Func);
        const funcScanner = new functionScan_1.FunctionScan();
        if (funcStmt instanceof Decl.Func) {
            const scanResult = funcScanner.scan(funcStmt.block);
            expect(scanResult).not.toBeUndefined();
            if (!typeGuards_1.empty(scanResult)) {
                expect(scanResult.return).toBe(false);
                expect(scanResult.requiredParameters).toBe(0);
                expect(scanResult.optionalParameters).toBe(0);
            }
        }
    });
    test('returning function', () => {
        const parseResult = parseSource(returnFunction);
        noParseErrors(parseResult);
        const { script } = parseResult.parse;
        expect(script.stmts.length).toBe(1);
        const [funcStmt] = script.stmts;
        expect(funcStmt).toBeInstanceOf(Decl.Func);
        const funcScanner = new functionScan_1.FunctionScan();
        if (funcStmt instanceof Decl.Func) {
            const scanResult = funcScanner.scan(funcStmt.block);
            expect(scanResult).not.toBeUndefined();
            if (!typeGuards_1.empty(scanResult)) {
                expect(scanResult.return).toBe(true);
                expect(scanResult.requiredParameters).toBe(0);
                expect(scanResult.optionalParameters).toBe(0);
            }
        }
    });
    test('complicated function', () => {
        const parseResult = parseSource(complicatedFunction);
        noParseErrors(parseResult);
        const { script } = parseResult.parse;
        expect(script.stmts.length).toBe(1);
        const [funcStmt] = script.stmts;
        expect(funcStmt).toBeInstanceOf(Decl.Func);
        const funcScanner = new functionScan_1.FunctionScan();
        if (funcStmt instanceof Decl.Func) {
            const scanResult = funcScanner.scan(funcStmt.block);
            expect(scanResult).not.toBeUndefined();
            if (!typeGuards_1.empty(scanResult)) {
                expect(scanResult.return).toBe(true);
                expect(scanResult.requiredParameters).toBe(3);
                expect(scanResult.optionalParameters).toBe(1);
            }
        }
    });
});
describe('Local Resolver', () => {
    test('local resolver test 1', () => {
        const source = 'set a to Thing:other[used1]:finally(used2, used3).';
        const result = parseSource(source);
        noParseErrors(result);
        const { script } = result.parse;
        const stmt = script.stmts[0];
        const resolver = new localResolver_1.LocalResolver();
        if (stmt instanceof Stmt.Set) {
            const resolverResult = resolver.resolveExpr(stmt.value);
            expect(resolverResult.length).toBe(4);
            const [thing, used1, used2, used3] = resolverResult;
            expect(thing.lexeme).toBe('Thing');
            expect(used1.lexeme).toBe('used1');
            expect(used2.lexeme).toBe('used2');
            expect(used3.lexeme).toBe('used3');
        }
        else {
            expect(true).toBeFalsy();
        }
    });
    test('local resolver test 2', () => {
        const source = 'set a to (first:suffix(nest:call(1), array["yo"]) ^ exponent) < example#3.';
        const result = parseSource(source);
        noParseErrors(result);
        const { script } = result.parse;
        const stmt = script.stmts[0];
        const resolver = new localResolver_1.LocalResolver();
        if (stmt instanceof Stmt.Set) {
            const resolverResult = resolver.resolveExpr(stmt.value);
            expect(resolverResult.length).toBe(5);
            const [first, nest, array, exponent, example] = resolverResult;
            expect(first.lexeme).toBe('first');
            expect(nest.lexeme).toBe('nest');
            expect(array.lexeme).toBe('array');
            expect(exponent.lexeme).toBe('exponent');
            expect(example.lexeme).toBe('example');
        }
        else {
            expect(true).toBeFalsy();
        }
    });
});
describe('Set Resolver', () => {
    test('set resolver test 1', () => {
        const source = 'set Thing(fart):other[used1]:finally(used2, used3):planet to 10.';
        const result = parseSource(source);
        noParseErrors(result);
        const { script } = result.parse;
        const stmt = script.stmts[0];
        const local = new localResolver_1.LocalResolver();
        const resolver = new setResolver_1.SetResolver(local);
        if (stmt instanceof Stmt.Set) {
            const { used, set } = resolver.resolveExpr(stmt.suffix);
            expect(set).not.toBeUndefined();
            if (!typeGuards_1.empty(set)) {
                expect(set.lexeme).toBe('Thing');
            }
            expect(used.length).toBe(4);
            const [fart, used1, used2, used3] = used;
            expect(fart.lexeme).toBe('fart');
            expect(used1.lexeme).toBe('used1');
            expect(used2.lexeme).toBe('used2');
            expect(used3.lexeme).toBe('used3');
        }
        else {
            expect(true).toBeFalsy();
        }
    });
});
//# sourceMappingURL=resolver.spec.js.map