"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const scanner_1 = require("../scanner/scanner");
const parser_1 = require("../parser/parser");
const symbolTableBuilder_1 = require("../analysis/symbolTableBuilder");
const standardLibrary_1 = require("../analysis/standardLibrary");
const preResolver_1 = require("../analysis/preResolver");
const resolver_1 = require("../analysis/resolver");
const typeChecker_1 = require("../typeChecker/typeChecker");
const types_1 = require("../analysis/types");
const typeGuards_1 = require("../utilities/typeGuards");
const boolean_1 = require("../typeChecker/types/primitives/boolean");
const initialize_1 = require("../typeChecker/types/primitives/initialize");
const initialize_2 = require("../typeChecker/types/orbital/initialize");
const scalar_1 = require("../typeChecker/types/primitives/scalar");
const string_1 = require("../typeChecker/types/primitives/string");
const fakeUri = 'C:\\fake.ks';
initialize_1.primitiveInitializer();
initialize_2.oribitalInitializer();
// parse source
const parseSource = (source) => {
    const scanner = new scanner_1.Scanner(source, fakeUri);
    const scan = scanner.scanTokens();
    const parser = new parser_1.Parser(fakeUri, scan.tokens);
    const parse = parser.parse();
    return { scan, parse };
};
const checkSource = (source, standardLib = false) => {
    const result = parseSource(source);
    const symbolTableBuilder = new symbolTableBuilder_1.SymbolTableBuilder(fakeUri);
    if (standardLib) {
        symbolTableBuilder.linkTable(standardLibrary_1.standardLibraryBuilder(0 /* lowercase */));
    }
    const functionResolver = new preResolver_1.PreResolver(result.parse.script, symbolTableBuilder);
    const resolver = new resolver_1.Resolver(result.parse.script, symbolTableBuilder);
    const preResolverError = functionResolver.resolve();
    const resolverErrors = resolver.resolve();
    const ususedErrors = symbolTableBuilder.findUnused();
    const checker = new typeChecker_1.TypeChecker(result.parse.script);
    const typeCheckError = checker.check();
    return {
        ...result,
        resolveDiagnostics: preResolverError.concat(resolverErrors, ususedErrors),
        typeCheckDiagnostics: typeCheckError,
        table: symbolTableBuilder.build(),
    };
};
const noErrors = (result) => {
    expect(result.scan.scanErrors.length).toBe(0);
    expect(result.parse.parseErrors.length).toBe(0);
    expect(result.resolveDiagnostics.length).toBe(0);
    console.log(result.typeCheckDiagnostics);
    expect(result.typeCheckDiagnostics.length).toBe(0);
};
const literalSource = `
local b1 is true.
local b2 is false.

local d1 is 10.0.
local d2 is 4e6.

local i is 5.

local fi is file.ks.
local s is "example".

print fi.
print b1.
print i.
print b2.
print d1.
print d2.
print s.
`;
const identiferSource = `
local b1 is true.
local b2 is false.

local d1 is 10.0.
local d2 is 4e6.

local i1 is 5.

local fi1 is file.ks.
local s1 is "example".

local b3 is b1.
local b4 is b2.

local d3 is d1.
local d4 is d2.

local i2 is i1.

local fi2 is fi1.
local s2 is s1.

print fi2.
print i2.
print b3.
print b4.
print d3.
print d4.
print s2.
`;
const symbolTests = (symbols, name, symbolKind, targetType) => {
    expect(symbols.has(name)).toBe(true);
    const nameWrap = symbols.get(name);
    expect(nameWrap).not.toBeUndefined();
    const nameUnWrap = typeGuards_1.unWrap(nameWrap);
    expect(nameUnWrap.name.lexeme).toBe(name);
    expect(nameUnWrap.tag).toBe(symbolKind);
    expect(nameUnWrap.name.tracker).not.toBeUndefined();
    const nameTrack = typeGuards_1.unWrap(nameUnWrap.name.tracker);
    expect(nameTrack.declared.type).toBe(targetType);
};
describe('Basic inferring', () => {
    test('Literal inferring', () => {
        const results = checkSource(literalSource);
        noErrors(results);
        const { table } = results;
        const symbols = table.fileSymbols();
        const names = new Map(symbols.map((s) => [s.name.lexeme, s]));
        symbolTests(names, 'b1', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'b2', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'd1', types_1.KsSymbolKind.variable, scalar_1.doubleType);
        symbolTests(names, 'd2', types_1.KsSymbolKind.variable, scalar_1.doubleType);
        symbolTests(names, 'i', types_1.KsSymbolKind.variable, scalar_1.integarType);
        symbolTests(names, 's', types_1.KsSymbolKind.variable, string_1.stringType);
        symbolTests(names, 'fi', types_1.KsSymbolKind.variable, string_1.stringType);
    });
    test('Identifier inferring', () => {
        const results = checkSource(identiferSource);
        noErrors(results);
        const { table } = results;
        const symbols = table.fileSymbols();
        const names = new Map(symbols.map((s) => [s.name.lexeme, s]));
        symbolTests(names, 'b3', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'b4', types_1.KsSymbolKind.variable, boolean_1.booleanType);
        symbolTests(names, 'd3', types_1.KsSymbolKind.variable, scalar_1.doubleType);
        symbolTests(names, 'd4', types_1.KsSymbolKind.variable, scalar_1.doubleType);
        symbolTests(names, 'i2', types_1.KsSymbolKind.variable, scalar_1.integarType);
        symbolTests(names, 's2', types_1.KsSymbolKind.variable, string_1.stringType);
        symbolTests(names, 'fi2', types_1.KsSymbolKind.variable, string_1.stringType);
    });
});
//# sourceMappingURL=typeChecking.spec.js.map