import { IGenericArgumentType, ArgumentType, IGenericSuffixType, IGenericVariadicType, CallType, IGenericBasicType, IConstantType, IBasicType, ISuffixType, IVariadicType, IFunctionType, Operator, TypeKind } from './types/types';
import { SuffixTracker } from '../analysis/suffixTracker';
/**
 * This represents a generic type, typically the containers of kos
 */
export declare class GenericBasicType implements IGenericBasicType {
    readonly name: string;
    /**
     * A memoized mapping of this genertic type to concrete types
     */
    private concreteTypes;
    /**
     * Operators that are applicable for this type
     */
    operators: Map<Operator, IBasicType>;
    /**
     * Suffixes attach to this type
     */
    suffixes: Map<string, IGenericSuffixType>;
    /**
     * Does this type inherent from another type
     */
    inherentsFrom?: IGenericArgumentType;
    /**
     * Constructor a generic type
     * @param name name of generic type
     */
    constructor(name: string);
    /**
     * Convert this type into a type string
     */
    toTypeString(): string;
    /**
     * Convert this type into it's concrete representation
     * @param type type parameter
     */
    toConcreteType(type: ArgumentType): ArgumentType;
    /**
     * Is this a full type
     */
    readonly fullType: boolean;
    /**
     * What is the type kind of this type
     */
    readonly kind: TypeKind.basic;
}
/**
 * This represents a generic suffix type, typically suffixes of containers in kos
 */
export declare class GenericSuffixType implements IGenericSuffixType {
    readonly name: string;
    readonly callType: CallType;
    readonly params: IGenericArgumentType[] | IGenericVariadicType;
    readonly returns: IGenericArgumentType;
    /**
     * A memoized mapping of this genertic type to concrete types
     */
    private concreteTypes;
    /**
     * Construct a generic suffix type
     * @param name name of the type
     * @param callType call type of this suffix
     * @param params parameters for this suffix
     * @param returns return type of this suffix
     */
    constructor(name: string, callType: CallType, params: IGenericArgumentType[] | IGenericVariadicType, returns: IGenericArgumentType);
    /**
     * Convert this type to a type string
     */
    toTypeString(): string;
    /**
     * Convert this type into it's concrete representation
     * @param type type parameter
     */
    toConcreteType(type: ArgumentType): ISuffixType;
    /**
     * Is this type a full type
     */
    readonly fullType: boolean;
    /**
     * What is the kind of this type
     */
    readonly kind: TypeKind.suffix;
    /**
     * Generate new parameters types for suffixes with calls
     * @param params parameters to convert to concrete types
     * @param type type parameter
     */
    private newParameters;
}
/**
 * This represents a type
 */
export declare class BasicType implements IBasicType {
    readonly name: string;
    /**
     * Suffixes attach to this type
     */
    suffixes: Map<string, ISuffixType>;
    /**
     * Does this type inherent from another type
     */
    inherentsFrom?: ArgumentType;
    /**
     * Operators that are applicable for this type
     */
    operators: Map<Operator, IBasicType>;
    /**
     * Type constructor
     * @param name name of the new type
     */
    constructor(name: string);
    /**
     * Convert this type into it's concrete representation
     * @param _ type parameter
     */
    toConcreteType(_: ArgumentType): ArgumentType;
    /**
     * Convert this type into a type string
     */
    toTypeString(): string;
    /**
     * Is this a full type
     */
    readonly fullType: true;
    /**
     * What is the kind of this type
     */
    readonly kind: TypeKind.basic;
}
/**
 * This represents a suffix type
 */
export declare class SuffixType implements ISuffixType {
    readonly name: string;
    readonly callType: CallType;
    readonly params: ArgumentType[] | IVariadicType;
    readonly returns: ArgumentType;
    /**
     * The suffix tracker for this type
     */
    private tracker;
    /**
     * Construct a suffix type
     * @param name name of the type
     * @param callType call type of this suffix
     * @param params parameters for this suffix
     * @param returns return type of this suffix
     */
    constructor(name: string, callType: CallType, params: ArgumentType[] | IVariadicType, returns: ArgumentType);
    /**
     * Generate the type string for this suffix type
     */
    toTypeString(): string;
    /**
     * Convert this type into it's concrete representation
     * @param _ type parameter
     */
    toConcreteType(_: ArgumentType): ISuffixType;
    /**
     * Is this a full type
     */
    readonly fullType: true;
    /**
     * What is the kind of this type
     */
    readonly kind: TypeKind.suffix;
    getTracker(): SuffixTracker;
}
/**
 * Represents a constant type, or a type with a fixed value
 */
export declare class ConstantType<T> extends BasicType implements IConstantType<T> {
    readonly value: T;
    /**
     * Construct a constant type
     * @param name name of this constant type
     * @param value value of this constant type
     */
    constructor(name: string, value: T);
    /**
     * Create a type string from this type
     */
    toTypeString(): string;
}
/**
 * Represents a generic variadic type, typically for functions that take an
 * unspecified number of the same parameters
 */
export declare class GenericVariadicType implements IGenericVariadicType {
    readonly type: IGenericBasicType;
    /**
     * A memoized mapping of this genertic type to concrete types
     */
    private concreteTypes;
    /**
     * Construct a generic variadic type
     * @param type type parameter
     */
    constructor(type: IGenericBasicType);
    /**
     * Convert this type to a type string
     */
    toTypeString(): string;
    /**
     * Convert this type into it's concrete representation
     * @param type type parameter
     */
    toConcreteType(type: IBasicType): IVariadicType;
    /**
     * What is the kind of this type
     */
    readonly kind: TypeKind.variadic;
}
export declare class VariadicType extends GenericVariadicType implements IVariadicType {
    readonly type: IBasicType;
    constructor(type: IBasicType);
    toConcreteType(_: IBasicType): IVariadicType;
    readonly fullType: true;
    readonly kind: TypeKind.variadic;
}
export declare class FunctionType implements IFunctionType {
    readonly name: string;
    readonly callType: CallType.call | CallType.optionalCall;
    readonly params: ArgumentType[] | IVariadicType;
    readonly returns: ArgumentType;
    constructor(name: string, callType: CallType.call | CallType.optionalCall, params: ArgumentType[] | IVariadicType, returns: ArgumentType);
    toTypeString(): string;
    toConcreteType(_: IBasicType): IFunctionType;
    readonly kind: TypeKind.function;
    readonly fullType: true;
}
