"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const typeGuards_1 = require("../utilities/typeGuards");
const suffixTracker_1 = require("../analysis/suffixTracker");
const suffix_1 = require("../entities/suffix");
const typeCreators_1 = require("./typeCreators");
/**
 * This represents a generic type, typically the containers of kos
 */
class GenericBasicType {
    /**
     * Constructor a generic type
     * @param name name of generic type
     */
    constructor(name) {
        this.name = name;
        this.suffixes = new Map();
        this.concreteTypes = new Map();
        this.operators = new Map();
    }
    /**
     * Convert this type into a type string
     */
    toTypeString() {
        return `${this.name}<T>`;
    }
    /**
     * Convert this type into it's concrete representation
     * @param type type parameter
     */
    toConcreteType(type) {
        if (this === typeCreators_1.tType) {
            return type;
        }
        // check cache
        const cache = this.concreteTypes.get(type);
        if (!typeGuards_1.empty(cache)) {
            return cache;
        }
        const newType = new BasicType(`${this.name}<${type.toTypeString()}>`);
        this.concreteTypes.set(type, newType);
        const newInherentsFrom = !typeGuards_1.empty(this.inherentsFrom)
            ? this.inherentsFrom.toConcreteType(type)
            : undefined;
        // add suffixes and prototype
        for (const [name, suffixType] of this.suffixes.entries()) {
            newType.suffixes.set(name, suffixType.toConcreteType(type));
        }
        newType.operators = new Map(this.operators);
        newType.inherentsFrom = newInherentsFrom;
        return newType;
    }
    /**
     * Is this a full type
     */
    get fullType() {
        return false;
    }
    /**
     * What is the type kind of this type
     */
    get kind() {
        return 0 /* basic */;
    }
}
exports.GenericBasicType = GenericBasicType;
/**
 * This represents a generic suffix type, typically suffixes of containers in kos
 */
class GenericSuffixType {
    /**
     * Construct a generic suffix type
     * @param name name of the type
     * @param callType call type of this suffix
     * @param params parameters for this suffix
     * @param returns return type of this suffix
     */
    constructor(name, callType, params, returns) {
        this.name = name;
        this.callType = callType;
        this.params = params;
        this.returns = returns;
        this.concreteTypes = new Map();
    }
    /**
     * Convert this type to a type string
     */
    toTypeString() {
        const returnString = returnTypeString(this.returns);
        if (this.callType !== 2 /* call */ &&
            this.callType !== 3 /* optionalCall */) {
            return returnString;
        }
        const paramsString = parameterTypeString(this.params);
        return `<T>(${paramsString}) => ${returnString}`;
    }
    /**
     * Convert this type into it's concrete representation
     * @param type type parameter
     */
    toConcreteType(type) {
        // check cache
        const cache = this.concreteTypes.get(type);
        if (!typeGuards_1.empty(cache)) {
            return cache;
        }
        // generate concete parameters
        const newParams = this.newParameters(this.params, type);
        // generate concrete return
        const newReturns = this.returns.toConcreteType(type);
        const newType = new SuffixType(`${this.name}<${type.toTypeString()}>`, this.callType, newParams, newReturns);
        this.concreteTypes.set(type, newType);
        return newType;
    }
    /**
     * Is this type a full type
     */
    get fullType() {
        return false;
    }
    /**
     * What is the kind of this type
     */
    get kind() {
        return 2 /* suffix */;
    }
    /**
     * Generate new parameters types for suffixes with calls
     * @param params parameters to convert to concrete types
     * @param type type parameter
     */
    newParameters(params, type) {
        // check if variadic type
        if (!Array.isArray(params)) {
            return params.toConcreteType(type);
        }
        const newParams = [];
        for (const param of params) {
            newParams.push(param.toConcreteType(type));
        }
        return newParams;
    }
}
exports.GenericSuffixType = GenericSuffixType;
/**
 * This represents a type
 */
class BasicType {
    /**
     * Type constructor
     * @param name name of the new type
     */
    constructor(name) {
        this.name = name;
        this.suffixes = new Map();
        this.operators = new Map();
    }
    /**
     * Convert this type into it's concrete representation
     * @param _ type parameter
     */
    toConcreteType(_) {
        return this;
    }
    /**
     * Convert this type into a type string
     */
    toTypeString() {
        return this.name;
    }
    /**
     * Is this a full type
     */
    get fullType() {
        return true;
    }
    /**
     * What is the kind of this type
     */
    get kind() {
        return 0 /* basic */;
    }
}
exports.BasicType = BasicType;
/**
 * This represents a suffix type
 */
class SuffixType {
    /**
     * Construct a suffix type
     * @param name name of the type
     * @param callType call type of this suffix
     * @param params parameters for this suffix
     * @param returns return type of this suffix
     */
    constructor(name, callType, params, returns) {
        this.name = name;
        this.callType = callType;
        this.params = params;
        this.returns = returns;
        this.tracker = new suffixTracker_1.SuffixTracker(new suffix_1.KsSuffix(name), this);
    }
    /**
     * Generate the type string for this suffix type
     */
    toTypeString() {
        const returnString = returnTypeString(this.returns);
        if (this.callType !== 2 /* call */ &&
            this.callType !== 3 /* optionalCall */) {
            return returnString;
        }
        const paramsString = parameterTypeString(this.params);
        return `(${paramsString}) => ${returnString}`;
    }
    /**
     * Convert this type into it's concrete representation
     * @param _ type parameter
     */
    toConcreteType(_) {
        return this;
    }
    /**
     * Is this a full type
     */
    get fullType() {
        return true;
    }
    /**
     * What is the kind of this type
     */
    get kind() {
        return 2 /* suffix */;
    }
    getTracker() {
        return this.tracker;
    }
}
exports.SuffixType = SuffixType;
/**
 * Represents a constant type, or a type with a fixed value
 */
class ConstantType extends BasicType {
    /**
     * Construct a constant type
     * @param name name of this constant type
     * @param value value of this constant type
     */
    constructor(name, value) {
        super(name);
        this.value = value;
    }
    /**
     * Create a type string from this type
     */
    toTypeString() {
        return `${super.toTypeString()} = ${this.value}`;
    }
}
exports.ConstantType = ConstantType;
/**
 * Represents a generic variadic type, typically for functions that take an
 * unspecified number of the same parameters
 */
class GenericVariadicType {
    /**
     * Construct a generic variadic type
     * @param type type parameter
     */
    constructor(type) {
        this.type = type;
        this.concreteTypes = new Map();
    }
    /**
     * Convert this type to a type string
     */
    toTypeString() {
        return `...${this.type.toTypeString()}[]`;
    }
    /**
     * Convert this type into it's concrete representation
     * @param type type parameter
     */
    toConcreteType(type) {
        // check cache
        const cache = this.concreteTypes.get(type);
        if (!typeGuards_1.empty(cache)) {
            return cache;
        }
        const newType = new VariadicType(type);
        this.concreteTypes.set(type, newType);
        return newType;
    }
    /**
     * What is the kind of this type
     */
    get kind() {
        return 1 /* variadic */;
    }
}
exports.GenericVariadicType = GenericVariadicType;
class VariadicType extends GenericVariadicType {
    constructor(type) {
        super(type);
        this.type = type;
    }
    toConcreteType(_) {
        return this;
    }
    get fullType() {
        return true;
    }
    get kind() {
        return 1 /* variadic */;
    }
}
exports.VariadicType = VariadicType;
class FunctionType {
    constructor(name, callType, params, returns) {
        this.name = name;
        this.callType = callType;
        this.params = params;
        this.returns = returns;
    }
    toTypeString() {
        const returnString = returnTypeString(this.returns);
        const paramsString = parameterTypeString(this.params);
        return `(${paramsString}) => ${returnString}`;
    }
    toConcreteType(_) {
        return this;
    }
    get kind() {
        return 3 /* function */;
    }
    get fullType() {
        return true;
    }
}
exports.FunctionType = FunctionType;
const returnTypeString = (returns) => {
    return typeGuards_1.empty(returns) ? 'void' : returns.toTypeString();
};
const parameterTypeString = (params) => {
    // empty string for no params
    if (typeGuards_1.empty(params)) {
        return '';
    }
    // check if variadic type
    if (!Array.isArray(params)) {
        return params.toTypeString();
    }
    // string separated i
    return params.map(param => param.toTypeString()).join(', ');
};
//# sourceMappingURL=ksType.js.map