import { IExprVisitor, IStmtVisitor, ISuffixTermParamVisitor } from '../parser/types';
import * as SuffixTerm from '../parser/suffixTerm';
import * as Expr from '../parser/expr';
import * as Stmt from '../parser/stmt';
import * as Decl from '../parser/declare';
import { ITypeResultExpr } from './types';
import { Script } from '../entities/script';
import { ArgumentType } from './types/types';
import { Diagnostic } from 'vscode-languageserver';
import { SuffixTypeBuilder } from './suffixTypeNode';
declare type Diagnostics = Diagnostic[];
/**
 * The type checker attempts to identify places where the rules of the type
 * system are broken.
 */
export declare class TypeChecker implements IStmtVisitor<Diagnostics>, IExprVisitor<ITypeResultExpr<ArgumentType>>, ISuffixTermParamVisitor<SuffixTypeBuilder, Diagnostics> {
    /**
     * the logger to logging information
     */
    private readonly logger;
    /**
     * The tracer for logging the current stack tracer
     */
    private readonly tracer;
    /**
     * The script that is being type checked
     */
    private readonly script;
    /**
     * The cached bound check statement
     */
    private readonly checkStmtBind;
    constructor(script: Script, logger?: ILogger, tracer?: ITracer);
    /**
     * Check the source file for type errors
     */
    check(): Diagnostics;
    /**
     * check a collection of statements
     * @param stmts statements to check
     */
    private checkStmts;
    /**
     * Check an statement for errors
     * @param stmt statement to check
     */
    private checkStmt;
    /**
     * Check an expression for errors
     * @param expr expression to check
     */
    private checkExpr;
    /**
     * Check a suffix term for errors
     * @param suffixTerm suffix term to check
     * @param builder type resolved so far
     */
    private checkSuffixTerm;
    /**
     * Visit a variable declaration
     * @param decl variable declaration
     */
    visitDeclVariable(decl: Decl.Var): Diagnostics;
    /**
     * Vist a lock declaration
     * @param decl lock declaration
     */
    visitDeclLock(decl: Decl.Lock): Diagnostics;
    /**
     * Visit a function declaration
     * @param decl function declaration
     */
    visitDeclFunction(decl: Decl.Func): Diagnostics;
    /**
     * Visit a parameter declaration
     * @param decl parameter declaration
     */
    visitDeclParameter(decl: Decl.Param): Diagnostics;
    /**
     * Vist an invalid statement
     * @param stmt invalid statement
     */
    visitStmtInvalid(stmt: Stmt.Invalid): Diagnostics;
    /**
     * Vist a block statement
     * @param stmt statement block
     */
    visitBlock(stmt: Stmt.Block): Diagnostics;
    /**
     * Visit an statement expression
     * @param stmt statement expression
     */
    visitExpr(stmt: Stmt.ExprStmt): Diagnostics;
    /**
     * Visit an on off statement
     * @param stmt on / off statement
     */
    visitOnOff(stmt: Stmt.OnOff): Diagnostics;
    /**
     * Visit a command statement
     * @param _ command statement
     */
    visitCommand(_: Stmt.Command): Diagnostics;
    /**
     * Visit a command expression statement
     * @param stmt command expression statement
     */
    visitCommandExpr(stmt: Stmt.CommandExpr): Diagnostics;
    /**
     * Visit an unset statement
     * @param _ unset statement
     */
    visitUnset(_: Stmt.Unset): Diagnostics;
    /**
     * Visit an unlock statement
     * @param _ unlock statement
     */
    visitUnlock(_: Stmt.Unlock): Diagnostics;
    /**
     * Visit set statement
     * @param stmt set statement
     */
    visitSet(stmt: Stmt.Set): Diagnostics;
    visitLazyGlobal(_: Stmt.LazyGlobal): Diagnostics;
    visitIf(stmt: Stmt.If): Diagnostics;
    visitElse(stmt: Stmt.Else): Diagnostics;
    visitUntil(stmt: Stmt.Until): Diagnostics;
    visitFrom(stmt: Stmt.From): Diagnostics;
    visitWhen(stmt: Stmt.When): Diagnostics;
    visitReturn(stmt: Stmt.Return): Diagnostics;
    visitBreak(_: Stmt.Break): Diagnostics;
    visitSwitch(stmt: Stmt.Switch): Diagnostics;
    /**
     * Visit a for loop
     * @param stmt for loop statement
     */
    visitFor(stmt: Stmt.For): Diagnostics;
    /**
     * Visit on statment
     * @param stmt on statement
     */
    visitOn(stmt: Stmt.On): Diagnostics;
    /**
     * Visit toggle statement
     * @param stmt toggle statement
     */
    visitToggle(stmt: Stmt.Toggle): Diagnostics;
    /**
     * Visit wait statement
     * @param stmt wait statement
     */
    visitWait(stmt: Stmt.Wait): Diagnostics;
    /**
     * Visit log statement
     * @param stmt log statement
     */
    visitLog(stmt: Stmt.Log): Diagnostics;
    visitCopy(stmt: Stmt.Copy): Diagnostics;
    visitRename(stmt: Stmt.Rename): Diagnostics;
    visitDelete(stmt: Stmt.Delete): Diagnostics;
    visitRun(_: Stmt.Run): Diagnostics;
    visitRunPath(_: Stmt.RunPath): Diagnostics;
    visitRunPathOnce(_: Stmt.RunOncePath): Diagnostics;
    visitCompile(stmt: Stmt.Compile): Diagnostics;
    /**
     * Visit list statement
     * @param stmt list statement
     */
    visitList(stmt: Stmt.List): Diagnostics;
    visitEmpty(_: Stmt.Empty): Diagnostics;
    visitPrint(stmt: Stmt.Print): Diagnostics;
    visitExprInvalid(_: Expr.Invalid): ITypeResultExpr<ArgumentType>;
    visitBinary(expr: Expr.Binary): ITypeResultExpr<ArgumentType>;
    visitUnary(expr: Expr.Unary): ITypeResultExpr<ArgumentType>;
    visitFactor(expr: Expr.Factor): ITypeResultExpr<ArgumentType>;
    /**
     * Visit a suffix
     * @param expr suffix expression
     */
    visitSuffix(expr: Expr.Suffix): ITypeResultExpr<ArgumentType>;
    visitLambda(expr: Expr.Lambda): ITypeResultExpr<ArgumentType>;
    visitSuffixTrailer(suffixTerm: SuffixTerm.SuffixTrailer, builder: SuffixTypeBuilder): Diagnostics;
    visitSuffixTermInvalid(suffixTerm: SuffixTerm.Invalid, builder: SuffixTypeBuilder): Diagnostics;
    visitSuffixTerm(suffixTerm: SuffixTerm.SuffixTerm, builder: SuffixTypeBuilder): Diagnostics;
    /**
     * Visit a call expression and check for type errors
     * @param call the current call expresion
     * @param builder current resolved suffix expression
     */
    visitCall(call: SuffixTerm.Call, builder: SuffixTypeBuilder): Diagnostics;
    /**
     * Visit a function call site
     * @param call call suffix term expression
     * @param builder suffix type builder
     */
    private visitFunctionCall;
    /**
     * Resolve variadic call for type errors
     * @param params parameter types
     * @param call current call expression
     */
    private visitVaradicCall;
    /**
     * Check a normal call signature for type errors
     * @param params the parameter types
     * @param call the call expression
     */
    private visitNormalCall;
    /**
     * visit an array index suffix expression.
     * @param suffixTerm the current array index
     * @param builder the current type
     */
    visitArrayIndex(suffixTerm: SuffixTerm.ArrayIndex, builder: SuffixTypeBuilder): Diagnostics;
    /**
     * visit an array bracket suffix expression.
     * @param suffixTerm the current array bracket expression
     * @param builder the suffix type builder
     */
    visitArrayBracket(suffixTerm: SuffixTerm.ArrayBracket, builder: SuffixTypeBuilder): Diagnostics;
    /**
     * visit the suffix term for delgates. This will return a new delgate type
     * @param suffixTerm the current delgate node
     * @param builder the suffix type builder
     */
    visitDelegate(suffixTerm: SuffixTerm.Delegate, builder: SuffixTypeBuilder): Diagnostics;
    /**
     * visit the suffix term for literals.
     * @param suffixTerm literal syntax node
     * @param builder the suffix type builder
     */
    visitLiteral(suffixTerm: SuffixTerm.Literal, builder: SuffixTypeBuilder): Diagnostics;
    /**
     * visit the suffix term for identifier.
     * @param suffixTerm identifier syntax node
     * @param builder the suffix type builder
     */
    visitIdentifier(suffixTerm: SuffixTerm.Identifier, builder: SuffixTypeBuilder): Diagnostics;
    /**
     * visit the suffix term for grouping.
     * @param suffixTerm grouping syntax node
     * @param builder the suffix type builder
     */
    visitGrouping(suffixTerm: SuffixTerm.Grouping, builder: SuffixTypeBuilder): Diagnostics;
    /**
     * Get the tracker from an atom
     * @param atom atom
     */
    private atomTracker;
    /**
     * Is this tracker a basic tracker
     * @param tracker tracker to inspect
     */
    private isBasicTracker;
    /**
     * Check if the current operator is valid and it's resulting type
     * @param expr the operator expresion
     * @param leftResult the left type
     * @param rightResult the right type
     * @param operator the operator to consider
     */
    private checkOperator;
    /**
     * Return the resultant type with any errors
     * @param type current type
     * @param errors accumulated errors
     */
    private resultExpr;
    /**
     * Get the current result of the suffix type builder
     * @param builder suffix type builder
     */
    private builderResult;
    /**
     * Convert node to string and relevant range
     * @param node node to create message for
     */
    private nodeError;
}
export {};
