"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const SuffixTerm = __importStar(require("../parser/suffixTerm"));
const Expr = __importStar(require("../parser/expr"));
const Stmt = __importStar(require("../parser/stmt"));
const logger_1 = require("../utilities/logger");
const typeGuards_1 = require("../utilities/typeGuards");
const structure_1 = require("./types/primitives/structure");
const coerce_1 = require("./coerce");
const constants_1 = require("../utilities/constants");
const tokentypes_1 = require("../entities/tokentypes");
const node_1 = require("./types/node");
const typeCreators_1 = require("./typeCreators");
const lexicon_1 = require("./types/collections/lexicon");
const arrayUtils_1 = require("../utilities/arrayUtils");
const typeUitlities_1 = require("./typeUitlities");
const void_1 = require("./types/primitives/void");
const userList_1 = require("./types/collections/userList");
const boolean_1 = require("./types/primitives/boolean");
const string_1 = require("./types/primitives/string");
const scalar_1 = require("./types/primitives/scalar");
const typeHelpers_1 = require("./typeHelpers");
const delegate_1 = require("./types/primitives/delegate");
const typeNode_1 = require("./typeNode");
const types_1 = require("../analysis/types");
const positionUtils_1 = require("../utilities/positionUtils");
const list_1 = require("./types/collections/list");
const bodyTarget_1 = require("./types/orbital/bodyTarget");
const vesselTarget_1 = require("./types/orbital/vesselTarget");
const volume_1 = require("./types/io/volume");
const volumeItem_1 = require("./types/io/volumeItem");
const partModule_1 = require("./types/parts/partModule");
const part_1 = require("./types/parts/part");
const path_1 = require("./types/io/path");
const vscode_languageserver_1 = require("vscode-languageserver");
const diagnosticsUtils_1 = require("../utilities/diagnosticsUtils");
const suffixTypeNode_1 = require("./suffixTypeNode");
/**
 * The type checker attempts to identify places where the rules of the type
 * system are broken.
 */
class TypeChecker {
    constructor(script, logger = logger_1.mockLogger, tracer = logger_1.mockTracer) {
        /**
         * The cached bound check statement
         */
        this.checkStmtBind = this.checkStmt.bind(this);
        this.script = script;
        this.logger = logger;
        this.tracer = tracer;
    }
    /**
     * Check the source file for type errors
     */
    check() {
        // resolve the sequence of statements
        try {
            const splits = this.script.uri.split('/');
            const file = splits[splits.length - 1];
            this.logger.info(`Type Checking started for ${file}.`);
            const typeErrors = this.checkStmts(this.script.stmts);
            this.logger.info(`Type Checking finished for ${file}`);
            if (typeErrors.length) {
                this.logger.warn(`Type Checking encounted ${typeErrors.length} errors`);
            }
            return typeErrors;
        }
        catch (err) {
            this.logger.error(`Error occured in type checker ${err}`);
            this.tracer.log(err);
            return [];
        }
    }
    /**
     * check a collection of statements
     * @param stmts statements to check
     */
    checkStmts(stmts) {
        return accumulateErrors(stmts, this.checkStmtBind);
    }
    /**
     * Check an statement for errors
     * @param stmt statement to check
     */
    checkStmt(stmt) {
        return stmt.accept(this);
    }
    /**
     * Check an expression for errors
     * @param expr expression to check
     */
    checkExpr(expr) {
        return expr.accept(this);
    }
    /**
     * Check a suffix term for errors
     * @param suffixTerm suffix term to check
     * @param builder type resolved so far
     */
    checkSuffixTerm(suffixTerm, builder) {
        return suffixTerm.acceptParam(this, builder);
    }
    // ----------------------------- Declaration -----------------------------------------
    /**
     * Visit a variable declaration
     * @param decl variable declaration
     */
    visitDeclVariable(decl) {
        const result = this.checkExpr(decl.value);
        const { tracker } = decl.identifier;
        if (this.isBasicTracker(tracker)) {
            tracker.declareType(result.type);
        }
        return result.errors;
    }
    /**
     * Vist a lock declaration
     * @param decl lock declaration
     */
    visitDeclLock(decl) {
        const result = this.checkExpr(decl.value);
        const { tracker } = decl.identifier;
        if (this.isBasicTracker(tracker)) {
            tracker.declareType(result.type);
        }
        return result.errors;
    }
    /**
     * Visit a function declaration
     * @param decl function declaration
     */
    visitDeclFunction(decl) {
        const errors = this.checkStmt(decl.block);
        const { tracker } = decl.identifier;
        if (!typeGuards_1.empty(tracker) && tracker.kind === 0 /* basic */) {
            const { symbol } = tracker.declared;
            if (symbol.tag === types_1.KsSymbolKind.function) {
                const paramsTypes = [];
                // TODO eventually we should tag ks parameter to the function type
                for (let i = 0; i < symbol.requiredParameters; i += 1) {
                    paramsTypes.push(structure_1.structureType);
                }
                for (let i = 0; i < symbol.optionalParameters; i += 1) {
                    paramsTypes.push(structure_1.structureType);
                }
                const returnType = symbol.returnValue ? structure_1.structureType : void_1.voidType;
                const funcType = typeCreators_1.createFunctionType(tracker.declared.symbol.name.lookup, returnType, ...paramsTypes);
                tracker.declareType(funcType);
            }
        }
        return errors;
    }
    /**
     * Visit a parameter declaration
     * @param decl parameter declaration
     */
    visitDeclParameter(decl) {
        let errors = [];
        // loop over normal parameters
        for (const required of decl.requiredParameters) {
            const { tracker } = required.identifier;
            if (this.isBasicTracker(tracker)) {
                tracker.declareType(structure_1.structureType);
            }
        }
        // loop over defaulted parameters
        for (const optional of decl.optionalParameters) {
            const valueResult = this.checkExpr(optional.value);
            const { tracker } = optional.identifier;
            if (this.isBasicTracker(tracker)) {
                tracker.declareType(valueResult.type);
            }
            errors = errors.concat(valueResult.errors);
        }
        return errors;
    }
    // ----------------------------- Statements -----------------------------------------
    /**
     * Vist an invalid statement
     * @param stmt invalid statement
     */
    visitStmtInvalid(stmt) {
        if (typeGuards_1.empty(stmt.partial)) {
            return [];
        }
        const errors = [];
        // check parsed partial nodes
        for (const node of Object.values(stmt.partial)) {
            if (node instanceof Stmt.Stmt) {
                errors.push(...this.checkStmt(node));
            }
            if (node instanceof Expr.Expr) {
                const result = this.checkExpr(node);
                errors.push(...result.errors);
            }
        }
        return errors;
    }
    /**
     * Vist a block statement
     * @param stmt statement block
     */
    visitBlock(stmt) {
        return accumulateErrors(stmt.stmts, this.checkStmtBind);
    }
    /**
     * Visit an statement expression
     * @param stmt statement expression
     */
    visitExpr(stmt) {
        const result = this.checkExpr(stmt.suffix);
        return result.errors;
    }
    /**
     * Visit an on off statement
     * @param stmt on / off statement
     */
    visitOnOff(stmt) {
        const result = this.checkExpr(stmt.suffix);
        return result.errors;
    }
    /**
     * Visit a command statement
     * @param _ command statement
     */
    visitCommand(_) {
        return [];
    }
    /**
     * Visit a command expression statement
     * @param stmt command expression statement
     */
    visitCommandExpr(stmt) {
        const result = this.checkExpr(stmt.expr);
        const errors = result.errors;
        switch (stmt.command.type) {
            // commands for adding and removing nodes
            case tokentypes_1.TokenType.add:
            case tokentypes_1.TokenType.remove:
                // expression must be a node type for node commands
                if (!coerce_1.coerce(result.type, node_1.nodeType)) {
                    const command = stmt.command.type === tokentypes_1.TokenType.add ? 'add' : 'remove';
                    errors.push(diagnosticsUtils_1.createDiagnostic(stmt.expr, `${command} expected a node.` +
                        ' Node may not able to be  be coerced into node type', vscode_languageserver_1.DiagnosticSeverity.Hint));
                }
                break;
            // command to edit a file
            case tokentypes_1.TokenType.edit:
                if (!coerce_1.coerce(result.type, node_1.nodeType)) {
                    errors.push(diagnosticsUtils_1.createDiagnostic(stmt.expr, 'Path may not be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
                }
                break;
            default:
                throw new Error('Unexpected token type found in command expression');
        }
        return result.errors;
    }
    /**
     * Visit an unset statement
     * @param _ unset statement
     */
    visitUnset(_) {
        return [];
    }
    /**
     * Visit an unlock statement
     * @param _ unlock statement
     */
    visitUnlock(_) {
        return [];
    }
    // visit set
    /**
     * Visit set statement
     * @param stmt set statement
     */
    visitSet(stmt) {
        const exprResult = this.checkExpr(stmt.value);
        const errors = exprResult.errors;
        const suffixResult = this.checkExpr(stmt.suffix);
        errors.push(...suffixResult.errors);
        // check if suffix is settable
        if (!stmt.suffix.isSettable()) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.suffix, `Cannot set ${this.nodeError(stmt.suffix)} as it is a call`, vscode_languageserver_1.DiagnosticSeverity.Hint));
            return errors;
        }
        const { atom, trailers } = stmt.suffix.suffixTerm;
        // if a suffix trailer exists we are a full suffix
        if (!typeGuards_1.empty(stmt.suffix.trailer) || trailers.length > 0) {
            if (!coerce_1.coerce(exprResult.type, suffixResult.type)) {
                errors.push(diagnosticsUtils_1.createDiagnostic(stmt.suffix, `Cannot set suffix ${this.nodeError(stmt.suffix)}` +
                    `of type ${suffixResult.type.name} to ${exprResult.type.name}`, vscode_languageserver_1.DiagnosticSeverity.Hint));
            }
            return errors;
        }
        if (atom instanceof SuffixTerm.Identifier) {
            const { tracker } = atom.token;
            if (this.isBasicTracker(tracker)) {
                // update declare or set type
                if (tracker.declared.symbol.name === atom.token) {
                    tracker.declareType(exprResult.type);
                }
                else {
                    tracker.setType(atom.token, exprResult.type);
                }
            }
            else {
                errors.push(diagnosticsUtils_1.createDiagnostic(stmt.value, 'TODO visit set, this should not occur', vscode_languageserver_1.DiagnosticSeverity.Hint));
            }
        }
        else {
            // was not found to be a valid target of setting
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.suffix, `Cannot set ${stmt.suffix.toString()}, must be identifier, or suffix`, vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors;
    }
    // visit lazy global directive
    visitLazyGlobal(_) {
        return [];
    }
    // visit if statement
    visitIf(stmt) {
        const conditionResult = this.checkExpr(stmt.condition);
        const errors = conditionResult.errors;
        if (!coerce_1.coerce(conditionResult.type, boolean_1.booleanType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.condition, 'Condition may not able to be  be coerced into boolean type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return typeGuards_1.empty(stmt.elseStmt)
            ? errors.concat(this.checkStmt(stmt.body))
            : errors.concat(this.checkStmt(stmt.body), this.checkStmt(stmt.elseStmt));
    }
    // visit else statement
    visitElse(stmt) {
        return this.checkStmt(stmt.body);
    }
    // visit until statement
    visitUntil(stmt) {
        const conditionResult = this.checkExpr(stmt.condition);
        const errors = conditionResult.errors;
        if (!coerce_1.coerce(conditionResult.type, boolean_1.booleanType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.condition, 'Condition may not able to be coerced into boolean type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors.concat(this.checkStmt(stmt.body));
    }
    // visit from loop
    visitFrom(stmt) {
        let errors = this.checkStmt(stmt.initializer);
        const conditionResult = this.checkExpr(stmt.condition);
        errors = errors.concat(conditionResult.errors);
        if (!coerce_1.coerce(conditionResult.type, boolean_1.booleanType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.condition, 'Condition may not able to be coerced into boolean type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors.concat(this.checkStmt(stmt.increment), this.checkStmt(stmt.body));
    }
    // vist when statment
    visitWhen(stmt) {
        const conditionResult = this.checkExpr(stmt.condition);
        const errors = conditionResult.errors;
        if (!coerce_1.coerce(conditionResult.type, boolean_1.booleanType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.condition, 'Condition may not able to be coerced into boolean type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors.concat(this.checkStmt(stmt.body));
    }
    // visit return
    visitReturn(stmt) {
        const errors = [];
        if (!typeGuards_1.empty(stmt.value)) {
            // TODO maybe update function type?
        }
        return errors;
    }
    // visit break
    visitBreak(_) {
        return [];
    }
    // visit switch
    visitSwitch(stmt) {
        const result = this.checkExpr(stmt.target);
        let errors = result.errors;
        if (coerce_1.coerce(result.type, string_1.stringType)) {
            errors = errors.concat(diagnosticsUtils_1.createDiagnostic(stmt.target, 'May not be a string identifer for volume', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors;
    }
    /**
     * Visit a for loop
     * @param stmt for loop statement
     */
    visitFor(stmt) {
        const result = this.checkExpr(stmt.collection);
        let errors = [];
        const { type } = result;
        if (type.kind !== 0 /* basic */ || !typeUitlities_1.hasSuffix(type, constants_1.iterator)) {
            errors = errors.concat(diagnosticsUtils_1.createDiagnostic(stmt.collection, 'May not be a valid enumerable type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        // TODO may be able to detect if type is really pure and not mixed
        const { tracker } = stmt.element;
        if (this.isBasicTracker(tracker)) {
            const collectionIterator = typeUitlities_1.getSuffix(type, constants_1.iterator);
            if (!typeGuards_1.empty(collectionIterator)) {
                const value = typeUitlities_1.getSuffix(collectionIterator.returns, 'value');
                tracker.setType(stmt.element, (value && value.returns) || structure_1.structureType);
            }
            else {
                tracker.setType(stmt.element, structure_1.structureType);
            }
        }
        return errors.concat(this.checkStmt(stmt.body));
    }
    /**
     * Visit on statment
     * @param stmt on statement
     */
    visitOn(stmt) {
        const result = this.checkExpr(stmt.suffix);
        const errors = [];
        if (!coerce_1.coerce(result.type, boolean_1.booleanType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.suffix, 'Condition may not able to be coerced into boolean type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors.concat(this.checkStmt(stmt.body));
    }
    /**
     * Visit toggle statement
     * @param stmt toggle statement
     */
    visitToggle(stmt) {
        const result = this.checkExpr(stmt.suffix);
        const errors = result.errors;
        if (!coerce_1.coerce(result.type, boolean_1.booleanType)) {
            // can only toggle boolean values
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.suffix, 'Toggle requires a boolean type. ' +
                'This may not able to be coerced into boolean type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return result.errors;
    }
    /**
     * Visit wait statement
     * @param stmt wait statement
     */
    visitWait(stmt) {
        const result = this.checkExpr(stmt.expr);
        const errors = result.errors;
        if (typeGuards_1.empty(stmt.until)) {
            // no until wait a set amount of time
            if (!coerce_1.coerce(result.type, scalar_1.scalarType)) {
                errors.push(diagnosticsUtils_1.createDiagnostic(stmt.expr, 'Wait requires a scalar type. ' +
                    'This may not able to be coerced into scalar type', vscode_languageserver_1.DiagnosticSeverity.Hint));
            }
        }
        else {
            // wait until condition
            if (!coerce_1.coerce(result.type, boolean_1.booleanType)) {
                errors.push(diagnosticsUtils_1.createDiagnostic(stmt.expr, 'Wait requires a boolean type. ' +
                    'This may not able to be coerced into boolean type', vscode_languageserver_1.DiagnosticSeverity.Hint));
            }
        }
        return errors;
    }
    /**
     * Visit log statement
     * @param stmt log statement
     */
    visitLog(stmt) {
        const exprResult = this.checkExpr(stmt.expr);
        const logResult = this.checkExpr(stmt.target);
        const errors = exprResult.errors.concat(logResult.errors);
        if (!coerce_1.coerce(exprResult.type, string_1.stringType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.expr, 'Can only log a string type. ' +
                'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        if (!coerce_1.coerce(exprResult.type, string_1.stringType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.expr, 'Can only log to a path. ', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors;
    }
    // visit copy
    visitCopy(stmt) {
        const sourceResult = this.checkExpr(stmt.target);
        const targetResult = this.checkExpr(stmt.destination);
        const errors = sourceResult.errors.concat(targetResult.errors);
        if (!coerce_1.coerce(sourceResult.type, string_1.stringType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.target, 'Can only copy from a string or bare path. ' +
                'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        if (!coerce_1.coerce(sourceResult.type, string_1.stringType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.destination, 'Can only copy to a string or bare path. ' +
                'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors;
    }
    // visit rename
    visitRename(stmt) {
        const targetResult = this.checkExpr(stmt.target);
        const alternativeResult = this.checkExpr(stmt.alternative);
        const errors = targetResult.errors.concat(alternativeResult.errors);
        if (!coerce_1.coerce(targetResult.type, string_1.stringType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.target, 'Can only rename from a string or bare path. ' +
                'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        if (!coerce_1.coerce(targetResult.type, string_1.stringType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.alternative, 'Can only rename to a string or bare path. ' +
                'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors;
    }
    visitDelete(stmt) {
        const targetResult = this.checkExpr(stmt.target);
        const errors = targetResult.errors;
        if (!coerce_1.coerce(targetResult.type, string_1.stringType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.target, 'Can only delete from a string or bare path. ' +
                'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        if (typeGuards_1.empty(stmt.volume)) {
            return errors;
        }
        const volumeResult = this.checkExpr(stmt.volume);
        if (!coerce_1.coerce(targetResult.type, string_1.stringType) &&
            !coerce_1.coerce(targetResult.type, path_1.pathType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.volume, 'Can only rename to a string or bare path. ' +
                'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors.concat(volumeResult.errors);
    }
    visitRun(_) {
        return [];
    }
    visitRunPath(_) {
        return [];
    }
    visitRunPathOnce(_) {
        return [];
    }
    visitCompile(stmt) {
        const targetResult = this.checkExpr(stmt.target);
        const errors = targetResult.errors;
        if (!coerce_1.coerce(targetResult.type, string_1.stringType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.target, 'Can only compile from a string or bare path. ' +
                'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        if (typeGuards_1.empty(stmt.destination)) {
            return errors;
        }
        const destinationResult = this.checkExpr(stmt.destination);
        if (!coerce_1.coerce(destinationResult.type, string_1.stringType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.destination, 'Can only compile to a string or bare path. ' +
                'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors.concat(destinationResult.errors);
    }
    /**
     * Visit list statement
     * @param stmt list statement
     */
    visitList(stmt) {
        const { target, collection } = stmt;
        if (typeGuards_1.empty(target) || typeGuards_1.empty(collection)) {
            return [];
        }
        let finalType;
        const errors = [];
        // determine the type that list returns
        switch (collection.lookup) {
            case 'bodies':
                finalType = bodyTarget_1.bodyTargetType;
                break;
            case 'targets':
                finalType = vesselTarget_1.vesselTargetType;
                break;
            case 'resources':
            case 'parts':
            case 'engines':
            case 'sensors':
            case 'elements':
            case 'dockingports':
                finalType = part_1.partType;
                break;
            case 'files':
                finalType = volumeItem_1.volumeItemType;
                break;
            case 'fonts':
                finalType = string_1.stringType;
                break;
            case 'volumes':
                finalType = volume_1.volumeType;
                break;
            case 'processors':
                finalType = partModule_1.partModuleType;
                break;
            default:
                finalType = structure_1.structureType;
                errors.push(diagnosticsUtils_1.createDiagnostic(collection, 'Not a valid list identifier', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        const { tracker } = target;
        if (this.isBasicTracker(tracker)) {
            tracker.setType(target, list_1.listType.toConcreteType(finalType));
        }
        return errors;
    }
    // visit empty statement
    visitEmpty(_) {
        return [];
    }
    // vist print statement
    visitPrint(stmt) {
        const result = this.checkExpr(stmt.expr);
        const errors = result.errors;
        if (!coerce_1.coerce(result.type, structure_1.structureType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(stmt.expr, 'Cannot print a function, can only print structures', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors;
    }
    // visit invalid expression
    visitExprInvalid(_) {
        return { type: structure_1.structureType, errors: [] };
    }
    // ----------------------------- Expressions -----------------------------------------
    visitBinary(expr) {
        const rightResult = this.checkExpr(expr.right);
        const leftResult = this.checkExpr(expr.left);
        switch (expr.operator.type) {
            case tokentypes_1.TokenType.minus:
                return this.checkOperator(expr, leftResult, rightResult, 1 /* subtract */);
            case tokentypes_1.TokenType.multi:
                return this.checkOperator(expr, leftResult, rightResult, 2 /* multiply */);
            case tokentypes_1.TokenType.div:
                return this.checkOperator(expr, leftResult, rightResult, 3 /* divide */);
            case tokentypes_1.TokenType.plus:
                return this.checkOperator(expr, leftResult, rightResult, 0 /* plus */);
            case tokentypes_1.TokenType.less:
                return this.checkOperator(expr, leftResult, rightResult, 6 /* lessThan */);
            case tokentypes_1.TokenType.lessEqual:
                return this.checkOperator(expr, leftResult, rightResult, 8 /* lessThanEqual */);
            case tokentypes_1.TokenType.greater:
                return this.checkOperator(expr, leftResult, rightResult, 5 /* greaterThan */);
            case tokentypes_1.TokenType.greaterEqual:
                return this.checkOperator(expr, leftResult, rightResult, 7 /* greaterThanEqual */);
            case tokentypes_1.TokenType.and:
            case tokentypes_1.TokenType.or:
                const errors = leftResult.errors.concat(rightResult.errors);
                if (!typeUitlities_1.isSubType(leftResult.type, boolean_1.booleanType) ||
                    !typeUitlities_1.isSubType(leftResult.type, boolean_1.booleanType)) {
                    errors.push(diagnosticsUtils_1.createDiagnostic(expr, '"and" and "or" require boolean types. May not be able to coerce one or other', vscode_languageserver_1.DiagnosticSeverity.Hint));
                }
                return { errors, type: boolean_1.booleanType };
            case tokentypes_1.TokenType.equal:
                return this.checkOperator(expr, leftResult, rightResult, 10 /* equal */);
            case tokentypes_1.TokenType.notEqual:
                return this.checkOperator(expr, leftResult, rightResult, 9 /* notEqual */);
        }
        throw new Error(`Unexpected token ${expr.operator.typeString} type encountered in binary expression`);
    }
    visitUnary(expr) {
        const result = this.checkExpr(expr.factor);
        const errors = result.errors;
        let finalType = undefined;
        switch (expr.operator.type) {
            case tokentypes_1.TokenType.plus:
            case tokentypes_1.TokenType.minus:
                // TODO check if this is true
                if (!coerce_1.coerce(result.type, scalar_1.scalarType)) {
                    errors.push(diagnosticsUtils_1.createDiagnostic(expr.factor, '+/- only valid for a scalar type. ' +
                        'This may not able to be coerced into scalar type', vscode_languageserver_1.DiagnosticSeverity.Hint));
                }
                finalType = scalar_1.scalarType;
                break;
            case tokentypes_1.TokenType.not:
                if (!coerce_1.coerce(result.type, boolean_1.booleanType)) {
                    errors.push(diagnosticsUtils_1.createDiagnostic(expr.factor, 'Can only "not" a boolean type. ' +
                        'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
                }
                finalType = boolean_1.booleanType;
                break;
            case tokentypes_1.TokenType.defined:
                finalType = boolean_1.booleanType;
                break;
            default:
                throw new Error(`Invalid Token ${expr.operator.typeString} for unary operator.`);
        }
        return { errors, type: finalType };
    }
    visitFactor(expr) {
        const suffixResult = this.checkExpr(expr.suffix);
        const exponentResult = this.checkExpr(expr.exponent);
        const errors = suffixResult.errors.concat(exponentResult.errors);
        if (!coerce_1.coerce(suffixResult.type, scalar_1.scalarType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(expr.suffix, 'Can only use scalars as base of power' +
                'This may not able to be coerced into scalar type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        if (!coerce_1.coerce(exponentResult.type, scalar_1.scalarType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(expr.exponent, 'Can only use scalars as exponent of power' +
                'This may not able to be coerced into scalar type', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return { errors, type: scalar_1.scalarType };
    }
    /**
     * Visit a suffix
     * @param expr suffix expression
     */
    visitSuffix(expr) {
        const { suffixTerm, trailer } = expr;
        const [firstTrailer, ...remainingTrailers] = suffixTerm.trailers;
        const builder = new suffixTypeNode_1.SuffixTypeBuilder();
        const errors = this.checkSuffixTerm(suffixTerm.atom, builder);
        if (!typeGuards_1.empty(firstTrailer)) {
            // handle case were suffix is actually a function call
            if (firstTrailer instanceof SuffixTerm.Call) {
                const tracker = this.atomTracker(suffixTerm.atom);
                errors.push(...this.visitFunctionCall(firstTrailer, tracker, builder));
            }
            else {
                errors.push(...this.checkSuffixTerm(firstTrailer, builder));
            }
            // handle remaining suffix term trailers
            for (const trailer of remainingTrailers) {
                errors.push(...this.checkSuffixTerm(trailer, builder));
            }
        }
        // if we have a trailer check that as well
        if (!typeGuards_1.empty(trailer)) {
            errors.push(...this.checkSuffixTerm(trailer, builder));
        }
        return this.resultExpr(this.builderResult(builder), errors);
    }
    visitLambda(expr) {
        const errors = this.checkStmt(expr.block);
        return this.resultExpr(delegate_1.delegateType, errors);
    }
    // ----------------------------- Suffix -----------------------------------------
    visitSuffixTrailer(suffixTerm, builder) {
        // check suffix term and trailers
        const errors = this.checkSuffixTerm(suffixTerm.suffixTerm, builder);
        // if no trailer exist attempt to return
        if (!typeGuards_1.empty(suffixTerm.trailer)) {
            errors.push(...this.checkSuffixTerm(suffixTerm.trailer, builder));
        }
        return errors;
    }
    visitSuffixTermInvalid(suffixTerm, builder) {
        builder.nodes.push(new typeNode_1.TypeNode(typeHelpers_1.suffixError, suffixTerm));
        return [];
    }
    visitSuffixTerm(suffixTerm, builder) {
        // check the atom
        const errors = this.checkSuffixTerm(suffixTerm.atom, builder);
        // add any errors from trailers
        for (const trailer of suffixTerm.trailers) {
            errors.push(...this.checkSuffixTerm(trailer, builder));
        }
        return errors;
    }
    /**
     * Visit a call expression and check for type errors
     * @param call the current call expresion
     * @param builder current resolved suffix expression
     */
    visitCall(call, builder) {
        if (!builder.isTrailer()) {
            throw new Error("TODO shouldn't be able to get here visitCall");
        }
        const type = builder.current();
        const errors = [];
        // if the current type isn't a suffix
        if (type.kind === 2 /* suffix */) {
            builder.nodes.push(new typeNode_1.TypeNode(type, call));
            call.open.tracker = type.getTracker();
            call.close.tracker = type.getTracker();
        }
        else {
            builder.nodes.push(new typeNode_1.TypeNode(typeHelpers_1.suffixError, call));
            call.open.tracker = typeHelpers_1.suffixError.getTracker();
            call.close.tracker = typeHelpers_1.suffixError.getTracker();
            errors.push(diagnosticsUtils_1.createDiagnostic(call, `${type.name} has no call signature`, vscode_languageserver_1.DiagnosticSeverity.Hint));
            return errors;
        }
        // handle variadic
        if (!Array.isArray(type.params)) {
            return this.visitVaradicCall(type.params, call);
        }
        // handle normal functions
        return this.visitNormalCall(type.params, call);
    }
    /**
     * Visit a function call site
     * @param call call suffix term expression
     * @param builder suffix type builder
     */
    visitFunctionCall(call, tracker, builder) {
        const type = builder.current();
        const errors = [];
        // check if previous identifier resolves to a function
        // TODO figure out function trackers
        if (type.kind === 3 /* function */) {
            builder.nodes.push(new typeNode_1.TypeNode(type, call));
            call.open.tracker = tracker;
            call.close.tracker = tracker;
        }
        else {
            builder.nodes.push(new typeNode_1.TypeNode(typeHelpers_1.functionError, call));
            errors.push(diagnosticsUtils_1.createDiagnostic(call, `Type ${type.name} does not have a call signature`, vscode_languageserver_1.DiagnosticSeverity.Hint));
            return errors;
        }
        // handle nomral or varadic calls
        if (!Array.isArray(type.params)) {
            errors.push(...this.visitVaradicCall(type.params, call));
        }
        else {
            errors.push(...this.visitNormalCall(type.params, call));
        }
        return errors;
    }
    /**
     * Resolve variadic call for type errors
     * @param params parameter types
     * @param call current call expression
     */
    visitVaradicCall(params, call) {
        const errors = [];
        for (const arg of call.args) {
            // determine type of each argument
            const result = this.checkExpr(arg);
            errors.push(...result.errors);
            // add diagnostic if argument cannot be matched to parameter type
            if (!coerce_1.coerce(result.type, params.type)) {
                errors.push(diagnosticsUtils_1.createDiagnostic(arg, `Function argument could not be coerced into ${params.type.name}`, vscode_languageserver_1.DiagnosticSeverity.Hint));
            }
        }
        return errors;
    }
    /**
     * Check a normal call signature for type errors
     * @param params the parameter types
     * @param call the call expression
     */
    visitNormalCall(params, call) {
        const errors = [];
        for (const [arg, param] of arrayUtils_1.zip(call.args, params)) {
            // determine type of each argument
            const result = this.checkExpr(arg);
            errors.push(...result.errors);
            // add diagnostic if argument cannot be matched to parameter type
            if (!coerce_1.coerce(result.type, param)) {
                errors.push(diagnosticsUtils_1.createDiagnostic(arg, `Function argument could not be coerced into ${param.name}`, vscode_languageserver_1.DiagnosticSeverity.Hint));
            }
        }
        // check argument length
        if (call.args.length !== params.length) {
            errors.push(diagnosticsUtils_1.createDiagnostic(call.close, `Function expected ${params.length} parameters but was called with ${call.args} arguments`, vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        return errors;
    }
    /**
     * visit an array index suffix expression.
     * @param suffixTerm the current array index
     * @param builder the current type
     */
    visitArrayIndex(suffixTerm, builder) {
        if (!builder.isTrailer()) {
            throw new Error("TODO shouldn't be able to get here visitArrayIndex");
        }
        const type = builder.current();
        const errors = [];
        builder.nodes.push(new typeNode_1.TypeNode(typeHelpers_1.arrayIndexer, suffixTerm));
        // TODO confirm indexable types
        // Only lists are indexable with '#'
        if (!coerce_1.coerce(type, userList_1.userListType)) {
            errors.push(diagnosticsUtils_1.createDiagnostic(suffixTerm, 'indexing with # requires a list', vscode_languageserver_1.DiagnosticSeverity.Hint));
            return errors;
        }
        switch (suffixTerm.indexer.type) {
            // If index is integer we're already in good shape
            case tokentypes_1.TokenType.integer:
                return errors;
            // If index is identify check that it holds a integarType
            case tokentypes_1.TokenType.identifier:
                const { tracker } = suffixTerm.indexer;
                if (this.isBasicTracker(tracker)) {
                    const type = tracker.getType(suffixTerm.indexer);
                    if (!typeGuards_1.empty(type) && coerce_1.coerce(type, scalar_1.integarType)) {
                        return errors;
                    }
                }
                errors.push(diagnosticsUtils_1.createDiagnostic(suffixTerm.indexer, `${suffixTerm.indexer.lexeme} is not a scalar type. ` +
                    'Can only use scalar to index with #', vscode_languageserver_1.DiagnosticSeverity.Hint));
                return errors;
            // All other cases are unallowed
            default:
                errors.push(diagnosticsUtils_1.createDiagnostic(suffixTerm.indexer, 'Can only index an array with # scalars or variables', vscode_languageserver_1.DiagnosticSeverity.Hint));
                return errors;
        }
    }
    /**
     * visit an array bracket suffix expression.
     * @param suffixTerm the current array bracket expression
     * @param builder the suffix type builder
     */
    visitArrayBracket(suffixTerm, builder) {
        if (!builder.isTrailer()) {
            throw new Error("TODO shouldn't be able to get here visitArrayBracket");
        }
        const indexResult = this.checkExpr(suffixTerm.index);
        const errors = indexResult.errors;
        const type = builder.current();
        let indexer = undefined;
        // if we know the collection type is a list we need a scalar indexer
        if (coerce_1.coerce(type, userList_1.userListType)) {
            indexer = typeHelpers_1.arrayBracketIndexer(type, scalar_1.scalarType, structure_1.structureType);
            builder.nodes.push(new typeNode_1.TypeNode(indexer, suffixTerm));
            if (!coerce_1.coerce(indexResult.type, scalar_1.scalarType)) {
                errors.push(diagnosticsUtils_1.createDiagnostic(suffixTerm.index, 'Can only use scalars as list index' +
                    'This may not able to be coerced into scalar type', vscode_languageserver_1.DiagnosticSeverity.Hint));
            }
        }
        // if we know the collection type is a lexicon we need a string indexer
        if (coerce_1.coerce(type, lexicon_1.lexiconType)) {
            indexer = typeHelpers_1.arrayBracketIndexer(type, string_1.stringType, structure_1.structureType);
            builder.nodes.push(new typeNode_1.TypeNode(indexer, suffixTerm));
            if (!coerce_1.coerce(indexResult.type, string_1.stringType)) {
                errors.push(diagnosticsUtils_1.createDiagnostic(suffixTerm.index, 'Can only use a string as a lexicon index.' +
                    'This may not able to be coerced into string type', vscode_languageserver_1.DiagnosticSeverity.Hint));
            }
        }
        // if we know the collection type is a string we need a scalar indexer
        if (coerce_1.coerce(type, string_1.stringType)) {
            indexer = typeHelpers_1.arrayBracketIndexer(type, scalar_1.scalarType, string_1.stringType);
            builder.nodes.push(new typeNode_1.TypeNode(indexer, suffixTerm));
            if (!coerce_1.coerce(indexResult.type, scalar_1.scalarType)) {
                errors.push(diagnosticsUtils_1.createDiagnostic(suffixTerm.index, 'Can only use a scalar as a string index.' +
                    'This may not able to be coerced into scalar type', vscode_languageserver_1.DiagnosticSeverity.Hint));
            }
        }
        // if couldn't coerce into one of our collection types
        // insert error node
        if (typeGuards_1.empty(indexer)) {
            builder.nodes.push(new typeNode_1.TypeNode(typeHelpers_1.suffixError, suffixTerm));
            errors.push(diagnosticsUtils_1.createDiagnostic(suffixTerm, 'Can only index a list, lexicon or string', vscode_languageserver_1.DiagnosticSeverity.Hint));
            suffixTerm.open.tracker = typeHelpers_1.suffixError.getTracker();
            suffixTerm.close.tracker = typeHelpers_1.suffixError.getTracker();
        }
        else {
            suffixTerm.open.tracker = indexer.getTracker();
            suffixTerm.close.tracker = indexer.getTracker();
        }
        return errors;
    }
    /**
     * visit the suffix term for delgates. This will return a new delgate type
     * @param suffixTerm the current delgate node
     * @param builder the suffix type builder
     */
    visitDelegate(suffixTerm, builder) {
        if (!builder.isTrailer()) {
            throw new Error("TODO shouldn't be able to get here visitDelegate");
        }
        const type = builder.current();
        const errors = [];
        if (type.kind !== 3 /* function */) {
            errors.push(diagnosticsUtils_1.createDiagnostic(suffixTerm, 'Can only create delegate of functions', vscode_languageserver_1.DiagnosticSeverity.Hint));
        }
        builder.nodes.push(new typeNode_1.TypeNode(typeHelpers_1.delegateCreation, suffixTerm));
        suffixTerm.atSign.tracker = typeHelpers_1.delegateCreation.getTracker();
        return errors;
    }
    /**
     * visit the suffix term for literals.
     * @param suffixTerm literal syntax node
     * @param builder the suffix type builder
     */
    visitLiteral(suffixTerm, builder) {
        if (builder.isTrailer()) {
            throw new Error("TODO shouldn't be able to get here visitLiteral");
        }
        // if we're at the base push the new type node onto the builder
        switch (suffixTerm.token.type) {
            case tokentypes_1.TokenType.true:
            case tokentypes_1.TokenType.false:
                builder.nodes.push(new typeNode_1.TypeNode(boolean_1.booleanType, suffixTerm));
                return [];
            case tokentypes_1.TokenType.integer:
                builder.nodes.push(new typeNode_1.TypeNode(scalar_1.integarType, suffixTerm));
                return [];
            case tokentypes_1.TokenType.double:
                builder.nodes.push(new typeNode_1.TypeNode(scalar_1.doubleType, suffixTerm));
                return [];
            case tokentypes_1.TokenType.string:
            case tokentypes_1.TokenType.fileIdentifier:
                builder.nodes.push(new typeNode_1.TypeNode(string_1.stringType, suffixTerm));
                return [];
            default:
                throw new Error('TODO invalid literal token found visitLiteral');
        }
    }
    /**
     * visit the suffix term for identifier.
     * @param suffixTerm identifier syntax node
     * @param builder the suffix type builder
     */
    visitIdentifier(suffixTerm, builder) {
        // if we're a trailer check for suffixes
        if (builder.isTrailer()) {
            const type = builder.current();
            const suffix = typeUitlities_1.getSuffix(this.builderResult(builder), suffixTerm.token.lookup);
            // may need to pass sommething in about if we're in get set context
            if (!typeGuards_1.empty(suffix)) {
                // assign type tracker
                suffixTerm.token.tracker = suffix.getTracker();
                // push new node onto builder
                builder.nodes.push(new typeNode_1.TypeNode(suffix, suffixTerm));
                return [];
            }
            // assign error suffix type
            suffixTerm.token.tracker = typeHelpers_1.suffixError.getTracker();
            // add suffix error to builder
            builder.nodes.push(new typeNode_1.TypeNode(typeHelpers_1.suffixError, suffixTerm));
            // indicate suffix not found
            return [
                diagnosticsUtils_1.createDiagnostic(suffixTerm, `Unable to find suffix ${suffixTerm.token.lookup} on type ${type.toTypeString()}`, vscode_languageserver_1.DiagnosticSeverity.Hint),
            ];
        }
        const { tracker } = suffixTerm.token;
        // make sure we have a basic tracker
        if (this.isBasicTracker(tracker)) {
            const type = tracker.getType(suffixTerm.toLocation(this.script.uri));
            // if type is found at this location add it to builder
            if (!typeGuards_1.empty(type)) {
                builder.nodes.push(new typeNode_1.TypeNode(type, suffixTerm));
                return [];
            }
            // if we can't find the type here default to structure
            builder.nodes.push(new typeNode_1.TypeNode(structure_1.structureType, suffixTerm));
            return [
                diagnosticsUtils_1.createDiagnostic(suffixTerm, `Cannot determine type for ${suffixTerm.token.lexeme}.`, vscode_languageserver_1.DiagnosticSeverity.Hint),
            ];
        }
        // in theory we should never get here
        builder.nodes.push(new typeNode_1.TypeNode(structure_1.structureType, suffixTerm));
        // no error as this is likely a resolver error
        return [];
    }
    /**
     * visit the suffix term for grouping.
     * @param suffixTerm grouping syntax node
     * @param builder the suffix type builder
     */
    visitGrouping(suffixTerm, builder) {
        if (builder.isTrailer()) {
            throw new Error("TODO shouldn't be able to get here visitGrouping");
        }
        const { type, errors } = this.checkExpr(suffixTerm.expr);
        // push result of grouping onto builder
        builder.nodes.push(new typeNode_1.TypeNode(type, suffixTerm));
        return errors;
    }
    // ----------------------------- Helpers -----------------------------------------
    /**
     * Get the tracker from an atom
     * @param atom atom
     */
    atomTracker(atom) {
        if (atom instanceof SuffixTerm.Identifier) {
            return atom.token.tracker;
        }
        return undefined;
    }
    /**
     * Is this tracker a basic tracker
     * @param tracker tracker to inspect
     */
    isBasicTracker(tracker) {
        return !typeGuards_1.empty(tracker) && tracker.kind === 0 /* basic */;
    }
    /**
     * Check if the current operator is valid and it's resulting type
     * @param expr the operator expresion
     * @param leftResult the left type
     * @param rightResult the right type
     * @param operator the operator to consider
     */
    checkOperator(expr, leftResult, rightResult, operator) {
        const leftType = leftResult.type;
        const rightType = rightResult.type;
        const errors = leftResult.errors.concat(rightResult.errors);
        let calcType = undefined;
        // TODO could be more efficient
        if (typeUitlities_1.isSubType(leftType, scalar_1.scalarType) && typeUitlities_1.isSubType(rightType, scalar_1.scalarType)) {
            calcType = scalar_1.scalarType;
        }
        else if (typeUitlities_1.isSubType(leftType, string_1.stringType) ||
            typeUitlities_1.isSubType(rightType, string_1.stringType)) {
            calcType = string_1.stringType;
        }
        else if (typeUitlities_1.isSubType(leftType, boolean_1.booleanType) ||
            typeUitlities_1.isSubType(rightType, boolean_1.booleanType)) {
            calcType = boolean_1.booleanType;
        }
        else {
            const leftOp = typeUitlities_1.hasOperator(leftType, operator);
            const rightOp = typeUitlities_1.hasOperator(rightType, operator);
            if (typeGuards_1.empty(leftOp) && typeGuards_1.empty(rightOp)) {
                return {
                    type: structure_1.structureType,
                    errors: errors.concat(diagnosticsUtils_1.createDiagnostic(expr, `${leftType.name} nor ${rightType.name} have TODO operator`, vscode_languageserver_1.DiagnosticSeverity.Hint)),
                };
            }
            if (!typeGuards_1.empty(leftOp)) {
                return { errors, type: leftOp };
            }
            if (!typeGuards_1.empty(rightOp)) {
                return { errors, type: rightOp };
            }
            return { errors, type: structure_1.structureType };
        }
        const returnType = calcType.operators.get(operator);
        if (typeGuards_1.empty(returnType)) {
            return {
                type: structure_1.structureType,
                errors: errors.concat(diagnosticsUtils_1.createDiagnostic(expr, `${calcType.name} type does not have TODO operator`, vscode_languageserver_1.DiagnosticSeverity.Hint)),
            };
        }
        return { errors, type: returnType };
    }
    /**
     * Return the resultant type with any errors
     * @param type current type
     * @param errors accumulated errors
     */
    resultExpr(type, ...errors) {
        return {
            type,
            errors: [].concat(...errors),
        };
    }
    /**
     * Get the current result of the suffix type builder
     * @param builder suffix type builder
     */
    builderResult(builder) {
        const current = builder.current();
        switch (current.kind) {
            case 0 /* basic */:
                return current;
            case 3 /* function */:
            case 2 /* suffix */:
                return current.returns;
        }
    }
    /**
     * Convert node to string and relevant range
     * @param node node to create message for
     */
    nodeError(node) {
        return `${node.toString()} ${positionUtils_1.rangeToString(node)}`;
    }
}
exports.TypeChecker = TypeChecker;
const accumulateErrors = (items, checker) => {
    const errors = [];
    for (const item of items) {
        errors.push(...checker(item));
    }
    return errors;
};
//# sourceMappingURL=typeChecker.js.map