"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const typeGuards_1 = require("../utilities/typeGuards");
const tokentypes_1 = require("../entities/tokentypes");
const boolean_1 = require("./types/primitives/boolean");
const scalar_1 = require("./types/primitives/scalar");
const string_1 = require("./types/primitives/string");
/**
 * Retrieve the type of the follow token
 * @param token token to retreive
 */
exports.tokenTrackedType = (token) => {
    // check literals and other tokens
    switch (token.type) {
        case tokentypes_1.TokenType.true:
        case tokentypes_1.TokenType.false:
            return boolean_1.booleanType;
        case tokentypes_1.TokenType.integer:
            return scalar_1.integarType;
        case tokentypes_1.TokenType.double:
            return scalar_1.doubleType;
        case tokentypes_1.TokenType.string:
        case tokentypes_1.TokenType.fileIdentifier:
            return string_1.stringType;
        default:
            // if not a literally we need to lookup tracker
            const { tracker } = token;
            if (typeGuards_1.empty(tracker)) {
                return undefined;
            }
            return tracker.getType({ uri: token.uri, range: token });
    }
};
/**
 * check if the target call type is compatable with real call type
 * @param queryCallType real call type
 * @param targetCallType query call type
 */
exports.isCorrectCallType = (queryCallType, targetCallType) => {
    switch (queryCallType) {
        case 3 /* optionalCall */:
            return (targetCallType === 0 /* get */ ||
                targetCallType === 2 /* call */ ||
                targetCallType === 3 /* optionalCall */);
        case 0 /* get */:
        case 1 /* set */:
        case 2 /* call */:
            return targetCallType === queryCallType;
    }
};
/**
 * check to see type is a sub type of target type
 * @param queryType query type
 * @param targetType target type
 */
exports.isSubType = (queryType, targetType) => {
    if (queryType.kind === 0 /* basic */ && targetType.kind === 0 /* basic */) {
        return moveDownPrototype(queryType, false, currentType => {
            if (currentType === targetType) {
                return true;
            }
            return undefined;
        });
    }
    return queryType === targetType;
};
/**
 * Does the given type have the requested operator
 * @param type type
 * @param operator operator
 */
exports.hasOperator = (type, operator) => {
    if (type.kind === 0 /* basic */) {
        return moveDownPrototype(type, undefined, currentType => {
            if (!typeGuards_1.empty(currentType.operators.has(operator))) {
                return type;
            }
            return undefined;
        });
    }
    return undefined;
};
/**
 * Does the given type have the requested suffix
 * @param type type
 * @param suffix suffix string
 */
exports.hasSuffix = (type, suffix) => {
    if (type.kind === 0 /* basic */) {
        return moveDownPrototype(type, false, currentType => {
            if (currentType.suffixes.has(suffix)) {
                return true;
            }
            return undefined;
        });
    }
    return false;
};
/**
 * Get the provided suffix from the type if it exists
 * @param type type
 * @param suffix suffix string
 */
exports.getSuffix = (type, suffix) => {
    if (type.kind === 0 /* basic */) {
        return moveDownPrototype(type, undefined, currentType => {
            return currentType.suffixes.get(suffix);
        });
    }
    return undefined;
};
/**
 * Retreive all suffixes from the given type
 * @param type type
 */
exports.allSuffixes = (type) => {
    const suffixes = new Map();
    switch (type.kind) {
        // if basic type get all suffixes on type
        case 0 /* basic */:
            moveDownPrototype(type, false, currentType => {
                for (const [name, suffix] of currentType.suffixes) {
                    if (!suffixes.has(name)) {
                        suffixes.set(name, suffix);
                    }
                }
                return undefined;
            });
            break;
        // TODO may move logic outside of this function if
        // a gettable suffix get all suffixes on return type
        case 2 /* suffix */:
            switch (type.callType) {
                case 0 /* get */:
                case 1 /* set */:
                case 3 /* optionalCall */:
                    moveDownPrototype(type.returns, false, currentType => {
                        for (const [name, suffix] of currentType.suffixes) {
                            if (!suffixes.has(name)) {
                                suffixes.set(name, suffix);
                            }
                        }
                        return undefined;
                    });
                    break;
                default:
                    break;
            }
            break;
        default:
    }
    return Array.from(suffixes.values());
};
/**
 * Is the generic variadic type a full variadic type
 * @param type maybe full variadic type
 */
exports.isFullVarType = (type) => {
    return exports.isFullType(type.type);
};
/**
 * Is the generic arguemnt type a full argument type
 * @param type maybe full argument type
 */
exports.isFullType = (type) => {
    return type.fullType;
};
/**
 * Add type to prototype chain
 * @param type type to add prototype
 * @param prototype prototype
 */
exports.addPrototype = (type, prototype) => {
    type.inherentsFrom = prototype;
};
/**
 * Add operator to type
 * @param type type to add operator
 * @param operators operators
 */
exports.addOperators = (type, ...operators) => {
    for (const [operator, returnType] of operators) {
        if (type.operators.has(operator)) {
            throw new Error(`duplicate operator ${operator} added to type`);
        }
        type.operators.set(operator, returnType);
    }
};
/**
 * Add suffixes to type
 * @param type type to add suffixes
 * @param suffixes suffixes
 */
exports.addSuffixes = (type, ...suffixes) => {
    for (const suffix of suffixes) {
        if (type.suffixes.has(suffix.name)) {
            throw new Error(`duplicate suffix ${suffix.name} added to type`);
        }
        type.suffixes.set(suffix.name, suffix);
    }
};
/**
 * Helper function to move down prototype chain
 * @param type type to query
 * @param nullValue null if function does not return
 * @param func query function
 */
const moveDownPrototype = (type, nullValue, func) => {
    let currentType = type;
    while (true) {
        const result = func(currentType);
        if (!typeGuards_1.empty(result)) {
            return result;
        }
        if (typeGuards_1.empty(currentType.inherentsFrom)) {
            return nullValue;
        }
        currentType = currentType.inherentsFrom;
    }
};
//# sourceMappingURL=typeUitlities.js.map