"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const Stmt = __importStar(require("../parser/stmt"));
const SuffixTerm = __importStar(require("../parser/suffixTerm"));
const Expr = __importStar(require("../parser/expr"));
const path_1 = require("path");
const typeGuards_1 = require("./typeGuards");
const tokentypes_1 = require("../entities/tokentypes");
/**
 * Class to resolve run statements or calls to file paths
 */
class PathResolver {
    constructor(volume0Path, volume0Uri) {
        this.volume0Path = volume0Path;
        this.volume0Uri = volume0Uri;
        if (path_1.sep === '\\') {
            this.replacer = /\\/g;
        }
        else {
            this.replacer = new RegExp(path_1.sep, 'g');
        }
    }
    /**
     * Is the resolve ready to resolve paths
     */
    get ready() {
        return !typeGuards_1.empty(this.volume0Path) && !typeGuards_1.empty(this.volume0Uri);
    }
    /**
     * Resolve uri to load data
     * @param caller location of caller
     * @param path provided path in run statement or class
     */
    resolveUri(caller, path) {
        if (typeGuards_1.empty(path) || typeGuards_1.empty(this.volume0Path) || typeGuards_1.empty(this.volume0Uri)) {
            return undefined;
        }
        // get relative run path from file
        const relativePath = path_1.relative(this.volume0Uri, path_1.dirname(caller.uri)).replace('%20', ' ');
        // check if the scripts reads from volume 0 "disk"
        // TODO no idea what to do for ship volumes
        const [possibleVolumne, ...remaining] = path.split('/');
        if (possibleVolumne.startsWith('0:')) {
            // if of style 0:first\remaining...
            if (possibleVolumne.length > 2) {
                const first = possibleVolumne.slice(2);
                return this.loadData(caller, first, ...remaining);
            }
            // else of style 0:\remaining...
            return this.loadData(caller, ...remaining);
        }
        // if no volumne do a relative lookup
        return this.loadData(caller, relativePath, possibleVolumne, ...remaining);
    }
    /**
     * Creates a load data payload from a caller and path segments
     * @param caller call location
     * @param pathSegments path segments
     */
    loadData(caller, ...pathSegments) {
        if (typeGuards_1.empty(this.volume0Path))
            return undefined;
        return {
            caller: { start: caller.range.start, end: caller.range.end },
            path: path_1.join(this.volume0Path, ...pathSegments),
            uri: [this.volume0Uri, this.pathToUri(...pathSegments)].join('/'),
        };
    }
    /**
     * Convert a path to a uri
     * @param pathSegments path segments to convert
     */
    pathToUri(...pathSegments) {
        return path_1.join(...pathSegments).replace(this.replacer, '/');
    }
    /**
     * Get the relative path to volume 0: for a uri
     * @param uri requested uri
     */
    relativePath(uri) {
        if (typeGuards_1.empty(this.volume0Uri)) {
            return uri;
        }
        return path_1.relative(this.volume0Uri, uri).replace('%20', ' ');
    }
}
exports.PathResolver = PathResolver;
/**
 * Get io path, currently only supports string literals
 * @param stmt io statements
 */
exports.ioPath = (stmt) => {
    const { target } = stmt;
    if (target instanceof SuffixTerm.Literal) {
        return literalPath(target);
    }
    return undefined;
};
/**
 * based on run type determine how to get file path
 * @param stmt run statement
 */
exports.runPath = (stmt) => {
    if (stmt instanceof Stmt.Run) {
        const { identifier } = stmt;
        switch (identifier.type) {
            case tokentypes_1.TokenType.string:
                return identifier.literal;
            case tokentypes_1.TokenType.fileIdentifier:
            case tokentypes_1.TokenType.identifier:
                return identifier.lexeme;
            default:
                return undefined;
        }
    }
    // for run path varients check for literal
    const { expr } = stmt;
    if (expr instanceof Expr.Suffix) {
        if (expr.suffixTerm.atom instanceof SuffixTerm.Literal) {
            return literalPath(expr.suffixTerm.atom);
        }
    }
    return undefined;
};
/**
 * determine which string to return for the filepath from literal
 * @param expr literal expression
 */
const literalPath = (expr) => {
    const { token } = expr;
    switch (token.type) {
        case tokentypes_1.TokenType.string:
            return token.literal;
        case tokentypes_1.TokenType.fileIdentifier:
            return token.lexeme;
    }
    return undefined;
};
//# sourceMappingURL=pathResolver.js.map