"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Is this position after the next
 * @param pos1 the target position
 * @param pos2 the other position
 */
exports.positionAfter = (pos1, pos2) => {
    if (pos1.line > pos2.line) {
        return true;
    }
    if (pos1.line === pos2.line) {
        if (pos1.character > pos2.character) {
            return true;
        }
    }
    return false;
};
/**
 * Is this position equal to or after the next
 * @param pos1 the target position
 * @param pos2 the other position
 */
exports.positionAfterEqual = (pos1, pos2) => {
    return exports.positionAfter(pos1, pos2) || exports.positionEqual(pos1, pos2);
};
/**
 * Is this position before the next
 * @param pos1 the target position
 * @param pos2 the other position
 */
exports.positionBefore = (pos1, pos2) => {
    if (pos1.line < pos2.line) {
        return true;
    }
    if (pos1.line === pos2.line) {
        if (pos1.character < pos2.character) {
            return true;
        }
    }
    return false;
};
/**
 * Is this position equal or before the next
 * @param pos1 the target position
 * @param pos2 the other position
 */
exports.positionBeforeEqual = (pos1, pos2) => {
    return exports.positionBefore(pos1, pos2) || exports.positionEqual(pos1, pos2);
};
/**
 * Is this position equal to the next
 * @param pos1 the target position
 * @param pos2 the other position
 */
exports.positionEqual = (pos1, pos2) => {
    return pos1.line === pos2.line && pos1.character === pos2.character;
};
/**
 * Is this range equal to the next
 * @param range1 the target range
 * @param range2 the other range
 */
exports.rangeEqual = (range1, range2) => {
    if (!exports.positionEqual(range1.start, range2.start))
        return false;
    return exports.positionEqual(range1.end, range2.end);
};
/**
 * Does this range contain another prange
 * @param range1 the target range
 * @param range2 the query range
 */
exports.rangeContains = (range1, range2) => {
    if (range2.start.line < range1.start.line)
        return false;
    if (range2.start.line === range1.start.line && range2.start.character < range1.start.character) {
        return false;
    }
    if (range2.end.line > range1.end.line)
        return false;
    if (range2.end.line === range1.end.line && range2.end.character > range1.end.character) {
        return false;
    }
    return true;
};
/**
 * Does this range contain this position
 * @param range the target range
 * @param pos the query position
 */
exports.rangeContainsPos = (range, pos) => {
    if (pos.line < range.start.line)
        return false;
    if (pos.line === range.start.line && pos.character < range.start.character) {
        return false;
    }
    if (pos.line > range.end.line)
        return false;
    if (pos.line === range.end.line && pos.character > range.end.character) {
        return false;
    }
    return true;
};
/**
 * Is this range before this position
 * @param range the target range
 * @param pos the query position
 */
exports.rangeBefore = (range, pos) => {
    if (pos.line > range.end.line)
        return true;
    if (pos.line === range.end.line) {
        if (pos.character > range.end.character)
            return true;
    }
    return false;
};
/**
 * Is this range after this position
 * @param range the target range
 * @param pos the query position
 */
exports.rangeAfter = (range, pos) => {
    if (pos.line < range.start.line)
        return true;
    if (pos.line === range.start.line) {
        if (pos.character < range.start.character)
            return true;
    }
    return false;
};
/**
 * This is used to sort a set of ranges in to order
 * @param range1 first range
 * @param range2 second range
 */
exports.rangeOrder = (range1, range2) => {
    // check line number first
    if (range1.end.line < range2.start.line)
        return -1;
    if (range1.start.line > range2.end.line)
        return 1;
    // must have equal lines so check characters
    if (range1.end.character < range2.start.character)
        return -1;
    if (range1.start.character > range2.end.character)
        return 1;
    // ranges must have overlap TODO for now treat as equal
    return 0;
};
/**
 * Does this range intersect the next
 * @param range1 the target range
 * @param range2 the other range
 */
exports.rangeIntersection = (range1, range2) => {
    if (exports.rangeAfter(range1, range2.end))
        return false;
    if (exports.rangeBefore(range2, range2.end))
        return false;
    return true;
};
/**
 * Generate a human readable string of the range
 * @param range range to string
 */
exports.rangeToString = (range) => {
    const sameLine = range.start.line === range.end.line;
    if (range.start.line === range.end.line) {
        const sameCharacter = range.start.character === range.end.character;
        const character = sameLine && sameCharacter
            ? (range.start.character + 1).toString()
            : `${range.start.character + 1}-${range.end.character + 1}`;
        return `line: ${range.start.line + 1} character: ${character}`;
    }
    return `${exports.positionToString(range.start)} to ${exports.positionToString(range.end)}`;
};
/**
 * Generate a human readable string of this position
 * @param pos position to string
 */
exports.positionToString = (pos) => {
    return `line: ${pos.line + 1} character: ${pos.character + 1}`;
};
/**
 * Binary search a collection of ordered ranges
 * @param ranges sorted ranges to search
 * @param pos query position
 */
exports.binarySearch = (ranges, pos) => {
    const index = exports.binarySearchIndex(ranges, pos);
    return Array.isArray(index) ? undefined : ranges[index];
};
/**
 * Binary search right on a collection of ordered ranges
 * @param ranges sorted ranges to search
 * @param pos query position
 */
exports.binaryRight = (ranges, pos) => {
    if (exports.rangeAfter(ranges[0], pos)) {
        return undefined;
    }
    const index = exports.binarySearchIndex(ranges, pos);
    return Array.isArray(index) ? ranges[index[1]] : ranges[index];
};
/**
 * Binary search left on a collection of ordered ranges
 * @param ranges sorted ranges to search
 * @param pos query position
 */
exports.binaryLeft = (ranges, pos) => {
    if (ranges.length === 0 || exports.rangeBefore(ranges[ranges.length - 1], pos)) {
        return undefined;
    }
    const index = exports.binarySearchIndex(ranges, pos);
    return Array.isArray(index) ? ranges[index[0]] : ranges[index];
};
/**
 * Binary search with provided function to generate a range
 * @param ranges sorted collection of items that can generate a range
 * @param pos query position
 * @param key key function to generate a range
 */
exports.binaryRightKey = (ranges, pos, key) => {
    if (ranges.length === 0 || exports.rangeAfter(key(ranges[0]), pos)) {
        return undefined;
    }
    const index = exports.binarySearchKeyIndex(ranges, pos, key);
    return Array.isArray(index) ? ranges[index[1]] : ranges[index];
};
/**
 * Binary search left with provided function to generate a range
 * @param ranges sorted collection of items that can generate a range
 * @param pos query position
 * @param key key function to generate a range
 */
exports.binaryLeftKey = (ranges, pos, key) => {
    if (ranges.length === 0 || exports.rangeBefore(key(ranges[ranges.length - 1]), pos)) {
        return undefined;
    }
    const index = exports.binarySearchKeyIndex(ranges, pos, key);
    return Array.isArray(index) ? ranges[index[0]] : ranges[index];
};
/**
 * Binary search for index with provided function to generate a range.
 * Returns array if between ranges
 * @param ranges ordered collection of ranges
 * @param pos query position
 */
exports.binarySearchIndex = (ranges, pos) => {
    let left = 0;
    let right = ranges.length - 1;
    while (left <= right) {
        const mid = Math.floor((right + left) / 2);
        if (exports.rangeBefore(ranges[mid], pos)) {
            left = mid + 1;
        }
        else if (exports.rangeAfter(ranges[mid], pos)) {
            right = mid - 1;
        }
        else if (exports.rangeContainsPos(ranges[mid], pos)) {
            return mid;
        }
        else {
            return [left, right];
        }
    }
    return left < right
        ? [left, right]
        : [right, left];
};
/**
 * Binary search for index of matching item using . Returns array if between ranges
 * @param ranges ordered collection of ranges
 * @param pos query position
 */
exports.binarySearchKeyIndex = (ranges, pos, key) => {
    let left = 0;
    let right = ranges.length - 1;
    while (left <= right) {
        const mid = Math.floor((right + left) / 2);
        if (exports.rangeBefore(key(ranges[mid]), pos)) {
            left = mid + 1;
        }
        else if (exports.rangeAfter(key(ranges[mid]), pos)) {
            right = mid - 1;
        }
        else if (exports.rangeContainsPos(key(ranges[mid]), pos)) {
            return mid;
        }
        else {
            return [left, right];
        }
    }
    return left < right
        ? [left, right]
        : [right, left];
};
/**
 * Are these locations equal
 * @param loc1 target location
 * @param loc2 other location
 */
exports.locationEqual = (loc1, loc2) => {
    if (loc1.uri !== loc2.uri)
        return false;
    return exports.rangeEqual(loc1.range, loc2.range);
};
/**
 * Does location 1 contain location 2
 * @param loc1 target location
 * @param loc2 query location
 */
exports.locationContains = (loc1, loc2) => {
    if (loc1.uri !== loc2.uri)
        return false;
    return exports.rangeContains(loc1.range, loc2.range);
};
//# sourceMappingURL=positionUtils.js.map