"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const vscode_languageserver_1 = require("vscode-languageserver");
const typeGuards_1 = require("./typeGuards");
const typeUitlities_1 = require("../typeChecker/typeUitlities");
const types_1 = require("../analysis/types");
const clean_1 = require("./clean");
const constants_1 = require("./constants");
/**
 * Get the connection primitives based on the request connection type
 * @param connectionType connection type
 */
exports.getConnectionPrimitives = (connectionType) => {
    let reader;
    let writer;
    switch (connectionType) {
        case '--node-ipc':
            reader = new vscode_languageserver_1.IPCMessageReader(process);
            writer = new vscode_languageserver_1.IPCMessageWriter(process);
            break;
        case '--stdio':
            reader = new vscode_languageserver_1.StreamMessageReader(process.stdin);
            writer = new vscode_languageserver_1.StreamMessageWriter(process.stdout);
            break;
        default:
            throw new Error('');
    }
    writer.onError(([error, message, code]) => {
        console.log(error);
        console.log(message);
        console.log(code);
    });
    reader.onError(error => {
        console.log(error);
    });
    return {
        reader,
        writer,
    };
};
/**
 * Get the connection primitives based on the request connection type
 * @param connectionType connection type
 */
exports.updateServer = (server) => {
    const { clientConfig } = server;
    let caseKind = 2 /* camelcase */;
    switch (clientConfig.completionCase) {
        case 'lowercase':
            caseKind = 0 /* lowercase */;
            break;
        case 'uppercase':
            caseKind = 1 /* uppercase */;
            break;
        case 'camelcase':
            caseKind = 2 /* camelcase */;
            break;
        case 'pascalcase':
            caseKind = 3 /* pascalcase */;
            break;
    }
    let logLevel = 4 /* error */;
    switch (clientConfig.trace.server.level) {
        case 'verbose':
            logLevel = 0 /* verbose */;
            break;
        case 'info':
            logLevel = 1 /* info */;
            break;
        case 'log':
            logLevel = 1 /* info */;
            break;
        case 'warn':
            logLevel = 3 /* warn */;
            break;
        case 'error':
            logLevel = 4 /* error */;
            break;
        case 'none':
            logLevel = 5 /* none */;
            break;
    }
    server.keywords = constants_1.keywordCompletions(caseKind);
    server.analyzer.logger.level = logLevel;
    server.analyzer.setCase(caseKind);
};
/**
 * Get a list of all symbols currently in scope at the given line
 * @param analyzer analyzer instance
 * @param documentPosition the current position in the document
 * @param keywordCompletions a list of keywords to always concat
 */
exports.symbolCompletionItems = (analyzer, documentPosition, keywordCompletions) => {
    const { position } = documentPosition;
    const { uri } = documentPosition.textDocument;
    // get all symbols currently in scope
    const entities = analyzer.getScopedSymbols(position, uri);
    // generate completions
    return entities
        .map(entity => {
        let kind = undefined;
        switch (entity.tag) {
            case types_1.KsSymbolKind.function:
                kind = vscode_languageserver_1.CompletionItemKind.Function;
                break;
            case types_1.KsSymbolKind.parameter:
                kind = vscode_languageserver_1.CompletionItemKind.Variable;
                break;
            case types_1.KsSymbolKind.lock:
                kind = vscode_languageserver_1.CompletionItemKind.Reference;
                break;
            case types_1.KsSymbolKind.variable:
                kind = vscode_languageserver_1.CompletionItemKind.Variable;
                break;
            default:
                throw new Error('Unknown entity type');
        }
        let typeString = 'structure';
        const { tracker } = entity.name;
        if (!typeGuards_1.empty(tracker)) {
            const type = tracker.getType({ uri, range: entity.name });
            if (!typeGuards_1.empty(type)) {
                typeString = type.toTypeString();
            }
        }
        return {
            kind,
            label: entity.name.lexeme,
            detail: `${entity.name.lexeme}: ${typeString}`,
            data: clean_1.cleanToken(entity.name),
        };
    })
        .concat(keywordCompletions)
        .map(completion => clean_1.cleanCompletion(completion));
};
/**
 * Get a list of all known suffixes give a suffix completion trigger
 * @param analyzer analyzer instance
 * @param documentPosition the current position in the document
 */
exports.suffixCompletionItems = (analyzer, documentPosition) => {
    const { position } = documentPosition;
    const { uri } = documentPosition.textDocument;
    // TODO more robust method
    const token = analyzer.getToken({ line: position.line, character: position.character - 2 }, uri);
    if (typeGuards_1.empty(token)) {
        return [];
    }
    const type = typeUitlities_1.tokenTrackedType(token);
    // if type not found exit
    if (typeGuards_1.empty(type)) {
        return [];
    }
    // get all suffixes on the predicted type
    const suffixes = typeUitlities_1.allSuffixes(type);
    // generate completions
    return suffixes.map(suffix => {
        switch (suffix.callType) {
            case 2 /* call */:
            case 3 /* optionalCall */:
                return {
                    kind: vscode_languageserver_1.CompletionItemKind.Method,
                    label: suffix.name,
                    detail: `${suffix.name}: ${suffix.toTypeString()}`,
                };
            case 0 /* get */:
            case 1 /* set */:
                return {
                    kind: vscode_languageserver_1.CompletionItemKind.Property,
                    label: suffix.name,
                    detail: `${suffix.name}: ${suffix.toTypeString()}`,
                };
            default:
                throw new Error('Unanticipated call type found');
        }
    });
};
/**
 * Get all symbols in the current document
 * @param analyzer analyzer instance
 * @param documentSymbol document identifier
 */
exports.documentSymbols = (analyzer, documentSymbol) => {
    const { uri } = documentSymbol.textDocument;
    const entities = analyzer.getAllFileSymbols(uri);
    return entities.map(entity => {
        let kind = undefined;
        switch (entity.tag) {
            case types_1.KsSymbolKind.function:
                kind = vscode_languageserver_1.SymbolKind.Function;
                break;
            case types_1.KsSymbolKind.parameter:
                kind = vscode_languageserver_1.SymbolKind.Variable;
                break;
            case types_1.KsSymbolKind.lock:
                kind = vscode_languageserver_1.SymbolKind.Object;
                break;
            case types_1.KsSymbolKind.variable:
                kind = vscode_languageserver_1.SymbolKind.Variable;
                break;
            default:
                throw new Error('Unknown symbol type');
        }
        return {
            kind,
            name: entity.name.lexeme,
            location: clean_1.cleanLocation({
                uri: entity.name.uri || uri,
                range: entity.name,
            }),
        };
    });
};
//# sourceMappingURL=serverUtils.js.map